//
// $XConsortium$
//

//
// Copyright (c) 1993-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

//
// Marshal buffer implementation
//

#include <Ox/env.h>
#include <Ox/exchange.h>
#include <Ox/ix_rep.h>
#include <Ox/marshal.h>
#include <Ox/schema.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#if defined(linux) || defined(__alpha)
extern "C" void exit(int);
#endif

//
// Error handling -- needs to be internationalized
//
// Most of these functions are inline because they are only called
// once and inlining should result in less code.  Error functions
// called more than once, such as overflow or underflow, should not
// be inline.
//

static void err_overflow() {
    fprintf(stderr, "MarshalBuffer overflow\n");
    exit(1);
}

static void err_underflow() {
    fprintf(stderr, "MarshalBuffer underflow\n");
    exit(1);
}

static inline void err_dispatch_receive(CORBA::TypeId t) {
    fprintf(stderr, "No receive function for id %08x\n", *t);
    exit(1);
}

#define buf_words(n) (((n) + sizeof(Long) - 1) / sizeof(Long))

struct GenericSequence {
    Long maximum;
    Long length;
    void* buffer;
    Boolean release;
};

void OxMarshalBuffer::init() {
    cur_ = &data_[0];
    end_ = &data_[sizeof(data_) / sizeof(data_[0])];
}

void OxMarshalBuffer::destroy() { }

void OxMarshalBuffer::make_out_in() {
    end_ = cur_;
    cur_ = &data_[0];
}

void OxMarshalBuffer::put32(Long n) {
    if (cur_ >= end_) {
	err_overflow();
    }
    *cur_++ = n;
}

Long OxMarshalBuffer::get32() {
    if (cur_ >= end_) {
	err_underflow();
    }
    Long n = *cur_;
    ++cur_;
    return n;
}

void OxMarshalBuffer::put_boolean(Boolean b) { put32(Long(b)); }
void OxMarshalBuffer::put_char(char c) { put32(Long(c)); }
void OxMarshalBuffer::put_octet(unsigned char c) { put32(Long(c)); }
void OxMarshalBuffer::put_short(short s) { put32(Long(s)); }
void OxMarshalBuffer::put_unsigned_short(unsigned short s) {
    put32(Long(s));
}
void OxMarshalBuffer::put_long(Long n) { put32(n); }
void OxMarshalBuffer::put_unsigned_long(ULong n) { put32(Long(n)); }

void OxMarshalBuffer::put_longlong(LongLong n) {
    Long w = buf_words(sizeof(n));
    if (cur_ + w >= end_) {
	err_overflow();
    }
    *(LongLong*)cur_ = n;
    move(w);
}

void OxMarshalBuffer::put_unsigned_longlong(ULongLong n) {
    Long w = buf_words(sizeof(n));
    if (cur_ + w >= end_) {
	err_overflow();
    }
    *(ULongLong*)cur_ = n;
    move(w);
}

void OxMarshalBuffer::put_float(float f) {
    Long w = buf_words(sizeof(f));
    if (cur_ + w >= end_) {
	err_overflow();
    }
    *(float*)cur_ = f;
    move(w);
}

void OxMarshalBuffer::put_double(double d) {
    Long w = buf_words(sizeof(d));
    if (cur_ + w >= end_) {
	err_overflow();
    }
    *(double*)cur_ = d;
    move(w);
}

//
// Marshal a string.  A nil pointer is marshalled as
// a string of length 0, and a string of length 0 is
// marshalled as a string of length 1 with a null character.
//

void OxMarshalBuffer::put_string(const char* s) {
    char* p = (char*)cur_;
    const char* src;
    ULong n;
    if (s == nil) {
	src = "\0\1";
	n = 2;
    } else if (s[0] == nil) {
	src = "\0\2";
	n = 2;
    } else {
	src = s;
	n = strlen(src) + 1;
    }
    if (p + n >= (char*)end_) {
	err_overflow();
    }
    memmove(p, src, int(n));
    move(n);
}

void OxMarshalBuffer::put_seq_hdr(const void* v) {
    const GenericSequence* seq = (const GenericSequence*)v;
    put_long(seq->maximum);
    put_long(seq->length);
}

//
// This function assumes the sequence element type is primitive,
// e.g., char, short, long.
//

void OxMarshalBuffer::put_seq(const void* v, Long size) {
    const GenericSequence* seq = (const GenericSequence*)v;
    put_long(seq->maximum);
    put_long(seq->length);
    Long n = seq->length * size;
    Long w = buf_words(n);
    if (cur_ + w >= end_) {
	err_overflow();
    }
    memcpy(cur_, seq->buffer, int(n));
    move(w);
}

void OxMarshalBuffer::put_object(CORBA::Object_ptr o) {
    if (is_nil(o)) {
	_OxExchangeId_put_nil(*this);
    } else {
	o->_exchange()->marshal(*this, o);
    }
}

Boolean OxMarshalBuffer::get_boolean() { return Boolean(get32()); }
char OxMarshalBuffer::get_char() { return char(get32()); }
unsigned char OxMarshalBuffer::get_octet() { return (unsigned char)(get32()); }
short OxMarshalBuffer::get_short() { return short(get32()); }
unsigned short OxMarshalBuffer::get_unsigned_short() {
    return (unsigned short)(get32());
}
Long OxMarshalBuffer::get_long() { return Long(get32()); }
ULong OxMarshalBuffer::get_unsigned_long() {
    return (ULong)get32();
}

LongLong OxMarshalBuffer::get_longlong() {
    LongLong n = *(LongLong*)cur_;
    move(buf_words(sizeof(n)));
    return n;
}

ULongLong OxMarshalBuffer::get_unsigned_longlong() {
    ULongLong n = *(ULongLong*)cur_;
    move(buf_words(sizeof(n)));
    return n;
}

float OxMarshalBuffer::get_float() {
    float f = *(float*)cur_;
    move(buf_words(sizeof(f)));
    return f;
}

double OxMarshalBuffer::get_double() {
    double d = *(double*)cur_;
    move(buf_words(sizeof(d)));
    return d;
}

//
// Unmarshal a string.  Need to handle the special cases here
// as in put_string.
//

char* OxMarshalBuffer::get_string() {
    char* p = (char*)cur_;
    char* s;
    ULong n;
    if (p[0] == '\0') {
	n = 2;
	if (p[1] == '\1') {
	    s = nil;
	} else {
	    s = new char[1];
	    s[0] = '\0';
	}
    } else {
	n = strlen(p) + 1;
	s = new char[n];
	strcpy(s, p);
    }
    move(n);
    return s;
}

//
// Get a sequence header.  We assume that unmarshalling a sequence
// always requires allocating the buffer on the heap, so we always
// set the release flag to true.
//

void OxMarshalBuffer::get_seq_hdr(void* v) {
    GenericSequence* seq = (GenericSequence*)v;
    seq->maximum = get_long();
    seq->length = get_long();
    seq->release = true;
}

//
// This function assumes the sequence element type is primitive,
// e.g., char, short, long.
//

void OxMarshalBuffer::get_seq(void* v, Long size) {
    GenericSequence* seq = (GenericSequence*)v;
    seq->maximum = get_long();
    seq->length = get_long();
    seq->release = true;
    Long n = seq->length * size;
    Long w = buf_words(n);
    if (cur_ + w >= end_) {
	err_underflow();
    }
    seq->buffer = new char[seq->maximum * size];
    memcpy(seq->buffer, cur_, int(n));
    move(w);
}

void* OxMarshalBuffer::get_object(OxStubCreator f) {
    return _OxExchangeId_get(*this, f);
}

//
// Offset table for moving the function table index.  Parameters
// with type _t_addr get two entries, one for put and one for get,
// while those with type _t_object get one entry for the stub creator.
// All other types do not have an entry.
//
static unsigned char func_delta[] = {
    2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1
};

//
// Perform a remote call using the object's exchange to setup the
// buffer and do the actual invocation.  The argument descriptors
// provide the necessary information to marshal in/inout parameters and
// unmarshal out/inout parameters and a return value.
//

void OxMarshalBuffer::invoke(
    CORBA::Object_ptr obj, ArgInfo& info, ArgValue* value, OxEnv* e
) {
    OxExchange* exch = obj->_exchange();
    init();
    env(e);
    exch->op(*this, obj, info.id, info.method);
    Long i;
    Long argc = info.desc[0];
    ArgDesc rdesc = info.desc[1] >> 2;
    ArgDesc* dp = &info.desc[2];
    // marshal functions are put/get pairs
    ArgMarshal* fp = &info.func[0];
    ArgValue* ap = &value[1];
    for (i = 1; i < argc; i++, dp++, ap++) {
	ArgDesc d = *dp;
	ArgDesc t = d >> 2;
	switch (d & 0x3) {
	case /* in */ 1:
	    put_in_param(t, fp, ap, false);
	    break;
	case /* inout */ 3:
	    put_inout_param(t, fp, ap, false);
	    break;
	}
	fp += func_delta[t];
    }
    ULong r = exch->invoke(*this, rdesc == _t_void_oneway);
    if (rdesc != _t_void_oneway) {
	if (r == 0) {
	    // normal return
	    dp = &info.desc[2];
	    fp = &info.func[0];
	    ap = &value[1];
	    for (i = 1; i < argc; i++, dp++, ap++) {
		ArgDesc d = *dp;
		ArgDesc t = d >> 2;
		ULong mode = d & 0x3;
		if (mode != 1) {
		    /* out or inout */
		    get_param(t, fp, ap, mode == 3);
		}
		fp += func_delta[t];
	    }
	    get_result(rdesc, fp, value);
	} else {
	    // raised exception

	    // make sure returned pointers are nil
	    value->u_vaddr = 0;

	    OxSchema* s = OxSchema::schema();
	    OxSchema::InterfaceDescriptor* d = s->get_descriptor(r, *this);
	    Long index = get_long();
	    OxEnv* ev = env();
	    if (ev == nil) {
		// use some default environment?
		return;
	    }
	    if (d == nil) {
		// should raise a system exception here
	    } else {
		ev->exception(s->get_exception(*this, d, index));
	    }
	}
    }
}

void OxMarshalBuffer::put_in_param(
    ArgDesc d, ArgMarshal* f, ArgValue* a, Boolean release
) {
    CORBA::Object_ptr obj;
    switch (d) {
    case _t_addr: (*MarshalFunc(f[0]))(this, a->u_vaddr); break;
    case _t_void: /* shouldn't happen */ break;
    case _t_void_oneway: /* shouldn't happen */ break;
    case _t_boolean: put_boolean(a->u_boolean); break;
    case _t_char: put_char(a->u_char); break;
    case _t_octet: put_octet(a->u_octet); break;
    case _t_short: put_short(a->u_short); break;
    case _t_unsigned_short: put_unsigned_short(a->u_unsigned_short); break;
    case _t_long: put_long(a->u_long); break;
    case _t_unsigned_long: put_unsigned_long(a->u_unsigned_long); break;
    case _t_longlong: put_longlong(a->u_longlong); break;
    case _t_unsigned_longlong:
	put_unsigned_longlong(a->u_unsigned_longlong);
	break;
    case _t_float: put_float(a->u_float); break;
    case _t_double: put_double(a->u_double); break;
    case _t_string:
	put_string(a->u_const_string);
	if (release) {
	    delete a->u_string;
	}
	break;
    case _t_object:
	obj = a->u_objref;
	put_object(obj);
	if (release) {
	    CORBA::release(obj); 
	}
	break;
    }
}

void OxMarshalBuffer::put_inout_param(
    ArgDesc d, ArgMarshal* f, ArgValue* a, Boolean release
) {
    CORBA::Object_ptr obj;
    char* str;
    void* p = a->u_vaddr;
    switch (d) {
    case _t_addr: (*MarshalFunc(f[0]))(this, p); break;
    case _t_void: /* shouldn't happen */ break;
    case _t_void_oneway: /* shouldn't happen */ break;
    case _t_boolean: put_boolean(*(Boolean*)p); break;
    case _t_char: put_char(*(char*)p); break;
    case _t_octet: put_octet(*(unsigned char*)p); break;
    case _t_short: put_short(*(short*)p); break;
    case _t_unsigned_short: put_unsigned_short(*(unsigned short*)p); break;
    case _t_long: put_long(*(Long*)p); break;
    case _t_unsigned_long: put_unsigned_long(*(ULong*)p); break;
    case _t_longlong: put_longlong(*(LongLong*)p); break;
    case _t_unsigned_longlong: put_unsigned_longlong(*(ULongLong*)p); break;
    case _t_float: put_float(*(float*)p); break;
    case _t_double: put_double(*(double*)p); break;
    case _t_string:
	str = *(char**)p;
	put_string(str);
	if (release) {
	    delete str;
	}
	break;

    //
    // This won't work for all multiple inheritance cases.
    // Passing inout object references correctly requires
    // that we store both the proxy address as a void* and
    // the proxy value as a CORBA::Object_ptr.  Yuck.
    //
    case _t_object:
	obj = *(CORBA::Object_ptr*)p;
	put_object(obj);
	if (release) {
	    CORBA::release(obj);
	}
	break;
    }
}

void OxMarshalBuffer::get_param(
    ArgDesc d, ArgMarshal* f, ArgValue* a, Boolean release
) {
    void* p = a->u_vaddr;
    switch (d) {
    case _t_addr: (*MarshalFunc(f[1]))(this, p); break;
    case _t_void: /* shouldn't happen */ break;
    case _t_void_oneway: /* shouldn't happen */ break;
    case _t_boolean: *(Boolean*)p = get_boolean(); break;
    case _t_char: *(char*)p = get_char(); break;
    case _t_octet: *(unsigned char*)p = get_octet(); break;
    case _t_short: *(short*)p = get_short(); break;
    case _t_unsigned_short: *(unsigned short*)p = get_unsigned_short(); break;
    case _t_long: *(Long*)p = get_long(); break;
    case _t_unsigned_long: *(ULong*)p = get_unsigned_long(); break;
    case _t_longlong: *(LongLong*)p = get_longlong(); break;
    case _t_unsigned_longlong: *(ULongLong*)p = get_unsigned_longlong(); break;
    case _t_float: *(float*)p = get_float(); break;
    case _t_double: *(double*)p = get_double(); break;
    case _t_string:
	if (release) {
	    char* str = *(char**)p;
	    delete str;
	}
	*(char**)p = get_string();
	break;
    case _t_object:
	if (release) {
	    CORBA::release(*(CORBA::Object_ptr*)p);
	}
	*(void**)p = _OxExchangeId_get(*this, OxStubCreator(*f));
	break;
    }
}

void OxMarshalBuffer::get_result(ArgDesc d, ArgMarshal* f, ArgValue* a) {
    switch (d) {
    case _t_addr: (*MarshalFunc(f[1]))(this, a->u_vaddr); break;
    case _t_void: /* function with no return value */ break;
    case _t_void_oneway: /* shouldn't happen */ break;
    case _t_boolean: a->u_boolean = get_boolean(); break;
    case _t_char: a->u_char = get_char(); break;
    case _t_octet: a->u_octet = get_octet(); break;
    case _t_short: a->u_short = get_short(); break;
    case _t_unsigned_short: a->u_unsigned_short = get_unsigned_short(); break;
    case _t_long: a->u_long = get_long(); break;
    case _t_unsigned_long: a->u_unsigned_long = get_unsigned_long(); break;
    case _t_longlong: a->u_longlong = get_longlong(); break;
    case _t_unsigned_longlong:
	a->u_unsigned_longlong = get_unsigned_longlong();
	break;
    case _t_float: a->u_float = get_float(); break;
    case _t_double: a->u_double = get_double(); break;
    case _t_string: a->u_string = get_string();break;
    case _t_object:
	a->u_vaddr = _OxExchangeId_get(*this, OxStubCreator(*f));
	break;
    }
}

//
// Release the storage associated with an "in" parameter
// after the call has completed on the server side.
//

void OxMarshalBuffer::release_param(ArgDesc d, ArgValue* a) {
    char* str;
    switch (d) {
    case _t_string:
	str = *(char**)a->u_vaddr;
	delete str;
	break;
    case _t_object:
//--- 	CORBA::release(*(CORBA::Object_ptr*)a->u_vaddr);   // $$$ hack
	break;
    }
}

//
// Before calling the implementation, unmarshal the in/inout parameters.
// If this function returns false then something bad happened
// while reading the parameters, in which case we won't want
// to call the function.
//

Boolean OxMarshalBuffer::receive(ArgInfo& info, ArgValue* value) {
    Long i;
    Long argc = info.desc[0];
    ArgDesc* dp = &info.desc[2];
    ArgMarshal* fp = &info.func[0];
    ArgValue* ap = &value[1];
    for (i = 1; i < argc; i++, dp++, ap++) {
	ArgDesc d = *dp;
	ArgDesc t = d >> 2;
	if ((d & 0x1) != 0) {
	    // in or inout parameter
	    get_param(t, fp, ap, false);
	}
	fp += func_delta[t];
    }
    return true;
}

//
// After calling the implementation, marshal the out/inout parameters and
// return value.
//

void OxMarshalBuffer::reply(ArgInfo& info, ArgValue* value) {
    init();
    if (env_ != nil) {
	CORBA::Exception* e = env_->exception();
	if (e != nil) {
	    e->_put(*this);
	    env_->clear_exception();
	    return;
	}
    }
    put_long(0);
    Long i;
    Long argc = info.desc[0];
    // desc[1] describes return value
    ArgDesc* dp = &info.desc[2];
    ArgMarshal* fp = &info.func[0];
    ArgValue* ap = &value[1];
    for (i = 1; i < argc; i++, dp++, ap++) {
	ArgDesc d = *dp;
	ArgDesc t = d >> 2;
	switch (d & 0x3) {
	case /* in */ 1:
	    release_param(t, ap);
	    break;
	case /* out */ 2:
	case /* inout */ 3:
	    put_inout_param(t, fp, ap, true);
	    break;
	}
	fp += func_delta[t];
    }
    put_in_param(info.desc[1] >> 2, fp, value, true);
}

//
// Check that a given pointer was successfully allocated
// off the heap.
//

Boolean OxMarshalBuffer::allocated(void* ptr) {
    // Should set a flag so that receive/invoke knows
    // that an unmarshal failed.
    return ptr != nil;
}

void OxMarshalBuffer::dispatch(CORBA::Object_ptr obj) {
    OxSchema* s = OxSchema::schema();
    ULong tbuf[10];
    CORBA::TypeId t = s->get_type_id(
	*this, tbuf, sizeof(tbuf) / sizeof(tbuf[0])
    );
    OxTypeObj_ReceiveFunc f = s->receiver(t);
    if (f == nil) {
	err_dispatch_receive(t);
    } else {
	ULong m = get_unsigned_long();
	if (m == OxForeignExchange_narrow_op) {
	    init();
	    put_long(0);
	    put_long(true); // !!! if we can narrow to ourself? -denis
	} else {
	    (*f)(obj, m, *this, *this);
	}
    }
    if (t != tbuf) {
	delete t;
    }
}
