//
// Copyright (c) 1992-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names
// of Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#include "dish.h"
#include <Fresco/defs/fresco.h>  //$$$  FIX
#include <stdlib.h>
#include <string.h>

static char* err_argument_number = "Wrong number of arguments.";
static char* err_argument_type = "Wrong argument type: ";

// class Dish

Dish* Dish::instance_;
Dish* Dish::instance() { return instance_; }

Dish::Dish(Tcl_Interp* interp) {
    instance_ = this;
    char* v = Tcl_GetVar(interp, "tcl_interactive", TCL_GLOBAL_ONLY);
    interactive_ = (v[0] == '1');
    verbose_ = false;
    context_ = nil;    // $$$ Won't work, we have no DII for NamingContext. 
}

Dish::~Dish() {
    CORBA::release(context_);
}

void Dish::cleanup() { delete instance_; }

void Dish::restore_args(Tcl_Interp* interp, int& argc, char**& argv) {
    int tcl_argc;
    char** tcl_argv;
    Tcl_SplitList(
        interp,
        Tcl_GetVar(interp, "argv", TCL_GLOBAL_ONLY),
        &tcl_argc, &tcl_argv
    );
    argc = tcl_argc + 1;
    argv = new char*[argc + 1];
    char* argv0 = Tcl_GetVar(interp, "argv0", TCL_GLOBAL_ONLY);
    argv[0] = argv0;
    for (int i = 0; i < tcl_argc; i++) {
	const char* arg = tcl_argv[i];
	int length = strlen(arg);
	char* copy = new char[length + 1];
	sprintf(copy, "%s", arg);
	argv[i + 1] = copy;
    }
    argv[argc] = 0;
    delete [] tcl_argv;
}

#define proc(str, data, func) Tcl_CreateCommand(interp, str, func, (ClientData) data, NULL)

void Dish::init(Tcl_Interp* interp) {
    proc("Dish_verbose", this, verbose);
    proc("Dish_scope", this, scope);
    proc("Dish_object", this, object);
    proc("Dish_invoke", this, invoke);
    proc("Dish_invoke_tmp", this, invoke_tmp);
    proc("Dish_duplicate", this, duplicate);
    proc("Dish_release", this, release);
    proc("Dish_get_current_context", this, get_current_context);
    proc("Dish_resolve", this, resolve_cmd);
    proc("Dish_isa", this, isa);
    proc("Dish_is_a", this, isa);
    proc("Dish_char_string", this, char_string);
    proc("Dish_drawing_kit", this, drawing_kit);

    //
    // Short-hand
    //
    proc("verbose", this, verbose);
    proc("object", this, object);
    proc("invoke", this, invoke);
    proc("invoke_tmp", this, invoke_tmp);
    proc("duplicate", this, duplicate);
    proc("release", this, release);
    proc("get_current_context", this, get_current_context);
    proc("resolve", this, resolve_cmd);
    proc("isa", this, isa);
    proc("is_a", this, isa);
    proc("char_string", this, char_string);
    proc("drawing_kit", this, drawing_kit);
}

//
// Default implementation of resolve just returns a nil object reference.
//

CORBA::Object_ptr Dish::resolve(const char* /* name */) { return nil; }

//
// Convert an argument string to an object reference, if possible.
// The string conversion might be type-specific (using convert_to_object),
// a string of the form _dish_<addr>, or a command defined by the
// CORBA_object command.
//

Boolean Dish::arg_to_object(
    Tcl_Interp* interp, const char* arg, CORBA::Type_ptr type,
    CORBA::Object_ptr& obj
) {
    Boolean b = true;
    if (arg == nil) {
	obj = nil;
    } else {
	if (!string_to_object(arg, obj)) {
	    Tcl_CmdInfo info;
	    if (Tcl_GetCommandInfo(interp, (char*)arg, &info) == 1 &&
		info.proc == object_invoke
	    ) {
		obj = (CORBA::Object_ptr)info.clientData;
	    } else {
		b = false;
	    }
	}
    }
    return b;
}

Boolean Dish::string_to_object(const char* s, CORBA::Object_ptr& obj) {
    Boolean b = true;
    if (s[0] == '0' && s[1] == '\0') {
	obj = nil;
    } else {
#if pformat
	b = sscanf(s, "_dish_%p", &obj) == 1;
#else
	//
	// Some old systems don't understand %p
	//
	long val;
	b = sscanf(s, "_dish_%lx", &val) == 1;
	if (b) {
	    obj = (CORBA::Object_ptr)(void*)val;
	}
#endif
    }
    return b;
}

char* Dish::object_to_string(CORBA::Object_ptr obj, char* result) {
    if (is_not_nil(obj)) {
#if pformat
	sprintf(result, "_dish_%p", obj);
#else
	long val = (long)obj;
	sprintf(result, "_dish_%lx", val);
#endif
    } else {
	sprintf(result, "0");
    }
    return result; 
}

Boolean Dish::dispatch(
    CORBA::Object_ptr obj, int argc, char* argv[], Tcl_Interp* interp
) {
    Boolean b = false;
    if (verbose_) {
	printf("%s", argv[0]);
	for (int i = 1; i < argc; i++) {
	    printf(" %s", argv[i]);
	}
	printf("\n");
    }
    if (argc > 1) {
	char* op = argv[1];
	CORBA::Request_var req = obj->_request(op);
	CORBA::Type::OpInfo op_info;
	CORBA::Request::CallStatus status = req->op_info(op_info);
	if (status == CORBA::Request::ok) {
	    if (parse_args(argc - 2, &argv[2], req, op_info, interp)) {
		status = req->invoke();
		if (status == CORBA::Request::ok) {
		    get_results(argc - 2, &argv[2], req, op_info, interp);
		    b = true;
		    if (verbose_) {
			printf("%s = '%s'\n", op, interp->result);
		    }
		} else {
		    set_call_status(req, status, obj, argc, op, interp);
		}
	    }
	} else {
	    set_call_status(req, status, obj, argc, op, interp);
	}
    } else {
	Tcl_AppendResult(interp, err_argument_number, 0);
    }
    return b;
}

void Dish::set_call_status(
    CORBA::Request_ptr req, CORBA::Request::CallStatus status,
    CORBA::Object_ptr obj, int argc, char* op, Tcl_Interp* interp
) {
    char obj_string[100];
    object_to_string(obj, obj_string);
    char* type_name = obj->_get_interface()->name();
    CORBA::Type_var except;
    switch (status) {
    case CORBA::Request::unknown_operation:
	Tcl_AppendResult(
	    interp, "Error: no operation '", type_name, "::", op,
	    "' for object ", obj_string, 0
	);
	break;
    case CORBA::Request::ambiguous_operation:
	Tcl_AppendResult(
	    interp, "Error: ambigous operation '", op, "' for object ",
	    obj_string, " of type ", type_name, 0
        ); 
	break;
    case CORBA::Request::bad_argument_count: {
	char count[10];
	sprintf(count, "%d", argc - 2);
	Tcl_AppendResult(
	    interp, "Error: bad argument count ",  count,
            " for '", type_name, "::", op, 0
	); 
	break;
    }
    case CORBA::Request::bad_argument_type:
	Tcl_AppendResult(
	    interp, "Error: bad argument type for '", type_name, "::", op, 0
	); 
	break;
    case CORBA::Request::invoke_failed:
	except = req->except_info();
	if (is_nil(except)) {
	    Tcl_AppendResult(
		interp, "Error: invocation failed for ", type_name, "::", op,
		0
	    );
	} else {
	    get_result(except, req, interp);
	}
	break;
    }
}

Boolean Dish::parse_args(
    int argc, char* argv[], CORBA::Request_ptr req,
    CORBA::Type::OpInfo& op_info, Tcl_Interp* interp
) {
    Boolean b = false;
    if (op_info.params.length() == ULong(argc)) {
	for (int i = 0; i < argc; i++) {
	    CORBA::Type::ParamInfo& p = op_info.params[i];
	    if (p.mode != CORBA::Type::param_out) {
		if (!add_arg(p.type, argv[i], req, interp)) {
		    Tcl_AppendResult(
			interp, err_argument_type,
			"\nExpecting: ", p.type->name(),
			", got `", argv[i], "'", 0
		    );
		    break;
		}
	    }
	}
	b = (i == argc);
    } else {
	Tcl_AppendResult(interp, err_argument_number, 0);
    }
    return b;
}

char* Dish::try_expand(char* arg, Tcl_Interp* interp) {
    char* expanded = Tcl_GetVar(interp, arg, 0);
    return expanded == nil ? arg : expanded;
}

#define add_primitive_arg(T,call,method) { \
    b = true; \
    T n = 0; \
    if (arg != nil) { \
	char* ptr; \
	n = (T)call; \
	b = (*ptr == '\0'); \
    } \
    if (b) { \
	req->method(n); \
    } \
}

#define IntCall strtol(arg, &ptr, 0)
#define UIntCall strtoul(arg, &ptr, 0)
#define FltCall strtod(arg, &ptr)

Boolean Dish::add_arg(
    CORBA::Type_ptr type, char* arg, CORBA::Request_ptr req, Tcl_Interp* interp
) {
    Boolean b = false;
    if (arg != nil) {
	arg = try_expand(arg, interp);
    }
    Boolean b_value;
    Char c_value;
    const char* s_value;
    switch (type->kind()) {
    case CORBA::Type::tk_boolean:
	b = true;
	b_value = false;
	if (arg != nil) {
	    if (strcmp(arg, "1") == 0) {
		b_value = true;
		b = true;
	    } else if (strcmp(arg, "0") == 0) {
		b = true;
	    } else {
		b = false;
	    }
	}
	if (b) {
	    req->put_boolean(b_value);
	}
	break;
    case CORBA::Type::tk_char:
	c_value = '\0';
	if (arg != nil) {
	    c_value = arg[0];
	}
	req->put_char(c_value);
	b = true;
	break;
    case CORBA::Type::tk_short:
	add_primitive_arg(Short,IntCall,put_short);
	break;
    case CORBA::Type::tk_ushort:
	add_primitive_arg(UShort,UIntCall,put_unsigned_short);
	break;
    case CORBA::Type::tk_long:
	add_primitive_arg(Long,IntCall,put_long);
	break;
    case CORBA::Type::tk_ulong:
	add_primitive_arg(ULong,UIntCall,put_unsigned_long);
	break;
    case CORBA::Type::tk_float:
	add_primitive_arg(Float,FltCall,put_float);
	break;
    case CORBA::Type::tk_double:
	add_primitive_arg(Double,FltCall,put_double);
	break;
    case CORBA::Type::tk_string:
	s_value = arg;
	if (s_value != nil && s_value[0] == '\0') {
	    s_value = nil;
	}
	req->put_string(s_value);
	b = true;
	break;
    case CORBA::Type::tk_objref:
	b = add_interface(type, arg, req, interp);
	break;
    case CORBA::Type::tk_array:
	b = add_array(type, arg, req, interp);
	break;
    case CORBA::Type::tk_enum:
	// todo: allow enum literals
	// should check that given value is valid
	add_primitive_arg(ULong,UIntCall,put_unsigned_long);
	break;
    case CORBA::Type::tk_struct:
	b = add_struct(type, arg, req, interp);
	break;
    case CORBA::Type::tk_sequence:
	b = add_sequence(type, arg, req, interp);
	break;
    case CORBA::Type::tk_TypeCode:
	b = add_arg(type->typedef_info(), arg, req, interp);
	break;
    default:
	fprintf(
	    stderr,
	    "internal error: unknown type %d in add_arg\n",
	    type->kind()
	);
	abort();
    }
    return b;
}

marshal_aggregate_body(add_interface, type, interp) {
    Boolean b = false;
    CORBA::Object_ptr obj;
    if (arg_to_object(interp, arg, type, obj)) {
	//
	// Should check non-nil o matches parameter type.
	//
	req->put_object(obj);
	b = true;
    }
    return b;
}

marshal_aggregate_body(add_array, array, interp) {
    Boolean b = false;
    Long element_count;
    CORBA::Type_var element_type;
    array->array_info(element_type.out(), element_count);
    if (arg != nil) {
	int argc;
	char** argv;
	if (split_list(interp, arg, argc, argv, "array")) {
	    if (argc == element_count) {
		req->begin_aggregate();
		for (int i = 0; i < argc; i++) {
		    if (!add_arg(element_type, argv[i], req, interp)) {
			char error[100];
			sprintf(
			    error, "%s: expected '%s', got '%s'",
			    "Array element type mismatch",
			    element_type->name(), argv[i]
			);
			Tcl_AppendResult(interp, error, 0);
			break;
		    }
		}
		req->end_aggregate();
		b = i == argc;
		delete [] argv;
	    } else {
		char error[100];
		sprintf(
		    error, "Array size mismatch: expected '%d', got '%d'\n.",
		    element_count, argc
                ); 
	        Tcl_AppendResult(interp, error, 0); 
	    }
	}
    } else {
	req->begin_aggregate();
	for (int i = 0; i < element_count; i++) {
	    add_arg(element_type, nil, req, interp);
	}
	req->end_aggregate();
	b = true;
    }
    return b;
}

marshal_aggregate_body(add_struct, t, interp) {
    Boolean b = false;
    long field_count = t->members();
    if (arg != nil) {
	int argc;
	char** argv;
	if (split_list(interp, arg, argc, argv, "struct")) {
	    if (argc == field_count) {
		req->begin_aggregate();
		for (int i = 0; i < argc; i++) {
		    CORBA::Type_ptr m = t->member_info(i);
		    if (!add_arg(m, argv[i], req, interp)) {
			char error[100];
			sprintf(
			    error,
			    "Field type mismatch: expected '%s', got '%s'",
			    m->name(), argv[i]
			);
			Tcl_AppendResult(interp, error, 0);
			break;
		    }
		}
		req->end_aggregate();
		b = i == argc;
		delete [] argv;
	    } else {
		char error[100];
		sprintf(
		    error,
		    "Struct field count mismatch: expected %d, got %d.",
		    field_count, argc
	        ); 
		Tcl_AppendResult(interp, error, 0);
	    }
	}
    } else {
	req->begin_aggregate();
	for (int i = 0; i < field_count; i++) {
	    add_arg(t->member_info(i), nil, req, interp);
	}
	req->end_aggregate();
	b = true;
    }
    return b;
}

marshal_aggregate_body(add_sequence, sequence, interp) {
    Boolean b = false;
    Long length;
    CORBA::Type_var element_type;
    sequence->sequence_info(element_type.out(), length);
    if (arg != nil && arg[0] != '\0') {
	int argc;
	char** argv;
	if (split_list(interp, arg, argc, argv, "sequence")) {
	    req->begin_aggregate();
	    // put maximum and length
	    req->put_long(argc);
	    req->put_long(argc);
	    for (int i = 0; i < argc; i++) {
		if (!add_arg(element_type, argv[i], req, interp)) {
		    char error[100];
		    sprintf(
			error, "%s: expected '%s', got '%s'",
			"Sequence element type mismatch",
			element_type->name(), argv[i]
		    );
		    Tcl_AppendResult(interp, error, 0);
		    break;
		}
	    }
	    req->end_aggregate();
	    b = (i == argc);
	    delete [] argv;
	}
    } else {
	req->begin_aggregate();
	req->put_long(0);
	req->put_long(0);
	req->put_long(0);
	req->end_aggregate();
	b = true;
    }
    return b;
}

CORBA::Boolean Dish::split_list(
    Tcl_Interp* interp, const char* arg, int& argc, char**& argv, char* type
) {
    // todo: this doesn't handle "t scale scale" when scale isn't set
    // cast workaround for Tcl prototype bug
    if (Tcl_SplitList(interp, (char*)arg, &argc, &argv) == TCL_OK) {
	return true;
    }
    Tcl_AppendResult(
	 interp,
	 "Expecting ", type, " to be represented as tcl list, got '",
	 arg, "'", 0
    ); 
    return false;
}
	
void Dish::get_results(
    int argc, char* argv[], CORBA::Request_ptr req, CORBA::Type::OpInfo& op, 
    Tcl_Interp* interp
) {
    for (int i = 0; i < argc; i++) {
	CORBA::Type::ParamInfo& p = op.params[i];
	if (p.mode != CORBA::Type::param_in) {
	    get_result(p.type, req, interp);
	    char* arg = argv[i];
	    if (arg != nil) {
		Tcl_SetVar(interp, arg, interp->result, TCL_LIST_ELEMENT);
		Tcl_ResetResult(interp);
	    }
	}
    }
    if (op.result->kind() != CORBA::Type::tk_void) {
	get_result(op.result, req, interp);
    }
}

void Dish::get_result(
    CORBA::Type_ptr type, CORBA::Request_ptr req, Tcl_Interp* interp
) {
    CORBA::Type::TCKind tag = type->kind();
    switch (tag) {
    case CORBA::Type::tk_objref:
	get_interface(type, req, interp);
	break;
    case CORBA::Type::tk_array:
	get_array(type, req, interp);
	break;
    case CORBA::Type::tk_enum:
	get_enum(type, req, interp);
	break;
    case CORBA::Type::tk_struct:
	get_struct(type, req, interp);
	break;
    case CORBA::Type::tk_except:
	get_except(type, req, interp);
	break;
    case CORBA::Type::tk_sequence: 
	get_sequence(type, req, interp);
	break;
    case CORBA::Type::tk_TypeCode: 
	get_result(type->typedef_info(), req, interp);
	break;
    default:
	get_concrete(tag, req, interp);
	break;
    }
}

void Dish::get_concrete(
    CORBA::Type::TCKind tag, CORBA::Request_ptr req, Tcl_Interp* interp
) {
    char result[200];
    switch(tag) {
    case CORBA::Type::tk_boolean:
	sprintf(result, "%s", req->get_boolean() ? "true" : "false");
	break;
    case CORBA::Type::tk_char:
	sprintf(result, "%c", req->get_char());
	break;
    case CORBA::Type::tk_octet:
	sprintf(result, "%d", req->get_octet());
	break;
    case CORBA::Type::tk_short:
	sprintf(result, "%d", req->get_short());
	break;
    case CORBA::Type::tk_ushort:
	sprintf(result, "%d", req->get_unsigned_short());
	break;
    case CORBA::Type::tk_long:
	sprintf(result, "%ld", req->get_long());
	break;
    case CORBA::Type::tk_ulong:
	sprintf(result, "%ld", req->get_unsigned_long());
	break;
    case CORBA::Type::tk_float:
	sprintf(result, "%g", req->get_float());
	break;
    case CORBA::Type::tk_double:
	sprintf(result, "%g", req->get_double());
	break;
    case CORBA::Type::tk_string:
	sprintf(result, "%s", req->get_string());
	break;
    default:
	fprintf(stderr, "Dish::get_concrete: unknown type");
	fflush(stderr);
	abort();
    }
    Tcl_AppendResult(interp, result, 0);
}

void Dish::get_string(CORBA::Request_ptr req, Tcl_Interp* interp) {
    char* result = req->get_string();
    Tcl_AppendResult(interp, result, 0);
    delete result;
}

//
// Should check the interface actual type is compatible with the formal.
//

unmarshal_aggregate_body(get_interface, type, interp) {
    CORBA::Object_ptr obj = CORBA::Object_ptr(req->get_object());
    char buffer[100];
    Tcl_AppendResult(interp, object_to_string(obj, buffer), 0);
}

unmarshal_aggregate_body(get_array, array, interp) {
    Tcl_DString result;
    Tcl_DStringInit(&result);
    Long element_count;
    CORBA::Type_var element_type;
    array->array_info(element_type.out(), element_count);
    for (int i = 0; i < element_count; i++) {
	get_result(element_type, req, interp);
	result.string = Tcl_DStringAppendElement(&result, interp->result);
	Tcl_ResetResult(interp);
    }
    Tcl_DStringResult(interp, &result);
    Tcl_DStringFree(&result);
}

unmarshal_aggregate_body(get_enum, type, interp) {
    // todo: use enum literals
    char result[10];
    sprintf(result, "%ld", req->get_unsigned_long());
    Tcl_AppendResult(interp, result, 0);
}

unmarshal_aggregate_body(get_struct, struct_type, interp) {
    Tcl_DString result;
    Tcl_DStringInit(&result);
    long field_count = struct_type->members();
    for (int i = 0; i < field_count; i++) {
	get_result(struct_type->member_info(i), req, interp);
	result.string = Tcl_DStringAppendElement(&result, interp->result);
	Tcl_ResetResult(interp);
    }
    Tcl_DStringResult(interp, &result);
    Tcl_DStringFree(&result);
}

unmarshal_aggregate_body(get_except, except, interp) {
    Tcl_DString result;
    Tcl_DStringInit(&result);
    result.string = Tcl_DStringAppendElement(&result, except->name());
    Long field_count = except->members();
    for (Long i = 0; i < field_count; i++) {
	get_result(except->member_info(i), req, interp);
	result.string = Tcl_DStringAppendElement(&result, interp->result);
	Tcl_ResetResult(interp);
    }
    Tcl_DStringResult(interp, &result);
    Tcl_DStringFree(&result);
}

unmarshal_aggregate_body(get_sequence, sequence, interp) {
    Tcl_DString result;
    Tcl_DStringInit(&result);
    Long size;
    CORBA::Type_var element_type;
    sequence->sequence_info(element_type.out(), size);
    size = req->get_long();
    long length = req->get_long();
    for (int i = 0; i < length; i++) {
	get_result(element_type, req, interp);
	result.string = Tcl_DStringAppendElement(&result, interp->result);
	Tcl_ResetResult(interp);
    }
    Tcl_DStringResult(interp, &result);
    Tcl_DStringFree(&result);
}

//
//  One Tcl interpreter is required per thread.
//

Tcl_Interp* Dish::clone(Tcl_Interp* interp) {
    Tcl_Interp* new_interp = Tcl_CreateInterp();
    // Dec alpha has race condition here if you use NULL rather
    // than name_buf.
    char name_buf[L_tmpnam];
    char* name = tmpnam(name_buf);
    FILE* f = fopen(name, "w");
    save_globals(interp, f);
    save_procs(interp, f);
    fclose(f);

    init(new_interp);
    Tcl_EvalFile(new_interp, name);
    return new_interp;
}

void Dish::save_globals(Tcl_Interp* interp, FILE* f) {
    Tcl_Eval(interp, "info globals");
    char* result = interp->result;
    if (result != 0) {
	int argc;
	char** argv;
	if (Tcl_SplitList(interp, interp->result, &argc, &argv) == TCL_OK) {
	    for (int i = 0; i < argc; i++) {
		save_global(interp, argv[i], f);
	    }
	    delete [] argv;
	}
    }
    Tcl_ResetResult(interp);
}

void Dish::save_global(Tcl_Interp* interp, char* var,  FILE* f) {
    char* value = Tcl_GetVar(interp, var, TCL_GLOBAL_ONLY);
    if (value != 0) {
	int argc;
	char** argv;
	if (Tcl_SplitList(interp, value, &argc, &argv) == TCL_OK) {
	    if (argc == 1) {
		fprintf(f, "set %s %s\n", var, value);
	    } else if (argc > 1) {
		fprintf(f, "set %s {%s}\n", var, value);
	    }
	    delete [] argv;
	}
    }
}

void Dish::save_procs(Tcl_Interp* interp, FILE* f) {
    Tcl_Eval(interp, "info procs");
    char* result = interp->result;
    if (result != 0) {
	int argc;
	char** argv;
	if (Tcl_SplitList(interp, interp->result, &argc, &argv) == TCL_OK) {
	    for (int i = 0; i < argc; i++) {
		save_proc(interp, argv[i], f);
	    }
	    delete [] argv;
	}
    }
    Tcl_ResetResult(interp);
}

void Dish::save_proc(Tcl_Interp* interp, char* proc, FILE* f) {
    fprintf(f, "proc %s ", proc);
    char command[100];
    sprintf(command, "info args %s", proc);
    Tcl_Eval(interp, command);
    int argc;
    char** argv;
    if (Tcl_SplitList(interp, interp->result, &argc, &argv) == TCL_OK) {
	fprintf(f, "{");
	char* var = "dish_xx_default";
	for (int i = 0; i < argc; i++) {
	    char *arg = argv[i];
	    sprintf(command, "info default %s %s %s", proc, arg, var);
	    Tcl_Eval(interp, command);
	    if (i > 0) {
		fprintf(f, " ");
	    }
	    if (interp->result[0] == '1') {
		fprintf(f, "{%s %s}", arg, Tcl_GetVar(interp, var, 0));
	    } else {
		fprintf(f, "%s", arg);
	    }
	}
	fprintf(f, "}");
	delete [] argv;
	Tcl_UnsetVar(interp, var, 0);

	sprintf(command, "info body %s", proc);
	Tcl_Eval(interp, command);
	fprintf(f, " {%s}\n", interp->result);
    }
}

//
// Dish-specific Tcl commands
//

Tcl_proc_body(Dish::verbose, data, argv) {
    int status = TCL_ERROR;
    if (argc != 2) {
	Tcl_AppendResult(interp, err_argument_number, 0);
    } else {
	Dish* dish = (Dish*)data;
	dish->verbose_ = (strcmp(argv[1], "on") == 0);
	status = TCL_OK;
    }
    return status;
}

//
// Just here for backward compatibility.
//

int Dish::scope(ClientData, Tcl_Interp*, int, char* []) {
    return TCL_OK;
}

Tcl_proc_body(Dish::object, data, argv) {
    int status = TCL_ERROR;
    if (argc != 3) {
	Tcl_AppendResult(interp, err_argument_number, 0);
    } else {
	Dish* dish = (Dish*)data;
	CORBA::Object_ptr obj;
	if (dish->string_to_object(argv[2], obj)) {
	    if (dish->verbose_) {
		printf("cmd %s for %s\n", argv[1], argv[2]);
	    }
	    Tcl_CreateCommand(interp, argv[1], object_invoke, (ClientData) obj, unobject);
	    Tcl_AppendResult(interp, argv[2], 0);
	    status = TCL_OK;
	}
    }
    return status;
}

//
// Handle the removal of a command created by "CORBA_object"
// by releasing the associated object.
//

void Dish::unobject(ClientData data) {
    CORBA::release((CORBA::Object_ptr)data);
}

//
// Invoke an operation on a CORBA object.  Note this operation cannot
// be called explicitly, but only after calling CORBA_object to define
// an invoke command.  The data associated with the command is a string
// containing the address of the target object.
//

Tcl_proc_body(Dish::object_invoke, data, argv) {
    int status = TCL_ERROR;
    Dish* dish = instance();
    CORBA::Object_ptr obj = (CORBA::Object_ptr)data;
    // check for nil obj?
    if (dish->dispatch(obj, argc, argv, interp)) {
	status = TCL_OK;
    }
    return status;
}

//
// Invoke an operation directly on a temporary that was returned
// from a previous call.  The object reference is assumed is released
// after the call.
//

Tcl_proc_body(Dish::invoke_tmp, data, argv) {
    int status = TCL_ERROR;
    Dish* dish = (Dish*)data;
    CORBA::Object_ptr obj;
    if (dish->string_to_object(argv[1], obj) && is_not_nil(obj)) {
	if (dish->dispatch(obj, argc - 1, &argv[1], interp)) {
	    status = TCL_OK;
	}
	CORBA::release(obj);
    }
    return status;
}

//
// Invoke an operation on the given object, generally referenced
// through a Tcl variable.
//

Tcl_proc_body(Dish::invoke, data, argv) {
    int status = TCL_ERROR;
    Dish* dish = (Dish*)data;
    CORBA::Object_ptr obj;
    if (dish->string_to_object(argv[1], obj) && is_not_nil(obj)) {
	if (dish->dispatch(obj, argc - 1, &argv[1], interp)) {
	    status = TCL_OK;
	} else {
	    Tcl_AppendResult(interp, "Can't dispatch ", argv[1], 0);
	}
    } else {
	Tcl_AppendResult(interp, "Not an object: ", argv[1], 0);
    }
    return status;
}

//
// Duplicate an object reference.  Only necessary when copying
// a Tcl variable that refers to a CORBA object.
//

Tcl_proc_body(Dish::duplicate, data, argv) {
    int status = TCL_ERROR;
    if (argc != 2) {
	Tcl_AppendResult(interp, err_argument_number, 0);
    } else {
	Dish* dish = (Dish*)data;
	CORBA::Object_ptr obj;
	if (dish->string_to_object(argv[1], obj)) {
	    obj = CORBA::Object::_duplicate(obj);
	    status = TCL_OK;
	} else {
	    Tcl_AppendResult(interp, "Not an object: ", argv[1], 0);
	}
    }
    return status;
}

//
// Release a CORBA object reference.  We just need to remove
// the associated Tcl command, which as a side effect will
// call Dish::unobject and release the objref.
//

Tcl_proc_body(Dish::release, data, argv) {
    int status = TCL_ERROR;
    if (argc == 1) {
	Tcl_AppendResult(interp, err_argument_number, 0);
    } else {
	status = TCL_OK;
	for (int i = 1; status == TCL_OK && i < argc; i++) {
	    status = Tcl_DeleteCommand(interp, argv[i]);
	}
    }
    return status;
}

//
// Return the current naming context.
//

Tcl_proc_body(Dish::get_current_context, data, argv) {
    int status = TCL_ERROR;
    if (argc != 1) {
	Tcl_AppendResult(interp, err_argument_number, 0);
    } else {
	char buffer[100];
	Dish* dish = (Dish*)data;
	Tcl_AppendResult(
	    interp, dish->object_to_string(dish->context(), buffer), 0
	);
	status = TCL_OK;
    }
    return status;
}

Tcl_proc_body(Dish::resolve_cmd, data, argv) {
    int status = TCL_ERROR;
    if (argc != 2) {
	Tcl_AppendResult(interp, err_argument_number, 0);
    } else {
	char buffer[100];
	Dish* dish = (Dish*)data;
	CORBA::Object_ptr obj = dish->resolve(argv[1]);
	Tcl_AppendResult(interp, dish->object_to_string(obj, buffer), 0);
	status = TCL_OK;
    }
    return status;
}

static Boolean check(OxSchema::InterfaceDescriptor* d, const char* name) {
    if (strcmp(d->hdr.name, name) == 0) {
	return true;
    }
    OxSchema::InterfaceDescriptor** p = d->parents;
    if (p != nil) {
	for (; *p != nil; p++) {
	    if (check(*p, name)) {
		return true;
	    }
	}
    }
    return false;
}

Tcl_proc_body(Dish::isa, data, argv) {
    int status = TCL_ERROR;
    if (argc != 3) {
	Tcl_AppendResult(interp, err_argument_number, 0);
    } else {
	Dish* dish = (Dish*)data;
	CORBA::Object_ptr obj;
	if (dish->string_to_object(argv[1], obj)) {
	    if (dish->verbose_) {
		printf("isa %s %s\n", argv[1], argv[2]);
	    }
	    OxSchema* s = OxSchema::schema();
	    OxSchema::InterfaceDescriptor* d = (
		s->descriptor(obj->_exchange()->tid(obj))
	    );
	    if (d == nil) {
		Tcl_AppendResult(interp, "Unknown type for ", argv[1], 0);
	    } else {
		Tcl_AppendResult(interp, check(d, argv[2]) ? "1" : "0", 0);
		status = TCL_OK;
	    }
	}
    }
    return status;
}

Tcl_proc_body(Dish::char_string, data, argv) {
    int status = TCL_ERROR;
    if (argc != 2) {
	Tcl_AppendResult(interp, err_argument_number, 0);
    } else {
	Dish* dish = (Dish*)data;
	char buffer[100];
	CORBA::Object_ptr obj;
	obj = FrescoLib::string_copy(argv[1]);
	Tcl_AppendResult(interp, dish->object_to_string(obj, buffer), 0);
	status = TCL_OK;
    }
    return status;
}

Tcl_proc_body(Dish::drawing_kit, data, argv) {
    // $$$ Bug here.
    // This can only be invoked as 'object drawing [drawing_kit]'
    // and not 'object drawing drawing_kit' or, better still
    // 'object r1 [drawing_kit bounding_box]'
    
    int status = TCL_ERROR;
    if (argc != 1) {
	Tcl_AppendResult(interp, err_argument_number, 0);
    } else {
	Dish* dish = (Dish*)data;
	char buffer[100];
	CORBA::Object_ptr obj;
	obj = FrescoLib::current_context()->default_drawing_kit();
	Tcl_AppendResult(interp, dish->object_to_string(obj, buffer), 0);
	status = TCL_OK;
    }
    return status;
}


#if defined(sun) && !defined(SVR4)
// Ensure that the linker includes the shared libraries.
#include <Fresco/figures/figures.h>
#include <Fresco/layouts/layouts.h>
#include <Fresco/widgets/widgets.h>
void dd() {
    FigureKit::_narrow(0);
    LayoutKit::_narrow(0);
    WidgetKit::_narrow(0);
}
#endif

static void clean_upper(ClientData , Tcl_Interp*) {
    Dish::cleanup();
}

extern Dish* create_dish(Tcl_Interp*);

//
// This is the best way to go, but requires Tcl 7.4 or newer.
//

#if TCL_MAJOR_VERSION >= 7 && TCL_MINOR_VERSION >= 4

int Tcl_AppInit(Tcl_Interp* interp) {
    Tcl_CallWhenDeleted(interp, clean_upper, (ClientData)dish);
    Dish* dish = create_dish(interp);
    dish->init(interp);
    return TCL_OK;
}

int main(int argc, char** argv) {
    Tcl_Main(argc, argv);
    return 0;			/* Needed only to prevent compiler warning. */
}

#else /* Tcl 7.3 or older */

//
// This is for cfront-based compilers that need to call _main to
// get static initializers to run.
//

#if defined(sun) && !defined(SVR4) && !defined(__GNUC__)
extern "C" void _main();

void init_static_initializers() {
    _main();
}
#else
void init_static_initializers() { }
#endif

//
//  Called by main() defined in tcl lib.
//

int Tcl_AppInit(Tcl_Interp* interp) {
    init_static_initializers();
    atexit(&Dish::cleanup);
    Dish* dish = create_dish(interp);
    dish->init(interp);
    return TCL_OK;
}

#endif /* Tcl 7.3 or older */
