//
// Copyright (c) 1992-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names
// of Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#include "dish.h"
#include <Fresco/defs/fresco.h>
#include <Fresco/defs_impl/action.h>
#include <Fresco/defs_impl/charstr.h>
#include <Fresco/defs_impl/glyphs.h>
#include <Fresco/defs_impl/names.h>
#include <Fresco/defs_impl/viewers.h>
#include <Fresco/defs_impl/threads.h>
#include <stdlib.h>
#include <string.h>

#if defined(sun) && !defined(SVR4)
#define pformat 0
#else
#define pformat 1
#endif

static char* err_argument_number = "Wrong number of arguments.";
static char* err_argument_type = "Wrong argument type: ";

class FrescoDish : public Dish {
public:
    FrescoDish(Tcl_Interp*);
    ~FrescoDish();
    
    virtual void init(Tcl_Interp*);
    virtual CORBA::Object_ptr resolve(const char* name);
private:
    Tcl_proc(action);
    Tcl_proc(debug);
    Tcl_proc(delay);
    Tcl_proc(run);
    Tcl_proc(spawn);

    Naming::NamingContext_ptr ncontext_;
};

Dish* create_dish(Tcl_Interp* interp) { return new FrescoDish(interp); }

class FrescoMain : public ActionImpl {
public:
    FrescoMain(Viewer_ptr, Glyph_ptr);
    ~FrescoMain();

    void execute();
private:
    Viewer_var viewer_;
    Glyph_var glyph_;
};

class DishAction : public ActionImpl {
public:
    DishAction(Tcl_Interp*, char* command);
    ~DishAction();

    void spawn(Tcl_Interp*);
    void execute();
private:
    Tcl_Interp* interp_;
    Tcl_Interp* spawned_;
    char* command_;
};

FrescoDish::FrescoDish(Tcl_Interp* interp) : Dish(interp) {
    int argc;
    char** argv;
    restore_args(interp, argc, argv);
    ncontext_ = FrescoLib::init_default_context("Dish", argc, argv);
    context_ = ncontext_;
}

FrescoDish::~FrescoDish() { }

#define proc(str, data, func) Tcl_CreateCommand(interp, str, func, (ClientData) data, NULL)

void FrescoDish::init(Tcl_Interp* interp) {
    Dish::init(interp);
    proc("run", this, run);
    proc("spawn", this, spawn);
    proc("action", this, action);
    proc("delay", this, delay);
    proc("debug", this, debug);
}

CORBA::Object_ptr FrescoDish::resolve(const char* name) {
    return ncontext_->resolve(*NamingLib::to_name(name));
}

Tcl_proc_body(FrescoDish::run, data, argv) {
    int status = TCL_ERROR;
    Dish* dish = (Dish*)data;
    if (argc != 3) {
        Tcl_AppendResult(interp, err_argument_number, 0);
    } else {
	CORBA::Object_ptr obj;
	char* arg1 = dish->try_expand(argv[1], interp);
	if (!dish->arg_to_object(interp, arg1, nil, obj)) {
	    Tcl_AppendResult(interp, err_argument_type, arg1, 0); 
	} else {
	    Viewer_ptr viewer = Viewer::_narrow(obj);
	    char* arg2 = dish->try_expand(argv[2], interp);
	    if (!dish->arg_to_object(interp, arg2, nil, obj)) {
		Tcl_AppendResult(interp, err_argument_type, arg2, 0); 
	    } else {
		Glyph_ptr glyph = Glyph::_narrow(obj);
		if (is_nil(viewer) && is_nil(glyph)) {
		    Tcl_AppendResult(interp, "nil viewer and glyph", 0); 
		} else {
		    Thread_ptr t = nil;
		    if (dish->interactive()) {
			t = ThreadLib::thread_kit_ptr()->thread_ptr(
			    new FrescoMain(viewer, glyph)
			);
		    }
		    if (is_not_nil(t)) {
			t->run();
		    } else {
			FrescoLib::run(viewer, glyph);
		    }
		    status = TCL_OK;
		}
	    }
	}
    }
    if (status == TCL_ERROR) {
	Tcl_AppendResult(interp, ".\nExpected `run viewer glyph'.", 0); 
    }
    return status;
}

Tcl_proc_body(FrescoDish::spawn, data, argv) {
    int status = TCL_ERROR;
    if (argc == 2) {
	Dish* dish = (Dish*)data;
	CORBA::Object_ptr obj;
	if (dish->string_to_object(argv[1], obj)) {
	    Action_ptr a = Action::_narrow(obj);
	    if (is_not_nil(a)) {
		// should use narrow
		DishAction* da = (DishAction*)a;
		if (is_not_nil(da)) {
		    da->spawn(dish->clone(interp));
		}
		Thread_ptr t = ThreadLib::thread_kit_ptr()->thread_ptr(a);
		if (is_not_nil(t)) {
		    t->run();
#if pformat
		    sprintf(interp->result, "%p", Thread_ptr(t));
#else
		    sprintf(interp->result, "%lx", Thread_ptr(t));
#endif
		    status = TCL_OK;
		} else {
		    Tcl_AppendResult(interp, "No threads available", 0);
		}
	    } else {
		Tcl_AppendResult(interp, err_argument_type, ": ", argv[1], 0); 
	    }
	} else {
	    Tcl_AppendResult(interp, err_argument_type, ": ", argv[1], 0); 
	}
    } else {
        Tcl_AppendResult(interp, err_argument_number, 0);
    }
    return status;
}

Tcl_proc_body(FrescoDish::action, data, argv) {
    int status = TCL_ERROR;
    if (argc > 1) {
	Action_ptr a = new DishAction(interp, Tcl_Merge(argc - 1, &argv[1]));
	char buffer[100];
	Dish* dish = (Dish*)data;
	Tcl_AppendResult(interp, dish->object_to_string(a, buffer), 0);
	status = TCL_OK;
    } else {
	Tcl_AppendResult(
            interp, err_argument_number, ".\nExpected: 'action command'.", 0
        );
    }
    return status;
}

Tcl_proc_body(FrescoDish::delay, data, argv) {
    int status = TCL_ERROR;
    if (argc == 2) {
	Dish* dish = (Dish*)data;
	double n;
	if (
	    Tcl_GetDouble(
		interp, dish->try_expand(argv[1], interp), &n
	    ) == TCL_OK
	) {
	    if (FrescoLib::delay(Float(n))) {
		status = TCL_OK;
	    } else {
		Tcl_AppendResult(interp, "Failed in delay", 0);
	    }
	}
    } else {
	Tcl_AppendResult(interp, err_argument_number, 0);
    }
    if ((status & TCL_ERROR) != 0) {
	Tcl_AppendResult(interp, ".\nExpected: 'delay float-seconds'.", 0);
    }
    return status;
}

Tcl_proc_body(FrescoDish::debug, data, argv) {
    int status = TCL_ERROR;
    CORBA::Object_ptr object;
    Glyph_ptr g;
    Dish* dish = (Dish*)data;

    if (argc != 3) {
	Tcl_AppendResult(interp, err_argument_number, 0);
    } else if (
	!dish->string_to_object(dish->try_expand(argv[1], interp), object)
    ) {
	Tcl_AppendResult(
	    interp, "Argument 1 ", argv[1], " is not a dish object.", 0
	);
    } else if (is_nil(g = Glyph::_narrow(object))) {
	Tcl_AppendResult(
	    interp, "Argument 1 ", argv[1], " is not a glyph.", 0
	);
    } else {
	char* l = new char[strlen(argv[2]) + 1];
	sprintf(l, "%s", argv[2]);
	char buffer[100];
	Tcl_AppendResult(
	    interp, dish->object_to_string(
		new DebugGlyph(g, l, DebugGlyph::trace_request_traverse),
		buffer
	    ), 0
	);
	status = TCL_OK;
    }

    if (status == TCL_ERROR) {
	Tcl_AppendResult(interp, "  Expected: debug <glyph> label.", 0);
    }
    return status;
}

// class FrescoMain

FrescoMain::FrescoMain(Viewer_ptr viewer, Glyph_ptr glyph) {
    viewer_ = Viewer::_duplicate(viewer);
    glyph_ = Glyph::_duplicate(glyph);
}

FrescoMain::~FrescoMain() { }

void FrescoMain::execute() {
    FrescoLib::run(viewer_, glyph_);
}

// class DishAction

DishAction::DishAction(Tcl_Interp* interp, char* command) {
    interp_ = interp;
    spawned_ = nil;
    command_ = command;
}

DishAction::~DishAction() {
    Tcl_DeleteInterp(spawned_);
    delete command_;
}

void DishAction::spawn(Tcl_Interp* interp) {
    interp_ = interp;
    spawned_ = interp;
}

void DishAction::execute() {
    if (Tcl_Eval(interp_, command_) != TCL_OK) {
	fprintf(
	    stderr, "DishAction '%s' failed: %s\n", command_, interp_->result
	);
	fflush(stdout);
	abort();
    }
}
