//
// Copyright (c) 1994 Silicon Graphics, Inc.
// Copyright (c) 1994 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
// OF THIS SOFTWARE.
//

/*
 * commands.cxx
 */
#include "commands.h"
#include "figviewer.h"
#include "manipulators.h"
#include "selection.h"
#include "visitors.h"
#include <Fresco/defs/fresco.h>
#include <Fresco/Ox/schema.h>
#include <stdlib.h>

struct KeyElem {
    Glyph* key;
    Glyph* object;
};

declareList(KeyList, KeyElem);
implementList(KeyList, KeyElem);

GlyphImplMap::GlyphImplMap () {
    keylist_ = new KeyList;
}

GlyphImplMap::~GlyphImplMap () {
    delete keylist_;
}

void GlyphImplMap::clear () {
    keylist_->remove_all();
}

void GlyphImplMap::map (Glyph* key, Glyph* obj) {
    KeyElem keyelem;
    keyelem.key = key;
    keyelem.object = obj;
    keylist_->append(keyelem);
}

Glyph* GlyphImplMap::find (Glyph* key) {
    for (long i = 0; i < keylist_->count(); i++) {
        KeyElem& keyelem = keylist_->item_ref(i);
        if (keyelem.key == key) {
            return keyelem.object;
        }
    }
    return nil;
}

//+ CmdData(CmdData)
void* CmdData::_this_ptr() { return this; }
_Ix_extern__tid(CmdData);
void* CmdData::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(CmdData) == t) return this;
    void* p;
    if (p = XfFrescoObject::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(CmdData);
_Ix_ImplementGetTid(CmdData,_Ix__tid(CmdData))
_Ix_ImplementNarrow(CmdData*,CmdData,_Ix_NullParameter,CmdData,0)
_Ix_ImplementDuplicate(CmdData*,CmdData,_Ix_NullParameter,_Ix__tid(CmdData),0)
//+

CmdData::CmdData () {}

CmdData::~CmdData () {}

implementTable(CmdDataTable, long, CmdData*);

implementPtrList(CmdList, FCommand);

//+ FCommand(FCommand)
void* FCommand::_this_ptr() { return this; }
_Ix_extern__tid(FCommand);
void* FCommand::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(FCommand) == t) return this;
    void* p;
    if (p = XfAction::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(FCommand);
_Ix_ImplementGetTid(FCommand,_Ix__tid(FCommand))
_Ix_ImplementNarrow(FCommand*,FCommand,_Ix_NullParameter,FCommand,0)
_Ix_ImplementDuplicate(FCommand*,FCommand,_Ix_NullParameter,_Ix__tid(FCommand),0)
//+

FCommand::FCommand (FigViewer* figviewer) {
    figviewer_ = figviewer;
    data_ = new CmdDataTable(32);
    selection_ = nil;
}

void FCommand::store (long key, CmdData* data) {
    data_->insert(key, data);
}

CmdData* FCommand::recall (long key) {
    CmdData* d = nil;
    data_->find(d, key);
    return d;
}

void FCommand::execute () {
    Selection* sel = figviewer_->selection()->copy();
    for (long i = 0; i < sel->count(); i++) {
        Manipulator* m = sel->item(i);
        m->execute(this);
    }
    delete sel;
}

FCommand::~FCommand () {
    delete selection_;
    TableIterator(CmdDataTable)* ti = new TableIterator(CmdDataTable)(*data_);
    for (; ti->more(); ti->next()) {
        CmdData*& data = ti->cur_value();
        delete data;
    }
    delete data_;
    delete ti;
}

void FCommand::log () {
    /* Logging is not supported.  All FCommands are irreversible */
}

//+ GroupCmd(GroupCmd)
void* GroupCmd::_this_ptr() { return this; }
_Ix_extern__tid(GroupCmd);
void* GroupCmd::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(GroupCmd) == t) return this;
    void* p;
    if (p = FCommand::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(GroupCmd);
_Ix_ImplementGetTid(GroupCmd,_Ix__tid(GroupCmd))
_Ix_ImplementNarrow(GroupCmd*,GroupCmd,_Ix_NullParameter,GroupCmd,0)
_Ix_ImplementDuplicate(GroupCmd*,GroupCmd,_Ix_NullParameter,_Ix__tid(GroupCmd),0)
//+

GroupCmd::GroupCmd (FigViewer* fv, Manipulator* m) : FCommand (fv) {
    m_ = m;
}

GroupCmd::~GroupCmd () {
    delete m_;
}

void GroupCmd::execute () {
    if (is_nil(m_)) {
        figviewer_->root()->execute(this);
    } else {
        m_->execute(this);
    }
}

//+ UngroupCmd(UngroupCmd)
void* UngroupCmd::_this_ptr() { return this; }
_Ix_extern__tid(UngroupCmd);
void* UngroupCmd::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(UngroupCmd) == t) return this;
    void* p;
    if (p = FCommand::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(UngroupCmd);
_Ix_ImplementGetTid(UngroupCmd,_Ix__tid(UngroupCmd))
_Ix_ImplementNarrow(UngroupCmd*,UngroupCmd,_Ix_NullParameter,UngroupCmd,0)
_Ix_ImplementDuplicate(UngroupCmd*,UngroupCmd,_Ix_NullParameter,_Ix__tid(UngroupCmd),0)
//+

UngroupCmd::UngroupCmd (FigViewer* fv) : FCommand (fv) {}

UngroupCmd::~UngroupCmd () {}

//+ DeleteCmd(DeleteCmd)
void* DeleteCmd::_this_ptr() { return this; }
_Ix_extern__tid(DeleteCmd);
void* DeleteCmd::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(DeleteCmd) == t) return this;
    void* p;
    if (p = FCommand::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(DeleteCmd);
_Ix_ImplementGetTid(DeleteCmd,_Ix__tid(DeleteCmd))
_Ix_ImplementNarrow(DeleteCmd*,DeleteCmd,_Ix_NullParameter,DeleteCmd,0)
_Ix_ImplementDuplicate(DeleteCmd*,DeleteCmd,_Ix_NullParameter,_Ix__tid(DeleteCmd),0)
//+

DeleteCmd::DeleteCmd (FigViewer* fv) : FCommand (fv) {}

DeleteCmd::~DeleteCmd () {}

//+ InstanceCmd(InstanceCmd)
void* InstanceCmd::_this_ptr() { return this; }
_Ix_extern__tid(InstanceCmd);
void* InstanceCmd::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(InstanceCmd) == t) return this;
    void* p;
    if (p = FCommand::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(InstanceCmd);
_Ix_ImplementGetTid(InstanceCmd,_Ix__tid(InstanceCmd))
_Ix_ImplementNarrow(InstanceCmd*,InstanceCmd,_Ix_NullParameter,InstanceCmd,0)
_Ix_ImplementDuplicate(InstanceCmd*,InstanceCmd,_Ix_NullParameter,_Ix__tid(InstanceCmd),0)
//+

InstanceCmd::InstanceCmd (FigViewer* fv) : FCommand (fv) {}

InstanceCmd::~InstanceCmd () {}

//+ CopyCmd(CopyCmd)
void* CopyCmd::_this_ptr() { return this; }
_Ix_extern__tid(CopyCmd);
void* CopyCmd::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(CopyCmd) == t) return this;
    void* p;
    if (p = FCommand::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(CopyCmd);
_Ix_ImplementGetTid(CopyCmd,_Ix__tid(CopyCmd))
_Ix_ImplementNarrow(CopyCmd*,CopyCmd,_Ix_NullParameter,CopyCmd,0)
_Ix_ImplementDuplicate(CopyCmd*,CopyCmd,_Ix_NullParameter,_Ix__tid(CopyCmd),0)
//+

CopyCmd::CopyCmd (FigViewer* fv) : FCommand (fv) {}

CopyCmd::~CopyCmd () {}

GlyphImplMap CopyCmd::glyphmap_;

//+ NarrowCmd(NarrowCmd)
void* NarrowCmd::_this_ptr() { return this; }
_Ix_extern__tid(NarrowCmd);
void* NarrowCmd::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(NarrowCmd) == t) return this;
    void* p;
    if (p = FCommand::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(NarrowCmd);
_Ix_ImplementGetTid(NarrowCmd,_Ix__tid(NarrowCmd))
_Ix_ImplementNarrow(NarrowCmd*,NarrowCmd,_Ix_NullParameter,NarrowCmd,0)
_Ix_ImplementDuplicate(NarrowCmd*,NarrowCmd,_Ix_NullParameter,_Ix__tid(NarrowCmd),0)
//+

NarrowCmd::NarrowCmd (FigViewer* fv) : FCommand(fv) {
    delta_ = 0;
}

NarrowCmd::NarrowCmd(FigViewer* fv, Long delta) : FCommand(fv) {
    delta_ = delta;
}

NarrowCmd::~NarrowCmd () {}

Long NarrowCmd::delta() { return delta_; }

//+ WidenCmd(WidenCmd)
void* WidenCmd::_this_ptr() { return this; }
_Ix_extern__tid(WidenCmd);
void* WidenCmd::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(WidenCmd) == t) return this;
    void* p;
    if (p = FCommand::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(WidenCmd);
_Ix_ImplementGetTid(WidenCmd,_Ix__tid(WidenCmd))
_Ix_ImplementNarrow(WidenCmd*,WidenCmd,_Ix_NullParameter,WidenCmd,0)
_Ix_ImplementDuplicate(WidenCmd*,WidenCmd,_Ix_NullParameter,_Ix__tid(WidenCmd),0)
//+

WidenCmd::WidenCmd(FigViewer* fv) : FCommand(fv) {
    delta_ = 0;
}

WidenCmd::~WidenCmd() { }

Long WidenCmd::delta() { return delta_; }

void WidenCmd::execute() {
    WidenGlyphList* glist = figviewer_->widen_list();
    Selection* sel = figviewer_->selection();
    SelectCmd unselectcmd(figviewer_, false);
    for (Long i = 0; i < sel->count(); i++) {
	sel->item(i)->execute(&unselectcmd);
    }
    if (glist != nil) {
	Long count = 0, level = glist->count() - 1;
	Manipulator* m;
	for (;;) {
	    count++;
	    Glyph_ptr g = glist->item(--level);
	    if (is_not_nil(Viewer::_narrow(g))) {
		CORBA::release(g);
		break;
	    }
	    if (is_not_nil(m = PolyManip::_narrow(g)) ||
		is_not_nil(m = BoxManip::_narrow(g))
	    ) {
		delta_ = -count;
		m->execute(this);
		CORBA::release(m);
		break;
	    }
        }
    }
}

void WidenCmd::unexecute () {
    if (delta_) {
	figviewer_->narrow(-delta_);
    }
}

//+ NaturalCmd(NaturalCmd)
void* NaturalCmd::_this_ptr() { return this; }
_Ix_extern__tid(NaturalCmd);
void* NaturalCmd::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(NaturalCmd) == t) return this;
    void* p;
    if (p = FCommand::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(NaturalCmd);
_Ix_ImplementGetTid(NaturalCmd,_Ix__tid(NaturalCmd))
_Ix_ImplementNarrow(NaturalCmd*,NaturalCmd,_Ix_NullParameter,NaturalCmd,0)
_Ix_ImplementDuplicate(NaturalCmd*,NaturalCmd,_Ix_NullParameter,_Ix__tid(NaturalCmd),0)
//+

NaturalCmd::NaturalCmd (FigViewer* fv) : FCommand (fv) {}

NaturalCmd::~NaturalCmd () {}

//+ SelectCmd(SelectCmd)
void* SelectCmd::_this_ptr() { return this; }
_Ix_extern__tid(SelectCmd);
void* SelectCmd::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(SelectCmd) == t) return this;
    void* p;
    if (p = FCommand::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(SelectCmd);
_Ix_ImplementGetTid(SelectCmd,_Ix__tid(SelectCmd))
_Ix_ImplementNarrow(SelectCmd*,SelectCmd,_Ix_NullParameter,SelectCmd,0)
_Ix_ImplementDuplicate(SelectCmd*,SelectCmd,_Ix_NullParameter,_Ix__tid(SelectCmd),0)
//+

SelectCmd::SelectCmd (FigViewer* fv, Boolean s) : FCommand (fv) {
    selected_ = s;
}

SelectCmd::~SelectCmd () {}

//+ MacroCmd(MacroCmd)
void* MacroCmd::_this_ptr() { return this; }
_Ix_extern__tid(MacroCmd);
void* MacroCmd::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(MacroCmd) == t) return this;
    void* p;
    if (p = FCommand::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(MacroCmd);
_Ix_ImplementGetTid(MacroCmd,_Ix__tid(MacroCmd))
_Ix_ImplementNarrow(MacroCmd*,MacroCmd,_Ix_NullParameter,MacroCmd,0)
_Ix_ImplementDuplicate(MacroCmd*,MacroCmd,_Ix_NullParameter,_Ix__tid(MacroCmd),0)
//+

MacroCmd::MacroCmd (FigViewer* fv) : FCommand(fv) {
    cmdlist_ = new CmdList;
}

MacroCmd::~MacroCmd () {
    for (long i = 0; i < cmdlist_->count(); i++) {
        FCommand* cmd = cmdlist_->item(i);
        delete cmd;
    }
    delete cmdlist_;
}

Boolean MacroCmd::reversible () {
    for (long i = 0; i < cmdlist_->count(); i++) {
        if (cmdlist_->item(i)->reversible()) {
            return true;
        }
    }
    return false;
}
 
void MacroCmd::execute () {       
    for (long i = 0; i < cmdlist_->count(); i++) {
        cmdlist_->item(i)->execute();
    }
}

void MacroCmd::unexecute () {       
    for (long i = cmdlist_->count()-1; i >= 0; i--) {
        cmdlist_->item(i)->execute();
    }
}

void MacroCmd::prepend (FCommand* cmd) {
    cmdlist_->prepend(cmd);
}

void MacroCmd::append (FCommand* cmd) {
    cmdlist_->append(cmd);
}

void MacroCmd::insert (long index, FCommand* cmd) {
    cmdlist_->insert(index, cmd);
}

void MacroCmd::remove(long index) {
    cmdlist_->remove(index);
}

void MacroCmd::remove_all() {
    cmdlist_->remove_all();
}

long MacroCmd::count () {
    return cmdlist_->count();
}

FCommand* MacroCmd::command (long index) {
    return cmdlist_->item(index);
}

//+ SelectInfoCmd(SelectInfoCmd)
void* SelectInfoCmd::_this_ptr() { return this; }
_Ix_extern__tid(SelectInfoCmd);
void* SelectInfoCmd::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(SelectInfoCmd) == t) return this;
    void* p;
    if (p = FCommand::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(SelectInfoCmd);
_Ix_ImplementGetTid(SelectInfoCmd,_Ix__tid(SelectInfoCmd))
_Ix_ImplementNarrow(SelectInfoCmd*,SelectInfoCmd,_Ix_NullParameter,SelectInfoCmd,0)
_Ix_ImplementDuplicate(SelectInfoCmd*,SelectInfoCmd,_Ix_NullParameter,_Ix__tid(SelectInfoCmd),0)
//+

SelectInfoCmd::SelectInfoCmd (FigViewer* fv) : FCommand (fv) {
    sinfo_ = nil;
}

SelectInfoCmd::~SelectInfoCmd () {}

//+ NewCmd(NewCmd)
void* NewCmd::_this_ptr() { return this; }
_Ix_extern__tid(NewCmd);
void* NewCmd::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(NewCmd) == t) return this;
    void* p;
    if (p = FCommand::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(NewCmd);
_Ix_ImplementGetTid(NewCmd,_Ix__tid(NewCmd))
_Ix_ImplementNarrow(NewCmd*,NewCmd,_Ix_NullParameter,NewCmd,0)
_Ix_ImplementDuplicate(NewCmd*,NewCmd,_Ix_NullParameter,_Ix__tid(NewCmd),0)
//+

NewCmd::NewCmd (FigViewer* fv, Manipulator* root) : FCommand (fv) {
    root_ = Manipulator::_duplicate(root);
}

NewCmd::~NewCmd () {
    CORBA::release(root_);
}

void NewCmd::execute () {
    Manipulator* r = root_->deep_copy();
    figviewer_->root(r);
    CORBA::release(r);
}

//+ QuitCmd(QuitCmd)
void* QuitCmd::_this_ptr() { return this; }
_Ix_extern__tid(QuitCmd);
void* QuitCmd::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(QuitCmd) == t) return this;
    void* p;
    if (p = FCommand::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(QuitCmd);
_Ix_ImplementGetTid(QuitCmd,_Ix__tid(QuitCmd))
_Ix_ImplementNarrow(QuitCmd*,QuitCmd,_Ix_NullParameter,QuitCmd,0)
_Ix_ImplementDuplicate(QuitCmd*,QuitCmd,_Ix_NullParameter,_Ix__tid(QuitCmd),0)
//+

QuitCmd::QuitCmd (FigViewer* fv) : FCommand (fv) {}

QuitCmd::~QuitCmd () {}

void QuitCmd::execute () {
    exit(0);
}

//+ SelectAllCmd(SelectAllCmd)
void* SelectAllCmd::_this_ptr() { return this; }
_Ix_extern__tid(SelectAllCmd);
void* SelectAllCmd::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(SelectAllCmd) == t) return this;
    void* p;
    if (p = FCommand::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(SelectAllCmd);
_Ix_ImplementGetTid(SelectAllCmd,_Ix__tid(SelectAllCmd))
_Ix_ImplementNarrow(SelectAllCmd*,SelectAllCmd,_Ix_NullParameter,SelectAllCmd,0)
_Ix_ImplementDuplicate(SelectAllCmd*,SelectAllCmd,_Ix_NullParameter,_Ix__tid(SelectAllCmd),0)
//+

SelectAllCmd::SelectAllCmd (FigViewer* fv) : FCommand (fv) {}

SelectAllCmd::~SelectAllCmd () {}

void SelectAllCmd::execute () {
    SelectCmd unselect(figviewer_, false);
    Selection* sel = figviewer_->selection()->copy();
    for (long i = 0; i < sel->count(); i++) {
        sel->item(i)->execute(&unselect);
    }
    delete sel;
    SelectCmd select_all(figviewer_);
    CmdVisitor cv(&select_all);
    Manipulator* root = figviewer_->root();
    Glyph* b = root->body();
    cv.visit_children(b);
    CORBA::release(root);
    CORBA::release(b);
}

//+ AboutCmd(AboutCmd)
void* AboutCmd::_this_ptr() { return this; }
_Ix_extern__tid(AboutCmd);
void* AboutCmd::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(AboutCmd) == t) return this;
    void* p;
    if (p = FCommand::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(AboutCmd);
_Ix_ImplementGetTid(AboutCmd,_Ix__tid(AboutCmd))
_Ix_ImplementNarrow(AboutCmd*,AboutCmd,_Ix_NullParameter,AboutCmd,0)
_Ix_ImplementDuplicate(AboutCmd*,AboutCmd,_Ix_NullParameter,_Ix__tid(AboutCmd),0)
//+

AboutCmd::AboutCmd (
    FigViewer* fv, DialogKit_ptr d, FigureKit_ptr f
) : FCommand (fv) {
    dialogs_ = DialogKit::_duplicate(d);
    figures_ = FigureKit::_duplicate(f);
}

AboutCmd::~AboutCmd () {
    CORBA::release(dialogs_);
    CORBA::release(figures_);
}

static Glyph_ptr label (const char* str, FigureKit_ptr figures) {
    FigureStyle_var fstyle = figures->default_style();
    return figures->label(fstyle, FrescoLib::tmp_string_ref(str));
}

void AboutCmd::execute () {
    MessageDialog_var dialog = dialogs_->information(
        FrescoLib::tmp_string_ref("FDraw Version 0.8"),
        Glyph_var(
            label(
                "By Steven Tang, Steven Churchill, Denis Bohm, and Mark Linton.",
                figures_
            )
        )
    );
    dialog->hide_choice(MessageDialog::help);

    Glyph::AllocationInfoSeq a;
    figviewer_->allocations(a);
    dialog->post_modal(Window_var(a[0].damaged->window_ptr()));
}
