//
// Copyright (c) 1994 Silicon Graphics, Inc.
// Copyright (c) 1994 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
// OF THIS SOFTWARE.
//

/*
 * fdraw.c
 */

#include "commands.h"
#include "fdraw.h"
#include "figviewer.h"
#include "manipulators.h"
#include "selection.h"
#include "tools.h"
#include "visitors.h"
#include <Fresco/defs/fresco.h>
#include <Fresco/defs_impl/action.h>
#include <Fresco/defs_impl/region.h>
#include <Fresco/defs_impl/transform.h>
#include <Fresco/layouts/layouts.h>
#include <stdio.h>

class FDrawImpl {
public:
    FigViewer* viewer_;
    Glyph_ptr deck_;
    Glyph_ptr card_[2];
    Boolean editing_;

    Glyph::Requisition req_;
    FigureKit* figures_;
    WidgetKit* widgets_;
    LayoutKit* layouts_;
    DialogKit* dialogs_;
    DrawingKit* drawings_;
    MenuKit* menus_;
};

class TransformSetter : public MonoGlyph {
public:
    TransformSetter(Glyph_ptr g, Transform_ptr t);
    ~TransformSetter();
    
    XfTransform_ptr transformation(); //+ XfGlyph::transformation
    void traverse(XfGlyphTraversal_ptr t); //+ XfGlyph::traverse
    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
    void child_allocate(Glyph::AllocationInfo&);
protected:
    Transform_ptr t_;
    RegionImpl natural_;
};

TransformSetter::TransformSetter(Glyph_ptr g, Transform_ptr t) {
    body(g);
    t_ = Transform::_duplicate(t);
}

TransformSetter::~TransformSetter() { 
    CORBA::release(t_);
}

//+ TransformSetter(XfGlyph::transformation)
XfTransform_ptr TransformSetter::transformation() {
    return Transform::_duplicate(t_);
}

//+ TransformSetter(XfGlyph::request)
void TransformSetter::request(XfGlyph::Requisition& r) {
    MonoGlyph::request(r);

    RegionImpl region;
    region.lower_.x = -r.x.natural*r.x.align;
    region.upper_.x = region.lower_.x + r.x.natural;
    region.xalign_ = r.x.align;
 
    region.lower_.y = -r.y.natural*r.y.align;
    region.upper_.y = region.lower_.y + r.y.natural;
    region.yalign_ = r.y.align;
    region.defined_ = true;
 
    natural_.copy(&region);
    region.apply_transform(t_);
 
    Coord x_lead = -region.lower_.x, x_trail = region.upper_.x;
    Coord y_lead = -region.lower_.y, y_trail = region.upper_.y;
    GlyphImpl::require_lead_trail(
        r.x, x_lead, x_lead, x_lead, x_trail, x_trail, x_trail
    );
    GlyphImpl::require_lead_trail(
        r.y, y_lead, y_lead, y_lead, y_trail, y_trail, y_trail
    );
}

//+ TransformSetter(XfGlyph::traverse)
void TransformSetter::traverse(XfGlyphTraversal_ptr t) {
    Painter_var p = t->current_painter();
    p->push_matrix();
    Glyph::AllocationInfo a;
    a.transformation = nil;
    a.allocation = t->allocation();
    a.damaged = nil;
    child_allocate(a);
    p->premultiply(a.transformation);
    MonoGlyph::traverse(t);
    p->pop_matrix();
}

void TransformSetter::child_allocate(Glyph::AllocationInfo& a) {
    if (is_nil(a.transformation)) {
        a.transformation = new TransformImpl;
    }
    TransformImpl tx;
    tx.copy(t_);
    if (is_not_nil(a.allocation)) {
        Vertex lower, upper;
        a.allocation->bounds(lower, upper);
        tx.translate(lower);
        a.allocation->copy(&natural_);
    }
    a.transformation->premultiply(&tx);
}

class Requestor : public MonoGlyph {
    //. A Requestor always returns a constant requisition. It is
    //. useful as a wrapper for glyphs whose requisition changes
    //. but who wish to be laid out as though their requisition
    //. were constant.
public:
    Requestor(
	Glyph_ptr g, Alignment xalign = .5, Alignment yalign = .5,
	Coord xspan = 1, Coord yspan = 1
    );
    Requestor(Glyph_ptr g, const Glyph::Requisition& r);
    ~Requestor();
    void request(XfGlyph::Requisition& r); //+ XfGlyph::request

protected:
    Glyph::Requisition req_;
};

Requestor::Requestor(Glyph_ptr g, const Glyph::Requisition& r) {
    body(g);
    req_ = r;
}

Requestor::~Requestor() { }

Requestor::Requestor(
    Glyph_ptr g, Alignment xalign, Alignment yalign, Coord xspan, Coord yspan
) {
    body(g);
    GlyphImpl::default_requisition(req_);
    Glyph::Requirement* rx = GlyphImpl::requirement(req_, X_axis);
    Glyph::Requirement* ry = GlyphImpl::requirement(req_, Y_axis);
    rx->align = xalign;
    ry->align = yalign;
    rx->natural = xspan;
    ry->natural = yspan; 
}

//+ Requestor(XfGlyph::request)
void Requestor::request(XfGlyph::Requisition& r) {
    r = req_;
}

class ToolSetter : public ActionImpl{
public:
    ToolSetter(FigViewer*, Tool*);
    ~ToolSetter();

    virtual void execute();
private:
    FigViewer* viewer_;
    Tool* tool_;
};

ToolSetter::ToolSetter(FigViewer* v, Tool* t) {
    viewer_ = v;
    tool_ = t;
}

ToolSetter::~ToolSetter() {
    delete tool_;
}

void ToolSetter::execute () {
    viewer_->current_tool(tool_);
}

FDraw::FDraw () : ViewerImpl(true) {
    FrescoContext* fc = FrescoLib::current_context();
    impl_ = new FDrawImpl;
    impl_->figures_ = FigureKit::_narrow(
        FrescoLib::resolve_string(fc, "figures_impl")
    );
    impl_->layouts_ = LayoutKit::_narrow(
        FrescoLib::resolve_string(fc, "layouts_impl")
    );
    impl_->widgets_ = WidgetKit::_narrow(
        FrescoLib::resolve_string(fc, "widgets_impl")
    );
    impl_->menus_ = impl_->widgets_->menu_kit();
    impl_->dialogs_ = impl_->widgets_->dialog_kit();
    impl_->drawings_ = fc->default_drawing_kit();
    impl_->viewer_ = new FigViewer(&impl_->editing_);

    Manipulator::init_kits(
        impl_->figures_, impl_->widgets_, impl_->layouts_, impl_->dialogs_,
        impl_->drawings_, impl_->menus_
    );

    append_viewer(impl_->viewer_);
    impl_->editing_ = true;
    body(interior());
}

FDraw::~FDraw () {
    CORBA::release(Viewer_ptr(impl_->viewer_));
    CORBA::release(impl_->figures_);
    CORBA::release(impl_->widgets_);
    CORBA::release(impl_->layouts_);
    CORBA::release(impl_->menus_);
    CORBA::release(impl_->dialogs_);
    CORBA::release(impl_->drawings_);
    delete impl_;
}

#define do_command(Command) { Command cmd(impl_->viewer_); cmd.execute(); }

void FDraw::request(Glyph::Requisition& r) {
    if (impl_->editing_) {
        ViewerImpl::request(r);
        impl_->req_ = r;
    } else {
        r = impl_->req_;
    }
}

void FDraw::traverse (GlyphTraversal_ptr t) {
    ViewerImpl::traverse(t);
}

static Boolean viewer_iterate_handle (
    Viewer_ptr v, GlyphTraversal_ptr t, Event_ptr e
) {
    Boolean done = false;
    for (
        Viewer_var vv = v->first_viewer(); is_not_nil(vv) && !done; 
        vv = vv->next_viewer()
    ) {
        done = vv->handle(t, e);
    }
    return done;
}

void FDraw::run_edit(Boolean b) {
    if (b) {
        impl_->editing_ = false;
        flip_to(1);
        need_redraw();

        SelectCmd unselect(impl_->viewer_, false);
        Selection* sel = impl_->viewer_->selection()->copy();
        for (long i = 0; i < sel->count(); i++) {
            sel->item(i)->execute(&unselect);
        }
        delete sel;
    } else {
        impl_->editing_ = true;
        flip_to(0);
        Viewer_var parent = parent_viewer();
        parent->request_focus(this, false);
        need_redraw();
    }
}

Boolean FDraw::key_press (GlyphTraversal_ptr t, Event_ptr e) {
/*
    if (e->character() == '\033') {
        if (e->modifier_is_down(Event::control)) {
            Boolean ok = viewer_iterate_handle(this, t, e);
            if (!ok && !impl_->editing_) {
                impl_->editing_ = true;
                flip_to(0);
                Viewer_var parent = parent_viewer();
                parent->request_focus(this, false);
                need_redraw();
                ok = true;
            }
            return ok;
        } else {
            if (impl_->editing_) {
                impl_->editing_ = false;
                flip_to(1);
                need_redraw();

                SelectCmd unselect(impl_->viewer_, false);
                Selection* sel = impl_->viewer_->selection()->copy();
                for (long i = 0; i < sel->count(); i++) {
                    sel->item(i)->execute(&unselect);
                }
                delete sel;
                return true;
            } else {
                return viewer_iterate_handle(this, t, e);
            }
        }
    } else if (!impl_->editing_) {
        return viewer_iterate_handle(this, t, e);
    }
    switch (e->character()) {
    case 'a':
        do_command(SelectAllCmd);
	break;
    case 'd':
	do_command(CopyCmd)
	break;
    case 'z': {
        Viewer_var v = nil;
	Selection* s = impl_->viewer_->selection();
        if (s->count() == 1) {
            v = s->item(0)->viewer();
        }
        if (is_nil(v) || e->modifier_is_down(Event::control)) {
            v = Viewer::_duplicate(this);
        }
        Viewer_var parent = v->parent_viewer();
        parent->request_focus(v, false);
	} break;
    case 'g':
	do_command(GroupCmd)
	break;
    case 'i':
	do_command(InstanceCmd)
	break;
    case '\4':
	do_command(DeleteCmd)
	break;
    }
    return true;
*/
    return viewer_iterate_handle(this, t, e);
}

class RunEditAction : public ActionImpl {
public:
    RunEditAction(FDraw* fdraw, Boolean run_edit) { fdraw_ = fdraw; run_edit_ = run_edit; }
    virtual void execute() { fdraw_->run_edit(run_edit_); }
private:
    FDraw* fdraw_;
    Boolean run_edit_;
};

Glyph_ptr FDraw::interior () {
    impl_->deck_ = impl_->layouts_->deck();

    /* bottom layer: run mode */
    impl_->card_[1] = impl_->layouts_->vbox();
    MenuItem_var mi = impl_->menus_->menu_item_string(
        FrescoLib::tmp_string_ref("Run Menu"), 0
    );
    MenuItem* menu = impl_->menus_->menu_item_string(
        FrescoLib::tmp_string_ref("File "), 'F'
    );
    CharString_var nil_accelerator = FrescoLib::tmp_string_ref("");
    menu->append(
        impl_->menus_->button_item_string(
            FrescoLib::tmp_string_ref("Edit"), 'E', nil_accelerator, 
            new RunEditAction(this, false)
        )
    );
    mi->append(menu);
    mi->append(impl_->menus_->separator_item());
    impl_->card_[1]->append(mi->menu_bar());
    impl_->card_[1]->append(impl_->layouts_->align(
	Glyph_var(impl_->layouts_->natural(impl_->viewer_, 5*72.0, 7*72.0)), 0.0, 1.0
    ));

    /* top layer: edit mode */
    impl_->card_[0] = impl_->layouts_->vbox();

    Glyph_var hbox = impl_->layouts_->hbox();
    
    hbox->append(Glyph_var(impl_->widgets_->outset_frame(Glyph_var(tools()))));
    hbox->append(
	Glyph_var(impl_->layouts_->valign(
	    Glyph_var(impl_->layouts_->natural(impl_->viewer_, 5*72.0, 7*72.0)), 1.0
	))
    );
    impl_->card_[0]->append(Glyph_var(commands()));
    impl_->card_[0]->append(hbox);

    impl_->deck_->append(impl_->card_[0]);

    return impl_->widgets_->inset_frame(impl_->deck_);
}
    
declareActionCallback(Command)
implementActionCallback(Command)

Glyph_ptr FDraw::commands() {
    MenuItem_var mi = impl_->menus_->menu_item_string(
        FrescoLib::tmp_string_ref("Edit Menu"), 0
    );
    mi->append(file_menu());
    mi->append(edit_menu());
    mi->append(structure_menu());
    mi->append(impl_->menus_->separator_item());
    mi->append(help_menu());

    return mi->menu_bar();
}

MenuItem_ptr FDraw::file_menu() {
    MenuItem* menu = impl_->menus_->menu_item_string(
        FrescoLib::tmp_string_ref("File "), 'F'
    );
    Glyph* pf = impl_->figures_->group();
    PolyManip* root = new PolyManip(pf);

    CharString_var nil_accelerator = FrescoLib::tmp_string_ref("");
    menu->append(
        impl_->menus_->button_item_string(
            FrescoLib::tmp_string_ref("Run"), 'R', nil_accelerator, 
            new RunEditAction(this, true)
        )
    );
    menu->append(
        impl_->menus_->button_item_string(
            FrescoLib::tmp_string_ref("New"), 'N', nil_accelerator, 
            new NewCmd(impl_->viewer_, root)
        )
    );
    menu->append(
        impl_->menus_->button_item_string(
            FrescoLib::tmp_string_ref("Open"), 'O', nil_accelerator, nil
        )
    );
    menu->append(
        impl_->menus_->button_item_string(
            FrescoLib::tmp_string_ref("Save"), 'S', nil_accelerator, nil
        )
    );
    menu->append(
        impl_->menus_->button_item_string(
            FrescoLib::tmp_string_ref("SaveAs"), 'A', nil_accelerator, nil
        )
    );
    menu->append(
        impl_->menus_->button_item_string(
            FrescoLib::tmp_string_ref("Print"), 'P', nil_accelerator, nil
        )
    );
    menu->append(
        impl_->menus_->button_item_string(
            FrescoLib::tmp_string_ref("Exit"), 'x', nil_accelerator,
            new QuitCmd(impl_->viewer_)
        )
    );
    CORBA::release(pf);
    CORBA::release(root);
    return menu;
}

MenuItem_ptr FDraw::edit_menu() {
    MenuItem* menu = impl_->menus_->menu_item_string(
        FrescoLib::tmp_string_ref("Edit "), 'E'
    );
    CharString_var nil_accelerator = FrescoLib::tmp_string_ref("");

    menu->append(
        impl_->menus_->button_item_string(
            FrescoLib::tmp_string_ref("Undo"), 'U', nil_accelerator, nil
        )
    );
    menu->append(
        impl_->menus_->button_item_string(
            FrescoLib::tmp_string_ref("Redo"), 'R', nil_accelerator, nil
        )
    );
    //menu->append(impl_->menus_->separator_item());

    menu->append(
        impl_->menus_->button_item_string(
            FrescoLib::tmp_string_ref("Select All"), 'a', nil_accelerator,
            new SelectAllCmd(impl_->viewer_)
        )
    );
    menu->append(
        impl_->menus_->button_item_string(
            FrescoLib::tmp_string_ref("Cut"), 't', nil_accelerator, nil
        )
    );
    menu->append(
        impl_->menus_->button_item_string(
            FrescoLib::tmp_string_ref("Copy"), 'c', nil_accelerator, nil
        )
    );
    menu->append(
        impl_->menus_->button_item_string(
            FrescoLib::tmp_string_ref("Paste"), 'p', nil_accelerator, nil
        )
    );
    menu->append(
        impl_->menus_->button_item_string(
            FrescoLib::tmp_string_ref("Delete"), 'e', nil_accelerator,
            new DeleteCmd(impl_->viewer_)
        )
    );
    menu->append(
        impl_->menus_->button_item_string(
            FrescoLib::tmp_string_ref("Duplicate"), 'd', nil_accelerator,
            new CopyCmd(impl_->viewer_)
        )
    );
    menu->append(
        impl_->menus_->button_item_string(
            FrescoLib::tmp_string_ref("Instance"), 'i', nil_accelerator,
            new InstanceCmd(impl_->viewer_)
        )
    );
    menu->append(
        impl_->menus_->button_item_string(
            FrescoLib::tmp_string_ref("Natural"), 'n', nil_accelerator,
            new NaturalCmd(impl_->viewer_)
        )
    );

    return menu;
}

MenuItem_ptr FDraw::structure_menu() {
    MenuItem* menu = impl_->menus_->menu_item_string(
        FrescoLib::tmp_string_ref("Structure"), 'S'
    );
    CharString_var nil_accelerator = FrescoLib::tmp_string_ref("");
    menu->append(
        impl_->menus_->button_item_string(
            FrescoLib::tmp_string_ref("Group"), 'g', nil_accelerator,
            new GroupCmd(impl_->viewer_)
        )
    );
    menu->append(
        impl_->menus_->button_item_string(
            FrescoLib::tmp_string_ref("Ungroup"), 'u', nil_accelerator,
            new UngroupCmd(impl_->viewer_)
        )
    );
    menu->append(
        impl_->menus_->button_item_string(
            FrescoLib::tmp_string_ref("Widen"), 'w', nil_accelerator,
            new WidenCmd(impl_->viewer_)
        )
    );
    menu->append(
        impl_->menus_->button_item_string(
            FrescoLib::tmp_string_ref("HBox"), 'h', nil_accelerator,
            new GroupCmd(impl_->viewer_, new HBoxManip)
        )
    );
    menu->append(
        impl_->menus_->button_item_string(
            FrescoLib::tmp_string_ref("VBox"), 'v', nil_accelerator,
            new GroupCmd(impl_->viewer_, new VBoxManip)
        )
    );
    return menu;
}

MenuItem_ptr FDraw::help_menu () {
    MenuItem* menu = impl_->menus_->menu_item_string(
        FrescoLib::tmp_string_ref("Help"), 'H'
    );
    CharString_var nil_accelerator = FrescoLib::tmp_string_ref("");
    menu->append(
        impl_->menus_->button_item_string(
            FrescoLib::tmp_string_ref("On Version"), '$', nil_accelerator,
            new AboutCmd(impl_->viewer_, impl_->dialogs_, impl_->figures_)
        )
    );
    return menu;
}

static Glyph_ptr create_label(const char* label, FigureKit_ptr figures) {
    return figures->label(
        FigureStyle_var(figures->default_style()),
	FrescoLib::tmp_string_ref(label)
    );
}

Glyph_ptr FDraw::tools() {
    FigureStyle_var fs = impl_->figures_->default_style();

    SelectTool* selectTool = new SelectTool;
    MoveTool* moveTool = new MoveTool;
    ScaleTool* scaleTool = new ScaleTool;
    RotateTool* rotateTool = new RotateTool;
    ResizeTool* resizeTool = new ResizeTool;
    NarrowTool* narrowTool = new NarrowTool;
    
    Manipulator* line_manip = new LineManip;
    Manipulator* rect_manip = new RectManip;
    Manipulator* ellipse_manip = new EllipseManip;
    Manipulator* openbspline_manip = new Open_BSplineManip;
    Manipulator* closedbspline_manip = new Closed_BSplineManip;
    Manipulator* multiline_manip = new MultiLineManip;
    Manipulator* polygon_manip = new PolygonManip;

    ButtonManip* button_manip = new ButtonManip(&impl_->editing_);
    Manipulator* glue_manip = new GlueManip;
    Manipulator* fdraw_manip = new FDrawManip(&impl_->editing_, impl_->viewer_);
    Manipulator* field_manip = new FieldManip(&impl_->editing_, impl_->viewer_);

    CreateTool* line_tool = new CreateTool(line_manip);
    CreateTool* multiline_tool = new CreateTool(multiline_manip);
    CreateTool* openbspline_tool = new CreateTool(openbspline_manip);
    CreateTool* rect_tool = new CreateTool(rect_manip);
    CreateTool* ellipse_tool = new CreateTool(ellipse_manip);
    CreateTool* polygon_tool = new CreateTool(polygon_manip);
    CreateTool* closedbspline_tool = new CreateTool(closedbspline_manip);
    CreateTool* button_tool = new CreateTool(button_manip);
    CreateTool* glue_tool = new CreateTool(glue_manip);
    CreateTool* fdraw_tool = new CreateTool(fdraw_manip);
    CreateTool* field_tool = new CreateTool(field_manip);

    Vertex* vclosed = new Vertex[5];
    vclosed[0].x = 3, vclosed[0].y = 0;
    vclosed[1].x = 16, vclosed[1].y = 3;
    vclosed[2].x = 17, vclosed[2].y = 9;
    vclosed[3].x = 9, vclosed[3].y = 16;
    vclosed[4].x = 0, vclosed[4].y = 6;

    Vertex* vopen = new Vertex[4];
    vopen[0].x = 0, vopen[0].y = 0;
    vopen[1].x = 7.5, vopen[1].y = 3.75;
    vopen[2].x = 7.5, vopen[2].y = 11.25;
    vopen[3].x = 15, vopen[3].y = 15;

    FigureKit::Vertices v_closed(5, 5, vclosed);
    FigureKit::Vertices v_open(4, 4, vopen);

    Telltale_var group = impl_->widgets_->telltale_group();
    Coord fil = impl_->layouts_->fil()/10000;

    Button_var line_button = tool_button(
        Glyph_var(impl_->layouts_->hmargin_lr_flexible(
	    Glyph_var(impl_->figures_->line(
		fs, 0, 0, 15, 15
	    )), 0.0, fil, 0, 15.0, fil, 0
	)), line_tool, group
    );
    Button_var multiline_button = tool_button(
        Glyph_var(impl_->layouts_->hmargin_lr_flexible(
            Glyph_var(impl_->figures_->multiline(
                FigureKit::stroke, fs, v_open
	    )), 0.0, fil, 0, 15.0, fil, 0
        )), multiline_tool, group
    );
    Button_var openbspline_button = tool_button(
        Glyph_var(impl_->layouts_->hmargin_lr_flexible(
            Glyph_var(impl_->figures_->open_bspline(
                FigureKit::stroke, fs, v_open
	    )), 0.0, fil, 0, 15.0, fil, 0
        )), openbspline_tool, group
    );
    Button_var rect_button = tool_button(
        Glyph_var(impl_->layouts_->hmargin_lr_flexible(
            Glyph_var(impl_->figures_->rectangle(
                FigureKit::fill_stroke, fs, 0, 0, 18, 15
	    )), 0.0, fil, 0, 15.0, fil, 0
        )), rect_tool, group
    );
    Button_var ellipse_button = tool_button(
        Glyph_var(impl_->layouts_->hmargin_lr_flexible(
            Glyph_var(impl_->figures_->ellipse(
                FigureKit::fill_stroke, fs, 0, 0, 12, 8
	    )), 0.0, fil, 0, 15.0, fil, 0
        )), ellipse_tool, group
    );
    Button_var polygon_button = tool_button(
        Glyph_var(impl_->layouts_->hmargin_lr_flexible(
            Glyph_var(impl_->figures_->polygon(
                FigureKit::fill_stroke, fs, v_closed
	    )), 0.0, fil, 0, 15.0, fil, 0
        )), polygon_tool, group
    );
    vclosed[3].x = 9, vclosed[3].y = 18;
    Button_var closedbspline_button = tool_button(
        Glyph_var(impl_->layouts_->hmargin_lr_flexible(
            Glyph_var(impl_->figures_->closed_bspline(
                FigureKit::fill_stroke, fs, v_closed
	    )), 0.0, fil, 0, 15.0, fil, 0
        )), closedbspline_tool, group
    );

    TransformImpl* t = new TransformImpl;
    Vertex s; s.x = 0.55; s.y = 0.55; s.z = 0;
    t->scale(s);
    Button_var push_button = tool_button(
        Glyph_var(impl_->layouts_->hmargin_lr_flexible(
            Glyph_var(new TransformSetter(
                Glyph_var(impl_->widgets_->palette_button(
		    Glyph_var(impl_->layouts_->vspace(10)), nil
		)), t
	    )), 0.0, fil, 0, 15.0, fil, 0
        )), button_tool, group
    );
    Button_var fdraw = tool_button(
	Glyph_var(create_label("FDraw", impl_->figures_)), fdraw_tool, group
    );
    Button_var field = tool_button(
	Glyph_var(create_label("Field", impl_->figures_)), field_tool, group
    );
    Button_var glue = tool_button(
	Glyph_var(create_label("Glue", impl_->figures_)), glue_tool, group
    );
    Button_var choose = tool_button(
        Glyph_var(create_label("Select", impl_->figures_)), selectTool, group
    );
    Button_var move = tool_button(
        Glyph_var(create_label("Move", impl_->figures_)), moveTool, group
    );
    Button_var scale = tool_button(
        Glyph_var(create_label("Scale", impl_->figures_)), scaleTool, group
    );
    Button_var rotate = tool_button(
        Glyph_var(create_label("Rotate", impl_->figures_)), rotateTool, group
    );
    Button_var resize = tool_button(
        Glyph_var(create_label("Resize", impl_->figures_)), resizeTool, group
    );
    Button_var narrow = tool_button(
        Glyph_var(create_label("Narrow", impl_->figures_)), narrowTool, group
    );

    Telltale_var bs = choose->state();
    bs->set(Telltale::chosen);
    impl_->viewer_->current_tool(selectTool);

    Glyph_var vglue = impl_->layouts_->front(
        Glyph_var(impl_->layouts_->vfil()), 
        Glyph_var(impl_->widgets_->filler())
    );

    Glyph_ptr g = impl_->layouts_->vbox();
    g->append(Glyph_var(impl_->layouts_->hflexible(choose, fil, 0)));
    g->append(Glyph_var(impl_->layouts_->hflexible(move, fil, 0)));
    g->append(Glyph_var(impl_->layouts_->hflexible(scale, fil, 0)));
    g->append(Glyph_var(impl_->layouts_->hflexible(rotate, fil, 0)));
    g->append(Glyph_var(impl_->layouts_->hflexible(resize, fil, 0)));
    g->append(Glyph_var(impl_->layouts_->hflexible(narrow, fil, 0)));
    g->append(Glyph_var(impl_->layouts_->hflexible(glue, fil, 0)));
    g->append(Glyph_var(impl_->layouts_->hflexible(fdraw, fil, 0)));
    g->append(Glyph_var(impl_->layouts_->hflexible(field, fil, 0)));
    g->append(Glyph_var(impl_->layouts_->hflexible(push_button, fil, 0)));
    g->append(Glyph_var(impl_->layouts_->hflexible(line_button, fil, 0)));
    g->append(Glyph_var(impl_->layouts_->hflexible(multiline_button, fil, 0)));
    g->append(Glyph_var(impl_->layouts_->hflexible(openbspline_button, fil, 0)));
    g->append(Glyph_var(impl_->layouts_->hflexible(rect_button, fil, 0)));
    g->append(Glyph_var(impl_->layouts_->hflexible(ellipse_button, fil, 0)));
    g->append(Glyph_var(impl_->layouts_->hflexible(polygon_button, fil, 0)));
    g->append(Glyph_var(impl_->layouts_->hflexible(closedbspline_button, fil, 0)));
    g->append(vglue);
    return g;
}

Button_ptr FDraw::tool_button(
    Glyph_ptr g, Tool* t, Telltale_ptr group
) {
    Button_ptr b = impl_->widgets_->palette_button(g, nil);
    Telltale_var bs = b->state();
    bs->clear(Telltale::toggle);
    bs->set(Telltale::choosable);
    bs->current(group);
    if (t != nil) {
	b->click_action(Action_var(new ToolSetter(impl_->viewer_, t)));
    }
    append_viewer(b);
    return b;
}

void FDraw::flip_to(Long card) {
    GlyphOffset_var g = impl_->deck_->first_child_offset();
    impl_->deck_->append(impl_->card_[card]);
    g->remove();
    impl_->deck_->need_redraw();
    impl_->deck_->need_resize();
}
