//
// Copyright (c) 1994 Silicon Graphics, Inc.
// Copyright (c) 1994 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
// OF THIS SOFTWARE.
//

/*
 * figviewer.cxx
 */

#include "commands.h"
#include "figviewer.h"
#include "manipulators.h"
#include "selection.h"
#include "tools.h"
#include "visitors.h"
#include <Fresco/defs/fcontext.h>
#include <Fresco/defs/fresco.h>
#include <Fresco/defs_impl/action.h>
#include <Fresco/defs_impl/glyphs.h>
#include <Fresco/defs_impl/region.h>
#include <Fresco/defs_impl/threads.h>
#include <Fresco/defs_impl/transform.h>
#include <Fresco/defs_impl/math.h>
#include <Fresco/defs_impl/memory.h>

implementPtrList(WidenGlyphList,Glyph)
declarePtrList(WidenOffsetList,GlyphOffset)
implementPtrList(WidenOffsetList,GlyphOffset)

static double zoom_factor = 600.0;

class EventAction : public ActionImpl {
public:
    virtual void event(Event_ptr) = 0;
protected:
    EventAction();
};

class ZoomAction : public EventAction {
public:
    ZoomAction(Transform_ptr, Glyph_ptr, Event_ptr);
    virtual ~ZoomAction();

    void execute();
    void event(Event_ptr);
private:
    Vertex center_;
    double zoom_;
    Transform_ptr t_;
    Glyph_ptr g_;
    Mutex_ptr lock_;
};

class GrabAction : public EventAction {
public:
    GrabAction(Transform_ptr, Glyph_ptr, Event_ptr);
    virtual ~GrabAction();

    void event(Event_ptr);
private:
    Vertex last_pos_;
    Transform_ptr t_;
    Glyph_ptr g_;
};

EventAction::EventAction() { }

ZoomAction::ZoomAction(Transform_ptr t, Glyph_ptr g, Event_ptr e) {
    t_ = Transform::_duplicate(t);
    g_ = Glyph::_duplicate(g);
    zoom_ = 1.0;
    center_.x = e->pointer_x();
    center_.y = e->pointer_y();
    center_.z = 0;
    t_->inverse_transform_vertex(center_);
    lock_ = ThreadKit_var(ThreadLib::thread_kit_ptr())->mutex_ptr();
}

ZoomAction::~ZoomAction() {
    CORBA::release(t_);
    CORBA::release(g_);
    CORBA::release(lock_);
}

void ZoomAction::execute() {
    for (;;) {
	if (!Math::equal(zoom_, 1.0, 1.0e-6)) {
	    Transform_var t = g_->transformation();
	    g_->need_redraw();
	    Vertex zoom;
	    Vertex center = center_;
	    center.x = -center.x;
	    center.y = -center.y;
	    center.z = -center.z;
	    t->translate(center);
	    lock_->acquire();
	    zoom.x = zoom_; zoom.y = zoom_; zoom.z = zoom_;
	    t->scale(zoom);
	    lock_->release();
	    t->translate(center_);
	    g_->need_redraw();
	}
	FrescoLib::delay(0.03);
    }
}

void ZoomAction::event(Event_ptr e) {
    Vertex center;
    center.x = e->pointer_x();
    center.y = e->pointer_y();
    center.z = 0;
    t_->inverse_transform_vertex(center);
    Coord dy = center_.y - center.y;
    lock_->acquire();
    if (dy > 0.0) {
	zoom_ = zoom_factor / (zoom_factor + dy);
    } else {
	zoom_ = (zoom_factor - dy) / zoom_factor;
    }
    lock_->release();
}

GrabAction::GrabAction(Transform_ptr t, Glyph_ptr g, Event_ptr e) {
    t_ = Transform::_duplicate(t);
    g_ = Glyph::_duplicate(g);
    last_pos_.x = e->pointer_x();
    last_pos_.y = e->pointer_y();
    last_pos_.z = 0;
    t_->inverse_transform_vertex(last_pos_);
}

GrabAction::~GrabAction() {
    CORBA::release(t_);
    CORBA::release(g_);
}

void GrabAction::event(Event_ptr e) {
    Vertex cur_pos, delta;
    cur_pos.x = e->pointer_x();
    cur_pos.y = e->pointer_y();
    cur_pos.z = 0;
    t_->inverse_transform_vertex(cur_pos);
    delta.x = cur_pos.x - last_pos_.x;
    delta.y = cur_pos.y - last_pos_.y;
    delta.z = 0;
    Transform_ptr t = g_->transformation();
    g_->need_redraw();
    t->translate(delta);
    g_->need_redraw();
    last_pos_ = cur_pos;
}

class ViewerDamage : public Damage {
public:
    ViewerDamage(Region_ptr a, Transform_ptr t, Damage_ptr d);
    ~ViewerDamage();

    OxExchange* _exchange();

    virtual void extend(Region_ptr r);
protected:
    Damage_var damage_;
    RegionImpl clip_;
};

ViewerDamage::ViewerDamage(Region_ptr a, Transform_ptr t, Damage_ptr d) {
    damage_ = d;
    clip_.copy(a);
    clip_.apply_transform(t);
    float tol = 0.1;
    clip_.lower_.x += tol;
    clip_.lower_.y += tol;
    clip_.upper_.x -= tol;
    clip_.upper_.y -= tol;
}

ViewerDamage::~ViewerDamage() { }

OxExchange* ViewerDamage::_exchange() { return damage_->_exchange(); }

void ViewerDamage::extend(Region_ptr r) {
    if (r->intersects(&clip_)) {
	RegionImpl region;
	region.copy(r);
	region.merge_intersect(&clip_);
	damage_->extend(&region);
    }
}

FigViewer::FigViewer(Boolean* editing) {
    FrescoContext* fc = FrescoLib::current_context();
    DrawingKit_var drawings = fc->default_drawing_kit();
    FigureKit_var figures = FigureKit::_narrow(
        FrescoLib::resolve_string(fc, "figures_impl")
    );

    curtool_ = nil;
    active_ = nil;
    editing_ = editing;

    sel_ = new Selection(10);

    Glyph* pf = figures->group();
    PolyManip* r = new PolyManip(pf);
    CORBA::release(pf);

    action_ = nil;
    thread_ = nil;
    tx_ = new TransformImpl;

    narrowgt_ = nil;
    offsets_ = nil;
    glyphs_ = nil;
    background_ = drawings->color_rgb(0.5, 0.5, 0.5);

    body(r);
}
 
FigViewer::~FigViewer() {
    CORBA::release(tx_);
    delete glyphs_;
    delete offsets_;
    delete sel_;
}

void FigViewer::root(Manipulator* r) {
    Glyph_var g = body();
    g->need_redraw();
    body(r);
    r->need_redraw();
}

//+ FigViewer(XfGlyph::allocations)
void FigViewer::allocations(XfGlyph::AllocationInfoSeq& a) {
    ViewerImpl::allocations(a);
}

//+ FigViewer(XfGlyph::traverse)
void FigViewer::traverse(XfGlyphTraversal_ptr t) {
    if (t->allocation_is_visible()) {
	t->begin_viewer(Viewer_ptr(this));
	Painter_var p = t->current_painter();
	Vertex lower, upper, origin;
	Boolean clipped = t->bounds(lower, upper, origin);
	if (clipped) {
	    p->push_clipping();
	    p->clip_rect(lower.x, lower.y, upper.x, upper.y);
	}
	if (t->op() == GlyphTraversal::draw) {
	    BgViewer::draw(t);
	}
	p->push_matrix();
	p->premultiply(tx_);
	t->traverse_child(&offset_, nil);
	if (t->is_pick_op() && is_nil(GlyphTraversal_var(t->picked()))) {
	    t->hit();
	}
	p->pop_matrix();
	if (clipped) {
	    p->pop_clipping();
	}
	t->end_viewer();
    }
}

Manipulator* FigViewer::root () { 
    return Manipulator::_narrow(body());
}

Selection* FigViewer::selection() {
    return sel_;
}

//+ FigViewer(XfGlyph::request)
void FigViewer::request(XfGlyph::Requisition& r) {
    Coord fil = 1.0e6;
    r.x.natural = 0; r.x.maximum = fil, r.x.minimum = 0, r.x.align = 0.5;
    r.y.natural = 0; r.y.maximum = fil, r.y.minimum = 0, r.y.align = 0.5;
}

//+ FigViewer(XfGlyph::extension)
void FigViewer::extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r) {
    GlyphImpl::default_extension(a, r);
}

Transform_ptr FigViewer::tform() {
    return Transform::_duplicate(tx_);
}

//+ FigViewer(XfGlyph::transformation)
XfTransform_ptr FigViewer::transformation() {
    return nil; // don't use tx_
}

Tool* FigViewer::current_tool() { return curtool_; }
void FigViewer::current_tool(Tool* t) { curtool_ = t; }

//+ FigViewer(XfGlyph::need_resize)
void FigViewer::need_resize() {
    /*
     * Do nothing because we assume that we've already handled
     * a resize appropriately.
     */
}

Boolean FigViewer::press(GlyphTraversal_ptr t, Event_ptr e) {
    if (!*editing_) {
        return false;
    }
    Boolean ok = false;
    if (is_nil(active_)) {
        SelectInfo* si = curtool_->create_manipulator(t, e, this);
        active_ = (si != nil) ? si->m_ : nil;
        if (is_not_nil(active_)) {
            ok = active_->grasp(*curtool_, *this, si, e);
        } else {
            ok = false;
        }
        delete si;
        if (ok) {
            grab(t);
        }
    } else {
        ok = active_->manipulate(e);
    }
    if(!ok && is_not_nil(active_)) {
        FCommand* command = active_->effect(e);
        if (is_not_nil(command) && command->reversible()) {
            command->log();
        }
        active_ = nil;
    }
    return true;
}

Boolean FigViewer::move(GlyphTraversal_ptr t, Event_ptr e) {
    if (!*editing_) {
        return false;
    }
    return drag(t, e);
}

Boolean FigViewer::drag(GlyphTraversal_ptr, Event_ptr e) {
    if (is_not_nil(action_)) {
	action_->event(e);
	return true;
    }
    if (!*editing_) {
        return false;
    }
    Boolean ok = false;
    if (is_not_nil(active_)) {
        ok = active_->manipulate(e);
        if (!ok) {
            FCommand* command = active_->effect(e);
            if (is_not_nil(command) && command->reversible()) {
                command->log();
            }
            active_ = nil;
            ungrab();
        }
    }
    return true;
}

Boolean FigViewer::release(GlyphTraversal_ptr gt, Event_ptr e) {
    if (is_not_nil(action_)) {
	//action_->event(e);
	CORBA::release(action_);
	action_ = nil;
	if (is_not_nil(thread_)) {
	    thread_->terminate();
	    CORBA::release(thread_);
	    thread_ = nil;
	}
	ungrab();
	return true;
    }
    if (!*editing_) {
        return false;
    }
    Boolean ok = false;
    if (is_not_nil(active_)) {
        ok = active_->manipulate(e);
        if (!ok) {
            FCommand* command = active_->effect(e);
            if (is_not_nil(command) && command->reversible()) {
                command->log();
            }
            active_ = nil;
            ungrab();
        } else {
	    grab(gt);
	}
    }
    return true;
}

void FigViewer::visit_trail(GlyphTraversal_ptr t) {
    Painter_var p = t->current_painter();
    p->premultiply(tx_);
}

void FigViewer::child_allocate(Glyph::AllocationInfo& a) {
    a.allocation = nil;
    a.transformation->premultiply(tx_);
}

static Transform_ptr transformation(WidenOffsetList* trail, Long level) {
    Glyph::AllocationInfo a;
    a.allocation = new RegionImpl;
    a.transformation = new TransformImpl;
    for (Long i=0; i <= level; i++) {
	GlyphOffset_ptr offset = trail->item(i);
	if (is_not_nil(offset)) {
	    offset->child_allocate(a);
	}
    }
    // avoid extra duplicate and release
    Transform_ptr t = a.transformation;
    a.transformation.ptr_ = nil;
    return t;
}

static Transform_ptr difference(
    WidenOffsetList* trail, Long level, Long offset
) {
    Transform_var tx = new TransformImpl;
    Transform_ptr txoffset = new TransformImpl;
    tx->copy(Transform_var(transformation(trail, level)));
    txoffset->copy(Transform_var(transformation(trail, level+offset)));
    tx->invert();
    txoffset->postmultiply(tx);
    return txoffset;
}

void FigViewer::narrow(Long level) {
    Transform_var diff;
    Long orig_level = offsets_->count()-1;
    if (level > 0) {
	narrowgt_->forward(); // start at root.
	for (Long i = 1; i <= level; i++) {
	    narrowgt_->forward();
	    GlyphOffset_ptr offset = narrowgt_->current_offset();
	    offsets_->append(offset);
	    glyphs_->append(Glyph_var(offset->child()));
	}
	while (i-- != 0) {
	    narrowgt_->backward();
	}
	diff = difference(offsets_, orig_level, level);
	tx_->premultiply(diff);
    } else {
	diff = difference(offsets_, orig_level-2, 2);
	diff->invert();
	tx_->premultiply(diff);
	while (level++ != 0) {
	    offsets_->remove(offsets_->count()-1);
	    glyphs_->remove(glyphs_->count()-1);
	}
    }
    body(glyphs_->item(glyphs_->count()-1));
    need_redraw();
}

void FigViewer::narrowgt(GlyphTraversal_ptr gt) {
    narrowgt_ = gt->trail();
    if (offsets_ == nil) {
	offsets_ = new WidenOffsetList;
	glyphs_ = new WidenGlyphList;
	offsets_->append(GlyphOffset_var(narrowgt_->current_offset())); 
	glyphs_->append(Glyph_var(narrowgt_->current_viewer()));
	narrowgt_->forward();
	glyphs_->append(Glyph_var(narrowgt_->current_glyph()));
	narrowgt_->backward();
    }
}

WidenGlyphList* FigViewer::widen_list() {
    return glyphs_;
}
