//
// Copyright (c) 1994 Silicon Graphics, Inc.
// Copyright (c) 1994 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
// OF THIS SOFTWARE.
//

/*
 * manipulators.cxx
 */

#include <assert.h>
#include <float.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#if defined(__sgi) || (defined(sun) && !defined(SVR4))
#include <values.h>
#endif

#include "commands.h"
#include "fdraw.h"
#include "figviewer.h"
#include "manipulators.h"
#include "selection.h"
#include "tools.h"
#include "visitors.h"

#include <Fresco/defs/drawing.h>
#include <Fresco/defs/fcontext.h>
#include <Fresco/defs/fresco.h>
#include <Fresco/defs/viewer.h>
#include <Fresco/figures/figures.h>
#include <Fresco/defs_impl/action.h>
#include <Fresco/defs_impl/region.h>
#include <Fresco/defs_impl/math.h>
#include <Fresco/widgets/widgets.h>
#include <Fresco/Ox/schema.h>

static const Coord dot = 4.0;
static const Coord half_dot = 2.0;
static const double tol = 0.000005;

#define mfil 10.0e6;

static Color_ptr black_color;

class PrintAction : public ActionImpl {
public:
    PrintAction(const char*);
    virtual ~PrintAction();

    virtual void execute(); 
private:
    char* string_;
};

PrintAction::PrintAction (const char* s) {
    string_ = new char[strlen(s) + 1];
    strcpy(string_, s);
}

PrintAction::~PrintAction () {
    delete string_;
}

void PrintAction::execute () {
    printf("%s\n", string_);
}

static void glyph_need_redraw(
    XfGlyph_ptr g, const XfGlyph::AllocationInfoSeq& a
) {
    RegionImpl r;
    XfGlyph::AllocationInfo clone;
    clone.transformation = new TransformImpl;

    for (Long i = 0; i < a._length; i++) {
	const XfGlyph::AllocationInfo& info = a[i];
	if (is_not_nil(info.damaged) && is_nil(info.allocation)) {
            r.defined_ = false;
            if (is_not_nil(info.transformation)) {
                clone.transformation->copy(info.transformation);
            } else {
                clone.transformation->load_identity();
            }
            if (is_not_nil(info.allocation)) {
                clone.allocation = new RegionImpl;
                clone.allocation->copy(info.allocation);
            }
            clone.damaged = info.damaged;
	    g->extension(clone, &r);
	    info.damaged->extend(&r);
	    if (is_not_nil(info.allocation)) {
		clone.allocation = nil;
	    }
	}
    }
}

static void parents_need_redraw(XfGlyph_ptr g) {
    OffsetVisitor ov;
    ov.visit_parents(g);
    for (Long i = 0; i < ov.offset_count(); i++) {
	XfGlyph_var p = ov.offset(i)->parent();
	XfGlyph::AllocationInfoSeq info;
	p->allocations(info);
	glyph_need_redraw(p, info);  // check for layout objects above
    }
}

static void _nat_allocation (XfGlyph::Requisition* r, RegionImpl* region) {
    region->xalign_ = r->x.align;
    region->lower_.x = 0;
    region->upper_.x = r->x.natural;
    region->yalign_ = r->y.align;
    region->lower_.y = 0;
    region->upper_.y = r->y.natural;
    region->lower_.z = 0;
    region->upper_.z = 0;
    region->zalign_ = 0;
    region->defined_ = true;
}

static void adjust_dotted_region (RegionImpl& region) {
    double delta = dot*0.5 + 4;
    region.lower_.x -= delta;
    region.lower_.y -= delta;
    region.upper_.x += delta;
    region.upper_.y += delta;
}    

static Float aspect_ratio(XfGlyph::Requisition* r) {
    Float aspect;
    if (r->x.natural == 0) {
        aspect = mfil;
    } else {
        aspect = r->y.natural/r->x.natural;
    }
    return aspect;
}

static void _translate (Transform_ptr tr, Vertex& v) {
    tr->translate(v);
}

static void _scale (Transform_ptr tr, Vertex& center, Vertex& s) {
    Vertex ncenter = center;
    ncenter.x = -ncenter.x;
    ncenter.y = -ncenter.y;
    ncenter.z = -ncenter.z;

    tr->translate(ncenter);
    tr->scale(s);
    tr->translate(center);
}

static void _rotate(Transform_ptr tr, double angle, Axis a, Vertex& center) {
    Vertex ncenter = center;
    ncenter.x = -ncenter.x;
    ncenter.y = -ncenter.y;
    ncenter.z = -ncenter.z;

    tr->translate(ncenter);
    tr->rotate(angle, a);
    tr->translate(center);
}

static void compose (XfGlyph_ptr m) {
    Transform_var tx = m->transformation();
    if (is_not_nil(tx)) {
        Transform::Matrix mat;
        tx->store_matrix(mat);
        mat[2][0] = 0; mat[2][1] = 0;
        tx->load_matrix(mat);
    }
}

static void local_extension(XfGlyph_ptr g, RegionImpl* r) {
    XfGlyph::AllocationInfo a;
    a.transformation = nil;
    a.allocation = nil;
    a.damaged = nil;
    r->defined_ = false;
    g->extension(a, r);
}

static void flexible_transform_requisition(
    XfGlyph::Requisition& req, Transform_ptr tx
) {
    RegionImpl nat, maxi, mini;
    
    nat.xalign_ = req.x.align;
    nat.lower_.x = -req.x.align * req.x.natural;
    nat.upper_.x = nat.lower_.x + req.x.natural;
    nat.yalign_ = req.y.align;
    nat.lower_.y = -req.y.align * req.y.natural;
    nat.upper_.y = nat.lower_.y + req.y.natural;
    nat.lower_.z = nat.upper_.z = 0;
    nat.defined_ = true;

    maxi.xalign_ = req.x.align;
    maxi.lower_.x = -req.x.align * req.x.maximum;
    maxi.upper_.x = maxi.lower_.x + req.x.maximum;
    maxi.yalign_ = req.y.align;
    maxi.lower_.y = -req.y.align * req.y.maximum;
    maxi.upper_.y = maxi.lower_.y + req.y.maximum;
    maxi.lower_.z = maxi.upper_.z = 0;
    maxi.defined_ = true;

    mini.xalign_ = req.x.align;
    mini.lower_.x = -req.x.align * req.x.minimum;
    mini.upper_.x = mini.lower_.x + req.x.minimum;
    mini.yalign_ = req.y.align;
    mini.lower_.y = -req.y.align * req.y.minimum;
    mini.upper_.y = mini.lower_.y + req.y.minimum;
    mini.lower_.z = mini.upper_.z = 0;
    mini.defined_ = true;

    if (!tx->identity()) {
        nat.apply_transform(tx);
        maxi.apply_transform(tx);
        mini.apply_transform(tx);
    }

    req.x.defined = true;
    req.x.natural = nat.upper_.x - nat.lower_.x;
    req.x.maximum = maxi.upper_.x - maxi.lower_.x;
    req.x.minimum = mini.upper_.x - mini.lower_.x;
    if (!Math::equal((double)req.x.natural, 0.0, 1.0e-6)) {
	req.x.align = -nat.lower_.x / req.x.natural;
    } else {
	req.x.align = 0;
    }

    req.y.defined = true;
    req.y.natural = nat.upper_.y - nat.lower_.y;
    req.y.maximum = maxi.upper_.y - maxi.lower_.y;
    req.y.minimum = mini.upper_.y - mini.lower_.y;
    if (!Math::equal((double)req.y.natural, 0.0, 1.0e-6)) {
	req.y.align = -nat.lower_.y / req.y.natural;
    } else {
	req.y.align = 0;
    }

    req.z.defined = false;
}

static void fixed_transform_requisition(
    XfGlyph::Requisition& req, Transform_ptr tx
) {
    RegionImpl nat;
    
    nat.xalign_ = req.x.align;
    nat.lower_.x = -req.x.align * req.x.natural;
    nat.upper_.x = nat.lower_.x + req.x.natural;
    nat.yalign_ = req.y.align;
    nat.lower_.y = -req.y.align * req.y.natural;
    nat.upper_.y = nat.lower_.y + req.y.natural;
    nat.lower_.z = nat.upper_.z = 0;
    nat.defined_ = true;

    if (!tx->identity()) {
        nat.apply_transform(tx);
    }
            
    Coord x_nat_lead = -nat.lower_.x;
    Coord x_nat_trail = nat.upper_.x;

    Coord y_nat_lead = -nat.lower_.y;
    Coord y_nat_trail = nat.upper_.y;

    XfGlyphImpl::require_lead_trail(
        req.x, x_nat_lead, x_nat_lead, x_nat_lead, 
        x_nat_trail, x_nat_trail, x_nat_trail
    );
    XfGlyphImpl::require_lead_trail(
        req.y, y_nat_lead, y_nat_lead, y_nat_lead, 
        y_nat_trail, y_nat_trail, y_nat_trail
    );
}

static Boolean rotated (Transform_ptr t) {
    Transform::Matrix m;
    t->store_matrix(m);
    return (m[0][1] != 0 || m[1][0] != 0);
}

static void set_span(
    RegionImpl& r, Axis a, Coord origin, Coord length
) {
    Coord begin = origin;
    Coord end = begin + length;
    r.defined_ = true;
    switch (a) {
    case X_axis:
	r.lower_.x = begin;
	r.upper_.x = end;
	break;
    case Y_axis:
	r.lower_.y = begin;
	r.upper_.y = end;
	break;
    case Z_axis:
	r.lower_.z = begin;
	r.upper_.z = end;
	break;
    }
}

static double compute_squeeze(const XfGlyph::Requirement& r, Coord length) {
    double f;
    Coord nat = r.natural;
    if (length > nat && r.maximum > nat) {
        f = (length - nat) / (r.maximum - nat);
    } else if (length < nat && r.minimum < nat) {
        f = (nat - length) / (nat - r.minimum);
    } else {
        f = 0;
    }
    return f;
}

static void compute_allocations(
    Axis a, XfGlyph::Requisition& total,
    long n, XfGlyph::Requisition* requests, Region_ptr given,
    RegionImpl* result
) {
    XfGlyph::Requirement* r;
    Region::BoundingSpan s;
    r = XfGlyphImpl::requirement(total, a);
    given->span(a, s);
    Coord length = s.length;
    Boolean growing = length > r->natural;
    Boolean shrinking = length < r->natural;
    double f = compute_squeeze(*r, length);
    Coord p = 0;
    for (long i = 0; i < n; i++) {
        r = XfGlyphImpl::requirement(requests[i], a);
        Coord cspan = r->natural;
        if (growing) {
            cspan += f * (r->maximum - r->natural);
        } else if (shrinking) {
            cspan -= f * (r->natural - r->minimum);
        }
        set_span(result[i], a, p, cspan);
        p += cspan;
    }
}

static void compensate (double a, double& x, double& y) {
    if (a > 0.0) {
        if (Math::equal(x, 0.0, 1.0e-6)) {
            y = a * x;
        } else {
            double aspect = y/x;
            if (aspect > a) {
                y = a * x;
            } else if (aspect < a) {
                x = y / a;
            }
        }
    }
}

static Vertex adjust_allocation (
    RegionImpl* region, Transform_ptr t, XfGlyph::Requisition* req
) {
    Vertex delta;
    delta.x = 0; delta.y = 0; delta.z = 0;
    if (!rotated(t)) {
        TransformImpl tx;
        tx.copy(t);
        tx.invert();
        region->apply_transform(&tx);
	region->xalign_ = req->x.align;
	region->yalign_ = req->y.align;
	region->zalign_ = req->z.align;
    } else {
	Vertex center;
	double x_len, y_len;
	center.x = (region->lower_.x + region->upper_.x) * 0.5;
	center.y = (region->lower_.y + region->upper_.y) * 0.5;

	Transform::Matrix m;
	t->store_matrix(m);

	XfGlyph::Requisition r[2], total;
	RegionImpl a[2];
	double a0 = -1; double a1 = -1;
	if (!Math::equal((double)m[0][0], 0.0, 1e-6)) {
	    a0 = Math::abs(m[0][1] / m[0][0]);
	}
	if (!Math::equal((double)m[1][0], 0.0, 1e-6)) {
	    a1 = Math::abs(m[1][1] / m[1][0]);
	}

	r[0].x.natural = Math::abs(req->x.natural*m[0][0]);
	r[0].x.maximum = Math::abs(req->x.maximum*m[0][0]);
	r[0].x.minimum = Math::abs(req->x.minimum*m[0][0]);
	r[0].x.defined = true;
	r[0].y.natural = Math::abs(req->x.natural*m[0][1]);
	r[0].y.maximum = Math::abs(req->x.maximum*m[0][1]);
	r[0].y.minimum = Math::abs(req->x.minimum*m[0][1]);
	r[0].y.defined = true;

	r[1].x.natural = Math::abs(req->y.natural*m[1][0]);
	r[1].x.maximum = Math::abs(req->y.maximum*m[1][0]);
	r[1].x.minimum = Math::abs(req->y.minimum*m[1][0]);
	r[1].x.defined = true;
	r[1].y.natural = Math::abs(req->y.natural*m[1][1]);
	r[1].y.maximum = Math::abs(req->y.maximum*m[1][1]);
	r[1].y.minimum = Math::abs(req->y.minimum*m[1][1]);
	r[1].y.defined = true;

	total.x.natural = r[0].x.natural + r[1].x.natural;
	total.x.maximum = r[0].x.maximum + r[1].x.maximum;
	total.x.minimum = r[0].x.minimum + r[1].x.minimum;
	total.x.defined = true;
	total.y.natural = r[0].y.natural + r[1].y.natural;
	total.y.maximum = r[0].y.maximum + r[1].y.maximum;
	total.y.minimum = r[0].y.minimum + r[1].y.minimum;
	total.y.defined = true;

	compute_allocations(X_axis, total, 2, r, region, a);
	compute_allocations(Y_axis, total, 2, r, region, a);

	double x0, y0, x1, y1;
	x0 = a[0].upper_.x - a[0].lower_.x;
	y0 = a[0].upper_.y - a[0].lower_.y;
	x1 = a[1].upper_.x - a[1].lower_.x;
	y1 = a[1].upper_.y - a[1].lower_.y;
	compensate(a0, x0, y0);
	compensate(a1, x1, y1);

	x_len = sqrt(x0*x0 + y0*y0)/sqrt(m[0][0]*m[0][0]+m[0][1]*m[0][1]);
	y_len = sqrt(x1*x1 + y1*y1)/sqrt(m[1][0]*m[1][0]+m[1][1]*m[1][1]);

	region->xalign_ = req->x.align;
	region->yalign_ = req->y.align;
	region->zalign_ = req->z.align;

	region->lower_.x = -region->xalign_ * x_len;
	region->lower_.y = -region->yalign_ * y_len;
	region->upper_.x = region->lower_.x + x_len;
	region->upper_.y = region->lower_.y + y_len;

	delta.x = (region->lower_.x+region->upper_.x)*0.5;
	delta.y = (region->lower_.y+region->upper_.y)*0.5;
	t->transform_vertex(delta);
	delta.x = center.x - delta.x;
	delta.y = center.y - delta.y;
    }
    return delta;
}

static void draw_handle_box(
    DrawingKit_ptr d, Painter_ptr po, Coord l, Coord b, Coord r, Coord t
) {
    po->current_color(black_color);
    po->push_matrix();
    po->current_matrix(Transform_var(d->identity_matrix()));

    Coord cx, cy;
    cx = l; cy = b;
    po->fill_rect(cx - half_dot, cy - half_dot, cx + half_dot, cy + half_dot); 

    cx = l; cy = (b+t)/2;
    po->fill_rect(cx - half_dot, cy - half_dot, cx + half_dot, cy + half_dot); 

    cx = l; cy = t;
    po->fill_rect(cx - half_dot, cy - half_dot, cx + half_dot, cy + half_dot); 

    cx = (l+r)*0.5; cy = t;
    po->fill_rect(cx - half_dot, cy - half_dot, cx + half_dot, cy + half_dot); 

    cx = r; cy = t;
    po->fill_rect(cx - half_dot, cy - half_dot, cx + half_dot, cy + half_dot); 

    cx = r; cy = (t+b)*0.5;
    po->fill_rect(cx - half_dot, cy - half_dot, cx + half_dot, cy + half_dot); 

    cx = r; cy = b;
    po->fill_rect(cx - half_dot, cy - half_dot, cx + half_dot, cy + half_dot); 

    cx = (r+l)*0.5; cy = b;
    po->fill_rect(cx - half_dot, cy - half_dot, cx + half_dot, cy + half_dot); 

    po->pop_matrix();
}

FigureKit* Manipulator::figures_;
WidgetKit* Manipulator::widgets_;
LayoutKit* Manipulator::layouts_;
DialogKit* Manipulator::dialogs_;
DrawingKit* Manipulator::drawings_;
MenuKit* Manipulator::menus_;

void Manipulator::init_kits (
    FigureKit* figures, WidgetKit* widgets, LayoutKit* layouts, DialogKit* dialogs,
    DrawingKit* drawings, MenuKit* menus    
) {
    figures_ = figures;
    widgets_ = widgets;
    layouts_ = layouts;
    dialogs_ = dialogs;
    drawings_ = drawings;
    menus_ = menus;
    black_color = drawings_->color_rgb(0.0, 0.0, 0.0);
}

implementPtrList(ManipList, Manipulator);

class ManipInfo {
public:
    ManipInfo(SelectInfo*);
    virtual ~ManipInfo();
public:
    ToolType tooltype_;
    Vertex center_;  // traversal independent center vertex
    Vertex pt_;  // traversal independent event position
    Coord ax_, ay_; // initial anchor position for multiple selections
    Coord tool_ax_, tool_ay_; // tool anchor position
    TransformImpl t_; // traversal transform inverted;
    Boolean obj_manip_;  // manipulate the actual object
    Boolean manipulating_; // block off command interpretation during manipulation

    SelectInfo* sinfo_;
};

ManipInfo::ManipInfo (SelectInfo* sinfo) {
    sinfo_ = sinfo->copy();
    tooltype_ = undefined_tool;
    obj_manip_ = false;
    manipulating_ = false;
}

ManipInfo::~ManipInfo () {
    delete sinfo_;
}

class Inserter : public GlyphVisitor {
public:
    Inserter(ManipList& m);
    Boolean visit(XfGlyph_ptr, XfGlyphOffset_ptr);
private:
    ManipList* m_;
};

Inserter::Inserter (ManipList& m) {
    m_ = &m;
}

Boolean Inserter::visit (XfGlyph_ptr, XfGlyphOffset_ptr go) {
    for (long i = 0; i < m_->count(); i++) {
        go->insert(m_->item(i));
    }
    return true;
}

//- Transformer*
//+ Transformer : Allocator
class Transformer : public Allocator {
public:
    ~Transformer();
    _Ix_InterfaceOps(Transformer*)
//+
public:
    Transformer(Transformer*);
    Transformer();
    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
    void extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r); //+ XfGlyph::extension
    void traverse(XfGlyphTraversal_ptr t); //+ XfGlyph::traverse
    XfTransform_ptr transformation(); //+ XfGlyph::transformation
    void child_allocate(XfGlyph::AllocationInfo& a);

    Region_ptr nat_allocation();
    void need_resize();
protected:
    TransformImpl tx_;
};

//+ Transformer(Transformer)
void* Transformer::_this_ptr() { return this; }
_Ix_extern__tid(Transformer);
void* Transformer::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(Transformer) == t) return this;
    void* p;
    if (p = XfGlyph::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(Transformer);
_Ix_ImplementGetTid(Transformer,_Ix__tid(Transformer))
_Ix_ImplementNarrow(Transformer*,Transformer,_Ix_NullParameter,Transformer,0)
_Ix_ImplementDuplicate(Transformer*,Transformer,_Ix_NullParameter,_Ix__tid(Transformer),0)
//+

Transformer::Transformer (Transformer* t) {
    tx_.copy(&t->tx_);
}
    
Transformer::Transformer () { }

Transformer::~Transformer () { }

void Transformer::need_resize () {
    Allocator::need_resize();
}

Region_ptr Transformer::nat_allocation () {
    return Region::_duplicate(&nat_);
}

//+ Transformer(XfGlyph::extension)
void Transformer::extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r) {
    Allocator::extension(a, r);
}

//+ Transformer(XfGlyph::request)
void Transformer::request(XfGlyph::Requisition& r) {
    Allocator::request(r);  // transform is handled by a layout manipulator
}

//+ Transformer(XfGlyph::traverse)
void Transformer::traverse(XfGlyphTraversal_ptr t) {
    Painter_var p = t->current_painter();
    p->push_matrix();
    p->premultiply(&tx_);
    Region_var a = t->allocation();
    if (is_not_nil(a)) {
	t->traverse_child(&offset_, a);
    } else {
        Allocator::traverse(t);
    }
    p->pop_matrix();
}

//+ Transformer(XfGlyph::transformation)
XfTransform_ptr Transformer::transformation() {
    return Transform::_duplicate(&tx_);
}

void Transformer::child_allocate(XfGlyph::AllocationInfo& a) {
    if (is_nil(a.transformation)) {
        a.transformation = new TransformImpl;
    }
    a.transformation->premultiply(&tx_);
    if (is_nil(a.allocation)) {
        Allocator::child_allocate(a);
    }
}

//+ Manipulator(Manipulator)
void* Manipulator::_this_ptr() { return this; }
_Ix_extern__tid(Manipulator);
void* Manipulator::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(Manipulator) == t) return this;
    void* p;
    if (p = XfGlyph::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(Manipulator);
_Ix_ImplementGetTid(Manipulator,_Ix__tid(Manipulator))
_Ix_ImplementNarrow(Manipulator*,Manipulator,_Ix_NullParameter,Manipulator,0)
_Ix_ImplementDuplicate(Manipulator*,Manipulator,_Ix_NullParameter,_Ix__tid(Manipulator),0)
//+

Manipulator::Manipulator (Manipulator* m) {
    selected_ = false;
    info_ = nil;
    if (is_not_nil(m)) {
        tx_.copy(&m->tx_);
        selected_ = m->selected_;
        if (m->info_ != nil) {
            info_ = new ManipInfo(m->info_->sinfo_->copy());
        }
    }
}
    
Manipulator::~Manipulator () {
    delete info_;
}
    
void Manipulator::info (Manipulator* m, ManipInfo* i) {
    m->info_ = i;
}

ManipInfo* Manipulator::info (Manipulator* m) {
    return m->info_;
}

void Manipulator::init_info(Tool& tool, FigViewer&, SelectInfo* si) {
    if (info_ != nil && info_->sinfo_ == si) {
	return;
    }
    delete info_;
    info_ = new ManipInfo(si);

    Tool* t = nil;
    if (is_not_nil(t = SelectTool::_narrow(&tool))) {
	info_->tooltype_ = choose_tool;
    } else if (is_not_nil(t = CreateTool::_narrow(&tool))) {
	info_->tooltype_ = create_tool;
    } else if (is_not_nil(t = MoveTool::_narrow(&tool))) {
	info_->tooltype_ = move_tool;
    } else if (is_not_nil(t = ScaleTool::_narrow(&tool))) {
	info_->tooltype_ = scale_tool;
    } else if (is_not_nil(t = RotateTool::_narrow(&tool))) {
	info_->tooltype_ = rotate_tool;
    } else if (is_not_nil(t = ResizeTool::_narrow(&tool))) {
	info_->tooltype_ = resize_tool;
    } else if (is_not_nil(t = NarrowTool::_narrow(&tool))) {
	info_->tooltype_ = narrow_tool;
    } else {
	info_->tooltype_ = undefined_tool;
    }
    CORBA::release(t);
}

Viewer_ptr Manipulator::viewer() {
    return nil;
}

Boolean Manipulator::grasp(
    Tool& tool, FigViewer& fv, SelectInfo* si, Event_ptr e
) {
    Boolean ok = false;
    
    init_info(tool, fv, si);

    if (info_->tooltype_ != undefined_tool) {
        ok = true;
        XfGlyph_var g = body();
        info_->obj_manip_ = e->modifier_is_down(Event::control);
        tool.anchor(info_->tool_ax_, info_->tool_ay_);
        info_->t_.copy(&si->t_);
        if (info_->obj_manip_ && info_->tooltype_ != resize_tool) {
            SelectCmd s;
            CmdVisitor cv(&s);
            cv.visit_parents(g);
	    info_->t_.premultiply(&tx_);
        }

        Vertex lower, upper;
        XfGlyph::AllocationInfo a;
        RegionImpl r;

        if (info_->tooltype_ == create_tool) {
            info_->ax_ = e->pointer_x();
            info_->ay_ = e->pointer_y();
            info_->pt_.x = e->pointer_x();
            info_->pt_.y = e->pointer_y();
            info_->pt_.z = 0;
        } else {
            a.transformation = new TransformImpl;
            a.transformation->copy(&si->t_);

            extension(a, &r);
            r.bounds(lower, upper);
            info_->ax_ = (lower.x+upper.x)*0.5;
            info_->ay_ = (lower.y+upper.y)*0.5;
            info_->pt_.x = e->pointer_x() + info_->ax_ - info_->tool_ax_;
            info_->pt_.y = e->pointer_y() + info_->ay_ - info_->tool_ay_;
            info_->pt_.z = 0;
        }
        info_->t_.invert();
        info_->t_.transform_vertex(info_->pt_);

        if (is_not_nil(info_->sinfo_->a_)) {
            r.copy(info_->sinfo_->a_);
            if (info_->obj_manip_) {
                TransformImpl t;
                t.copy(&tx_);
                t.invert();
                r.apply_transform(&t);
            }
        } else {
            if (info_->obj_manip_) {
                local_extension(XfGlyph_var(body()), &r);
            } else {
                Boolean selected = selected_;
                selected_ = false;
                local_extension(this, &r);
                selected_ = selected;
            }
        }
        r.bounds(lower, upper);
        info_->center_.x = (lower.x+upper.x)*0.5;
        info_->center_.y = (lower.y+upper.y)*0.5;
        info_->center_.z = 0.0;

        XfGlyph::AllocationInfoSeq alist;
        if (!info_->obj_manip_) {
            allocations(alist);
        }
        if (info_->tooltype_ == create_tool) {
            Vertex delta;
 
            delta.x = info_->pt_.x - (lower.x+upper.x)*0.5;
            delta.y = info_->pt_.y - (lower.y+upper.y)*0.5;
            delta.z = 0.0;
 
            if (info_->obj_manip_) {
                parents_need_redraw(g);
                _translate(Transform_var(g->transformation()), delta);
                parents_need_redraw(g);
                g->need_resize();
            } else {
                glyph_need_redraw(this, alist);
                _translate(&tx_, delta);
                glyph_need_redraw(this, alist);
                if (is_not_nil(info_->sinfo_->a_)) {
                    compose(this);
                }
                need_resize();
            }
            info_->center_.x += delta.x;
            info_->center_.y += delta.y;
 
        } else if (info_->tooltype_ == choose_tool) {
            SelectCmd selectcmd(&fv);
            execute(&selectcmd);

        } else if (info_->tooltype_ == scale_tool) {
            Vertex scaler;
            scaler.x = Math::abs(
                (info_->pt_.x-info_->center_.x)*2.0/(upper.x-lower.x)
            );
            scaler.y = Math::abs(
                (info_->pt_.y-info_->center_.y)*2.0/(upper.y-lower.y)
            );
            scaler.z = 1.0;

            if (info_->obj_manip_) {
                parents_need_redraw(g);
                _scale(
		    Transform_var(g->transformation()), info_->center_, scaler
		);
                parents_need_redraw(g);
                g->need_resize();
            } else {
                glyph_need_redraw(this, alist);
                _scale(&tx_, info_->center_, scaler);
                glyph_need_redraw(this, alist);
                if (is_not_nil(info_->sinfo_->a_)) {
                    compose(this);
                }
                need_resize();
            }
	} else if (info_->tooltype_ == narrow_tool) {
	    NarrowTool* nt = NarrowTool::_narrow(&tool);
	    NarrowCmd narrowcmd(&fv, nt->delta());
	    execute(&narrowcmd);
	    CORBA::release(nt);
        }
        info_->manipulating_ = true;
    } else {
        delete info_;
        info_ = nil;
    }
    return ok;
}

Boolean Manipulator::manipulate (Event_ptr e) {
    switch (e->type()) {
    case Event::motion:
    case Event::enter:
    case Event::leave:
	break;
    case Event::focus_in:
    case Event::focus_out:
	return true;
    default:
	return false;
    }

    Vertex pt;
    pt.x = e->pointer_x() + info_->ax_ - info_->tool_ax_;
    pt.y = e->pointer_y() + info_->ay_ - info_->tool_ay_;
    pt.z = 0;
    info_->t_.transform_vertex(pt);

    XfGlyph_var g = body();
    XfGlyph::AllocationInfoSeq a;
    if (!info_->obj_manip_) {
        allocations(a);
    }
    switch (info_->tooltype_) {
    case create_tool:
    case move_tool:
        {
            Vertex orig = pt;

            pt.x -= info_->pt_.x;
            pt.y -= info_->pt_.y;

            if (info_->obj_manip_) {
                parents_need_redraw(g);
                _translate(Transform_var(g->transformation()), pt);
                parents_need_redraw(g);
                g->need_resize();
            } else {
                glyph_need_redraw(this, a);
                _translate(&tx_, pt);
                glyph_need_redraw(this, a);
                if (is_not_nil(info_->sinfo_->a_)) {
                    compose(this);
                }
                need_resize();
            }
            info_->center_.x += pt.x;
            info_->center_.y += pt.y;
            info_->pt_ = orig;
        }
        break;
    case rotate_tool:
        {
            double pi = 3.14159;
            double den = pt.x-info_->center_.x;
            if (Math::abs(den) < tol) {
                return true;
            }
            double cur = atan((pt.y-info_->center_.y)/den)/pi*180.0;
            double last = atan(
                (info_->pt_.y-info_->center_.y)/
                (info_->pt_.x-info_->center_.x)
            )/pi*180.0;
            if ((pt.x-info_->center_.x) < 0.0) {
                cur += 180.0;
            }
            if ((info_->pt_.x-info_->center_.x) < 0.0) {
                last += 180.0;
            }
            if (info_->obj_manip_) { 
                parents_need_redraw(g);
                _rotate(
                    Transform_var(g->transformation()),
		    cur-last, Z_axis, info_->center_
                );
                parents_need_redraw(g);
                g->need_resize();
            } else {
                glyph_need_redraw(this, a);
                _rotate(&tx_, cur-last, Z_axis, info_->center_);
                glyph_need_redraw(this, a);
                if (is_not_nil(info_->sinfo_->a_)) {
                    compose(this);
                }
                need_resize();
            }
            info_->pt_ = pt;
        }
        break;
    case scale_tool:
        {
            Vertex scaler;
            double den_x = info_->pt_.x-info_->center_.x;
            double den_y = info_->pt_.y-info_->center_.y;
            double num_x = pt.x-info_->center_.x;
            double num_y = pt.y-info_->center_.y;
            if (Math::abs(num_x) < tol || Math::abs(num_y) < tol) {
                return true;
            }
            scaler.x = num_x/den_x;
            scaler.y = num_y/den_y;
            scaler.z = 1.00;

            if (info_->obj_manip_) {
                parents_need_redraw(g);
                _scale(
		    Transform_var(g->transformation()), info_->center_, scaler
		);
                parents_need_redraw(g);
                g->need_resize();
            } else {
                glyph_need_redraw(this, a);
                _scale(&tx_, info_->center_, scaler);
                glyph_need_redraw(this, a);
                if (is_not_nil(info_->sinfo_->a_)) {
                    compose(this);
                }
                need_resize();
            }
            info_->pt_ = pt;
        }
        break;
    case choose_tool:
        break;
    default:
        break;
    }
    return true;
}

FCommand* Manipulator::effect (Event_ptr) {
    if (info_ != nil) {
        info_->manipulating_ = false;
        if (info_->obj_manip_) {
            selected_ = false;
            XfGlyph_var g = body();
            SelectCmd s(nil, false);
            CmdVisitor cv(&s);
            cv.visit_parents(g);
            selected_ = true;
        }
    }
    return nil;  // commands for undo and redo not supported yet
}

void Manipulator::draw_handles (XfGlyphTraversal_ptr gt) {
    if (selected_) {
        selected_ = false; // turn this off momentarily since it may affect
                           // bounding box computation
        Painter_var po = gt->current_painter();
        RegionImpl bbox;

        XfGlyph::AllocationInfo ga;
        ga.transformation = new TransformImpl;
        ga.transformation->copy(Transform_var(po->current_matrix()));
	Region_var a = gt->allocation();
        if (is_not_nil(a)) {
            ga.allocation = new RegionImpl;
            ga.allocation->copy(a);
        }
        extension(ga, &bbox);
        draw_handle_box(
            drawings_, po, bbox.lower_.x, bbox.lower_.y, bbox.upper_.x, bbox.upper_.y
        );
        selected_ = true;
    }
}

void Manipulator::execute (FCommand* cmd) {
    if (info_ != nil && info_->manipulating_) {
        return;
    }
    FigViewer* v = cmd->figviewer();
    SelectCmd* selectcmd = SelectCmd::_narrow(cmd);
    if (is_not_nil(selectcmd)) {
        Boolean s = selectcmd->selected();
        if (s != selected_) {
	    selected_ = true;
	    need_redraw();
            selected_ = s;
            if (!selected_) {
                delete info_;
                info_ = nil;
            }
        }
        if (v != nil) {
            if (selected_) {
                v->selection()->add(this);
            } else {
                v->selection()->remove(this);
            }
        }
	CORBA::release(selectcmd);
        return;
    }
    UngroupCmd* ungroupcmd = UngroupCmd::_narrow(cmd);
    if (is_not_nil(ungroupcmd)) {
        XfGlyph_var g = body();
        Counter counter;
        counter.visit_children(g);
        if (counter.count() == 0) {
            return;
        }

        TransformImpl t;
        t.copy(&tx_);
        t.premultiply(Transform_var(g->transformation()));
        ManipCopier mc;
        mc.visit_children(g);

        ManipList* ml = mc.manipulators();
        for (Long i = 0; i < ml->count(); i++) {
            Manipulator* man = ml->item(i);
            Transform_var tr = man->transformation();
            tr->postmultiply(&t);
        }
        Inserter inserter(*ml);
        inserter.visit_parents(this);
        SelectCmd selectcmd(v);
        for (i = 0; i < ml->count(); i++) {
            Manipulator* man = ml->item(i);
            man->execute(&selectcmd);
            //man->info_ = new ManipInfo(info_->sinfo_->copy());
        }
        SelectCmd unselectcmd(v, false);
        execute(&unselectcmd);
        Remover remover;
        remover.visit_parents(this);
	CORBA::release(ungroupcmd);
        return;
    }
    DeleteCmd* deletecmd = DeleteCmd::_narrow(cmd);
    if (is_not_nil(deletecmd)) {
        XfGlyph_var g = body();
        if (info_ != nil && info_->obj_manip_) {
            delete info_;
            info_ = nil;
            CmdVisitor cv(deletecmd);
            cv.visit_parents(g);
        } else {
            SelectCmd unselectcmd(v, false);
            execute(&unselectcmd);
            Remover remover;
            remover.visit_parents(this);
        }
	CORBA::release(deletecmd);
        return;
    }   
    InstanceCmd* instancecmd = InstanceCmd::_narrow(cmd);
    CopyCmd* copycmd = CopyCmd::_narrow(cmd);
    if (is_not_nil(instancecmd) || is_not_nil(copycmd)) {
        Manipulator* man;
        if (is_not_nil(instancecmd)) {
            man = shallow_copy();
        } else {
            CopyCmd::glyphmap_.clear();
            man = deep_copy();
        }
        SelectCmd unselectcmd(v, false);
        execute(&unselectcmd);

        Transform_var tr = man->transformation();

        Vertex delta;
        delta.x = 4.0;
        delta.y = 4.0;
        delta.z = 0.0;

        tr->translate(delta);
        Appender app(man);
        app.visit_parents(this);

        SelectCmd selectcmd(v);
        man->execute(&selectcmd);
	man->need_redraw();
	CORBA::release(instancecmd);
	CORBA::release(copycmd);
        return;
    }
    NarrowCmd* narrowcmd = NarrowCmd::_narrow(cmd);
    if (is_not_nil(narrowcmd)) {
	v->narrow(narrowcmd->delta());
	CORBA::release(narrowcmd);
        return;
    }
    WidenCmd* widencmd = WidenCmd::_narrow(cmd);
    if (is_not_nil(widencmd)) {
	WidenGlyphList* glist = v->widen_list();
	Manipulator* m = Manipulator::_narrow(glist->item(glist->count() - 1));
	v->narrow(widencmd->delta());
	if (is_not_nil(m)) {
	    SelectCmd selectcmd(v);
	    m->execute(&selectcmd);
            CORBA::release(m);
	}
	CORBA::release(widencmd);
    }
    SelectInfoCmd* scmd = SelectInfoCmd::_narrow(cmd);
    if (is_not_nil(scmd)) {
        scmd->select_info(info_->sinfo_);
	CORBA::release(scmd);
        return;
    }
    GroupCmd* groupcmd = GroupCmd::_narrow(cmd);
    if (is_not_nil(groupcmd)) {
        Selection* sel = v->selection()->copy();
        long n = sel->count();
        if (n < 1) {
            delete sel;
	    CORBA::release(groupcmd);
            return;
        }
        Glyph* pf = figures_->group();
        Manipulator* pm = new PolyManip(pf);
        CORBA::release(pf);

        SelectCmd unselectcmd(v, false);
        SelectCmd selectcmd(v);
        pm->execute(&selectcmd);

        Manipulator* m0 = sel->item(0);
        pm->info_ = m0->info_;
	m0->info_ = nil;
        OffsetVisitor ov;
        ov.visit_parents(m0);

        Remover* remover = new Remover;
        for (long i = 0; i < sel->count(); i++) {
            Manipulator* m = sel->item(i);
            m->execute(&unselectcmd);
            remover->visit_parents(m);
            pf->append(m);
        }
        delete remover;
        for (long j = 0; j < ov.offset_count(); j++) {
            ov.offset(j)->parent()->append(pm);
        }
        pm->need_redraw();
	CORBA::release(groupcmd);
        return;
    }
}

void Manipulator::unexecute (FCommand*) {}

//+ Manipulator(XfGlyph::draw)
void Manipulator::draw(XfGlyphTraversal_ptr t) {
    draw_handles(t);
}

//+ Manipulator(XfGlyph::traverse)
void Manipulator::traverse(XfGlyphTraversal_ptr t) {
    Painter_var p = t->current_painter();
    if (is_not_nil(p)) {
        if (tx_.identity()) {
            MonoGlyph::traverse(t);
        } else {
            p->push_matrix();
            p->premultiply(&tx_);
            MonoGlyph::traverse(t);
            p->pop_matrix();
        }
    }
    if (t->op() == XfGlyphTraversal::draw) {
        draw(t);
    }
}

//+ Manipulator(XfGlyph::extension)
void Manipulator::extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r) {
    MonoGlyph::extension(a, r);
    if (selected_) {
        RegionImpl region;
        region.copy(r);
        adjust_dotted_region(region);
        r->copy(&region);
    }
}

//+ Manipulator(XfGlyph::request)
void Manipulator::request(XfGlyph::Requisition& r) {
    MonoGlyph::request(r);
    fixed_transform_requisition(r, &tx_);
}

//+ Manipulator(XfGlyph::transformation)
XfTransform_ptr Manipulator::transformation() {
    return Transform::_duplicate(&tx_);
}

void Manipulator::child_allocate(XfGlyph::AllocationInfo& a) {
    if (is_nil(a.transformation)) {
	a.transformation = new TransformImpl;
    }
    a.transformation->premultiply(&tx_);
}

//+ FigureManip(FigureManip)
void* FigureManip::_this_ptr() { return this; }
_Ix_extern__tid(FigureManip);
void* FigureManip::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(FigureManip) == t) return this;
    void* p;
    if (p = Manipulator::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(FigureManip);
_Ix_ImplementGetTid(FigureManip,_Ix__tid(FigureManip))
_Ix_ImplementNarrow(FigureManip*,FigureManip,_Ix_NullParameter,FigureManip,0)
_Ix_ImplementDuplicate(FigureManip*,FigureManip,_Ix_NullParameter,_Ix__tid(FigureManip),0)
//+

FigureManip::FigureManip (FigureManip* m) : Manipulator (m) {
    figure_ = nil;
    if (is_not_nil(m) && is_not_nil(m->figure_)) {
        figure_body(m->figure_);
    }
}

Figure* FigureManip::figure_body () { return figure_; }

void FigureManip::figure_body (Figure* f) {
    // assumes ref counting
    XfGlyph::_duplicate(f);
    CORBA::release(figure_);
    figure_ = f;
    if (is_not_nil(figure_)) {
        body(figure_);
    }
}

FigureManip::~FigureManip () {
    CORBA::release(figure_);
}

Boolean FigureManip::grasp (
    Tool& tool, FigViewer& fv, SelectInfo* si, Event_ptr e
) {
    Boolean ok = Manipulator::grasp(tool, fv, si, e);
    if (info_->tooltype_ == create_tool) {
        Transform_ptr tr = &info_->sinfo_->t_;
        info_->t_.copy(tr);
        info_->t_.premultiply(&tx_);
        XfGlyph_var g = body();
        info_->t_.premultiply(Transform_var(g->transformation()));
        info_->t_.invert();

        info_->ax_ = e->pointer_x();
        info_->ay_ = e->pointer_y();
        info_->pt_.x = e->pointer_x();
        info_->pt_.y = e->pointer_y();
        info_->pt_.z = 0;
        info_->t_.transform_vertex(info_->pt_);
    }
    return ok;
}

//+ VertexManip(VertexManip)
void* VertexManip::_this_ptr() { return this; }
_Ix_extern__tid(VertexManip);
void* VertexManip::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(VertexManip) == t) return this;
    void* p;
    if (p = FigureManip::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(VertexManip);
_Ix_ImplementGetTid(VertexManip,_Ix__tid(VertexManip))
_Ix_ImplementNarrow(VertexManip*,VertexManip,_Ix_NullParameter,VertexManip,0)
_Ix_ImplementDuplicate(VertexManip*,VertexManip,_Ix_NullParameter,_Ix__tid(VertexManip),0)
//+

declareList(VertexList,Vertex)
implementList(VertexList,Vertex)

VertexManip::VertexManip (VertexManip* m) : FigureManip (m) {
    vlist_ = new VertexList;
    if (is_not_nil(m)) {
        for (long i = 0; i < m->vlist_->count(); i++) {
            vlist_->append(m->vlist_->item(i));
        }
        vmin_ = m->vmin_;
        vmax_ = m->vmax_;
    } else {
	vmin_.x = Coord(0);
	vmin_.y = Coord(0);
	vmin_.z = Coord(0);
	vmax_.x = Coord(0);
	vmax_.y = Coord(0);
	vmax_.z = Coord(0);
    }
}

VertexManip::~VertexManip () { delete vlist_; }

//+ VertexManip(XfGlyph::extension)
void VertexManip::extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r) {
    if (selected_ && vlist_->count() > 0) {
        TransformImpl t;
        RegionImpl region;
        if (is_not_nil(a.transformation)) {
            t.copy(a.transformation);
        }
        t.premultiply(&tx_);
        if (is_not_nil(Transform_var(figure_->transformation()))) {
            t.premultiply(Transform_var(figure_->transformation()));
        }
        region.defined_ = true;
	region.lower_ = vmin_;
	region.upper_ = vmax_;
	region.xalign_ = region.yalign_ = region.zalign_ = 0;
	region.apply_transform(&t);

        adjust_dotted_region(region);
        r->merge_union(&region);
    } else {
        Manipulator::extension(a, r);
    }
}

Boolean VertexManip::grasp (
    Tool& tool, FigViewer& fv, SelectInfo* si, Event_ptr e
) {
    Boolean ok = FigureManip::grasp(tool, fv, si, e);
    if (info_->tooltype_ == create_tool) {
        vlist_->remove_all();
        vlist_->append(info_->pt_);
        vlist_->append(info_->pt_);
    }
    return ok;
}

Boolean VertexManip::manipulate (Event_ptr e) {
    if (e->type() == Event::down && e->pointer_button() == 3) {
        return false;
    }
    Boolean ok = true;
    if (info_->tooltype_ == create_tool) {
        Vertex pt;
        XfGlyph::AllocationInfoSeq a;
        allocations(a);
        glyph_need_redraw(this, a);

        pt.x = e->pointer_x(); pt.y = e->pointer_y(); pt.z = 0;
        info_->t_.transform_vertex(pt);
        if (e->type() == Event::motion) {
            Vertex& last = vlist_->item_ref(vlist_->count()-1);
            last = pt;
        } else if (e->type() == Event::down) {
            if (e->pointer_button() == 1) {
                vlist_->append(pt);
            } else if (e->pointer_button() == 2) {
                if (vlist_->count() > 2) {
                    vlist_->remove(vlist_->count()-1);
                    Vertex& last = vlist_->item_ref(vlist_->count()-1);
                    last = pt;
                }
            }
        }
        recompute_shape();
        info_->pt_ = pt;

        glyph_need_redraw(this, a);
        if (is_not_nil(info_->sinfo_->a_)) {
            compose(this);
        }
        need_resize();
    } else {
        ok = Manipulator::manipulate(e);
    }
    return ok;
}

void VertexManip::draw_handles (XfGlyphTraversal_ptr gt) {
    if (selected_) {
        Painter_var po = gt->current_painter();
        po->current_color(black_color);
        Transform_var tx = figure_->transformation();
        po->push_matrix();
        po->premultiply(&tx_);
        po->premultiply(tx);
        TransformImpl ti;
        ti.copy(Transform_var(po->current_matrix()));
        po->current_matrix(Transform_var(drawings_->identity_matrix()));

        for (long i = 0; i < vlist_->count(); i++) {
            Vertex vtex = vlist_->item(i);
            ti.transform_vertex(vtex);
            po->fill_rect(
		vtex.x - half_dot, vtex.y - half_dot,
		vtex.x + half_dot, vtex.y + half_dot
	    );
        }
        po->pop_matrix();
    }
}

void VertexManip::recompute_shape () {
    figure_->reset();
    vmin_.x = FLT_MAX;
    vmin_.y = FLT_MAX;
    vmax_.x = -FLT_MAX;
    vmax_.y = -FLT_MAX;
    for (long i = 0; i < vlist_->count(); i++) {
        figure_->add_point(vlist_->item(i).x, vlist_->item(i).y);
        vmin_.x = Math::min(vmin_.x, vlist_->item(i).x);
        vmax_.x = Math::max(vmax_.x, vlist_->item(i).x);
        vmin_.y = Math::min(vmin_.y, vlist_->item(i).y);
        vmax_.y = Math::max(vmax_.y, vlist_->item(i).y);
    }
}

//+ LineManip(LineManip)
void* LineManip::_this_ptr() { return this; }
_Ix_extern__tid(LineManip);
void* LineManip::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(LineManip) == t) return this;
    void* p;
    if (p = VertexManip::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(LineManip);
_Ix_ImplementGetTid(LineManip,_Ix__tid(LineManip))
_Ix_ImplementNarrow(LineManip*,LineManip,_Ix_NullParameter,LineManip,0)
_Ix_ImplementDuplicate(LineManip*,LineManip,_Ix_NullParameter,_Ix__tid(LineManip),0)
//+

LineManip::LineManip (LineManip* m) : VertexManip (m) {
    if (is_nil(m) || is_nil(m->figure_)) {
        Figure* f = figures_->general_figure(
            FigureKit::stroke, FigureStyle_var(figures_->default_style()),
	    false, false
        );
        f->add_point(0, 0);
        figure_body(f);
    }
}

LineManip::~LineManip () {}

Manipulator* LineManip::shallow_copy () {
    LineManip* manip = new LineManip(this);
    return manip;
}

Manipulator* LineManip::deep_copy () {
    Figure* copy = Figure::_narrow(CopyCmd::glyphmap_.find(figure_));
    if (is_nil(copy)) {
        Figure* f = figure_->copy_figure();
        CopyCmd::glyphmap_.map(figure_, f);
        LineManip* manip =  new LineManip(this);
        manip->figure_body(f);
        CORBA::release(f);
        return manip;
    } else {
        LineManip* manip = new LineManip(this);
        manip->figure_body(copy);
        return manip;
    }
}

Boolean LineManip::manipulate (Event_ptr e) {
    switch (e->type()) {
    case Event::motion:
    case Event::enter:
    case Event::leave:
    case Event::up:
	break;
    case Event::focus_in:
    case Event::focus_out:
	return true;
    default:
	return false;
    }
    Boolean ok = true;
    if (info_->tooltype_ == create_tool) {
        Vertex pt;
        XfGlyph::AllocationInfoSeq a;
        allocations(a);
        glyph_need_redraw(this, a);

        pt.x = e->pointer_x(); pt.y = e->pointer_y(); pt.z = 0;
        info_->t_.transform_vertex(pt);
        Vertex& last = vlist_->item_ref(1);
        last = pt;
        if (e->type() == Event::up) {
            ok = false;
        }
        recompute_shape();
        info_->pt_ = pt;

        glyph_need_redraw(this, a);
        if (is_not_nil(info_->sinfo_->a_)) {
            compose(this);
        }
        need_resize();
    } else {
        ok = Manipulator::manipulate(e);
    }
    return ok;
}

//+ RectManip(RectManip)
void* RectManip::_this_ptr() { return this; }
_Ix_extern__tid(RectManip);
void* RectManip::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(RectManip) == t) return this;
    void* p;
    if (p = VertexManip::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(RectManip);
_Ix_ImplementGetTid(RectManip,_Ix__tid(RectManip))
_Ix_ImplementNarrow(RectManip*,RectManip,_Ix_NullParameter,RectManip,0)
_Ix_ImplementDuplicate(RectManip*,RectManip,_Ix_NullParameter,_Ix__tid(RectManip),0)
//+

RectManip::RectManip (RectManip* m) : VertexManip (m) {
    if (is_nil(m) || is_nil(m->figure_)) {
        Figure* f = figures_->general_figure(
            FigureKit::fill_stroke, FigureStyle_var(figures_->default_style()),
            true, false
        );
        f->add_point(0, 0);
        figure_body(f);
    }
}

RectManip::~RectManip () {}

Manipulator* RectManip::shallow_copy () {
    RectManip* manip = new RectManip(this);
    return manip;
}

Manipulator* RectManip::deep_copy () {
    Figure* copy = Figure::_narrow(CopyCmd::glyphmap_.find(figure_));
    if (is_nil(copy)) {
        Figure* f = figure_->copy_figure();
        CopyCmd::glyphmap_.map(figure_, f);
        RectManip* manip = new RectManip(this);
        manip->figure_body(f);
        CORBA::release(f);
        return manip;
    } else {
        RectManip* manip = new RectManip(this);
        manip->figure_body(copy);
        return manip;
    }
}

Boolean RectManip::manipulate (Event_ptr e) {
    switch (e->type()) {
    case Event::motion:
    case Event::enter:
    case Event::leave:
    case Event::up:
	break;
    case Event::focus_in:
    case Event::focus_out:
	return true;
    default:
	return false;
    }
    Boolean ok = true;
    if (info_->tooltype_ == create_tool) {
        Vertex pt, orig;
        XfGlyph::AllocationInfoSeq a;
        allocations(a);
        glyph_need_redraw(this, a);

        pt.x = e->pointer_x(); pt.y = e->pointer_y(); pt.z = 0;
        info_->t_.transform_vertex(pt);
        orig = vlist_->item(0);
        vlist_->remove_all();
        if (e->modifier_is_down(Event::shift)) {
            if (Math::abs(pt.x-orig.x) > Math::abs(pt.y-orig.y)) {
                if (
                    (pt.y-orig.y) > 0 && (pt.x-orig.x) < 0 ||
                    (pt.y-orig.y) < 0 && (pt.x-orig.x) > 0
                ) {
                    pt.y = -pt.x;
                } else {
                    pt.y = pt.x;
                }
            } else {
                if (
                    (pt.x-orig.x) > 0 && (pt.y-orig.y) < 0 ||
                    (pt.x-orig.x) < 0 && (pt.y-orig.y) > 0
                ) {
                    pt.x = -pt.y;
                } else {
                    pt.x = pt.y;
                }
            }
        }
        Vertex v = orig;
        vlist_->append(v);
        v.y = pt.y;
        vlist_->append(v);
        v.x = pt.x;
        vlist_->append(v);
        v.y = orig.y;
        vlist_->append(v);
        if (e->type() == Event::up) {
            ok = false;
        }
        recompute_shape();
        info_->pt_ = pt;

        glyph_need_redraw(this, a);
        if (is_not_nil(info_->sinfo_->a_)) {
            compose(this);
        }
        need_resize();
    } else {
        ok = Manipulator::manipulate(e);
    }
    return ok;
}

//+ EllipseManip(EllipseManip)
void* EllipseManip::_this_ptr() { return this; }
_Ix_extern__tid(EllipseManip);
void* EllipseManip::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(EllipseManip) == t) return this;
    void* p;
    if (p = VertexManip::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(EllipseManip);
_Ix_ImplementGetTid(EllipseManip,_Ix__tid(EllipseManip))
_Ix_ImplementNarrow(EllipseManip*,EllipseManip,_Ix_NullParameter,EllipseManip,0)
_Ix_ImplementDuplicate(EllipseManip*,EllipseManip,_Ix_NullParameter,_Ix__tid(EllipseManip),0)
//+

EllipseManip::EllipseManip (EllipseManip* m) : VertexManip (m) {
    if (is_nil(m) || is_nil(m->figure_)) {
        Figure* f = figures_->general_figure(
            FigureKit::fill_stroke, figures_->default_style(), true, true
        );
        f->add_point(0, 0);
        figure_body(f);
    }
}

EllipseManip::~EllipseManip () {}

//+ EllipseManip(XfGlyph::extension)
void EllipseManip::extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r) {
    Manipulator::extension(a, r);
}

Manipulator* EllipseManip::shallow_copy () {
    EllipseManip* manip = new EllipseManip(this);
    return manip;
}

Manipulator* EllipseManip::deep_copy () {
    Figure* copy = Figure::_narrow(CopyCmd::glyphmap_.find(figure_));
    if (is_nil(copy)) {
        Figure* f = figure_->copy_figure();
        CopyCmd::glyphmap_.map(figure_, f);
        EllipseManip* manip = new EllipseManip(this);
        manip->figure_body(f);
        CORBA::release(f);
        return manip;
    } else {
        EllipseManip* manip = new EllipseManip(this);
        manip->figure_body(copy);
        return manip;
    }
}

static const double p0 = 1.00000000;
static const double p1 = 0.89657547;   // cos 30 * sqrt(1 + tan 15 * tan 15)
static const double p2 = 0.70710678;   // cos 45
static const double p3 = 0.51763809;   // cos 60 * sqrt(1 + tan 15 * tan 15)
static const double p4 = 0.26794919;   // tan 15

Boolean EllipseManip::manipulate (Event_ptr e) {
    switch (e->type()) {
    case Event::motion:
    case Event::enter:
    case Event::leave:
    case Event::up:
	break;
    case Event::focus_in:
    case Event::focus_out:
	return true;
    default:
	return false;
    }
    Boolean ok = true;
    if (info_->tooltype_ == create_tool) {
        Vertex pt;
        XfGlyph::AllocationInfoSeq a;
        allocations(a);
        glyph_need_redraw(this, a);

        pt.x = e->pointer_x(); pt.y = e->pointer_y(); pt.z = 0;
        info_->t_.transform_vertex(pt);
        Vertex& last = vlist_->item_ref(1);
        Vertex& orig = vlist_->item_ref(0);
        if (e->type() == Event::up) {
            ok = false;
        }
        Coord r1 = Math::abs(pt.x - orig.x);
        Coord r2 = Math::abs(pt.y - orig.y);
        if (e->modifier_is_down(Event::shift)) {
            if (r1 > r2) {
                r2 = r1;
            } else {
                r1 = r2;
            }
            pt.x = r1 + orig.x;
            pt.y = r2 + orig.y;
        }
        last = pt;
        recompute_shape();
        info_->pt_ = pt;

        glyph_need_redraw(this, a);
        if (is_not_nil(info_->sinfo_->a_)) {
            compose(this);
        }
        need_resize();
    } else {
        ok = Manipulator::manipulate(e);
    }
    return ok;
}

void EllipseManip::recompute_shape () {
    figure_->reset();

    Vertex& pt = vlist_->item_ref(1);
    Vertex& orig = vlist_->item_ref(0);

    Coord r1 = pt.x - orig.x;
    Coord r2 = pt.y - orig.y;

    double px0 = p0 * r1, py0 = p0 * r2;
    double px1 = p1 * r1, py1 = p1 * r2;
    double px2 = p2 * r1, py2 = p2 * r2;
    double px3 = p3 * r1, py3 = p3 * r2;
    double px4 = p4 * r1, py4 = p4 * r2;
    
    figure_->add_point(orig.x + r1, orig.y);
    figure_->add_curve(
        orig.x + px2, orig.y + py2, orig.x + px0, 
        orig.y + py4, orig.x + px1, orig.y + py3
    );
    figure_->add_curve(
        orig.x, orig.y + r2, orig.x + px3, 
        orig.y + py1, orig.x + px4, orig.y + py0
    );
    figure_->add_curve(
        orig.x - px2, orig.y + py2, orig.x - px4, 
        orig.y + py0, orig.x - px3, orig.y + py1
    );
    figure_->add_curve(
        orig.x - r1, orig.y, orig.x - px1, 
        orig.y + py3, orig.x - px0, orig.y + py4
    );
    figure_->add_curve(
        orig.x - px2, orig.y - py2, orig.x - px0, 
        orig.y - py4, orig.x - px1, orig.y - py3
    );
    figure_->add_curve(
        orig.x, orig.y - r2, 
        orig.x - px3, orig.y - py1, orig.x - px4, orig.y - py0
    );
    figure_->add_curve(
        orig.x + px2, orig.y - py2, orig.x + px4,
        orig.y - py0, orig.x + px3, orig.y - py1
    );
    figure_->add_curve(
        orig.x + r1, orig.y, orig.x + px1, orig.y - py3, 
        orig.x + px0, orig.y - py4
    );
}

void EllipseManip::draw_handles (XfGlyphTraversal_ptr gt) {
    Manipulator::draw_handles(gt);
}

//+ Open_BSplineManip(Open_BSplineManip)
void* Open_BSplineManip::_this_ptr() { return this; }
_Ix_extern__tid(Open_BSplineManip);
void* Open_BSplineManip::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(Open_BSplineManip) == t) return this;
    void* p;
    if (p = VertexManip::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(Open_BSplineManip);
_Ix_ImplementGetTid(Open_BSplineManip,_Ix__tid(Open_BSplineManip))
_Ix_ImplementNarrow(Open_BSplineManip*,Open_BSplineManip,_Ix_NullParameter,Open_BSplineManip,0)
_Ix_ImplementDuplicate(Open_BSplineManip*,Open_BSplineManip,_Ix_NullParameter,_Ix__tid(Open_BSplineManip),0)
//+

Open_BSplineManip::Open_BSplineManip (Open_BSplineManip* m) : VertexManip (m) {
    if (is_nil(m) || is_nil(m->figure_)) {
        Figure* f = figures_->general_figure(
            FigureKit::stroke, figures_->default_style(), false, true
        );
        f->add_point(0, 0);
        figure_body(f);
    }
}

Open_BSplineManip::~Open_BSplineManip () {}

Manipulator* Open_BSplineManip::shallow_copy () {
    Open_BSplineManip* manip = new Open_BSplineManip(this);
    return manip;
}

Manipulator* Open_BSplineManip::deep_copy () {
    Figure* copy = Figure::_narrow(CopyCmd::glyphmap_.find(figure_));
    if (is_nil(copy)) {
        Figure* f = figure_->copy_figure();
        CopyCmd::glyphmap_.map(figure_, f);
        Open_BSplineManip* manip = new Open_BSplineManip(this);
        manip->figure_body(f);
        CORBA::release(f);
        return manip;
    } else {
        Open_BSplineManip* manip = new Open_BSplineManip(this);
        manip->figure_body(copy);
        return manip;
    }
}

void Open_BSplineManip::recompute_shape () {
    long n = vlist_->count();
    figure_->reset();
    figure_->Bspline_move_to(
        vlist_->item(0).x, vlist_->item(0).y, 
        vlist_->item(0).x, vlist_->item(0).y, 
        vlist_->item(0).x, vlist_->item(0).y
    );
    figure_->Bspline_curve_to(
        vlist_->item(0).x, vlist_->item(0).y, 
        vlist_->item(0).x, vlist_->item(0).y, 
        vlist_->item(1).x, vlist_->item(1).y
    );
    for (long i = 1; i < n - 1; ++i) {
        figure_->Bspline_curve_to(
            vlist_->item(i).x, vlist_->item(i).y, 
            vlist_->item(i-1).x, vlist_->item(i-1).y, 
            vlist_->item(i+1).x, vlist_->item(i+1).y
        );
    }
    figure_->Bspline_curve_to(
        vlist_->item(n-1).x, vlist_->item(n-1).y, 
        vlist_->item(n-2).x, vlist_->item(n-2).y, 
        vlist_->item(n-1).x, vlist_->item(n-1).y
    );
    figure_->Bspline_curve_to(
        vlist_->item(n-1).x, vlist_->item(n-1).y, 
        vlist_->item(n-1).x, vlist_->item(n-1).y, 
        vlist_->item(n-1).x, vlist_->item(n-1).y
    );
    vmin_.x = FLT_MAX;
    vmin_.y = FLT_MAX;
    vmax_.x = -FLT_MAX;
    vmax_.y = -FLT_MAX;
    for (long j = 0; j < vlist_->count(); j++) {
        vmin_.x = Math::min(vmin_.x, vlist_->item(j).x);
        vmax_.x = Math::max(vmax_.x, vlist_->item(j).x);
        vmin_.y = Math::min(vmin_.y, vlist_->item(j).y);
        vmax_.y = Math::max(vmax_.y, vlist_->item(j).y);
    }
}

//+ Closed_BSplineManip(Closed_BSplineManip)
void* Closed_BSplineManip::_this_ptr() { return this; }
_Ix_extern__tid(Closed_BSplineManip);
void* Closed_BSplineManip::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(Closed_BSplineManip) == t) return this;
    void* p;
    if (p = VertexManip::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(Closed_BSplineManip);
_Ix_ImplementGetTid(Closed_BSplineManip,_Ix__tid(Closed_BSplineManip))
_Ix_ImplementNarrow(Closed_BSplineManip*,Closed_BSplineManip,_Ix_NullParameter,Closed_BSplineManip,0)
_Ix_ImplementDuplicate(Closed_BSplineManip*,Closed_BSplineManip,_Ix_NullParameter,_Ix__tid(Closed_BSplineManip),0)
//+

Closed_BSplineManip::Closed_BSplineManip (Closed_BSplineManip* m) : VertexManip (m){
    if (is_nil(m) || is_nil(m->figure_)) {
        Figure* f = figures_->general_figure(
            FigureKit::fill_stroke, figures_->default_style(), true, true
        );
        f->add_point(0, 0);
        figure_body(f);
    }
}

Closed_BSplineManip::~Closed_BSplineManip () {}

Manipulator* Closed_BSplineManip::shallow_copy () {
    Closed_BSplineManip* manip = new Closed_BSplineManip(this);
    return manip;
}

Manipulator* Closed_BSplineManip::deep_copy () {
    Figure* copy = Figure::_narrow(CopyCmd::glyphmap_.find(figure_));
    if (is_nil(copy)) {
        Figure* f = figure_->copy_figure();
        CopyCmd::glyphmap_.map(figure_, f);
        Closed_BSplineManip* manip = new Closed_BSplineManip(this);
        manip->figure_body(f);
        CORBA::release(f);
        return manip;
    } else {
        Closed_BSplineManip* manip = new Closed_BSplineManip(this);
        manip->figure_body(copy);
        return manip;
    }
}

void Closed_BSplineManip::recompute_shape () {
    long n = vlist_->count();
    figure_->reset();
    figure_->Bspline_move_to(
        vlist_->item(0).x, vlist_->item(0).y, 
        vlist_->item(n-1).x, vlist_->item(n-1).y, 
        vlist_->item(1).x, vlist_->item(1).y
    );
    for (long i = 1; i < n - 1; ++i) {
        figure_->Bspline_curve_to(
            vlist_->item(i).x, vlist_->item(i).y, 
            vlist_->item(i-1).x, vlist_->item(i-1).y, 
            vlist_->item(i+1).x, vlist_->item(i+1).y
        );
    }
    figure_->Bspline_curve_to(
        vlist_->item(n-1).x, vlist_->item(n-1).y, 
        vlist_->item(n-2).x, vlist_->item(n-2).y, 
        vlist_->item(0).x, vlist_->item(0).y
    );
    figure_->Bspline_curve_to(
        vlist_->item(0).x, vlist_->item(0).y, 
        vlist_->item(n-1).x, vlist_->item(n-1).y, 
        vlist_->item(1).x, vlist_->item(1).y
    );
    vmin_.x = FLT_MAX;
    vmin_.y = FLT_MAX;
    vmax_.x = -FLT_MAX;
    vmax_.y = -FLT_MAX;
    for (long j = 0; j < vlist_->count(); j++) {
        vmin_.x = Math::min(vmin_.x, vlist_->item(j).x);
        vmax_.x = Math::max(vmax_.x, vlist_->item(j).x);
        vmin_.y = Math::min(vmin_.y, vlist_->item(j).y);
        vmax_.y = Math::max(vmax_.y, vlist_->item(j).y);
    }
}

//+ MultiLineManip(MultiLineManip)
void* MultiLineManip::_this_ptr() { return this; }
_Ix_extern__tid(MultiLineManip);
void* MultiLineManip::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(MultiLineManip) == t) return this;
    void* p;
    if (p = VertexManip::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(MultiLineManip);
_Ix_ImplementGetTid(MultiLineManip,_Ix__tid(MultiLineManip))
_Ix_ImplementNarrow(MultiLineManip*,MultiLineManip,_Ix_NullParameter,MultiLineManip,0)
_Ix_ImplementDuplicate(MultiLineManip*,MultiLineManip,_Ix_NullParameter,_Ix__tid(MultiLineManip),0)
//+

MultiLineManip::MultiLineManip (MultiLineManip* m) : VertexManip (m) {
    if (is_nil(m) || is_nil(m->figure_)) {
        Figure* f = figures_->general_figure(
            FigureKit::stroke, figures_->default_style(), false, false
        );
        f->add_point(0, 0);
        figure_body(f);
    }
}

MultiLineManip::~MultiLineManip () {}

Manipulator* MultiLineManip::shallow_copy () {
    MultiLineManip* manip = new MultiLineManip(this);
    return manip;
}

Manipulator* MultiLineManip::deep_copy () {
    Figure* copy = Figure::_narrow(CopyCmd::glyphmap_.find(figure_));
    if (is_nil(copy)) {
        Figure* f = figure_->copy_figure();
        CopyCmd::glyphmap_.map(figure_, f);
        MultiLineManip* manip = new MultiLineManip(this);
        manip->figure_body(f);
        CORBA::release(f);
        return manip;
    } else {
        MultiLineManip* manip = new MultiLineManip(this);
        manip->figure_body(copy);
        return manip;
    }
}

//+ PolygonManip(PolygonManip)
void* PolygonManip::_this_ptr() { return this; }
_Ix_extern__tid(PolygonManip);
void* PolygonManip::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(PolygonManip) == t) return this;
    void* p;
    if (p = VertexManip::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(PolygonManip);
_Ix_ImplementGetTid(PolygonManip,_Ix__tid(PolygonManip))
_Ix_ImplementNarrow(PolygonManip*,PolygonManip,_Ix_NullParameter,PolygonManip,0)
_Ix_ImplementDuplicate(PolygonManip*,PolygonManip,_Ix_NullParameter,_Ix__tid(PolygonManip),0)
//+

PolygonManip::PolygonManip (PolygonManip* m) : VertexManip (m) {
    if (is_nil(m) || is_nil(m->figure_)) {
        Figure* f = figures_->general_figure(
            FigureKit::fill_stroke, figures_->default_style(), true, false
        );
        f->add_point(0, 0);
        figure_body(f);
    }
}

PolygonManip::~PolygonManip () {}

Manipulator* PolygonManip::shallow_copy () {
    PolygonManip* manip = new PolygonManip(this);
    return manip;
}

Manipulator* PolygonManip::deep_copy () {
    Figure* copy = Figure::_narrow(CopyCmd::glyphmap_.find(figure_));
    if (is_nil(copy)) {
        Figure* f = figure_->copy_figure();
        CopyCmd::glyphmap_.map(figure_, f);
        PolygonManip* manip = new PolygonManip(this);
        manip->figure_body(f);
        CORBA::release(f);
        return manip;
    } else {
        PolygonManip* manip = new PolygonManip(this);
        manip->figure_body(copy);
        return manip;
    }
}

//+ PolyManip(PolyManip)
void* PolyManip::_this_ptr() { return this; }
_Ix_extern__tid(PolyManip);
void* PolyManip::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(PolyManip) == t) return this;
    void* p;
    if (p = Manipulator::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(PolyManip);
_Ix_ImplementGetTid(PolyManip,_Ix__tid(PolyManip))
_Ix_ImplementNarrow(PolyManip*,PolyManip,_Ix_NullParameter,PolyManip,0)
_Ix_ImplementDuplicate(PolyManip*,PolyManip,_Ix_NullParameter,_Ix__tid(PolyManip),0)
//+

PolyManip::PolyManip (Glyph* pf, PolyManip* m) : Manipulator (m) {
    // assumes ref counting
    XfGlyph::_duplicate(pf);
    polyfigure_ = pf;
    body(polyfigure_);
}

PolyManip::~PolyManip () {
    CORBA::release(polyfigure_);
}

void PolyManip::execute (FCommand* cmd) {
    if (info_ != nil && info_->manipulating_) {
        return;
    }
    FigViewer* v = cmd->figviewer();
    NaturalCmd* natcmd = NaturalCmd::_narrow(cmd);
    if (is_not_nil(natcmd)) {
        CmdVisitor cv(natcmd);
        cv.visit_children(polyfigure_);
        CORBA::release(natcmd);
    } else {
        Manipulator::execute(cmd);
    }
}

void PolyManip::unexecute (FCommand* cmd) {
    FigViewer* v = cmd->figviewer();
    NaturalCmd* natcmd = NaturalCmd::_narrow(cmd);
    if (is_not_nil(natcmd)) {
        CmdVisitor cv(natcmd, false);
        cv.visit_children(polyfigure_);
        CORBA::release(natcmd);
    } else {
        Manipulator::unexecute(cmd);
    }
}

Manipulator* PolyManip::shallow_copy () {
    return new PolyManip(polyfigure_, this);
}

Manipulator* PolyManip::deep_copy () {
    Glyph* copy = CopyCmd::glyphmap_.find(polyfigure_);
    if (is_nil(copy)) {
        Glyph* f = polyfigure_->clone_glyph();
        assert(! is_nil(f));
        CopyCmd::glyphmap_.map(polyfigure_, f);

        ManipCopier dcopier(false);
        XfGlyph_var g = body();
        dcopier.visit_children(g);
        ManipList* ml = dcopier.manipulators();
        for (long i = 0; i < ml->count(); i++) {
            Manipulator* man = ml->item(i);
            f->append(man);
        }
        Manipulator* m = new PolyManip(f, this);
        CORBA::release(f);
        return m;
    } else {
        return new PolyManip(copy, this);
    }
}

//+ LayoutManip(LayoutManip)
void* LayoutManip::_this_ptr() { return this; }
_Ix_extern__tid(LayoutManip);
void* LayoutManip::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(LayoutManip) == t) return this;
    void* p;
    if (p = Manipulator::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(LayoutManip);
_Ix_ImplementGetTid(LayoutManip,_Ix__tid(LayoutManip))
_Ix_ImplementNarrow(LayoutManip*,LayoutManip,_Ix_NullParameter,LayoutManip,0)
_Ix_ImplementDuplicate(LayoutManip*,LayoutManip,_Ix_NullParameter,_Ix__tid(LayoutManip),0)
//+

LayoutManip::LayoutManip (
    Transformer* t, LayoutManip* m
) : Manipulator (m) {
    transformer_ = t;
    resize_info_ = nil;
    if (is_not_nil(t)) {
        body(transformer_);
    }
    if (is_not_nil(m) && m->resize_info_ != nil) {
        resize_info_ = new ResizeInfo;
        *resize_info_ = *m->resize_info_;
    }
}

LayoutManip::~LayoutManip () {
    delete resize_info_;
}

void LayoutManip::child_allocate(XfGlyph::AllocationInfo& a) {
    if (resize_info_ != nil || is_not_nil(a.allocation)) {
	RegionImpl rr;
	if (resize_info_ != nil) {
	    XfGlyph::AllocationInfo ga;
	    ga.allocation = Region::_duplicate(a.allocation);
	    resize_extension(ga, rr);
	} else {
	    rr.copy(a.allocation);
	}
	XfGlyph::Requisition r;
	MonoGlyph::request(r);
	TransformImpl tx;
	tx.copy(&tx_);
	tx.premultiply(transformer_->transformation());
	Vertex delta = adjust_allocation(&rr, &tx, &r);
	if (is_nil(a.allocation)) {
	    a.allocation = new RegionImpl;
	}
        if (is_nil(a.transformation)) {
            a.transformation = new TransformImpl;
        }
        a.transformation->premultiply(&tx);
	a.transformation->translate(delta);
	a.allocation->copy(&rr);
    } else {
        Manipulator::child_allocate(a);
    }
}

//+ LayoutManip(XfGlyph::request)
void LayoutManip::request(XfGlyph::Requisition& r) {
    MonoGlyph::request(r);
    TransformImpl tx;
    tx.copy(&tx_);
    tx.premultiply(transformer_->transformation());
    flexible_transform_requisition(r, &tx);
}

//+ LayoutManip(XfGlyph::allocations)
void LayoutManip::allocations(XfGlyph::AllocationInfoSeq& a) {
    Manipulator::allocations(a);
}

void LayoutManip::draw_handles (XfGlyphTraversal_ptr gt) {
    Manipulator::draw_handles(gt);
}

static inline Coord rectify(Coord a) { return (a > 0) ? a : 0; }

static void constrain_size(Coord& len, Coord maxi, Coord mini) {
    len = Math::max(len, mini);
    len = Math::min(len, maxi);
}

Boolean LayoutManip::grasp (
    Tool& tool, FigViewer& fv, SelectInfo* si, Event_ptr e
) {
    init_info(tool, fv, si);
    if (info_->tooltype_ != move_tool && info_->tooltype_ != choose_tool) {
	NaturalCmd natcmd(&fv);
	execute(&natcmd);
    }
    Boolean ok = Manipulator::grasp(tool, fv, si, e);
    if (info_->tooltype_ == resize_tool) {
	RegionImpl r;
        XfGlyph::AllocationInfoSeq a;
	allocations(a);
	Vertex center;
	glyph_need_redraw(this, a);
	local_extension(transformer_, &r);
	r.apply_transform(&tx_);
	center.x = (r.lower_.x + r.upper_.x) * 0.5;
	center.y = (r.lower_.y + r.upper_.y) * 0.5;
	resize_info_ = new ResizeInfo;
	request(resize_info_->req);
	resize_info_->width = Math::abs((info_->pt_.x - center.x) * 2);
	resize_info_->height = Math::abs((info_->pt_.y - center.y) * 2);
	constrain_size(
	    resize_info_->width,
	    resize_info_->req.x.maximum, resize_info_->req.x.minimum
	);
	constrain_size(
	    resize_info_->height,
	    resize_info_->req.y.maximum, resize_info_->req.y.minimum
	);
	glyph_need_redraw(this, a);
	info_->pt_ = center;
	if (is_not_nil(info_->sinfo_->a_)) {
	    compose(this);
	}
    }
    return ok;
}

Boolean LayoutManip::manipulate (Event_ptr e) {
    switch (e->type()) {
    case Event::motion:
    case Event::enter:
    case Event::leave:
	break;
    case Event::focus_in:
    case Event::focus_out:
	return true;
    default:
	return false;
    }
    Boolean ok = true;
    if (info_->tooltype_ == resize_tool) {
        Vertex pt;
        pt.x = e->pointer_x() + info_->ax_ - info_->tool_ax_;
        pt.y = e->pointer_y() + info_->ay_ - info_->tool_ay_;
        pt.z = 0;
        info_->t_.transform_vertex(pt);
        XfGlyph::AllocationInfoSeq a;
	allocations(a);
	glyph_need_redraw(this, a);
	resize_info_->width = Math::abs((pt.x - info_->pt_.x) * 2);
	resize_info_->height = Math::abs((pt.y - info_->pt_.y) * 2);
	constrain_size(
	    resize_info_->width,
	    resize_info_->req.x.maximum, resize_info_->req.x.minimum
	);
	constrain_size(
	    resize_info_->height,
	    resize_info_->req.y.maximum, resize_info_->req.y.minimum
	);
	glyph_need_redraw(this, a);
	if (is_not_nil(info_->sinfo_->a_)) {
	    compose(this);
	}
    } else {
        ok = Manipulator::manipulate(e);
    }
    return ok;
}

void LayoutManip::execute (FCommand* cmd) {
    if (info_ != nil && info_->manipulating_) {
        return;
    }
    NaturalCmd* natcmd = NaturalCmd::_narrow(cmd);
    if (is_not_nil(natcmd)) {
        if (resize_info_ != nil) {
            XfGlyph::AllocationInfoSeq a;
	    allocations(a);
	    glyph_need_redraw(this, a);
	    delete resize_info_;
	    resize_info_ = nil;
	    glyph_need_redraw(this, a);
        }
        CmdVisitor cv(natcmd);
        cv.visit_children(XfGlyph_var(transformer_->body()));
	CORBA::release(natcmd);
        return;
    }
    Manipulator::execute(cmd);
}

void LayoutManip::unexecute (FCommand* cmd) {
    Manipulator::unexecute(cmd);
}

void LayoutManip::resize_extension(
    const XfGlyph::AllocationInfo& a, RegionImpl& rr
) {
    if (is_not_nil(a.allocation)) {
	rr.copy(a.allocation);
    } else {
	local_extension(transformer_, &rr);
	rr.apply_transform(&tx_);
	rr.lower_.x = (rr.lower_.x + rr.upper_.x - resize_info_->width) * 0.5;
	rr.lower_.y = (rr.lower_.y + rr.upper_.y - resize_info_->height) * 0.5;
    }
    rr.upper_.x = rr.lower_.x + resize_info_->width;
    rr.upper_.y = rr.lower_.y + resize_info_->height;
    if (is_not_nil(a.transformation)) {
	rr.apply_transform(a.transformation);
    }
}
    
//+ LayoutManip(XfGlyph::extension)
void LayoutManip::extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r) {
    if (resize_info_ != nil) {
        RegionImpl rr;
	resize_extension(a, rr);
	if (selected_) {
	    adjust_dotted_region(rr);
	}
        r->copy(&rr);
    } else {
        Manipulator::extension(a, r);
    }
}

//+ LayoutManip(XfGlyph::traverse)
void LayoutManip::traverse(XfGlyphTraversal_ptr t) {
    Region_var a = t->allocation();
    if (resize_info_ != nil || is_not_nil(a)) {
        Painter_var p = t->current_painter();
        p->push_matrix();
        p->premultiply(&tx_);
        RegionImpl* rr = new RegionImpl;
        if (resize_info_ != nil) {
	    XfGlyph::AllocationInfo i;
	    // avoid extra duplicate and release
	    i.allocation.ptr_ = a;
	    resize_extension(i, *rr);
	    i.allocation.ptr_ = nil;
        } else {
            rr->copy(a);
        }
        XfGlyph::Requisition r;
        MonoGlyph::request(r);
	TransformImpl tx;
	tx.copy(&tx_);
	tx.premultiply(Transform_var(transformer_->transformation()));

	Vertex delta = adjust_allocation(rr, &tx, &r);
	TransformImpl transform;
	transform.translate(delta);
	p->premultiply(&transform);

        t->traverse_child(&offset_, rr);
        CORBA::release(rr);
        p->pop_matrix();
        if (t->op() == XfGlyphTraversal::draw) {
            draw(t);
        }
    } else {
        Manipulator::traverse(t);
    }
}

void LayoutManip::need_resize () {
    OffsetVisitor ov;
    ov.visit_children(XfGlyph_var(transformer_->body()));
    for (Long i = 0; i < ov.offset_count(); i++) {
        compose(ov.offset(i)->child());
    }
    Manipulator::need_resize();
}

//+ ViewerManip(ViewerManip)
void* ViewerManip::_this_ptr() { return this; }
_Ix_extern__tid(ViewerManip);
void* ViewerManip::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(ViewerManip) == t) return this;
    void* p;
    if (p = LayoutManip::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(ViewerManip);
_Ix_ImplementGetTid(ViewerManip,_Ix__tid(ViewerManip))
_Ix_ImplementNarrow(ViewerManip*,ViewerManip,_Ix_NullParameter,ViewerManip,0)
_Ix_ImplementDuplicate(ViewerManip*,ViewerManip,_Ix_NullParameter,_Ix__tid(ViewerManip),0)
//+

ViewerManip::ViewerManip(
    Boolean* e, Viewer_ptr v, Transformer* t, ViewerManip* m
) : LayoutManip(t, m) {
    viewer_ = v;  // do not duplicate to avoid circularity
    editing_ = e;
}

ViewerManip::~ViewerManip() { }

//+ ViewerManip(XfGlyph::traverse)
void ViewerManip::traverse(XfGlyphTraversal_ptr t) {
    switch (t->op()) {
    case GlyphTraversal::pick_top:
    case GlyphTraversal::pick_all:
    case GlyphTraversal::pick_any:
	if (*editing_) {
	    RegionImpl r;
	    local_extension(this, &r);
	    if (Painter_var(t->current_painter())->is_visible(&r)) {
		t->hit();
	    }
	} else {
	    LayoutManip::traverse(t);
	}
	break;
    default:
	LayoutManip::traverse(t);
	break;
    }
}

void ViewerManip::execute(FCommand* cmd) {
    if (info_ != nil && info_->manipulating_) {
        return;
    }
    UngroupCmd* ungroupcmd = UngroupCmd::_narrow(cmd);
    if (is_nil(ungroupcmd)) {
	LayoutManip::execute(cmd);
    } else {
	CORBA::release(ungroupcmd);
    }
}

void ViewerManip::unexecute(FCommand* cmd) {
    UngroupCmd* ungroupcmd = UngroupCmd::_narrow(cmd);
    if (is_nil(ungroupcmd)) {
	LayoutManip::unexecute(cmd);
    } else {
	CORBA::release(ungroupcmd);
    }
}

//+ ButtonManip(ButtonManip)
void* ButtonManip::_this_ptr() { return this; }
_Ix_extern__tid(ButtonManip);
void* ButtonManip::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(ButtonManip) == t) return this;
    void* p;
    if (p = ViewerManip::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(ButtonManip);
_Ix_ImplementGetTid(ButtonManip,_Ix__tid(ButtonManip))
_Ix_ImplementNarrow(ButtonManip*,ButtonManip,_Ix_NullParameter,ButtonManip,0)
_Ix_ImplementDuplicate(ButtonManip*,ButtonManip,_Ix_NullParameter,_Ix__tid(ButtonManip),0)
//+

ButtonManip::ButtonManip (
    Boolean* e, Transformer* t, ButtonManip* m
) : ViewerManip(e, nil, t, m) {
    if (is_nil(transformer_)) {
        Button_var br = widgets_->palette_button(layouts_->vspace(10), nil);
        transformer_ = new Transformer;
        transformer_->body(br);
        body(transformer_);
    }
}

ButtonManip::~ButtonManip () {}

Manipulator* ButtonManip::shallow_copy () {
    return new ButtonManip(editing_, transformer_, this);
}

Manipulator* ButtonManip::deep_copy () {
    Transformer* copy = Transformer::_narrow(CopyCmd::glyphmap_.find(transformer_));
    if (is_nil(copy)) {
        Button_var br = widgets_->palette_button(
            figures_->label(
                FigureStyle_var(figures_->default_style()), 
                FrescoLib::tmp_string_ref("Button")
            ), Action_var(new PrintAction("Pushed"))
        );
        Transformer* transformer = new Transformer(transformer_);
        CopyCmd::glyphmap_.map(transformer_, transformer);
        transformer->body(br);
        Manipulator* m = new ButtonManip(editing_, transformer, this);
        CORBA::release(transformer);
        return m;
    } else {
        return new ButtonManip(editing_, copy, this);
    }
}

//+ FDrawManip(FDrawManip)
void* FDrawManip::_this_ptr() { return this; }
_Ix_extern__tid(FDrawManip);
void* FDrawManip::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(FDrawManip) == t) return this;
    void* p;
    if (p = ViewerManip::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(FDrawManip);
_Ix_ImplementGetTid(FDrawManip,_Ix__tid(FDrawManip))
_Ix_ImplementNarrow(FDrawManip*,FDrawManip,_Ix_NullParameter,FDrawManip,0)
_Ix_ImplementDuplicate(FDrawManip*,FDrawManip,_Ix_NullParameter,_Ix__tid(FDrawManip),0)
//+

FDrawManip::FDrawManip(
    Boolean* e, Viewer_ptr v, FDraw* f, Transformer* t, FDrawManip* m
) : ViewerManip(e, v, t, m) {
    fdraw_ = f;
    if (is_nil(transformer_)) {
	transformer_ = new Transformer;
    }
}

FDrawManip::~FDrawManip() { 
    if (is_not_nil(fdraw_)) {
        fdraw_->remove_viewer();
    }
}

Manipulator* FDrawManip::shallow_copy() {
    return new FDrawManip(editing_, viewer_, nil, transformer_, this);
}

Manipulator* FDrawManip::deep_copy() {
    Transformer* copy = Transformer::_narrow(CopyCmd::glyphmap_.find(transformer_));
    if (is_nil(copy)) {
	Transformer* transformer = new Transformer(transformer_);
	CopyCmd::glyphmap_.map(transformer_, transformer);
        FDraw* fdraw = new FDraw;
	transformer->body(fdraw);
        viewer_->append_viewer(fdraw);
	Manipulator* m = new FDrawManip(
            editing_, viewer_, fdraw, transformer, this
        );
	CORBA::release(transformer);
	return m;
    } else {
	return new FDrawManip(editing_, viewer_, nil, copy, this);
    }
}

Viewer_ptr FDrawManip::viewer() {
    return Viewer::_duplicate(fdraw_);
}

//+ FieldManip(FieldManip)
void* FieldManip::_this_ptr() { return this; }
_Ix_extern__tid(FieldManip);
void* FieldManip::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(FieldManip) == t) return this;
    void* p;
    if (p = ViewerManip::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(FieldManip);
_Ix_ImplementGetTid(FieldManip,_Ix__tid(FieldManip))
_Ix_ImplementNarrow(FieldManip*,FieldManip,_Ix_NullParameter,FieldManip,0)
_Ix_ImplementDuplicate(FieldManip*,FieldManip,_Ix_NullParameter,_Ix__tid(FieldManip),0)
//+

FieldManip::FieldManip(
    Boolean* e, Viewer_ptr v, FieldEditor* f, Transformer* t, FieldManip* m
) : ViewerManip(e, v, t, m) {
    fedit_ = f;
    if (is_nil(transformer_)) {
	transformer_ = new Transformer;
    }
}

FieldManip::~FieldManip() {
    if (is_not_nil(fedit_)) {
        fedit_->remove_viewer();
    }
}

Manipulator* FieldManip::shallow_copy() {
    return new FieldManip(editing_, viewer_, nil, transformer_, this);
}

Manipulator* FieldManip::deep_copy() {
    Transformer* copy = Transformer::_narrow(CopyCmd::glyphmap_.find(transformer_));
    if (is_nil(copy)) {
	Transformer* transformer = new Transformer(transformer_);
	CopyCmd::glyphmap_.map(transformer_, transformer);
	FieldSubject_var text = widgets_->field_subject();
	FieldEditor* field_editor = widgets_->field_editor(text, 5, 0);
        Glyph_var fedit = layouts_->back(
	    Glyph_var(widgets_->outset_frame(
		Glyph_var(layouts_->margin(
		    Glyph_var(widgets_->inset_frame(
			field_editor
		    )),
		    2
		))
	    )),
	    Glyph_var(widgets_->filler())
	);
 	transformer->body(fedit);
        viewer_->append_viewer(field_editor);
	Manipulator* m = new FieldManip(
            editing_, viewer_, field_editor, transformer, this
        );
	CORBA::release(transformer);
	return m;
    } else {
	return new FieldManip(editing_, viewer_, nil, copy, this);
    }
}

Viewer_ptr FieldManip::viewer() {
    return Viewer::_duplicate(fedit_);
}

//+ BoxManip(BoxManip)
void* BoxManip::_this_ptr() { return this; }
_Ix_extern__tid(BoxManip);
void* BoxManip::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(BoxManip) == t) return this;
    void* p;
    if (p = LayoutManip::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(BoxManip);
_Ix_ImplementGetTid(BoxManip,_Ix__tid(BoxManip))
_Ix_ImplementNarrow(BoxManip*,BoxManip,_Ix_NullParameter,BoxManip,0)
_Ix_ImplementDuplicate(BoxManip*,BoxManip,_Ix_NullParameter,_Ix__tid(BoxManip),0)
//+

BoxManip::BoxManip (Transformer* tr, BoxManip* m) : LayoutManip(tr, m) {}

BoxManip::~BoxManip () {}

void BoxManip::execute (FCommand* cmd) {
    if (info_ != nil && info_->manipulating_) {
        return;
    }
    FigViewer* v = cmd->figviewer();
    UngroupCmd* ungroupcmd = UngroupCmd::_narrow(cmd);
    if (is_not_nil(ungroupcmd)) {
        XfGlyph_var box = transformer_->body();

        Region_var na = transformer_->nat_allocation();
        TAManipCopier mc(na);
        mc.visit_children(box);
        ManipList* ml = mc.manipulators();
        TransformList* tl = mc.transforms();

        Transform_var tt = transformer_->transformation();
        for (long i = 0; i < ml->count(); i++) {
            Manipulator* man = ml->item(i);
            Transform_var tr = man->transformation();
            tr->postmultiply(tl->item(i));
            tr->postmultiply(tt);
            tr->postmultiply(&tx_);
        }
        Inserter inserter(*ml);
        inserter.visit_parents(this);
        SelectCmd selectcmd(v);
        for (i = 0; i < ml->count(); i++) {
            Manipulator* man = ml->item(i);
            man->execute(&selectcmd);
            info(man, new ManipInfo(info_->sinfo_->copy()));
        }
        SelectCmd unselectcmd(v, false);
        execute(&unselectcmd);
        Remover remover;
        remover.visit_parents(this);
        CORBA::release(ungroupcmd);
        return;
    }
    GroupCmd* groupcmd = GroupCmd::_narrow(cmd);
    if (is_not_nil(groupcmd)) {
        Selection* sel = v->selection()->copy();
        long n = sel->count();
        if (n < 2) {
            delete sel;
            return;
        }
        CopyCmd::glyphmap_.clear();
        Manipulator* bm = deep_copy();
        XfGlyph_var box = XfGlyph_var(bm->body())->body();

        SelectCmd unselectcmd(v, false);
        Manipulator* m0 = sel->item(0);
        if (info(m0) != nil) {
            info(bm, new ManipInfo(info(m0)->sinfo_->copy()));
        }
        OffsetVisitor ov;
        ov.visit_parents(m0);

        Remover* remover = new Remover;

        RegionImpl r;
        Vertex center;
        for (long i = 0; i < sel->count(); i++) {
            Manipulator* m = sel->item(i);
            RegionImpl ri;
            local_extension(m, &ri);
            r.merge_union(&ri);
            m->execute(&unselectcmd);
            remover->visit_parents(m);
            compose(m);
            box->append(m);
        }
        delete remover;
        for (long j = 0; j < ov.offset_count(); j++) {
            ov.offset(j)->parent()->append(bm);
        }

        center.x = (r.upper_.x+r.lower_.x) * 0.5;
        center.y = (r.upper_.y+r.lower_.y) * 0.5;
        NaturalCmd nat(v);
        bm->execute(&nat);
        local_extension(bm, &r);
        center.x -= (r.upper_.x+r.lower_.x) * 0.5;
        center.y -= (r.upper_.y+r.lower_.y) * 0.5;
        bm->transformation()->translate(center);

        bm->need_resize();
        SelectCmd selectcmd(v);
        bm->execute(&selectcmd);
        CORBA::release(groupcmd);
        return;
    }
    LayoutManip::execute(cmd);
}

void BoxManip::unexecute (FCommand*) {}

//+ HBoxManip(HBoxManip)
void* HBoxManip::_this_ptr() { return this; }
_Ix_extern__tid(HBoxManip);
void* HBoxManip::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(HBoxManip) == t) return this;
    void* p;
    if (p = LayoutManip::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(HBoxManip);
_Ix_ImplementGetTid(HBoxManip,_Ix__tid(HBoxManip))
_Ix_ImplementNarrow(HBoxManip*,HBoxManip,_Ix_NullParameter,HBoxManip,0)
_Ix_ImplementDuplicate(HBoxManip*,HBoxManip,_Ix_NullParameter,_Ix__tid(HBoxManip),0)
//+

HBoxManip::HBoxManip (Transformer* t, HBoxManip* m) : BoxManip (t, m) {
    if (is_nil(t)) {
        Transformer* th = new Transformer;
        th->body(XfGlyph_var(layouts_->hbox()));
        transformer_ = th;
        body(transformer_);
    }
}

HBoxManip::~HBoxManip () {}

Manipulator* HBoxManip::shallow_copy () {
    return new HBoxManip(transformer_, this);
}

Manipulator* HBoxManip::deep_copy () {
    Transformer* copy = Transformer::_narrow(CopyCmd::glyphmap_.find(transformer_));
    if (is_nil(copy)) {
        XfGlyph_var box = layouts_->hbox();

        Transformer* transformer = new Transformer(transformer_);
        CopyCmd::glyphmap_.map(transformer_, transformer);
        
        transformer->body(box);
        ManipCopier dcopier(false);
        XfGlyph_var g = XfGlyph_var(body())->body();
        dcopier.visit_children(g);
        ManipList* ml = dcopier.manipulators();
        for (long i = 0; i < ml->count(); i++) {
            Manipulator* man = ml->item(i);
            box->append(man);
        }

        Manipulator* m = new HBoxManip(transformer, this);
	CORBA::release(transformer);
        return m;
    } else {
        return new HBoxManip(copy, this);
    }
}

//+ VBoxManip(VBoxManip)
void* VBoxManip::_this_ptr() { return this; }
_Ix_extern__tid(VBoxManip);
void* VBoxManip::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(VBoxManip) == t) return this;
    void* p;
    if (p = LayoutManip::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(VBoxManip);
_Ix_ImplementGetTid(VBoxManip,_Ix__tid(VBoxManip))
_Ix_ImplementNarrow(VBoxManip*,VBoxManip,_Ix_NullParameter,VBoxManip,0)
_Ix_ImplementDuplicate(VBoxManip*,VBoxManip,_Ix_NullParameter,_Ix__tid(VBoxManip),0)
//+

VBoxManip::VBoxManip (Transformer* t, VBoxManip* m) : BoxManip (t, m) {
    if (is_nil(t)) {
        Transformer* tv = new Transformer;
        tv->body(XfGlyph_var(layouts_->vbox()));
        transformer_ = tv;
        body(transformer_);
    }
}

VBoxManip::~VBoxManip () {}

Manipulator* VBoxManip::shallow_copy () {
    return new VBoxManip(transformer_, this);
}

Manipulator* VBoxManip::deep_copy () {
    Transformer* copy = Transformer::_narrow(CopyCmd::glyphmap_.find(transformer_));
    if (is_nil(copy)) {
        XfGlyph_var box = layouts_->vbox();

        Transformer* transformer = new Transformer(transformer_);
        CopyCmd::glyphmap_.map(transformer_, transformer);
        transformer->body(box);

        ManipCopier dcopier(false);
        XfGlyph_var g = XfGlyph_var(body())->body();
        dcopier.visit_children(g);
        ManipList* ml = dcopier.manipulators();
        for (long i = 0; i < ml->count(); i++) {
            Manipulator* man = ml->item(i);
            box->append(man);
        }

        Manipulator* m = new VBoxManip(transformer, this);
	CORBA::release(transformer);
        return m;
    } else {
        return new VBoxManip(copy, this);
    }
}

//+ GlueManip(GlueManip)
void* GlueManip::_this_ptr() { return this; }
_Ix_extern__tid(GlueManip);
void* GlueManip::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(GlueManip) == t) return this;
    void* p;
    if (p = LayoutManip::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(GlueManip);
_Ix_ImplementGetTid(GlueManip,_Ix__tid(GlueManip))
_Ix_ImplementNarrow(GlueManip*,GlueManip,_Ix_NullParameter,GlueManip,0)
_Ix_ImplementDuplicate(GlueManip*,GlueManip,_Ix_NullParameter,_Ix__tid(GlueManip),0)
//+

static const Coord minor_len = 5.0;

GlueManip::GlueManip(Transformer* t, GlueManip* m) : LayoutManip(t, m) {
    if (is_nil(transformer_)) {
	XfGlyph_var glue = layouts_->hfil();
	transformer_ = new Transformer;
	transformer_->body(glue);
	nat_ = 0.0;
	fil_ = true;
	axis_ = X_axis;
	body(transformer_);
    } else {
	if (is_not_nil(m)) {
	    nat_ = m->nat_;
	    fil_ = m->fil_;
	    axis_ = m->axis_;
	}
    }
}

GlueManip::~GlueManip() { }

//+ GlueManip(XfGlyph::request)
void GlueManip::request(XfGlyph::Requisition& r) {
    LayoutManip::request(r);
}

//+ GlueManip(XfGlyph::traverse)
void GlueManip::traverse(XfGlyphTraversal_ptr t) {
    switch (t->op()) {
    case GlyphTraversal::pick_top:
    case GlyphTraversal::pick_all:
    case GlyphTraversal::pick_any:
	{
	    RegionImpl r;
	    Region_var a = t->allocation();
	    if (is_not_nil(a)) {
		r.copy(a);
	    } else {
		local_extension(this, &r);
	    }
	    if (Painter_var(t->current_painter())->is_visible(&r)) {
		t->hit();
	    }
	}
	break;
    case GlyphTraversal::draw:
	draw(t);
	break;
    default:
	LayoutManip::traverse(t);
	break;
    }
}

static Color_ptr glue_color;

//+ GlueManip(XfGlyph::draw)
void GlueManip::draw(XfGlyphTraversal_ptr t) {
    Painter_var p = t->current_painter();
    if (is_nil(glue_color)) {
        FrescoContext* fc = FrescoLib::current_context();
	glue_color = drawings_->color_rgb(0.8, 0.8, 0.8);
    }
    p->current_color(glue_color);
    p->push_matrix();
    p->premultiply(&tx_);
    p->premultiply(Transform_var(transformer_->transformation()));
    Region_var a = t->allocation();
    if (resize_info_ != nil || is_not_nil(a)) {
	RegionImpl* rr = new RegionImpl;
	if (resize_info_ != nil) {
	    XfGlyph::AllocationInfo i;
	    // avoid extra duplicate + release
	    i.allocation.ptr_ = a;
	    resize_extension(i, *rr);
	    i.allocation.ptr_ = nil;
	} else {
	    rr->copy(a);
	}
	XfGlyph::Requisition r;
	MonoGlyph::request(r);
	TransformImpl tx;
	tx.copy(&tx_);
	tx.premultiply(Transform_var(transformer_->transformation()));

	Vertex delta = adjust_allocation(rr, &tx, &r);
	TransformImpl transform;
	transform.translate(delta);
	p->premultiply(&transform);

	p->fill_rect(rr->lower_.x, rr->lower_.y, rr->upper_.x, rr->upper_.y);
	CORBA::release(rr);
    } else {
	if (axis_ == X_axis) {
	    p->fill_rect(0, -minor_len, nat_, minor_len);
	} else {
	    p->fill_rect(-minor_len, 0, minor_len, nat_);
	}
    }
    p->pop_matrix();
    LayoutManip::draw(t);
}

//+ GlueManip(XfGlyph::extension)
void GlueManip::extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r) {
    if (resize_info_ != nil || is_not_nil(a.allocation)) {
	LayoutManip::extension(a, r);
    } else {
	XfGlyph::AllocationInfo ga;
	ga.transformation = a.transformation;
	ga.damaged = a.damaged;
	if (is_nil(ga.transformation)) {
	    ga.transformation = new TransformImpl;
	}
	ga.transformation->premultiply(&tx_);
	ga.transformation->premultiply(transformer_->transformation());
	RegionImpl region;
	region.lower_.x = 0;
	region.lower_.y = 0;
	region.lower_.z = 0;
	region.upper_.z = 0;
	region.xalign_ = 0;
	region.yalign_ = 0;
	region.zalign_ = 0;
	region.defined_ = true;
	if (axis_ == X_axis) {
	    region.upper_.x = nat_;
	    region.upper_.y = minor_len;
	    region.lower_.y = -minor_len;
	} else {
	    region.upper_.x = minor_len;
	    region.lower_.x = -minor_len;
	    region.upper_.y = nat_;
	}
	region.apply_transform(ga.transformation);
	if (selected_) {
	    adjust_dotted_region(region);
	}
	r->copy(&region);
    }
}

static XfGlyph_ptr create_glue(Coord n, Axis a, Boolean ff, LayoutKit* layouts) {
    Coord nat = Math::abs(n);
    XfGlyph::Requisition r;
    Coord fil = layouts->fil();
    r.x.defined = true;
    r.y.defined = true;
    if (n > 0) {
	r.x.align = 0;
	r.y.align = 0;
    } else {
	r.x.align = 1.0;
	r.y.align = 1.0;
    }
    r.z.defined = false;
    if (a == X_axis) {
	r.y.natural = 0;
	r.y.maximum = fil;
	r.y.minimum = 0;
	r.x.natural = nat;
	if (ff) {
	    r.x.maximum = fil;
	    r.x.minimum = 0;
	} else {
	    r.x.maximum = nat;
	    r.x.minimum = nat;
	}
    } else {
	r.x.natural = 0;
	r.x.maximum = fil;
	r.x.minimum = 0;
	r.y.natural = nat;
	if (ff) {
	    r.y.maximum = fil;
	    r.y.minimum = 0;
	} else {
	    r.y.maximum = nat;
	    r.y.minimum = nat;
	}
    }
    return layouts->glue_requisition(r);
}

Manipulator* GlueManip::shallow_copy() {
    return new GlueManip(transformer_, this);
}

Manipulator* GlueManip::deep_copy() {
    Transformer* copy = Transformer::_narrow(CopyCmd::glyphmap_.find(transformer_));
    if (is_nil(copy)) {
	XfGlyph_var g = create_glue(nat_, axis_, fil_, layouts_);
	Transformer* transformer = new Transformer(transformer_);
	CopyCmd::glyphmap_.map(transformer_, transformer);
	transformer->body(g);
	Manipulator* m = new GlueManip(transformer, this);
	CORBA::release(transformer);
	return m;
    } else {
	return new GlueManip(copy, this);
    }
}

Boolean GlueManip::manipulate(Event_ptr e) {
    if (e->type() != Event::motion) {
	return false;
    }
    if (info_->tooltype_ == create_tool) {
	Vertex pt;
	pt.x = e->pointer_x() + info_->ax_ - info_->tool_ax_;
	pt.y = e->pointer_y() + info_->ay_ - info_->tool_ay_;
	pt.z = 0;
	info_->t_.transform_vertex(pt);

	Coord dx, dy;

	dx = pt.x - info_->pt_.x;
	dy = pt.y - info_->pt_.y;

	XfGlyph::AllocationInfoSeq a;
	allocations(a);
	glyph_need_redraw(this, a);
	if (Math::abs(dx) > Math::abs(dy)) {
	    axis_ = X_axis;
	    nat_ = dx;
	    info_->center_.x = info_->pt_.x + dx * 0.5;
	} else {
	    axis_ = Y_axis;
	    nat_ = dy;
	    info_->center_.y = info_->pt_.y + dy * 0.5;
	}
	fil_ = !e->modifier_is_down(Event::shift);
	XfGlyph_var g = create_glue(nat_, axis_, fil_, layouts_);
	transformer_->body(g);
	glyph_need_redraw(this, a);
	if (is_not_nil(info_->sinfo_->a_)) {
	    compose(this);
	}
	need_resize();
    } else {
	return LayoutManip::manipulate(e);
    }
    return true;
}

FCommand* GlueManip::effect(Event_ptr e) {
    if (info_ != nil && info_->tooltype_ == create_tool) {
	transformer_->need_resize();
    }
    return LayoutManip::effect(e);
}

void GlueManip::execute(FCommand* cmd) {
    if (info_ != nil && info_->manipulating_) {
        return;
    }
    UngroupCmd* ungroupcmd = UngroupCmd::_narrow(cmd);
    if (is_nil(ungroupcmd)) {
	LayoutManip::execute(cmd);
    } else {
	CORBA::release(ungroupcmd);
    }
}

void GlueManip::unexecute(FCommand* cmd) {
    UngroupCmd* ungroupcmd = UngroupCmd::_narrow(cmd);
    if (is_nil(ungroupcmd)) {
	LayoutManip::unexecute(cmd);
    } else {
	CORBA::release(ungroupcmd);
    }
}

//+ MacroManip(MacroManip)
void* MacroManip::_this_ptr() { return this; }
_Ix_extern__tid(MacroManip);
void* MacroManip::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(MacroManip) == t) return this;
    void* p;
    if (p = Manipulator::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(MacroManip);
_Ix_ImplementGetTid(MacroManip,_Ix__tid(MacroManip))
_Ix_ImplementNarrow(MacroManip*,MacroManip,_Ix_NullParameter,MacroManip,0)
_Ix_ImplementDuplicate(MacroManip*,MacroManip,_Ix_NullParameter,_Ix__tid(MacroManip),0)
//+

MacroManip::MacroManip (MacroManip* m) : Manipulator(m) {
    maniplist_ = new ManipList;
}

MacroManip::~MacroManip () {
    delete maniplist_;
}

Boolean MacroManip::grasp(
    Tool& t, FigViewer& fv, SelectInfo* si, Event_ptr e
) {
    Boolean ok = true;
    for (long i = 0; i < maniplist_->count(); i++) {
        Manipulator* m = maniplist_->item(i);
        ok = m->grasp(t, fv, si, e) && ok;
    }
    return ok;
}

Boolean MacroManip::manipulate(Event_ptr e) {
    Boolean ok = true;
    for (long i = 0; i < maniplist_->count(); i++) {
        Manipulator* m = maniplist_->item(i);
        ok = m->manipulate(e) && ok;
    }
    return ok;
}

FCommand* MacroManip::effect (Event_ptr e) {
    MacroCmd* mc = new MacroCmd;
    for (long i = 0; i < maniplist_->count(); i++) {
        Manipulator* m = maniplist_->item(i);
        FCommand* cmd = m->effect(e);
        if (is_not_nil(cmd)) {
            mc->append(cmd);
        }
    }
    if (mc->count() == 0) {
        delete mc;
        mc = nil;
    }
    return mc;
}

void MacroManip::execute (FCommand* cmd) {
    if (info_ != nil && info_->manipulating_) {
        return;
    }
    for (long i = 0; i < maniplist_->count(); i++) {
        Manipulator* m = maniplist_->item(i);
        m->execute(cmd);
    }
}

void MacroManip::unexecute (FCommand* cmd) {
    for (long i = 0; i < maniplist_->count(); i++) {
        Manipulator* m = maniplist_->item(i);
        m->unexecute(cmd);
    }
}

Manipulator* MacroManip::shallow_copy () {
    MacroManip* macro = new MacroManip(this);
    for (long i = 0; i < maniplist_->count(); i++) {
        Manipulator* m = maniplist_->item(i);
        macro->append(m);
    }
    return macro;
}
    
Manipulator* MacroManip::deep_copy () {
    MacroManip* macro = new MacroManip(this);
    for (long i = 0; i < maniplist_->count(); i++) {
        Manipulator* m = maniplist_->item(i)->deep_copy();
        macro->append(m);
        CORBA::release(m);
    }
    return macro;
}
    
void MacroManip::add (Manipulator* manip) {
    maniplist_->append(manip);
}

void MacroManip::insert (long index, Manipulator* manip) {
    maniplist_->insert(index, manip);
}

void MacroManip::remove(long index) {
    maniplist_->remove(index);
}

void MacroManip::remove_all() {
    maniplist_->remove_all();
}

long MacroManip::count () {
    return maniplist_->count();
}

Manipulator* MacroManip::manipulator (long index) {
    return maniplist_->item(index);
}

