//
// Copyright (c) 1994 Silicon Graphics, Inc.
// Copyright (c) 1994 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
// OF THIS SOFTWARE.
//

/*
 * manipulator.h
 */

#ifndef _manipulator_h
#define _manipulator_h

#include <Fresco/defs_impl/glyphs.h>
#include <Fresco/defs_impl/list.h>
#include <Fresco/defs_impl/region.h>
#include <Fresco/defs_impl/transform.h>
#include <Fresco/defs/types.h>
#include <Fresco/defs/viewer.h>

class FDraw;
class Figure;
class FieldEditor;
class FigViewer;
class Manipulator;
class ManipList;
class FCommand;
class SelectInfo;
class Tool;
class Transformer;
class VertexList;

class FigureKit;
class WidgetKit;
class LayoutKit;
class DialogKit;
class DrawingKit;
class ManipInfo;
class MenuKit;

enum ToolType {
    create_tool, move_tool, alter_tool, rotate_tool, scale_tool, choose_tool,
    stretch_tool, resize_tool, narrow_tool, undefined_tool
};

declarePtrList(ManipList, Manipulator)

//- Manipulator*
//+ Manipulator : MonoGlyph
class Manipulator : public MonoGlyph {
public:
    ~Manipulator();
    _Ix_InterfaceOps(Manipulator*)
//+
public:
    virtual Boolean grasp(Tool&, FigViewer&, SelectInfo*, Event_ptr);
    virtual Boolean manipulate(Event_ptr);
    virtual FCommand* effect(Event_ptr);

    virtual void execute(FCommand*);
    virtual void unexecute(FCommand*);

    virtual Manipulator* shallow_copy() = 0;
    virtual Manipulator* deep_copy() = 0;

    virtual void draw_handles(XfGlyphTraversal_ptr);

    virtual Viewer_ptr viewer();

    void draw(XfGlyphTraversal_ptr t); //+ XfGlyph::draw
    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
    void extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r); //+ XfGlyph::extension
    void traverse(XfGlyphTraversal_ptr t); //+ XfGlyph::traverse
    XfTransform_ptr transformation(); //+ XfGlyph::transformation
    void child_allocate(XfGlyph::AllocationInfo& a);

    static void init_kits(
        FigureKit*, WidgetKit*, LayoutKit*, DialogKit*, DrawingKit*, MenuKit*
    );
protected:
    Manipulator(Manipulator* = nil);
    void info(Manipulator*, ManipInfo*);
    void init_info(Tool& tool, FigViewer&, SelectInfo*);
    ManipInfo* info(Manipulator*);
protected:
    Boolean selected_ : 1;
    TransformImpl tx_;
    ManipInfo* info_;  // cache information for manipulation

    static FigureKit* figures_;
    static WidgetKit* widgets_;
    static LayoutKit* layouts_;
    static DialogKit* dialogs_;
    static DrawingKit* drawings_;
    static MenuKit* menus_;
};

//- FigureManip*
//+ FigureManip : Manipulator
class FigureManip : public Manipulator {
public:
    ~FigureManip();
    _Ix_InterfaceOps(FigureManip*)
//+
public:
    virtual Boolean grasp(Tool&, FigViewer&, SelectInfo*, Event_ptr);
    virtual Manipulator* shallow_copy() = 0;
    virtual Manipulator* deep_copy() = 0;

    Figure* figure_body();
protected:
    FigureManip(FigureManip* m = nil);
    void figure_body(Figure*);
protected:
    Figure* figure_;
};

//- VertexManip*
//+ VertexManip : FigureManip
class VertexManip : public FigureManip {
public:
    ~VertexManip();
    _Ix_InterfaceOps(VertexManip*)
//+
protected:
    VertexManip(VertexManip* m = nil);

    void extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r); //+ XfGlyph::extension
    virtual Boolean grasp(Tool&, FigViewer&, SelectInfo*, Event_ptr);
    virtual Boolean manipulate(Event_ptr);

    virtual Manipulator* shallow_copy() = 0;
    virtual Manipulator* deep_copy() = 0;
    virtual void recompute_shape();

    virtual void draw_handles(XfGlyphTraversal_ptr);
protected:
    VertexList* vlist_;
    Vertex vmin_;
    Vertex vmax_;
};

//- LineManip*
//+ LineManip : VertexManip
class LineManip : public VertexManip {
public:
    ~LineManip();
    _Ix_InterfaceOps(LineManip*)
//+
public:
    LineManip(LineManip* m = nil);

    virtual Boolean manipulate(Event_ptr);
    virtual Manipulator* shallow_copy();
    virtual Manipulator* deep_copy();
};

//- RectManip*
//+ RectManip : VertexManip
class RectManip : public VertexManip {
public:
    ~RectManip();
    _Ix_InterfaceOps(RectManip*)
//+
public:
    RectManip(RectManip* m = nil);

    virtual Boolean manipulate(Event_ptr);
    virtual Manipulator* shallow_copy();
    virtual Manipulator* deep_copy();
};

//- EllipseManip*
//+ EllipseManip : VertexManip
class EllipseManip : public VertexManip {
public:
    ~EllipseManip();
    _Ix_InterfaceOps(EllipseManip*)
//+
public:
    EllipseManip(EllipseManip* m = nil);

    void extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r); //+ XfGlyph::extension
    virtual Boolean manipulate(Event_ptr);
    virtual Manipulator* shallow_copy();
    virtual Manipulator* deep_copy();
    virtual void draw_handles(XfGlyphTraversal_ptr);
    virtual void recompute_shape();
};

//- Open_BSplineManip*
//+ Open_BSplineManip : VertexManip
class Open_BSplineManip : public VertexManip {
public:
    ~Open_BSplineManip();
    _Ix_InterfaceOps(Open_BSplineManip*)
//+
public:
    Open_BSplineManip(Open_BSplineManip* m = nil);

    virtual Manipulator* shallow_copy();
    virtual Manipulator* deep_copy();
    virtual void recompute_shape();
};

//- Closed_BSplineManip*
//+ Closed_BSplineManip : VertexManip
class Closed_BSplineManip : public VertexManip {
public:
    ~Closed_BSplineManip();
    _Ix_InterfaceOps(Closed_BSplineManip*)
//+
public:
    Closed_BSplineManip(Closed_BSplineManip* m = nil);

    virtual Manipulator* shallow_copy();
    virtual Manipulator* deep_copy();
    virtual void recompute_shape();
};

//- MultiLineManip*
//+ MultiLineManip : VertexManip
class MultiLineManip : public VertexManip {
public:
    ~MultiLineManip();
    _Ix_InterfaceOps(MultiLineManip*)
//+
public:
    MultiLineManip(MultiLineManip* m = nil);

    virtual Manipulator* shallow_copy();
    virtual Manipulator* deep_copy();
};

//- PolygonManip*
//+ PolygonManip : VertexManip
class PolygonManip : public VertexManip {
public:
    ~PolygonManip();
    _Ix_InterfaceOps(PolygonManip*)
//+
public:
    PolygonManip(PolygonManip* m = nil);

    virtual Manipulator* shallow_copy();
    virtual Manipulator* deep_copy();
};

//- PolyManip*
//+ PolyManip : Manipulator
class PolyManip : public Manipulator {
public:
    ~PolyManip();
    _Ix_InterfaceOps(PolyManip*)
//+
public:
    PolyManip(Glyph*, PolyManip* m = nil);

    virtual void execute(FCommand*);
    virtual void unexecute(FCommand*);

    virtual Manipulator* shallow_copy();
    virtual Manipulator* deep_copy();

protected:
    Glyph* polyfigure_;
};

struct ResizeInfo {
    Coord width, height, depth;
    XfGlyph::Requisition req;
};

//- LayoutManip*
//+ LayoutManip : Manipulator
class LayoutManip : public Manipulator {
public:
    ~LayoutManip();
    _Ix_InterfaceOps(LayoutManip*)
//+
public:
    LayoutManip(
        Transformer* t = nil, LayoutManip* m = nil
    );

    virtual Boolean grasp(Tool&, FigViewer&, SelectInfo*, Event_ptr);
    virtual Boolean manipulate(Event_ptr);
    virtual void execute(FCommand*);
    virtual void unexecute(FCommand*);
    virtual void draw_handles(XfGlyphTraversal_ptr);

    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
    void extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r); //+ XfGlyph::extension
    void traverse(XfGlyphTraversal_ptr t); //+ XfGlyph::traverse

    virtual Manipulator* shallow_copy() = 0;
    virtual Manipulator* deep_copy() = 0;
    void need_resize(); //+ XfGlyph::need_resize
    void child_allocate(XfGlyph::AllocationInfo& a);
    void allocations(XfGlyph::AllocationInfoSeq& a); //+ XfGlyph::allocations
protected:
    void resize_extension(const XfGlyph::AllocationInfo&, RegionImpl&);
protected:
    Transformer* transformer_;
    ResizeInfo* resize_info_;
};

//- ViewerManip*
//+ ViewerManip : LayoutManip
class ViewerManip : public LayoutManip {
public:
    ~ViewerManip();
    _Ix_InterfaceOps(ViewerManip*)
//+
public:
    virtual void execute(FCommand*);
    virtual void unexecute(FCommand*);

    void traverse(XfGlyphTraversal_ptr t); //+ XfGlyph::traverse
protected:
    ViewerManip(
	Boolean*, Viewer_ptr, Transformer* t = nil, ViewerManip* m = nil
    );
protected:
    Boolean* editing_;
    Viewer_ptr viewer_;
};

//- ButtonManip*
//+ ButtonManip : ViewerManip
class ButtonManip : public ViewerManip {
public:
    ~ButtonManip();
    _Ix_InterfaceOps(ButtonManip*)
//+
public:
    ButtonManip(
        Boolean*, Transformer* t = nil, ButtonManip* m = nil
    );
    virtual Manipulator* shallow_copy();
    virtual Manipulator* deep_copy();
};

//- FDrawManip*
//+ FDrawManip : ViewerManip
class FDrawManip : public ViewerManip {
public:
    ~FDrawManip();
    _Ix_InterfaceOps(FDrawManip*)
//+
public:
    FDrawManip(
        Boolean*, Viewer_ptr, FDraw* f= nil, 
        Transformer* t = nil, FDrawManip* m = nil
    );

    virtual Manipulator* shallow_copy();
    virtual Manipulator* deep_copy();
    virtual Viewer_ptr viewer();
private:
    FDraw* fdraw_;
};

//- FieldManip*
//+ FieldManip : ViewerManip
class FieldManip : public ViewerManip {
public:
    ~FieldManip();
    _Ix_InterfaceOps(FieldManip*)
//+
public:
    FieldManip(
        Boolean*, Viewer_ptr, FieldEditor* f = nil,
        Transformer* t = nil, FieldManip* m = nil
    );
    virtual Manipulator* shallow_copy();
    virtual Manipulator* deep_copy();
    virtual Viewer_ptr viewer();
private:
    FieldEditor* fedit_;
};

//- BoxManip*
//+ BoxManip : LayoutManip
class BoxManip : public LayoutManip {
public:
    ~BoxManip();
    _Ix_InterfaceOps(BoxManip*)
//+
public:
    virtual void execute(FCommand*);
    virtual void unexecute(FCommand*);
    virtual Manipulator* shallow_copy() = 0;
    virtual Manipulator* deep_copy() = 0;
protected:
    BoxManip(Transformer*, BoxManip* m = nil);
};

//- HBoxManip*
//+ HBoxManip : BoxManip
class HBoxManip : public BoxManip {
public:
    ~HBoxManip();
    _Ix_InterfaceOps(HBoxManip*)
//+
public:
    HBoxManip(Transformer* t = nil, HBoxManip* m = nil);
    virtual Manipulator* shallow_copy();
    virtual Manipulator* deep_copy();
};

//- VBoxManip*
//+ VBoxManip : BoxManip
class VBoxManip : public BoxManip {
public:
    ~VBoxManip();
    _Ix_InterfaceOps(VBoxManip*)
//+
public:
    VBoxManip(Transformer* t = nil, VBoxManip* m = nil);
    virtual Manipulator* shallow_copy();
    virtual Manipulator* deep_copy();
};

//- GlueManip*
//+ GlueManip : LayoutManip
class GlueManip : public LayoutManip {
public:
    ~GlueManip();
    _Ix_InterfaceOps(GlueManip*)
//+
public:
    GlueManip(Transformer* t = nil, GlueManip* m = nil);

    virtual void execute(FCommand*);
    virtual void unexecute(FCommand*);

    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
    void traverse(XfGlyphTraversal_ptr t); //+ XfGlyph::traverse
    void draw(XfGlyphTraversal_ptr t); //+ XfGlyph::draw
    void extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r); //+ XfGlyph::extension

    virtual Boolean manipulate(Event_ptr);
    virtual FCommand* effect(Event_ptr);
    virtual Manipulator* shallow_copy();
    virtual Manipulator* deep_copy();
protected:
    Coord nat_;
    Boolean fil_;
    Axis axis_;
};

//- MacroManip*
//+ MacroManip : Manipulator
class MacroManip : public Manipulator {
public:
    ~MacroManip();
    _Ix_InterfaceOps(MacroManip*)
//+
public:
    MacroManip(MacroManip* m = nil);

    virtual Boolean grasp(Tool&, FigViewer& fv, SelectInfo*, Event_ptr);
    virtual Boolean manipulate(Event_ptr);
    virtual FCommand* effect(Event_ptr);

    virtual void execute(FCommand*);
    virtual void unexecute(FCommand*);

    virtual Manipulator* shallow_copy();
    virtual Manipulator* deep_copy();

    virtual void add(Manipulator*);
    virtual void insert(long index, Manipulator*);
    virtual void remove(long index);
    virtual void remove_all();
    virtual long count();

    Manipulator* manipulator(long);
protected:
    ManipList* maniplist_;
};


#endif
