//
// Copyright (c) 1994 Silicon Graphics, Inc.
// Copyright (c) 1994 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
// OF THIS SOFTWARE.
//

/*
 * tools.c
 */

#include "commands.h"
#include "figviewer.h"
#include "manipulators.h"
#include "selection.h"
#include "tools.h"
#include "visitors.h"
#include <Fresco/defs_impl/region.h>
#include <Fresco/defs_impl/traversal.h>
#include <Fresco/Ox/schema.h>

SelectInfo::SelectInfo () {
    m_ = nil;
    a_ = nil;
}

SelectInfo::~SelectInfo () {
    CORBA::release(a_);
}

SelectInfo* SelectInfo::copy () {
    SelectInfo* scopy = new SelectInfo;
    scopy->t_.copy(&t_);
    scopy->level_ = level_;
    scopy->m_ = m_;
    if (is_not_nil(a_)) {
        scopy->a_ = new RegionImpl;
        scopy->a_->copy(a_);
    } else {        
        scopy->a_ = nil;
    }
    return scopy;
}

declarePtrList(SelectInfoList, SelectInfo)
implementPtrList(SelectInfoList, SelectInfo)

static void get_manipulators(
    GlyphTraversal_ptr gt, SelectInfoList* slist, long l
) {
    if (gt->forward()) {
        GlyphOffset_ptr go = gt->current_offset();
        Glyph_var g = go->child();
        Manipulator* m = Manipulator::_narrow(g);
        if (is_not_nil(m)) {
            SelectInfo* sinfo = new SelectInfo;
            sinfo->m_ = m; 
            sinfo->t_.copy(Transform_var(gt->current_transform())); 
            sinfo->level_ = l;
	    Region_var a = gt->allocation();
            if (is_not_nil(a)) {
                sinfo->a_ = new RegionImpl;
                sinfo->a_->copy(a);
            }
            slist->append(sinfo);
            CORBA::release(m);
        }
        get_manipulators(gt, slist, ++l);
        gt->backward();
    }
}

static long level (Manipulator* m) {
    SelectInfoCmd scmd;
    m->execute(&scmd);
    SelectInfo* sinfo = scmd.select_info();
    if (sinfo != nil) {
        return sinfo->level_;
    }
    return -1;
}

static Manipulator* create_macro (Selection* sel) {
    if (sel->count() == 0) {
        return nil;
    } else if (sel->count() == 1) {
        return sel->item(0);
    } else {
        MacroManip* mm = new MacroManip;
        for (long i = 0; i < sel->count(); i++) {
            Manipulator* m = sel->item(i);
            mm->add(m);
        }
        return mm;
    }
}

static void unselect_all(Selection* sel) {
    SelectCmd unselect(nil, false);
    for (long i = 0; i < sel->count(); i++) {
        Manipulator* m = sel->item(i);
        m->execute(&unselect);
    }
    sel->remove_all();
}

static void unselect(Selection* sel, Manipulator* m) {
    SelectCmd unselect(nil, false);
    m->execute(&unselect);
    sel->remove(m);
}

//+ Tool(Tool)
void* Tool::_this_ptr() { return this; }
_Ix_extern__tid(Tool);
void* Tool::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(Tool) == t) return this;
    void* p;
    if (p = XfFrescoObject::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(Tool);
_Ix_ImplementGetTid(Tool,_Ix__tid(Tool))
_Ix_ImplementNarrow(Tool*,Tool,_Ix_NullParameter,Tool,0)
_Ix_ImplementDuplicate(Tool*,Tool,_Ix_NullParameter,_Ix__tid(Tool),0)
//+

Tool::Tool() { }
Tool::~Tool() { }

OxExchange* Tool::_exchange() { return &object_; }

void Tool::anchor (Coord& ax, Coord& ay) {
    ax = ax_;
    ay = ay_;
}

//+ SelectTool(SelectTool)
void* SelectTool::_this_ptr() { return this; }
_Ix_extern__tid(SelectTool);
void* SelectTool::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(SelectTool) == t) return this;
    void* p;
    if (p = Tool::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(SelectTool);
_Ix_ImplementGetTid(SelectTool,_Ix__tid(SelectTool))
_Ix_ImplementNarrow(SelectTool*,SelectTool,_Ix_NullParameter,SelectTool,0)
_Ix_ImplementDuplicate(SelectTool*,SelectTool,_Ix_NullParameter,_Ix__tid(SelectTool),0)
//+

SelectInfo* SelectTool::create_manipulator (
    GlyphTraversal_ptr gt, Event_ptr e, FigViewer* fv
) {
    ax_ = e->pointer_x();
    ay_ = e->pointer_y();

    SelectInfoList slist;
    get_manipulators(gt, &slist, 0);
    SelectInfo* target = nil;
    Selection* sel = fv->selection();
    if (slist.count() != 0) {
        if (sel->count() == 0) {
            target = slist.item(1);
            slist.remove(1);
        } else {
            Boolean selected = false;
            for (Long i = 1; i < slist.count(); i++) {
                if (sel->selected(slist.item(i)->m_)) {
                    selected = true;
                    break;
                }
            }
            if (selected) {
                if (e->modifier_is_down(Event::control)) {
                    unselect_all(sel);
                } else if (e->modifier_is_down(Event::shift)) {
                    unselect_all(sel);
		    if (slist.count() > 1) {
			NaturalCmd natcmd;
			slist.item(i)->m_->execute(&natcmd);
		    }
                    i = i % (slist.count() - 1) + 1;
                } else {
                    unselect(sel, slist.item(i)->m_);
                }
                target = slist.item(i);
                slist.remove(i);
            } else if (e->modifier_is_down(Event::shift)) {
                target = slist.item(1);
                slist.remove(1);
                Selection* sel_copy = sel->copy();
                for (Long i = 0; i < sel_copy->count(); i++) {
                    if (level(sel->item(i)) != target->level_) {
                        unselect(sel, sel->item(i));
                    }
                }
                delete sel_copy;

            } else {
                unselect_all(sel);
                target = slist.item(1);
                slist.remove(1);
            }
        }
        for (Long i = 0; i < slist.count(); i++) {
            delete slist.item(i);
        }
    } else {
	if (!e->modifier_is_down(Event::shift)) {
	    unselect_all(sel);
	}
    }
    if (target != nil) {
        Glyph::AllocationInfo a;
        RegionImpl r;
        a.transformation = new TransformImpl;
        a.transformation->copy(&target->t_);
        target->m_->extension(a, &r);
        ax_ = (r.lower_.x + r.upper_.x) * 0.5;
        ay_ = (r.lower_.y + r.upper_.y) * 0.5;
    }
    return target;
}

SelectTool::SelectTool () {}

SelectTool::~SelectTool () {}

//+ CreateTool(CreateTool)
void* CreateTool::_this_ptr() { return this; }
_Ix_extern__tid(CreateTool);
void* CreateTool::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(CreateTool) == t) return this;
    void* p;
    if (p = Tool::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(CreateTool);
_Ix_ImplementGetTid(CreateTool,_Ix__tid(CreateTool))
_Ix_ImplementNarrow(CreateTool*,CreateTool,_Ix_NullParameter,CreateTool,0)
_Ix_ImplementDuplicate(CreateTool*,CreateTool,_Ix_NullParameter,_Ix__tid(CreateTool),0)
//+

CreateTool::CreateTool (Manipulator* m) {
    manip_ = m;
}

CreateTool::~CreateTool () {
    CORBA::release(manip_);
}

SelectInfo* CreateTool::create_manipulator(
    GlyphTraversal_ptr gt, Event_ptr e, FigViewer* fv
) {
    ax_ = e->pointer_x();
    ay_ = e->pointer_y();

    unselect_all(fv->selection());
    CopyCmd::glyphmap_.clear();
    Manipulator* fm = manip_->deep_copy();
    Manipulator* m = fv->root();
    Glyph_var b = m->body();
    b->append(fm);

    SelectCmd selectcmd(fv);
    fm->execute(&selectcmd);

    SelectInfo* sinfo = new SelectInfo;
    sinfo->level_ = 2; // you just know you start from level 2
    sinfo->m_ = fm;
    sinfo->t_.copy(Transform_var(gt->current_transform()));
    sinfo->t_.premultiply(Transform_var(fv->tform()));
    sinfo->t_.premultiply(Transform_var(m->transformation()));

    return sinfo;
}

//+ CommonTool(CommonTool)
void* CommonTool::_this_ptr() { return this; }
_Ix_extern__tid(CommonTool);
void* CommonTool::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(CommonTool) == t) return this;
    void* p;
    if (p = Tool::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(CommonTool);
_Ix_ImplementGetTid(CommonTool,_Ix__tid(CommonTool))
_Ix_ImplementNarrow(CommonTool*,CommonTool,_Ix_NullParameter,CommonTool,0)
_Ix_ImplementDuplicate(CommonTool*,CommonTool,_Ix_NullParameter,_Ix__tid(CommonTool),0)
//+

CommonTool::CommonTool() { }
CommonTool::~CommonTool() { }

SelectInfo* CommonTool::create_manipulator(
    GlyphTraversal_ptr gt, Event_ptr e, FigViewer* fv
) {
    SelectTool select_tool;
    SelectInfo* si = select_tool.create_manipulator(gt, e, fv);
    select_tool.anchor(ax_, ay_);
    if (si != nil) {
	SelectCmd selectcmd(fv);
	si->m_->execute(&selectcmd);
	si->m_ = create_macro(fv->selection());
    }
    return si;
}

//+ MoveTool(MoveTool)
void* MoveTool::_this_ptr() { return this; }
_Ix_extern__tid(MoveTool);
void* MoveTool::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(MoveTool) == t) return this;
    void* p;
    if (p = CommonTool::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(MoveTool);
_Ix_ImplementGetTid(MoveTool,_Ix__tid(MoveTool))
_Ix_ImplementNarrow(MoveTool*,MoveTool,_Ix_NullParameter,MoveTool,0)
_Ix_ImplementDuplicate(MoveTool*,MoveTool,_Ix_NullParameter,_Ix__tid(MoveTool),0)
//+

MoveTool::MoveTool() { }
MoveTool::~MoveTool() { }

//+ ScaleTool(ScaleTool)
void* ScaleTool::_this_ptr() { return this; }
_Ix_extern__tid(ScaleTool);
void* ScaleTool::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(ScaleTool) == t) return this;
    void* p;
    if (p = CommonTool::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(ScaleTool);
_Ix_ImplementGetTid(ScaleTool,_Ix__tid(ScaleTool))
_Ix_ImplementNarrow(ScaleTool*,ScaleTool,_Ix_NullParameter,ScaleTool,0)
_Ix_ImplementDuplicate(ScaleTool*,ScaleTool,_Ix_NullParameter,_Ix__tid(ScaleTool),0)
//+

ScaleTool::ScaleTool() { }
ScaleTool::~ScaleTool() { }

//+ RotateTool(RotateTool)
void* RotateTool::_this_ptr() { return this; }
_Ix_extern__tid(RotateTool);
void* RotateTool::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(RotateTool) == t) return this;
    void* p;
    if (p = CommonTool::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(RotateTool);
_Ix_ImplementGetTid(RotateTool,_Ix__tid(RotateTool))
_Ix_ImplementNarrow(RotateTool*,RotateTool,_Ix_NullParameter,RotateTool,0)
_Ix_ImplementDuplicate(RotateTool*,RotateTool,_Ix_NullParameter,_Ix__tid(RotateTool),0)
//+

RotateTool::RotateTool() { }
RotateTool::~RotateTool() { }

//+ AlterTool(AlterTool)
void* AlterTool::_this_ptr() { return this; }
_Ix_extern__tid(AlterTool);
void* AlterTool::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(AlterTool) == t) return this;
    void* p;
    if (p = CommonTool::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(AlterTool);
_Ix_ImplementGetTid(AlterTool,_Ix__tid(AlterTool))
_Ix_ImplementNarrow(AlterTool*,AlterTool,_Ix_NullParameter,AlterTool,0)
_Ix_ImplementDuplicate(AlterTool*,AlterTool,_Ix_NullParameter,_Ix__tid(AlterTool),0)
//+

AlterTool::AlterTool() { }
AlterTool::~AlterTool() { }

//+ ResizeTool(ResizeTool)
void* ResizeTool::_this_ptr() { return this; }
_Ix_extern__tid(ResizeTool);
void* ResizeTool::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(ResizeTool) == t) return this;
    void* p;
    if (p = CommonTool::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(ResizeTool);
_Ix_ImplementGetTid(ResizeTool,_Ix__tid(ResizeTool))
_Ix_ImplementNarrow(ResizeTool*,ResizeTool,_Ix_NullParameter,ResizeTool,0)
_Ix_ImplementDuplicate(ResizeTool*,ResizeTool,_Ix_NullParameter,_Ix__tid(ResizeTool),0)
//+

ResizeTool::ResizeTool() { }
ResizeTool::~ResizeTool() { }

//+ NarrowTool(NarrowTool)
void* NarrowTool::_this_ptr() { return this; }
_Ix_extern__tid(NarrowTool);
void* NarrowTool::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(NarrowTool) == t) return this;
    void* p;
    if (p = CommonTool::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(NarrowTool);
_Ix_ImplementGetTid(NarrowTool,_Ix__tid(NarrowTool))
_Ix_ImplementNarrow(NarrowTool*,NarrowTool,_Ix_NullParameter,NarrowTool,0)
_Ix_ImplementDuplicate(NarrowTool*,NarrowTool,_Ix_NullParameter,_Ix__tid(NarrowTool),0)
//+

NarrowTool::NarrowTool() {
    delta_ = 0;
}

NarrowTool::~NarrowTool() { }

Long NarrowTool::delta() { return delta_; }

SelectInfo* NarrowTool::create_manipulator(
    GlyphTraversal_ptr gt, Event_ptr e, FigViewer* fv
) {
    SelectTool select_tool;
    SelectInfo* si = select_tool.create_manipulator(gt, e, fv);
    select_tool.anchor(ax_, ay_);
    if (si != nil) {
	unselect_all(fv->selection());
	if (is_not_nil(PolyManip::_narrow(si->m_)) ||
	    is_not_nil(BoxManip::_narrow(si->m_))) {
	    fv->narrowgt(gt);
	    delta_ = si->level_;
	    gt->forward();
	    for (long count = 1; count <= si->level_; count ++) {
		gt->forward();
	    }
	    Manipulator* m = nil;
	    while (is_nil(m)) {
		gt->forward();
		count++;
		m = Manipulator::_narrow(Glyph_var(gt->current_glyph()));
	    }
            si->m_ = m; 
            si->t_.copy(Transform_var(gt->current_transform())); 
            si->level_ = count;
	    Region_var a = gt->allocation();
            if (is_not_nil(a)) {
                si->a_ = new RegionImpl;
                si->a_->copy(a);
            }
	    while (count--) gt->backward();
	    SelectCmd selectcmd(fv);
	    si->m_->execute(&selectcmd);
            CORBA::release(si->m_);
	} else {
	    delete(si);
	    si = nil;
	}
    }
    return si;
}
