//
// Copyright (c) 1994 Silicon Graphics, Inc.
// Copyright (c) 1994 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
// OF THIS SOFTWARE.
//

/*
 * vistitors.cxx
 */

#include "commands.h"
#include "manipulators.h"
#include "visitors.h"
#include <Fresco/defs_impl/region.h>
#include <Fresco/defs_impl/transform.h>
#include <stdlib.h> 

GlyphVisitor::GlyphVisitor() { }
GlyphVisitor::~GlyphVisitor() { }

Boolean GlyphVisitor::visit(Glyph_ptr, GlyphOffset_ptr) { return false; }

void GlyphVisitor::visit_children(Glyph_ptr parent) {
    GlyphOffset_var g = parent->first_child_offset();
    for (; is_not_nil(g); g = g->next_child()) {
	if (!visit(g->child(), g)) {
	    break;
	}
    }
}

void GlyphVisitor::visit_parents(Glyph_ptr parent) {
    GlyphOffset_ptr static_parents[5];
    for (Long i = 0; i < 5; i++) {
        static_parents[i] = nil;
    }
    Glyph::OffsetSeq parents(5, 0, static_parents);
    parent->parent_offsets(parents);
    for (Long j = 0; j < parents._length; j++) {
	GlyphOffset_ptr g = parents._buffer[j];
	if (!visit(g->parent(), g)) {
	    break;
	}
    }
}

implementPtrList(AllocationList, Region);
implementPtrList(TransformList, Transform);

Appender::Appender(Glyph_ptr glyph, Boolean resize) {
    glyph_ = glyph;
    resize_ = resize;
}

Boolean Appender::visit(Glyph_ptr parent, GlyphOffset_ptr) {
    parent->append(glyph_);
    if (resize_) {
        parent->need_resize();
    }
    return true;
}

Appender::~Appender () {}

Remover::Remover() {}

Remover::~Remover() {
    for (ListItr(GlyphOffsetList) i(list_); i.more(); i.next()) {
	GlyphOffset_ptr offset = i.cur();
        Glyph_var parent = offset->parent();
        Glyph_var child = offset->child();
        child->need_redraw();
	offset->notify_observers();
	offset->remove();
        parent->need_resize();
    }
}

Boolean Remover::visit(Glyph_ptr, GlyphOffset_ptr offset) {
    list_.append(offset);
    return true;
}

Counter::Counter() {
    count_ = 0;
}

Boolean Counter::visit(Glyph_ptr, GlyphOffset_ptr) {
    count_++;
    return true;
}

CmdVisitor::CmdVisitor (FCommand* cmd, Boolean e) {
    cmd_ = cmd;
    execute_ = e;
}

CmdVisitor::~CmdVisitor () {}

Boolean CmdVisitor::visit (Glyph_ptr g, GlyphOffset_ptr) {
    Manipulator* m = Manipulator::_narrow(g);
    if (is_not_nil(m)) {
        if (execute_) {
            m->execute(cmd_);
        } else {
            m->unexecute(cmd_);
        }
	CORBA::release(m);
    }
    return true;
}
    
ManipCopier::ManipCopier (Boolean s) {
    maniplist_ = new ManipList;
    shallow_ = s;
}

ManipCopier::~ManipCopier () {
    for (Long i = 0; i < maniplist_->count(); i++) {
        CORBA::release(maniplist_->item(i));
    }
    delete maniplist_;
}

Boolean ManipCopier::visit (Glyph_ptr g, GlyphOffset_ptr) {
    Manipulator* m = Manipulator::_narrow(g);
    if (is_not_nil(m)) {
        if (shallow_) {
            maniplist_->append(m->shallow_copy());
        } else {
            maniplist_->append(m->deep_copy());
        }
	CORBA::release(m);
    }
    return true;
}

TAManipCopier::TAManipCopier (Region_ptr a, Boolean) {
    alist_ = new AllocationList;
    tlist_ = new TransformList;
    a_ = a;
}

TAManipCopier::~TAManipCopier () {
    for (long i = 0; i < alist_->count(); i++) {
        CORBA::release(alist_->item(i));
    }
    delete alist_;
    for (long j = 0; i < tlist_->count(); j++) {
        CORBA::release(tlist_->item(j));
    }
    delete tlist_;
}

Boolean TAManipCopier::visit (Glyph_ptr g, GlyphOffset_ptr go) {
    ManipCopier::visit(g, go);
    RegionImpl* r = new RegionImpl;
    TransformImpl* t = new TransformImpl;
    r->copy(a_);

    Glyph::AllocationInfo a;
    a.allocation = r;
    a.transformation = t;
    a.damaged = nil;
    
    go->child_allocate(a);
    
    alist_->append(Region::_duplicate(r));
    tlist_->append(Transform::_duplicate(t));

    return true;
}

OffsetVisitor::OffsetVisitor () {
    glist_ = new GlyphOffsetList;
}

OffsetVisitor::~OffsetVisitor () {
    for (long i = 0; i < glist_->count(); i++) {
        CORBA::release(glist_->item(i));
    }
    delete glist_;
}

Boolean OffsetVisitor::visit (Glyph_ptr, GlyphOffset_ptr go) {
    glist_->append(GlyphOffset::_duplicate(go));
    return true;
}

GlyphOffset_ptr OffsetVisitor::offset (long i) {
    GlyphOffset_ptr go = nil;
    if (i < glist_->count()) {
        go = glist_->item(i);
    }
    return go;
}

long OffsetVisitor::offset_count () {
    return glist_->count();
}
