//
// Copyright (c) 1993 Silicon Graphics, Inc.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the name of
// Silicon Graphics may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL SILICON GRAPHICS BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#ifndef Fresco_defs_fresco_h
#define Fresco_defs_fresco_h

#include <Fresco/defs/fcontext.h>
#include <Fresco/defs/types.h>
#include <Fresco/defs/viewer.h>

//- Option
struct export Option {
    //. An option describes a possible command-line argument in terms
    //. of a name that matches the argument string, a path that
    //. defines the root-style attribute to set, a format, and
    //. an optional implicit value.  If the format is "implicit" then
    //. the attribute will be set to the implicit value.
    //. If the format is "value" then the attribute will be set
    //. to the next argument on the command-line.  If the format
    //. is "path_value" then the next argument is a string containing
    //. the attribute path and value separated by a colon (":").

    enum Format { path_value, value, implicit };

    const char* name;
    const char* path;
    Option::Format format;
    const char* implicit_value;
};

//- FrescoLib
class export FrescoLib {
    //. FrescoLib is a module that provides convenience functions
    //. for accessing objects through the Fresco naming context,
    //. as well as creating strings.
public:
    //- init_context, init_default_context, init_display_context
    static FrescoContext* init_context(
	const char* class_name, int& argc, char** argv, Option* = nil
    );
    static FrescoContext* init_default_context(
	const char* class_name, int& argc, char** argv, Option* = nil
    );
    static FrescoContext* init_display_context(
	const char* display_name,
	const char* class_name, int& argc, char** argv, Option* = nil
    );
	//. These operations initialize and return the current
	//. naming context for accessing Fresco objects.
	//. The class_name and argument list is accessible from
	//. the nested "style" context.
	//.
	//. The operations differ in the way the treat information
	//. associated with a display connection.  The init_context
	//. operation does not open a display, init_default_context
	//. opens the default display, and init_display_context opens
	//. a specified display.  The name of the default display
	//. is platform-specific; on POSIX systems it can be set by the
	//. DISPLAY environment variable.
	//.
	//. Display names are of the form N:D.S, where N is
	//. an optional network host name.  If omitted, then
	//. the display is assumed to be on the local host.
	//. D is a display number on the host, and S is
	//. the number of the screen to use on the display
	//. by default.

    //- current_context
    static void current_context(FrescoContext*);
    static FrescoContext* current_context();
	//. Set or get the current naming context as a FrescoContext
	//. object, which provides access to naming system service
	//. as well as the information used to create the context.
	//. These operations assume that the object reference
	//. is statically allocated, and as such no memory management
	//. is implied.  The caller therefore should not release
	//. the returned kit object reference.

    //- run
    static void run(Viewer_ptr v, Glyph_ptr g);
	//. Start and run a "main loop" in the common case
	//. with a single viewer.  The target display is assumed
	//. from the display specified in a previous call
	//. to init_default_context or init_display_context.
	//. If the init_context operation was used to access
	//. the initial context, then the run operation will open
	//. a new connection to the default display.
	//.
	//. This operation creates and maps a main viewer
	//. on the default screen of the current display,
	//. if set by init_default_context or init_display_context,
	//. or otherwise the default display.
	//. 
	//. The main viewer uses the given viewer for input
	//. (unless the viewer is nil, in which case input is ignored) and
	//. the given glyph for output (unless the glyph is nil,
	//. in which case the given viewer is used for output).
	//. After mapping the viewer, run calls Display::run and
	//. waits for it to return.

    //- string_ref, tmp_string_ref, string_copy, tmp_string_copy
    static CharString_ptr string_ref(const char*);
    static CharString_ptr string_copy(const char*);
    static CharString_var tmp_string_ref(const char*);
    static CharString_var tmp_string_copy(const char*);
	//. Create a string from a null-terminated array
	//. of characters.  The string_ref operation simply
	//. references the data; the string_copy operation
	//. copies the data into its own storage that is freed
	//. when the string is deallocated.  The tmp_string_ref and
	//. tmp_string_copy operations return a managed rather than
	//. raw object reference.

    //- get_string, get_tmp_string
    //- get_fcontext_string, get_tmp_fcontext_string
    static CharString_ptr get_string(StyleContext_ptr, const char* name);
    static CharString_var get_tmp_string(StyleContext_ptr, const char* name);
    static CharString_ptr get_fcontext_string(
	FrescoContext*, const char* name
    );
    static CharString_var get_tmp_fcontext_string(
	FrescoContext*, const char* name
    );
	//. Return the string bound to the given name in
	//. either a given style or the root style associated
	//. with a given FrescoContext.  The operations return
	//. a nil string if the given name is not defined.
	//.
	//. The get_tmp_string operation returns
	//. a managed rather than a raw reference.  These operations
	//. are effectively short-hand for creating a string, calling
	//. resolve on the style, and calling read_string on the
	//. result if non-nil.

    //- resolve_string, tmp_resolve_string
    static CORBA::Object_ptr resolve_string(FrescoContext*, const char*);
    static CORBA::Object_var tmp_resolve_string(FrescoContext*, const char*);
	//. Return an object reference for the named object in the
	//. given context.  These functions are short-hand for converting
	//. the string to a name (using NamingLib::to_name) and passing
	//. the result to FrescoContext::resolve.

    //- keysym_to_string, string_to_keysym
    static char* keysym_to_string(Event::KeySym);
    static Event::KeySym string_to_keysym(const char*);
	//. Convert a string to or from a keysym.

    //- delay
    static Boolean delay(Float seconds);
	//. Pause execution the given number of seconds or fraction
	//. of a second.  Return true if execution was successfully
	//. suspended the requested duration.
protected:
    static FrescoContext* current_;
};

inline CharString_var FrescoLib::tmp_string_ref(const char* s) {
    return string_ref(s);
}

inline CharString_var FrescoLib::tmp_string_copy(const char* s) {
    return string_copy(s);
}

inline CharString_var FrescoLib::get_tmp_string(
    StyleContext_ptr s, const char* name
) {
    return get_string(s, name);
}

inline CharString_var FrescoLib::get_tmp_fcontext_string(
    FrescoContext* fc, const char* name
) {
    return get_fcontext_string(fc, name);
}

inline CORBA::Object_var FrescoLib::tmp_resolve_string(
    FrescoContext* fc, const char* name
) {
    return resolve_string(fc, name);
}

//
// This macro is just for brain-damaged C++ compilers like cfront that
// construct globals using the name of an early symbol in the file.
//

#define Fresco_file_id(name) \
char name = 0;

#endif
