/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1995 Silicon Graphics, Inc.
 * Copyright (c) 1995 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include <new.h>
#include <stdio.h>

#include <Fresco/Ox/schema.h>

#include <Fresco/defs/fresco.h>

#include <Fresco/defs_impl/dload.h>
#include <Fresco/defs_impl/table.h>

#include <Fresco/defs_Mac/MacCursor.h>
#include <Fresco/defs_Mac/MacDisplay.h>
#include <Fresco/defs_Mac/MacDrawing.h>
#include <Fresco/defs_Mac/MacEvent.h>
#include <Fresco/defs_Mac/MacLib.h>
#include <Fresco/defs_Mac/MacScreen.h>
#include <Fresco/defs_Mac/MacWindow.h>

#include <Sound.h>

// class MacDisplayContext

class MacDisplayContext : public DisplayContext {
public:
    MacDisplayContext() {}
    ~MacDisplayContext() {}

    OxExchange* _exchange() { return &exchange_; }

    //+ XfDisplayContext::*
    /* XfDisplayContext */
    XfDisplay_ptr open_string(const char* n);
    XfDisplay_ptr open_default();
    //+

    static MacDisplay* display_;
    
private:
    OxLibraryExchange exchange_;
};

dload(MacDisplayContext)

MacDisplay* MacDisplayContext::display_ = nil;

//+ MacDisplayContext(XfDisplayContext::open_string)
XfDisplay_ptr MacDisplayContext::open_string(const char* n) {
    return open_default();
}

//+ MacDisplayContext(XfDisplayContext::open_default)
XfDisplay_ptr MacDisplayContext::open_default() {
    if (display_ == nil) {
        display_ = new MacDisplay;
    }
    return display_;
}


// class MacDisplay

enum MenuBars { rMenuBar = 128 };
enum Menus { mApple = 128, mFile, mEdit };
enum Apple { miAbout = 1 };
enum File { miQuit = 3 };

enum Alerts { rAbout = 128 };

static void out_of_memory() {
	MacAssert("out of memory" == nil);
}

static void init_macintosh() {
	MaxApplZone();
	InitGraf(&qd.thePort);
	InitFonts();
	FlushEvents(everyEvent, 0);
	InitWindows();
	InitMenus();
	TEInit();
	InitDialogs(0L);
	InitCursor();
	
	set_new_handler(out_of_memory);
}

declareTable(MacDisplayWindowTable,WindowPtr,MacWindow*)
implementTable(MacDisplayWindowTable,WindowPtr,MacWindow*)

MacDisplay::MacDisplay() {
    init_macintosh();

    resume_id_ = 0;

	Handle menu_bar = GetNewMBar(rMenuBar);
	MacAssert(menu_bar);
	SetMenuBar(menu_bar);
	DisposeHandle(menu_bar);
	MenuHandle apple = GetMHandle(mApple);
	MacAssert(apple);
	AddResMenu(apple, 'DRVR');
	DrawMenuBar();

    if (double_click_ == 0.0) {
        double_click_ = GetDblTime() / 60.0;
    }

    nscreens_ = 1;
    screen_ = new MacScreen*[nscreens_];
    for (long i = 0; i < nscreens_; i++) {
        screen_[i] = new MacScreen(this, i);
    }

    wtable_ = new MacDisplayWindowTable(256);
}

MacDisplay::~MacDisplay() {
    for (long i = 0; i < nscreens_; i++) {
        CORBA::release(screen_[i]);
    }
    delete [] screen_;

    delete wtable_;
}

//+ MacDisplay(XfDisplay::drawing_kit)
XfDrawingKit_ptr MacDisplay::drawing_kit() {
    return new MacDrawingKit(this);
}

//+ MacDisplay(XfDisplay::number_of_screens)
XfDisplay::ScreenNumber MacDisplay::number_of_screens() {
    return nscreens_;
}

//+ MacDisplay(XfDisplay::default_screen)
XfScreen_ptr MacDisplay::default_screen() {
    return Screen::_duplicate(screen_[0]);
}

//+ MacDisplay(XfDisplay::display_screen)
XfScreen_ptr MacDisplay::display_screen(XfDisplay::ScreenNumber n) {
    MacAssert(n < nscreens_);
    return Screen::_duplicate(screen_[n]);
}

//+ MacDisplay(XfDisplay::cursor_from_data)
XfCursor_ptr MacDisplay::cursor_from_data(Short x, Short y, CursorPattern pat, CursorPattern mask) {
    return new MacCursor(x, y, pat, mask);
}

//+ MacDisplay(XfDisplay::cursor_from_bitmap)
XfCursor_ptr MacDisplay::cursor_from_bitmap(XfRaster_ptr b, XfRaster_ptr mask) {
    return new MacCursor(b, mask);
}

//+ MacDisplay(XfDisplay::cursor_from_font)
XfCursor_ptr MacDisplay::cursor_from_font(XfFont_ptr f, Long pat, Long mask) {
    return new MacCursor(f, pat, mask);
}

//+ MacDisplay(XfDisplay::cursor_from_index)
XfCursor_ptr MacDisplay::cursor_from_index(Long n) {
    return new MacCursor(n);
}

Boolean MacDisplay::run_one_event_bool() {
    Boolean prepared = prepare_to_read();
    if (prepared) {
        EventRecord event;
        Boolean result = GetNextEvent(everyEvent, &event);
        after_read();
        if (result) {
            dispatch(event);
        }
    }
    return prepared;
}

Boolean MacDisplay::peek() {
    EventRecord event;
    return EventAvail(everyEvent, &event);
}

// Convert between keysyms and strings.

static char* keysym_names[256] = {
    "", "lbutton", "rbutton", "cancel", "mbutton", "", "", "",
    "back", "tab", "", "", "clear", "return", "", "",
    "shift", "control", "menu", "pause", "capital", "", "", "",
    "", "", "", "escape", "", "", "", "",
    "space", "prior", "next", "end", "home", "left", "up", "right",
    "down", "select", "print", "execute", "snapshot", "insert", "delete", "help",
    "0", "1", "2", "3", "4", "5", "6", "7",
    "8", "9", "", "", "", "", "", "",
    "", "A", "B", "C", "D", "E", "F", "G",
    "H", "I", "J", "K", "L", "M", "N", "O",
    "P", "Q", "R", "S", "T", "U", "V", "W",
    "X", "Y", "Z", "lwin", "rwin", "apps", "", "",
    "numpad0", "numpad1", "numpad2", "numpad3", "numpad4", "numpad5", "numpad6", "numpad7",
    "numpad8", "numpad9", "multiply", "add", "separator", "subtract", "decimal", "divide",
    "f1", "f2", "f3", "f4", "f5", "f6", "f7", "f8",
    "f9", "f10", "f11", "f12", "f13", "f14", "f15", "f16",
    "f17", "f18", "f19", "f20", "f21", "f22", "f23", "f24",
    "", "", "", "", "", "", "", "",
    "numlock", "scroll", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "",
    "lshift", "rshift", "lcontrol", "rcontrol", "lmenu", "rmenu", "", "",
    "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "", "",
    "", "", "", "", "", "", "attn", "crsel",
    "exsel", "ereof", "play", "zoom", "noname", "pa1", "oem_clear", ""
};

//+ MacDisplay(XfDisplay::keysym_to_string)
char* MacDisplay::keysym_to_string(XfEvent::KeySym k) {
    return (0 <= k) && (k <= 256) ? keysym_names[k] : "";
}

declareTable(MacStringToKeysymTable,const char*,XfEvent::KeySym)
implementTable(MacStringToKeysymTable,const char*,XfEvent::KeySym)

//+ MacDisplay(XfDisplay::string_to_keysym)
XfEvent::KeySym MacDisplay::string_to_keysym(const char* s) {
    static MacStringToKeysymTable* string_to_keysym_table = nil;
    if (string_to_keysym_table	== nil) {
        string_to_keysym_table = new MacStringToKeysymTable(256);
        for (Event::KeySym i = 0; i < 256; ++i) {
            if (*keysym_names[i] != '\0') {
                string_to_keysym_table->insert(keysym_names[i], i);
            }
        }
    }
    Event::KeySym keysym;
    if (string_to_keysym_table->find(keysym, s)) {
        return keysym;
    } else {
        return 0; // !!! what to return in this case? -denis
    }
}

//+ MacDisplay(XfDisplay::flush)
void MacDisplay::flush() {
}

//+ MacDisplay(XfDisplay::flush_and_wait)
void MacDisplay::flush_and_wait() {
}

//+ MacDisplay(XfDisplay::ring_bell)
void MacDisplay::ring_bell(Float pct_loudness) {
    if (pct_loudness >= 0) {
        long volume;
        GetSysBeepVolume(&volume);
        SetSysBeepVolume(long(pct_loudness * 100.0f));
        SysBeep(30);
        SetSysBeepVolume(volume);
    }
}

//+ MacDisplay(XfDisplay::auto_repeat=b)
void MacDisplay::auto_repeat(Boolean b) {
    // !!! unimplemented
    DisplayBase::auto_repeat(b);
}

//+ MacDisplay(XfDisplay::auto_repeat?)
Boolean MacDisplay::auto_repeat() {
    // !!! unimplemented
    return DisplayBase::auto_repeat();
}

//+ MacDisplay(XfDisplay::key_click_volume=v)
void MacDisplay::key_click_volume(Float v) {
    // !!! unimplemented
    DisplayBase::key_click_volume(v);
}

//+ MacDisplay(XfDisplay::key_click_volume?)
Float MacDisplay::key_click_volume() {
    // !!! unimplemented
    return DisplayBase::key_click_volume();
}

//+ MacDisplay(XfDisplay::pointer_acceleration=a)
void MacDisplay::pointer_acceleration(Float a) {
    // !!! unimplemented
    DisplayBase::pointer_acceleration(a);
}

//+ MacDisplay(XfDisplay::pointer_acceleration?)
Float MacDisplay::pointer_acceleration() {
    // !!! unimplemented
    return DisplayBase::pointer_acceleration();
}

//+ MacDisplay(XfDisplay::pointer_threshold=t)
void MacDisplay::pointer_threshold(Long t) {
    // !!! unimplemented
    DisplayBase::pointer_threshold(t);
}

//+ MacDisplay(XfDisplay::pointer_threshold?)
Long MacDisplay::pointer_threshold() {
    // !!! unimplemented
    return DisplayBase::pointer_threshold();
}

//+ MacDisplay(XfDisplay::pointer_double_click_threshold=t)
void MacDisplay::pointer_double_click_threshold(Float t) {
    DisplayBase::pointer_double_click_threshold(t);
}

//+ MacDisplay(XfDisplay::pointer_double_click_threshold?)
Float MacDisplay::pointer_double_click_threshold() {
    return DisplayBase::pointer_double_click_threshold();
}

//
// Associate the given window id with the given window object.
//
void MacDisplay::bind(WindowPtr id, MacWindow* window) {
    lock_->acquire();
    wtable_->insert(id, window);
    lock_->release();
}

//
// Disassociate the given window id with the given window object.
//
void MacDisplay::unbind(WindowPtr id) {
    lock_->acquire();
    wtable_->remove(id);
    lock_->release();
}

void MacDisplay::do_adjust_menus() {
}

void MacDisplay::do_apple_menu(int menu_item) {
	if (menu_item == miAbout) {
		Alert(rAbout, nil);
	} else {
		Str255 name;
		MenuHandle apple = GetMHandle(mApple);
		MacAssert(apple);
		GetItem(apple, menu_item, name);
		OpenDeskAcc(name);
	}
}

void MacDisplay::do_file_menu(int menu_item) {
	switch (menu_item) {
		case miQuit:
			run(false);
		break;
	}
}

void MacDisplay::do_edit_menu(int menu_item) {
	if (! SystemEdit(menu_item - 1)) {
		SysBeep(5);
	}
}

void MacDisplay::do_menu_command(int menu_select) {
	int menu_id = HiWord(menu_select);
	int menu_item = LoWord(menu_select);
	
	switch (menu_id) {
		case mApple:
			do_apple_menu(menu_item);
		break;
		case mFile:
			do_file_menu(menu_item);
		break;
		case mEdit:
			do_edit_menu(menu_item);
		break;
	}
}

static Boolean is_movable_modal(WindowPtr) {
    return false;
}

WindowPtr MacDisplay::do_mouse_down(const EventRecord& event) {
    WindowPtr window;
    int part = FindWindow(event.where, &window);
    
    switch (part) {
	    case inMenuBar:{
			// make sure menu items are properly enabled/disabled
			do_adjust_menus();
			
			// let user choose a menu command
			do_menu_command(MenuSelect(event.where));
		} break;
		
	    case inSysWindow: { // cursor is in a desk accessory window
		    SystemClick(&event, window);
		} break;
		
	    case inContent: {
	        if (window != FrontWindow()) { // cursor is not in front window
			    if (is_movable_modal(FrontWindow())) { // front window is movable modal
				    SysBeep(30);
			    } else { // front window is not movable modal
				    SelectWindow(window); // make window active
		        }
		    } else { // cursor is in content region of active window
			    return window; // handle event in content region
			}
		} break;
		
	    case inDrag: { // cursor is in drag area
		    // if a movable modal is active, ignore click in an inactive title bar
		    if ((window != FrontWindow()) && is_movable_modal(FrontWindow())) {
			    SysBeep(30);
		    } else {
			    // let Window Manager drag window
			    DragWindow(window, event.where, &(*GetGrayRgn())->rgnBBox);
			    Rect r = window->portRect;
	            MacWindow* window = nil;
	            if (wtable_->find(window, (WindowPtr) event.message)) {
			        window->do_move(r.left, r.top);
			    }
			}
	    } break;
	    
	    case inGrow: { // cursor is in size box
	        // change window size
			Rect smallest;
			SetRect(&smallest, 64, 64, 65535, 65535);
			long size = GrowWindow(window, event.where, &smallest);
			if (size) {
	            MacWindow* window = nil;
	            if (wtable_->find(window, (WindowPtr) event.message)) {
	                window->do_size(LoWord(size), HiWord(size));
	            }
			}
		} break;
		
	    case inGoAway: { // cursor is in close box
		    // call TrackGoAway to handle mouse until button is released
		    if (TrackGoAway(window, event.where)) {
			    HideWindow(window); // handle close window
			}
		} break;
		
	    case inZoomIn:
	    case inZoomOut: { // cursor is in zoom box
		    // call TrackBox to handle mouse until button is released
		    if (TrackBox(window, event.where, part)) {
			    ZoomWindow(window, part, true); // handle zoom window
			}
	    } break;
    }
    
    return 0;
}

void MacDisplay::do_os_event(const EventRecord& event) {
	switch ((event.message & osEvtMessageMask) >> 24) {
		case suspendResumeMessage: {
			if (event.message & resumeFlag) {
				if (resume_id_ != 0) {
					SelectWindow(resume_id_);
					resume_id_ = 0;
				}
			} else {
				resume_id_ = FrontWindow();
				if (resume_id_ != 0) {
					// dim controls
				}
			}
		} break;
	}
}

void MacDisplay::do_disk_event(const EventRecord& event) {
	if (HiWord(event.message) != noErr) {
	    // attempt to mount was unsuccessful
		DILoad(); // load Disk Initialization Manager
		
		// notify the user
        Point point;
		point.h =120;
		point.v =120;
	    OSErr myErr = DIBadMount(point, event.message);
		DIUnload(); // unload Disk Initialization Manager
	} else {
        // attempt to mount was successful
        // record the drive number or do other processing
    }
}

void MacDisplay::dispatch(const EventRecord& event) {
    WindowPtr id;
    switch (event.what) {
		case mouseDown: {
		    id = do_mouse_down(event);
		} break;
		
		case keyDown:
		case autoKey: {
		    if (event.modifiers & cmdKey) {
		    	do_adjust_menus();
				do_menu_command(MenuKey(char(event.message & charCodeMask)));
			    id = 0;
			} else {
				FindWindow(event.where, &id);
			}
        } break;
        
		case mouseUp:
		case keyUp:
		case osEvt: {
		    FindWindow(event.where, &id);
		} break;
		
		case updateEvt: {
            MacWindow* window = nil;
            if (wtable_->find(window, (WindowPtr) event.message)) {
                window->do_update();
            }
		    id = 0;
		} break;
		
		case activateEvt: {
		    id = (WindowPtr) event.message;
		} break;
		
		case diskEvt: {
		    do_disk_event(event);
		    id = 0;
		} break;

        case nullEvent:
 		case kHighLevelEvent: {
 		    id = 0;
 		} break;
    }
    
    if (id != 0) {
        MacWindow* window = nil;
        if (wtable_->find(window, id)) {
            EventRecord e = event; // !!! this seems to be due to a bug in MW C++ 1.2
            MacEvent* event = new MacEvent((MacWindow*) window, e);
 	        if (! filtered(event)) {
	            window->handle_event(event);
            }
            CORBA::release(event);
        }
    }
}
