/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1995 Silicon Graphics, Inc.
 * Copyright (c) 1995 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include <string.h>

#include <Fresco/Ox/schema.h>

#include <Fresco/defs_impl/charstr.h>
#include <Fresco/defs_impl/glyphs.h>
#include <Fresco/defs_impl/math.h>
#include <Fresco/defs_impl/region.h>
#include <Fresco/defs_impl/traversal.h>
#include <Fresco/defs_impl/viewers.h>
	  
#include <Fresco/defs_Mac/MacCursor.h>
#include <Fresco/defs_Mac/MacDisplay.h>
#include <Fresco/defs_Mac/MacEvent.h>
#include <Fresco/defs_Mac/MacLib.h>
#include <Fresco/defs_Mac/MacPainter.h>
#include <Fresco/defs_Mac/MacScreen.h>
#include <Fresco/defs_Mac/MacWindow.h>

// class MacWindow

/*
 * A window doesn't hold a reference to the display or screen,
 * as the display's style could indirectly hold a reference to
 * the main viewer.  We could attach to the display to guarantee
 * safety, but for now we assume the display will live as long
 * if not longer than a window.
 */

MacWindow::MacWindow(
    MacDisplay* d, MacScreen* s, Viewer_ptr v
) : WindowBase(d, v) {
    window_ = 0;
    
    display_ = d;
    screen_ = s;

    buffer_bitmap_ = nil;
    painter_ = new MacPainter(s);
    draw_->current_painter(painter_);
}

MacWindow::~MacWindow() {
    if (buffer_bitmap_ != nil) {
        DisposeGWorld(buffer_bitmap_);
    }
    
    if (window_ != nil) {
        DisposeWindow(window_);
    }
    
    CORBA::release(painter_);
}

void MacWindow::decor_adjust(int& /* x */, int& /* y */, int& w, int& h) {
    w -= 15;
    h -= 15;
}

void MacWindow::do_update() {
	BeginUpdate(window_);
	
	Rect r = window_->portRect;
	ClipRect(&r);

	RgnHandle vr = window_->visRgn;
	Rect content;
	content.top = r.top;
	content.left = r.left;
	content.right = r.right - 15;
	content.bottom = r.bottom - 15;
	RgnHandle cr = NewRgn();
	RectRgn(cr, &content);
	RgnHandle ir = NewRgn();
	SectRgn(vr, cr, ir);
	if (! EqualRgn(vr, ir)) {
	    // only controls need to be updated
		RgnHandle wr = NewRgn();
		DiffRgn(vr, cr, wr);
		EraseRgn(wr);
		UpdateControls(window_, wr);
		DrawGrowIcon(window_);
		DisposeRgn(wr);
	}
	if (! EmptyRgn(ir)) {
		// damage content area
	    Coord left = screen_->to_coord(content.left - 1, X_axis);
	    Coord bottom = screen_->to_coord(height_ - content.bottom - 1, Y_axis);
	    Coord width = screen_->to_coord(content.right - content.left + 2, X_axis);
	    Coord height = screen_->to_coord(content.bottom - content.top + 2, Y_axis);
	    redraw(left, bottom, width, height);
	}
	DisposeRgn(cr);
	DisposeRgn(ir);
	
	EndUpdate(window_);
}

void MacWindow::do_size(int cx, int cy) {
    lock_->acquire();

    // release old double-buffer resources
    if (buffer_bitmap_) {
		DisposeGWorld(buffer_bitmap_);
		buffer_bitmap_ = nil;
    }

    // remember the window size
    width_ = cx;
    height_ = cy;

    // determine how big the interior wants to be
    Glyph::Requisition r;
    GlyphImpl::init_requisition(r);
    viewer_->request(r);

    // fill in the viewer allocation
    viewer_->allocate(screen_->to_pixel(width_, X_axis), screen_->to_pixel(height_, Y_axis), 0);

    // schedule entire area to be re-rendered
    incur();

    lock_->release();
}

void MacWindow::do_move(int x, int y) {
    left_ = x;
    top_ = y;
}

//+ MacWindow(XfWindow::screen_ptr)
XfScreen_ptr MacWindow::screen_ptr() {
    return Screen::_duplicate(screen_);
}

//+ MacWindow(XfWindow::configure)
void MacWindow::configure(const XfWindow::Placement& p) {
    lock_->acquire();

    WindowBase::configure(p);

    if (window_ != nil) {
        int x = left_;
        int y = top_;
        int w = width_;
        int h = height_;
        decor_adjust(x, y, w, h);
        MoveWindow(window_, x, y, FALSE);
    }

    lock_->release();
}

//+ MacWindow(XfWindow::map)
void MacWindow::map() {
    lock_->acquire();
    if (window_ == nil) {
        create_window();
    }
    MacAssert(window_ != nil);
    ShowWindow(window_);
    lock_->release();
}

//+ MacWindow(XfWindow::unmap)
void MacWindow::unmap() {
    lock_->acquire();
    HideWindow(window_);
    lock_->release();
}

//+ MacWindow(XfWindow::is_mapped)
Boolean MacWindow::is_mapped() {
    return  ((CWindowRecord*) window_)->visible;
}

//+ MacWindow(XfWindow::iconify)
void MacWindow::iconify() {
    unmap();
}

//+ MacWindow(XfWindow::deiconify)
void MacWindow::deiconify() {
    map();
}

//+ MacWindow(XfWindow::raise)
void MacWindow::raise() {
    lock_->acquire();
    SelectWindow(window_);
    lock_->release();
}

//+ MacWindow(XfWindow::lower)
void MacWindow::lower() {
    lock_->acquire();
    SendBehind(window_, nil);
    lock_->release();
}

/*
 * Initiate a draw traversal on the window's viewer.
 * Note that we keep the window locked during the traversal.
 * Not sure this is the best approach, but it is simple and
 * right not there doesn't appear to be any need to access
 * the window while drawing (except for damage, which is protected
 * by a different lock).
 */

//+ MacWindow(XfWindow::repair)
void MacWindow::repair() {
    lock_->acquire();

    // block other threads from changing the damage while we are 
    // determining what area we will update.  Stash the coordinates 
    // because we will let other threads change the damage before 
    // we are done updating.
    //
    damage_lock_->acquire();
    Coord dl = damage_->lower_.x;
    Coord db = damage_->lower_.y;
    Coord dr = damage_->upper_.x;
    Coord dt = damage_->upper_.y;

	CGrafPtr cg;
	GDHandle gd;
	GetGWorld(&cg, &gd);
	PixMapHandle pixmap = 0;
    if (double_buffer_) {
        if (buffer_bitmap_ == nil) {
			Rect c;
			c.top = 0;
			c.left = 0;
			c.bottom = width_;
			c.right = height_;
			QDErr error = NewGWorld(&buffer_bitmap_, 0, &c, nil, nil, nil);
			MacAssert(error == noErr);
			MacAssert(buffer_bitmap_ != nil);
		}
		
	    SetGWorld(buffer_bitmap_, nil);
	    pixmap = GetGWorldPixMap(buffer_bitmap_);
		Boolean good = LockPixels(pixmap);
		MacAssert(good); // !!! should turn off double buffering if this fails?
    } else {
	    SetGWorld((CGrafPort*) window_, GetMainDevice());
    }

    // --- prepare the painter with the appropriate device context ---
    painter_->begin_paint();
    painter_->push_clipping();
    painter_->clip_rect(dl, db, dr, dt);

    // The damage is marked undefined, and the lock is released to allow
    // other threads access again.  We have stashed the damage coordinates
    // in case they are needed for the double-buffering.
    damage_->defined_ = false;
    damage_lock_->release();

    render();

    // ---- finish with painter ----
    painter_->pop_clipping();
    painter_->end_paint();
    
    // ---- see if we need to blt ----
    if (double_buffer_) {
		PixMapHandle pixmap = GetGWorldPixMap(buffer_bitmap_);
		Boolean good = LockPixels(pixmap);
		MacAssert(good);
		Rect region;
		region.left = screen_->to_coord(dl, X_axis);
		region.top = height_ - screen_->to_coord(dt, Y_axis);
		region.right = screen_->to_coord(dr, X_axis);
		region.bottom = height_ - screen_->to_coord(db, Y_axis);
		CopyBits(
			&GrafPtr(double_buffer_)->portBits, &GrafPtr(window_)->portBits,
			&((CGrafPort*) double_buffer_)->portRect, &region, srcCopy, nil
		);
		UnlockPixels(pixmap);
    }
    
    if (pixmap != 0) {
	    UnlockPixels(pixmap);
	}
	SetGWorld(cg, gd);

    lock_->release();
}

//+ MacWindow(XfWindow::grab_pointer)
void MacWindow::grab_pointer(XfCursor_ptr c) {
    // !!! unimplemented
}

//+ MacWindow(XfWindow::ungrab_pointer)
void MacWindow::ungrab_pointer() {
    // !!! unimplemented
}

// class MacManagedWindow

MacManagedWindow::MacManagedWindow(
    MacDisplay* display, MacScreen* screen, Viewer_ptr viewer
) : MacWindow(display, screen, viewer) {
}
 
MacManagedWindow::~MacManagedWindow() {
}

void MacManagedWindow::new_window(void* proc) {
    PixelCoord x, y, w, h;
    get_client_geometry(x, y, w, h);
    if (! configured_) {
        // !!! We should follow the Macintosh rules for placing new windows. -denis
        x = 0;
        y = 0;
    }
    
    // Fill in the viewer allocation that we expect to get.
    viewer_->allocate(screen_->to_pixel(w, X_axis), screen_->to_pixel(h, Y_axis), 0);

    Rect rect;
    rect.left = x;
    rect.top = y;
    rect.right = x + w;
    rect.bottom = x + h;
    Str255 title;
    title[0] = '\0'; // !!! need to set title
    window_ = NewCWindow(nil, &rect, title, FALSE, proc, (WindowPtr) -1, TRUE, 0);
}

// class MacApplicationWindow

MacApplicationWindow::MacApplicationWindow(
    MacDisplay* display, MacScreen* screen, Viewer_ptr viewer
) : MacManagedWindow(display, screen, viewer) {
}
 
MacApplicationWindow::~MacApplicationWindow() {
}

void MacApplicationWindow::create_window() {
    new_window(zoomDocProc);
}

// class MacTopLevelWindow

MacTopLevelWindow::MacTopLevelWindow(
    MacDisplay* display, MacScreen* screen, Viewer_ptr viewer, Window_ptr group_leader
) : MacManagedWindow(display, screen, viewer) {
    group_leader_ = Window::_duplicate(group_leader);
}

MacTopLevelWindow::~MacTopLevelWindow() {
    CORBA::release(group_leader_);
}

void MacTopLevelWindow::create_window() {
    new_window(zoomDocProc);
}

// class MacTransientWindow

MacTransientWindow::MacTransientWindow(
    MacDisplay* display, MacScreen* screen, Viewer_ptr viewer, Window_ptr transient_for
) : MacManagedWindow(display, screen, viewer) {
    transient_for_ = Window::_duplicate(transient_for);
}

MacTransientWindow::~MacTransientWindow() {
    CORBA::release(transient_for_);
}

void MacTransientWindow::create_window() {
    new_window(movableDBoxProc);
}

// class MacPopupWindow

MacPopupWindow::MacPopupWindow(
    MacDisplay* display, MacScreen* screen, Viewer_ptr viewer
) : MacManagedWindow(display, screen, viewer) {
}
 
MacPopupWindow::~MacPopupWindow() {
}

void MacPopupWindow::create_window() {
    new_window(altDBoxProc);
}

// class MacIconWindow

MacIconWindow::MacIconWindow(
    MacDisplay* display, MacScreen* screen, Viewer_ptr viewer
) : MacManagedWindow(display, screen, viewer) {
}

MacIconWindow::~MacIconWindow() {
}

void MacIconWindow::create_window() {
    new_window(plainDBox);
}
