/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1993-95 Silicon Graphics, Inc.
 * Copyright (c) 1993-95 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include <Fresco/defs/fresco.h>

#include <Fresco/defs_impl/math.h>
#include <Fresco/defs_impl/memory.h>
#include <Fresco/defs_impl/region.h>
#include <Fresco/defs_impl/transform.h>

#include <Fresco/Ox/schema.h>
	  
#include <Fresco/defs_Win32/Wdrawing.h>
#include <Fresco/defs_Win32/Wpainter.h>
#include <Fresco/defs_Win32/Wscreen.h>

/*
 * Two useful array macros: determine the number of elements and
 * return the address of the end of the array (after last element).
 * We assume the compiler will fold any constant arithmetic.
 */

#define array_nelements(array) (sizeof(array) / sizeof(array[0]))
#define array_end(array) (&array[array_nelements(array)])

declareList(Win32ClippingStack,HRGN)
implementList(Win32ClippingStack,HRGN)

// class Win32TDB

class Win32TDB {
public:
    Win32TDB(XfRaster_ptr raster, XfTransform_ptr transform, unsigned long age);
    ~Win32TDB();

    XfTransform_ptr transform() { return &transform_; }

    unsigned long age() { return age_; }
    void age(unsigned long age) { age_ = age; }

    PixelCoord raster_origin_x() { return raster_origin_x_; }
    PixelCoord raster_origin_y() { return raster_origin_y_; }

    PixelCoord width() { return width_; }
    PixelCoord height() { return height_; }
    PixelCoord origin_x() { return origin_x_; }
    PixelCoord origin_y() { return origin_y_; }

    unsigned long size() { return width_ * height_ * 3; }

    void create();
    
    void* color_bits() { return color_bits_; }
    BITMAPINFO* color_info() { return &color_info_; }

    HBITMAP mask() { return mask_; }

private:
    XfRaster_ptr raster_;
    TransformImpl transform_;
    unsigned long age_;

    PixelCoord raster_origin_x_;
    PixelCoord raster_origin_y_;

    PixelCoord width_;
    PixelCoord height_;
    PixelCoord origin_x_;
    PixelCoord origin_y_;

    TransformImpl* it_;

    void* color_bits_;
    BITMAPINFO color_info_;

    HBITMAP mask_;
};

inline int pad(int c, int s) {
    return (s - (c & (s - 1))) & (s - 1);
}

Win32TDB::Win32TDB(XfRaster_ptr raster, XfTransform_ptr transform, unsigned long age) {
    raster_ = XfRaster::_duplicate(raster);
    transform_.copy(transform);
    age_ = age;
    color_bits_ = nil;
    mask_ = nil;

    Raster::Info ri;
    raster->raster_info(ri);
    raster_origin_x_ = ri.origin_x;
    raster_origin_y_ = ri.origin_y;

    XfVertex v1;
    v1.x = Coord(0);
    v1.y = Coord(0);
    v1.z = Coord(0);
    transform_.transform_vertex(v1);
    XfVertex v2;
    v2.x = Coord(0);
    v2.y = Coord(ri.rows - 1);
    v2.z = Coord(0);
    transform_.transform_vertex(v2);
    XfVertex v3;
    v3.x = Coord(ri.columns - 1);
    v3.y = Coord(ri.rows - 1);
    v3.z = Coord(0);
    transform_.transform_vertex(v3);
    XfVertex v4;
    v4.x = Coord(ri.columns - 1);
    v4.y = Coord(0);
    v4.z = Coord(0);
    transform_.transform_vertex(v4);

    Coord bbl = Math::min(v1.x, v2.x, v3.x, v4.x);
    Coord bbb = Math::min(v1.y, v2.y, v3.y, v4.y);
    Coord bbr = Math::max(v1.x, v2.x, v3.x, v4.x);
    Coord bbt = Math::max(v1.y, v2.y, v3.y, v4.y);

    width_ = PixelCoord(bbr - bbl) + 1;
    height_ = PixelCoord(bbt - bbb) + 1;

    origin_x_ = PixelCoord(-bbl);
    origin_y_ = PixelCoord(-bbb);

    it_ = new TransformImpl;
    it_->copy(&transform_);
    Vertex vlb;
    vlb.x = -bbl;
    vlb.y = -bbb;
    vlb.z = Coord(0);
    it_->translate(vlb);
    it_->invert();
}

void Win32TDB::create() {
    BITMAPINFOHEADER& ch = color_info_.bmiHeader;
    ch.biSize		= sizeof(BITMAPINFOHEADER);
    ch.biWidth		= width_;
    ch.biHeight		= height_;
    ch.biPlanes		= 1;
    ch.biBitCount	= 24;
    ch.biCompression	= BI_RGB;
    ch.biSizeImage	= 0;
    ch.biXPelsPerMeter	= 0; // not used?
    ch.biYPelsPerMeter	= 0; // not used?
    ch.biClrUsed	= 0; // not used?
    ch.biClrImportant	= 0; // not used?

    int color_pad = pad(ch.biWidth * 3, sizeof(LONG));
    long color_size = (ch.biWidth * 3 + color_pad) * ch.biHeight;
    color_bits_ = new unsigned char[color_size];

    BITMAP bm;
    bm.bmType		= 0;
    bm.bmWidth		= width_;
    bm.bmHeight		= height_;
    bm.bmWidthBytes	= ((bm.bmWidth + 15) / 16) * 2;
    bm.bmPlanes		= 1;
    bm.bmBitsPixel	= 1;
    long bm_size = bm.bmWidthBytes * bm.bmHeight;
    bm.bmBits = new unsigned char[bm_size];

    Boolean mask_pad = ((bm.bmWidth + 7) / 8) % 2;

    unsigned char* cp = (unsigned char*) color_bits_;
    unsigned char* mp = (unsigned char*) bm.bmBits;
    XfVertex v;
    XfColor* color;
    XfColor::Intensity r = XfColor::Intensity(0);
    XfColor::Intensity g = XfColor::Intensity(0);
    XfColor::Intensity b = XfColor::Intensity(0);
    Float blend;
    unsigned char mb;
    int index;
    for (XfRaster::Index row = 0; row < height_; ++row) {
        mb = 0xff;
	index = 7;
        for (XfRaster::Index col = 0; col < width_; ++col) {
	    v.x = Coord(col);
            v.y = Coord(row);
            v.z = Coord(0);
	    it_->transform_vertex(v);
	    PixelCoord raster_col = PixelCoord(v.x);
	    PixelCoord raster_row = PixelCoord(v.y);
	    if (
	        (0 <= raster_col) && (raster_col < width_) &&
	        (0 <= raster_row) && (raster_row < height_)
	    ) {
	        raster_->get_color_blend(raster_row, raster_col, color, blend);
	        color->rgb(r, g, b);
		MWassert(color != nil);
	        CORBA::release(color);
	    } else {
	        blend = 0.0f;
	    }

	    if (blend > 0.99f) {
	        mb &= ~(1 << index);
	    } else {
	        r = g = b = XfColor::Intensity(1);
	    }
	    if (index > 0) {
	        --index;
	    } else {
	        *mp++ = mb;
	        mb = 0xff;
	        index = 7;
	    }

	    *cp++ = (unsigned char) (b * 255.001);
	    *cp++ = (unsigned char) (g * 255.001);
	    *cp++ = (unsigned char) (r * 255.001);
	}
	for (int i = 0; i < color_pad; ++i) {
	    *cp++ = 0;
	}

	if (index != 7) {
	    *mp++ = mb;
	}
	if (mask_pad) {
	    *mp++ = 0xff;
	}
    }

    CORBA::release(it_);
    it_ = nil;

    CORBA::release(raster_);
    raster_ = nil;

    mask_ = CreateBitmapIndirect(&bm);
    MWassert(mask_);
    delete bm.bmBits;
}

Win32TDB::~Win32TDB() {
    if (! is_nil(it_)) {
        CORBA::release(it_);
    }

    delete color_bits_;

    if (mask_ != nil) {
        BOOL result = DeleteObject(mask_);
        MWassert(result);
    }
}

// class Win32DIB

class Win32DIB {
public:
    Win32DIB(XfRaster_ptr raster, ULong hash);
    Win32DIB(XfFont_ptr font, XfCharCode character, ULong hash);
    ~Win32DIB();

    XfFont_ptr font() { return font_; }
    XfCharCode character() { return character_; }

    XfRaster_ptr raster() { return raster_; }
    ULong hash() { return hash_; }

    Win32TDB* find(XfTransform_ptr transform, unsigned long age);
    Win32TDB* create(XfTransform_ptr transform, unsigned long age);
    void remove(unsigned long i);

    unsigned long count() { return count_; }
    Win32TDB* tdb(unsigned long i) { return buffer_[i]; }

private:
    XfFont_ptr font_;
    XfCharCode character_;

    XfRaster_ptr raster_;
    ULong hash_;

    Win32TDB** buffer_;
    unsigned long size_;
    unsigned long count_;
};

Win32DIB::Win32DIB(XfRaster_ptr raster, ULong hash) {
    font_ = nil;
    character_ = 0;

    raster_ = XfRaster::_duplicate(raster);
    hash_ = hash;

    size_ = 8;
    buffer_ = new Win32TDB*[size_];
    count_ = 0;
}

Win32DIB::Win32DIB(XfFont_ptr font, XfCharCode character, ULong hash) {
    font_ = XfFont::_duplicate(font);
    character_ = character;

    // !!! bitmap should be member function of XfFont.
    Win32Font* fi = Win32Font::_narrow(font);
    raster_ = fi->bitmap(character);
    CORBA::release(fi);
    hash_ = hash;

    size_ = 8;
    buffer_ = new Win32TDB*[size_];
    count_ = 0;
}

Win32DIB::~Win32DIB() {
    CORBA::release(font_);

    CORBA::release(raster_);

    for (unsigned long i = 0; i < count_; ++i) {
        delete buffer_[i];
    }
    delete buffer_;
}

Win32TDB* Win32DIB::find(XfTransform_ptr transform, unsigned long age) {
    for (unsigned long i = 0; i < count_; ++i) {
        Win32TDB* tdb = buffer_[i];
	if (transform->equal(tdb->transform())) {
	    tdb->age(age);
	    return tdb;
	}
    }
    return nil;
}

Win32TDB* Win32DIB::create(XfTransform_ptr transform, unsigned long age) {
    if (count_ == size_) {
        Win32TDB** old_buffer = buffer_;
	size_ = size_ + size_;
	buffer_ = new Win32TDB*[size_];
	Memory::copy(buffer_, old_buffer, sizeof(Win32TDB*) * count_);
    }

    Win32TDB* tdb = new Win32TDB(raster_, transform, age);
    buffer_[count_] = tdb;
    ++count_;

    return tdb;
}

void Win32DIB::remove(unsigned long i) {
    delete buffer_[i];
    for (unsigned long j = i; j < (count_ - 1); ++j) {
        buffer_[j] = buffer_[j + 1];
    }
    buffer_[count_ - 1] = nil;
    --count_;
}

// class Win32DIBCache

class Win32DIBCache {
public:
    Win32DIBCache();
    ~Win32DIBCache();

    Win32TDB* find(XfRaster_ptr raster, XfTransform_ptr transform);
    Win32TDB* find(XfFont_ptr font, XfCharCode character, XfTransform_ptr transform);

private:
    unsigned long slot(XfRaster_ptr raster, ULong hash);
    unsigned long slot(XfFont_ptr font, XfCharCode character, ULong hash);
    unsigned long slot(Win32DIB* dib, ULong hash);
    void rehash();
    Win32TDB* find(Win32DIB* dib, XfTransform_ptr transform);
    void remove(unsigned long i);

    Win32DIB** buffer_;
    unsigned long size_;
    unsigned long count_;
    unsigned long age_;
    unsigned long memory_size_;
    unsigned long memory_size_limit_;
};

Win32DIBCache::Win32DIBCache() {
    size_ = 64;
    buffer_ = new Win32DIB*[size_];
    Memory::zero(buffer_, size_ * sizeof(Win32DIB*));
    count_ = 0;
    age_ = 0;
    memory_size_ = 0;
    memory_size_limit_ = 4000000; // approx. 4 megabyte limit
}

Win32DIBCache::~Win32DIBCache() {
    for (unsigned long i = 0; i < size_; ++i) {
	Win32DIB* dib = buffer_[i];
	if (dib != nil) {
	    delete dib;
	}
    }
    delete buffer_;
}

inline unsigned long Win32DIBCache::slot(XfRaster_ptr raster, ULong hash) {
    unsigned long index = hash % size_;
    while (
        (buffer_[index] != nil) &&
        !is_eq(raster, buffer_[index]->raster())
    ) {
        index = (index + 1) % size_;
    }
    return index;
}

inline unsigned long Win32DIBCache::slot(XfFont_ptr font, XfCharCode character, ULong hash) {
    unsigned long index = hash % size_;
    while (
        (buffer_[index] != nil) &&
        !(is_eq(font, buffer_[index]->font()) && (buffer_[index]->character() == character))
    ) {
        index = (index + 1) % size_;
    }
    return index;
}

inline unsigned long Win32DIBCache::slot(Win32DIB* dib, ULong hash) {
    unsigned long index = hash % size_;
    while (
        (buffer_[index] != nil) &&
        (dib != buffer_[index])
    ) {
        index = (index + 1) % size_;
    }
    return index;
}

void Win32DIBCache::rehash() {
    Win32DIB** old_buffer = buffer_;
    unsigned long old_size = size_;
    size_ = count_ * 4;
    buffer_ = new Win32DIB*[size_];
    Memory::zero(buffer_, size_ * sizeof(Win32DIB*));
    for (unsigned long i = 0; i < old_size; ++i) {
        Win32DIB* dib = old_buffer[i];
        if (dib != nil) {
            buffer_[slot(dib, dib->hash())] = dib;
        }
    }
    delete old_buffer;
}

Win32TDB* Win32DIBCache::find(Win32DIB* dib, XfTransform_ptr transform) {
    // Remove the translation component for transformed object lookup.
    // We should also limit the scale and stretch blit...
    XfVertex v;
    v.x = v.y = v.z = Coord(0);
    transform->transform_vertex(v);
    v.x = -v.x;
    v.y = -v.y;
    v.z = -v.z;
    TransformImpl nt;
    nt.copy(transform);
    nt.translate(v);

    ++age_;
    Win32TDB* tdb = dib->find(&nt, age_);
    if (! tdb) {
        tdb = dib->create(&nt, age_);

	// Reduce the cache to stay within the limit.
	while ((memory_size_ > 0) && ((memory_size_ + tdb->size()) > memory_size_limit_)) {
	    Win32DIB* candidate_dib = nil;
	    unsigned long candidate_i = 0;
	    Win32TDB* candidate_tdb = nil;
	    unsigned long candidate_j = 0;
	    for (unsigned long i = 0; i < size_; ++i) {
	        Win32DIB* dib = buffer_[i];
		for (unsigned long j = 0; j < dib->count(); ++j) {
		    Win32TDB* tdb = dib->tdb(j);
		    if ((candidate_tdb == nil) || (tdb->age() < candidate_tdb->age())) {
			candidate_i = i;
		        candidate_dib = dib;
			candidate_j = j;
		        candidate_tdb = tdb;
		    }
		}
	    }
	    if ((candidate_tdb != nil) && (candidate_tdb != tdb)) {
		memory_size_ -= candidate_tdb->size();
	        candidate_dib->remove(candidate_j);
		if ((candidate_dib != dib) && (candidate_dib->count() == 0)) {
		    remove(candidate_j);
		}
	    }
	}

	memory_size_ += tdb->size();
	tdb->create();
    }
    return tdb;
}

Win32TDB* Win32DIBCache::find(XfRaster_ptr raster, XfTransform_ptr transform) {
    ULong hash = raster->hash();
    unsigned long index = slot(raster, hash);

    Win32DIB* dib = buffer_[index];
    if (dib == nil) {
        dib = new Win32DIB(raster, hash);
	buffer_[index] = dib;
	count_++;
	if ((count_ * 3) > size_) {
            rehash();
        }
    }

    return find(dib, transform);
}

Win32TDB* Win32DIBCache::find(XfFont_ptr font, XfCharCode character, XfTransform_ptr transform) {
    ULong hash = font->hash() ^ (character * 5);
    unsigned long index = slot(font, character, hash);

    Win32DIB* dib = buffer_[index];
    if (dib == nil) {
        dib = new Win32DIB(font, character, hash);
	buffer_[index] = dib;
	count_++;
	if ((count_ * 3) > size_) {
            rehash();
        }
    }

    return find(dib, transform);
}

void Win32DIBCache::remove(unsigned long index) {
    delete buffer_[index];
    buffer_[index] = nil;
    --count_;

    // Rehash if there might have been an element that collided with
    // this one.  We should be smarter and just rehash the section of
    // the table that could have had collisions	with this element.
    if (buffer_[(index + 1) % size_] != nil) {
        rehash();
    }
}

static Win32DIBCache cache;

static TransformImpl identity_transform;

// class Win32Painter

//+ XfWin32Painter(XfWin32Painter)
void* XfWin32Painter::_this_ptr() { return this; }
_Ix_extern__tid(XfWin32Painter);
void* XfWin32Painter::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(XfWin32Painter) == t) return this;
    void* p;
    if (p = XfPainter::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(XfWin32Painter);
_Ix_ImplementGetTid(XfWin32Painter,_Ix__tid(XfWin32Painter))
_Ix_ImplementNarrow(XfWin32Painter*,XfWin32Painter,_Ix_NullParameter,XfWin32Painter,0)
_Ix_ImplementDuplicate(XfWin32Painter*,XfWin32Painter,_Ix_NullParameter,_Ix__tid(XfWin32Painter),0)
//+

Win32Painter::Win32Painter(Win32Screen* s) {
    screen_ = s;

    clippers_ = new Win32ClippingStack;
    clipping_ = CreateRectRgn(0, 0, 0, 0);
    clippers_->append(clipping_);

    // ---- set brush template to a solid white ----
    brush_stats_.lbStyle = BS_SOLID;
    brush_stats_.lbColor = RGB(255,255,255);

    mode_ = 0;

    init_items();

    // The following variables are only valid within a begin_paint
    // and end_paint pair, but initialize them here just to be safe.
    drawable_ = nil;

    orig_pen_ = nil;
    orig_brush_ = nil;
    orig_font_ = nil;
}

Win32Painter::~Win32Painter() {            
    MWassert(drawable_ == nil);

    delete clippers_;
}

OxExchange* Win32Painter::_exchange() { return &object_; } 

void Win32Painter::begin_paint(HDC hdc, POINT& p) {
    PainterBase::begin_paint();

    drawable_ = hdc;

    orig_pen_ = (HPEN) GetCurrentObject(drawable_, OBJ_PEN);
    orig_brush_ = (HBRUSH) GetCurrentObject(drawable_, OBJ_BRUSH);
    orig_font_ = (HFONT) GetCurrentObject(drawable_, OBJ_FONT);

    // Try to use the advanced graphics mode.  This works for Windows NT
    // but not on Windows 95.  We remember the graphics mode so we know
    // if we need to do our own transformations or if the graphics
    // system will do it for us.
    int result = SetGraphicsMode(drawable_, GM_ADVANCED);
    if (result != 0) {
        mode_ = GM_ADVANCED;
    } else {
        int result = SetGraphicsMode(drawable_, GM_COMPATIBLE);
        MWassert(result);
        mode_ = GM_COMPATIBLE;
    }

    // ---- initialize clipping to cover entire area ----
    result = SelectClipRgn(drawable_, NULL);
    MWassert(result != ERROR);
    result = SetRectRgn(clipping_, 0, 0, p.x, p.y);
    MWassert(result);
    result = SelectClipRgn(drawable_, clipping_);
    MWassert(result != ERROR);

    // ---- use TWIPS mapping mode for convenient conversion ----
    result = SetMapMode(drawable_, MM_TWIPS);
    MWassert(result);
    result = DPtoLP(drawable_, &p, 1);
    MWassert(result);
    result = SetWindowOrgEx(drawable_, 0, -p.y, NULL);
    MWassert(result);

    // ---- text properties that apply to all text rendering ----
    SetTextAlign(drawable_, TA_LEFT | TA_BASELINE | TA_NOUPDATECP);
    SetBkMode(drawable_, TRANSPARENT);

    SetStretchBltMode(drawable_, COLORONCOLOR);
}

void Win32Painter::end_paint() {
    flush_text();

    PainterBase::end_paint();

    if (orig_pen_) {
        if (GetCurrentObject(drawable_, OBJ_PEN) != orig_pen_) {
            DeleteObject(SelectObject(drawable_, orig_pen_));
	}
        orig_pen_ = nil;
    }
    if (orig_brush_) {
        if (GetCurrentObject(drawable_, OBJ_BRUSH) != orig_brush_) {
            DeleteObject(SelectObject(drawable_, orig_brush_));
        }
        orig_brush_ = nil;
    }
    if (orig_font_) {
	// The font is owned by the Font object, so we don't need to delete
	// the previously selected font handle.  This will be done in the
	// destructor of the Font object when the font is no longer needed.
        SelectObject(drawable_, orig_font_);
        orig_font_ = nil;
    }

    drawable_ = nil;
}

//+ Win32Painter(XfPainter::to_coord)
XfCoord Win32Painter::to_coord(XfPixelCoord p, XfAxis a) {
    return screen_->to_coord(p, a);
}

//+ Win32Painter(XfPainter::to_pixels)
XfPixelCoord Win32Painter::to_pixels(XfCoord c, XfAxis a) {
    return screen_->to_pixels(c, a);
}

//+ Win32Painter(XfPainter::to_pixels_coord)
XfCoord Win32Painter::to_pixels_coord(XfCoord c, XfAxis a) {
    return screen_->to_pixels_coord(c, a);
}

//+ Win32Painter(XfPainter::current_brush=b)
void Win32Painter::current_brush(XfBrush_ptr b) {
    if (brush_ == b) {
        return;
    }

    PainterBase::current_brush(b);

    // !!! Need to handle scaling the brush when not in advanced graphics mode,
    // !!! but then we need to handle arbitrary transformations (like when the
    // !!! x and y scales are different).

    // ---- set new pen width ----
    LONG width = brush_ == nil ? 0 : coord_to_twips(brush_->width());
    HPEN hp = ExtCreatePen(
        PS_GEOMETRIC | PS_SOLID | PS_ENDCAP_FLAT | PS_JOIN_BEVEL,
        width, &brush_stats_, 0, NULL
    );
    DeleteObject(SelectObject(drawable_, hp));
}

//+ Win32Painter(XfPainter::current_color=c)
void Win32Painter::current_color(XfColor_ptr c) {
    if (color_ == c) {
        return;
    }

    flush_text();

    PainterBase::current_color(c);

    // need a table here from Color -> pixel
    Color::Intensity r, g, b;
    color_->rgb(r, g, b);
    COLORREF mc = RGB((BYTE) (r * float(255)), (BYTE) (g * float(255)), (BYTE) (b * float(255)));

    // get closest color
    screen_->find_color(nil, mc);

    // ----  set new brush color ----
    brush_stats_.lbStyle = BS_SOLID;
    brush_stats_.lbColor = mc;
    HBRUSH hb = CreateBrushIndirect(&brush_stats_);
    DeleteObject(SelectObject(drawable_, hb));

    // ---- set new pen color ----
    LONG width = brush_ == nil ? 0 : coord_to_twips(brush_->width());
    HPEN hp = ExtCreatePen(
        PS_GEOMETRIC | PS_SOLID | PS_ENDCAP_FLAT | PS_JOIN_BEVEL,
        width, &brush_stats_, 0, NULL
    );
    DeleteObject(SelectObject(drawable_, hp));

    // ---- set text color ----
    SetTextColor(drawable_, mc);
}

//+ Win32Painter(XfPainter::current_font=f)
void Win32Painter::current_font(XfFont_ptr f) {
    if (font_ == f) {
        return;
    }

    flush_text();

    PainterBase::current_font(f);

    Win32Font* font = Win32Font::_narrow(f);
    if (! is_nil(font)) {
        HFONT hf = font->mfont();
        if (hf) {
            SelectObject(drawable_, hf);
	}
    } else {
        // We should find a win32 font that matches the font spec...
        // !!! unimplemented
    }

    CORBA::release(font);
}

void Win32Painter::select_path() {
    flush_text();

    BeginPath(drawable_);

    if (mode_ == GM_ADVANCED) {
        for (int i = 0; i < path_.count(); ++i) {
            PathImpl::Point& p = path_.point(i);
            switch (p.type) {
	        case PathImpl::MoveTo: {
	            MoveToEx(drawable_, coord_to_twips(p.x), coord_to_twips(p.y), NULL);
	        } break;
	        case PathImpl::LineTo: {
	            LineTo(drawable_, coord_to_twips(p.x), coord_to_twips(p.y));
	        } break;
	        case PathImpl::CurveTo: {
                    PathImpl::Point& p1 = path_.point(i + 1);
                    PathImpl::Point& p2 = path_.point(i + 2);
	            POINT pts[3];
		    pts[2].x = coord_to_twips(p.x);
		    pts[2].y = coord_to_twips(p.y);
		    pts[0].x = coord_to_twips(p1.x);
		    pts[0].y = coord_to_twips(p1.y);
		    pts[1].x = coord_to_twips(p2.x);
		    pts[1].y = coord_to_twips(p2.y);
	            PolyBezierTo(drawable_, pts, 3);
                    i += 2;
	        } break;
	    }
	}
    } else {
        Transform::Matrix m;
        matrix_->store_matrix(m);
        m[0][0] *= 20.0f;
        m[0][1] *= 20.0f;
        m[1][0] *= 20.0f;
        m[1][1] *= 20.0f;
        m[2][0] *= 20.0f;
        m[2][1] *= 20.0f;
        path_.transform(m);

        for (int i = 0; i < path_.count(); ++i) {
            PathImpl::Point& p = path_.point(i);
            switch (p.type) {
	        case PathImpl::MoveTo: {
	            MoveToEx(drawable_, p.px, p.py, NULL);
	        } break;
	        case PathImpl::LineTo: {
	            LineTo(drawable_, p.px, p.py);
	        } break;
	        case PathImpl::CurveTo: {
                    PathImpl::Point& p1 = path_.point(i + 1);
                    PathImpl::Point& p2 = path_.point(i + 2);
	            POINT pts[3];
		    pts[2].x = p.px;
		    pts[2].y = p.py;
		    pts[0].x = p1.px;
		    pts[0].y = p1.py;
		    pts[1].x = p2.px;
		    pts[1].y = p2.py;
	            PolyBezierTo(drawable_, pts, 3);
                    i += 2;
	        } break;
	    }
        }
    }
        
    if (path_.is_closed()) {
        CloseFigure(drawable_);
    }
    EndPath(drawable_);
}

//+ Win32Painter(XfPainter::stroke)
void Win32Painter::stroke() {
    select_path();
    int result = StrokePath(drawable_);
    MWassert(result);
}

//+ Win32Painter(XfPainter::fill)
void Win32Painter::fill() {
    select_path();
    int result = FillPath(drawable_);
    MWassert(result);
}

// The text rendering is buffered so that a larger request is made over
// to the GDI subsystem.

void Win32Painter::init_items() {
    char_ = chars_;
    width_ = widths_;
    item_ = items_;
    item_->lpstr = char_;
    item_->n = 0;
    item_->x = 0;
    item_->y = 0;
    item_->uiFlags = nil;
    item_->pdx = width_;
}

void Win32Painter::flush_text() {
    if (char_ == chars_) {
        return;
    }
    int n = item_ - items_ + 1;
    BOOL result = PolyTextOut(drawable_, items_, n);
    if (! result) {
        for (int i = 0; i < n; ++i) {
            POLYTEXT& item = items_[i];
            result = TextOut(drawable_, item.x, item.y, item.lpstr, item.n);
            MWassert(result);
        }
    }
    init_items();
}

void Win32Painter::add_item() {
    POLYTEXT* i = item_;
    if (i->n > 0) {
        ++i;
        if (i >= array_end(items_)) {
            flush_text();
            i = item_;
        } else {
            item_ = i;
        }
        i->lpstr = char_;
        i->n = 0;
        i->x = 0;
        i->y = 0;
        i->uiFlags = nil;
        i->pdx = width_;
    }
}

/*
 * This function tests whether a matrix is only a translation.
 * Something like this probably should be provided directly
 * by the Transform interface.
 */
static Boolean is_translation(Transform_ptr t) {
    Transform::Matrix m;
    t->store_matrix(m);
    return m[0][0] == 1 && m[0][1] == 0 && m[1][0] == 0 && m[1][1] == 1;
}

//+ Win32Painter(XfPainter::character)
void Win32Painter::character(XfCharCode ch, XfCoord width, XfCoord x, XfCoord y) {
    if ((mode_ == GM_ADVANCED) || is_translation(matrix_)) {
        add_item();
        *width_ = 0; // coord_to_twips(width);
        width_++;
        if (mode_ == GM_ADVANCED) {
            item_->x = coord_to_twips(x);
            item_->y = coord_to_twips(y);
        } else {
            XfVertex v;
            v.x = x;
            v.y = y;
            v.z = Coord(0);
            matrix_->transform_vertex(v);
            item_->x = coord_to_twips(v.x);
            item_->y = coord_to_twips(v.y);
        }
        item_->n = 1;
        *char_ = ch;
        char_++;
    } else {
        mask_blt(x, y, cache.find(font_, ch, matrix_));
    }
}

HDC Win32Painter::prep_blt(Coord x, Coord y, Win32TDB* tdb, int& x0, int& y0, int& width, int& height) {
    if (mode_ == GM_ADVANCED) {
        x0 = coord_to_twips(x - to_coord(tdb->raster_origin_x(), X_axis));
        y0 = coord_to_twips(y - to_coord(tdb->raster_origin_y(), Y_axis));
    } else {
        XfVertex v;
        v.x = x - to_coord(tdb->raster_origin_x(), X_axis);
        v.y = y - to_coord(tdb->raster_origin_y(), Y_axis);
        v.z = Coord(0);
        matrix_->transform_vertex(v);
        x0 = coord_to_twips(v.x - to_coord(tdb->origin_x(), X_axis));
        y0 = coord_to_twips(v.y - to_coord(tdb->origin_y(), Y_axis));
    }
    width = coord_to_twips(to_coord(tdb->width(), X_axis));
    height = coord_to_twips(to_coord(tdb->height(), Y_axis));
    		    
    HDC hdc = CreateCompatibleDC(drawable_);
    MWassert(hdc);
    SelectObject(hdc, tdb->mask());

    SetBkColor(drawable_, RGB(255, 255, 255));
    COLORREF c = GetTextColor(drawable_);
    SetTextColor(drawable_, RGB(0, 0, 0));

    BOOL result = StretchBlt(
        drawable_, x0, y0, width, height, 
    	hdc, 0, 0, tdb->width(), tdb->height(),
	SRCAND
    );
    MWassert(result);

    SetBkColor(drawable_, RGB(0, 0, 0));
    SetTextColor(drawable_, c);

    return hdc;
}

void Win32Painter::post_blt(HDC hdc) {
    BOOL result = DeleteDC(hdc);
    MWassert(result);
}

void Win32Painter::mask_blt(Coord x, Coord y, Win32TDB* tdb) {
    flush_text();

    int x0, y0, width, height;
    HDC hdc = prep_blt(x, y, tdb, x0, y0, width, height);

    BOOL result = StretchBlt(
        drawable_, x0, y0, width, height, 
    	hdc, 0, 0, tdb->width(), tdb->height(),
	SRCPAINT
    );
    MWassert(result);

    post_blt(hdc);
}

//+ Win32Painter(XfPainter::image)
void Win32Painter::image(XfRaster_ptr r, XfCoord x, XfCoord y) {
    flush_text();

    Win32TDB* tdb = (mode_ == GM_ADVANCED ? cache.find(r, &identity_transform) : cache.find(r, matrix_));
    int x0, y0, width, height;
    HDC hdc = prep_blt(x, y, tdb, x0, y0, width, height);

    int lines = StretchDIBits(
        drawable_, x0, y0 + height, width, height, 
	0, 0, tdb->width(), tdb->height(),
	tdb->color_bits(), tdb->color_info(), 
	DIB_RGB_COLORS, SRCPAINT
    );
    MWassert(lines != GDI_ERROR);

    post_blt(hdc);
}

//+ Win32Painter(XfPainter::stencil)
void Win32Painter::stencil(XfRaster_ptr r, XfCoord x, XfCoord y) {
    mask_blt(x, y, mode_ == GM_ADVANCED ? cache.find(r, &identity_transform) : cache.find(r, matrix_));
}

void Win32Painter::update_world_transform() {
    if (mode_ == GM_ADVANCED) {
        Transform::Matrix m;
        matrix_->store_matrix(m);

        XFORM x;
        x.eM11 = m[0][0]; 
        x.eM12 = m[0][1]; 
        x.eM21 = m[1][0]; 
        x.eM22 = m[1][1]; 
        x.eDx = m[2][0] * 20.0f;
        x.eDy = m[2][1] * 20.0f;
        BOOL result = SetWorldTransform(drawable_, &x);
        MWassert(result);
    }
}

//+ Win32Painter(XfPainter::current_matrix)
void Win32Painter::current_matrix(Transform_ptr t) {
    flush_text();
    PainterBase::current_matrix(t);
    update_world_transform();
}

//+ Win32Painter(XfPainter::pop_matrix)
void Win32Painter::pop_matrix() {
    flush_text();
    PainterBase::pop_matrix();
    update_world_transform();
}

//+ Win32Painter(XfPainter::premultiply)
void Win32Painter::premultiply(XfTransform_ptr t) {
    flush_text();
    PainterBase::premultiply(t);
    update_world_transform();
}

//+ Win32Painter(XfPainter::clip)
void Win32Painter::clip() {
    select_path();

    // ---- set the new clipping ----
    HRGN rgn;
    rgn = PathToRegion(drawable_);
    MWassert(rgn);
    int result1 = ExtSelectClipRgn(drawable_, rgn, RGN_AND);
    MWassert(result1 != ERROR);

    // ---- replace the old region -----
    DeleteObject(clipping_);
    clipping_ = rgn;
}

//+ Win32Painter(XfPainter::push_clipping)
void Win32Painter::push_clipping() {
    flush_text();

    HRGN old_clip = clipping_;
    HRGN new_clip = CreateRectRgn(0, 0, 0, 0);
    CombineRgn(new_clip, old_clip, old_clip, RGN_COPY);
    clippers_->append(old_clip);
    clipping_ = new_clip;
    SelectClipRgn(drawable_, new_clip);
}

//+ Win32Painter(XfPainter::pop_clipping)
void Win32Painter::pop_clipping() {
    flush_text();

    int n = clippers_->count();
    MWassert(n > 0);

    HRGN clip = clipping_;
    DeleteObject(clip);

    clip = clippers_->item(n - 1);
    clippers_->remove(n - 1);
    clipping_ = clip;

    SelectClipRgn(drawable_, clip);
}

//+ Win32Painter(XfPainter::is_visible)
Boolean Win32Painter::is_visible(XfRegion_ptr r) {
    Boolean b = (clipping_ == 0);
    if (!b) {
        Vertex lower, upper;
        if (mode_ == GM_ADVANCED) {
            r->bounds(lower, upper);
	} else {
            RegionImpl region;
	    region.copy(r);
	    region.apply_transform(matrix_);
            region.bounds(lower, upper);
	}
        RECT tr;
        tr.left = coord_to_twips(lower.x);
        tr.top = coord_to_twips(upper.y);
        tr.right = coord_to_twips(upper.x);
        tr.bottom = coord_to_twips(lower.y);
        b = RectVisible(drawable_, &tr);
    }
    return b;
}

//+ Win32Painter(XfPainter::visible)
XfRegion_ptr Win32Painter::visible() {
    RECT box;
    int result = GetClipBox(drawable_, &box);
    MWassert(result != ERROR);

    RegionImpl* r = new RegionImpl;
    r->lower_.x = twips_to_coord(box.left);
    r->lower_.y = twips_to_coord(box.bottom);
    r->upper_.x = twips_to_coord(box.right);
    r->upper_.y = twips_to_coord(box.top);
    r->defined_ = true;
    if (mode_ != GM_ADVANCED) {
        TransformImpl it;
	it.copy(matrix_);
	it.invert();
	r->apply_transform(&it);
    }

    return r;
}
