/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1993-95 Silicon Graphics, Inc.
 * Copyright (c) 1993-95 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include <Fresco/defs_impl/math.h>
#include <Fresco/defs_impl/table.h>

#include <Fresco/Ox/schema.h>

#include <Fresco/defs_Win32/Wdisplay.h>
#include <Fresco/defs_Win32/Wscreen.h>
#include <Fresco/defs_Win32/Wwindow.h>

// class MColormap - A class to manage colormaps under windows

class MColormap {
public:
    MColormap();
    ~MColormap();

    // Add an entry into the palette for the given color rgb 
    // specification.  If the color already exists, nothing is done, 
    // otherwise it is added and a reference is returned.
    COLORREF add_entry(int r, int g, int b);

    // Looks for the given rgb color specification in the palette.  If
    // found, true is returned and the value set to the corresponding 
    // reference.
    Boolean find_entry(int r, int g, int b, COLORREF& value);

    HPALETTE handle() { return palette; }

private:
    class MColorTable* table;		// color lookup table
    HPALETTE palette; 			// the MS-Windows palette
    int numEntries;			// number of palette entries
    int palSize;			// size of the palette
};

static const int GROW_SIZE = 16;

static unsigned long key_to_hash(COLORREF r) { return (unsigned long) r; }

declareTable(MColorTable, COLORREF, COLORREF)
implementTable(MColorTable, COLORREF, COLORREF)

MColormap::MColormap() {
    // ---- lookup table for unique colors ----
    table = new MColorTable(512);

    // ---- create a temporary logical palette ----
    int allocSize = sizeof(LOGPALETTE) + (GROW_SIZE * sizeof(PALETTEENTRY));
    LOGPALETTE* lpal = (LOGPALETTE*) new char[allocSize];

    // ---- initialize it ----
    lpal->palNumEntries = GROW_SIZE;
    lpal->palVersion = 0x300;
    for (int i = 0; i < GROW_SIZE; i++) {
	lpal->palPalEntry[i].peRed = 0;
	lpal->palPalEntry[i].peGreen = 0;
	lpal->palPalEntry[i].peBlue = 0;
	lpal->palPalEntry[i].peFlags = 0;
    }

    // ---- create it ----
    palette = CreatePalette(lpal);
    delete lpal;
    numEntries = 0;
    palSize = GROW_SIZE;
}

MColormap::~MColormap() {
    DeleteObject(palette);
}

// -----------------------------------------------------------------------
//  Add an entry to the palette.
// -----------------------------------------------------------------------
COLORREF MColormap::add_entry(int r, int g, int b) {
    COLORREF color = PALETTERGB(LOBYTE(r), LOBYTE(g), LOBYTE(b));
    COLORREF bogus;

    // ---- see if it's already there ----
    if (table->find(bogus, color)) {
	return color;
    }

    // ---- extend the palette if necessary ----
    if (numEntries >= palSize) {
	palSize += GROW_SIZE;
	ResizePalette(palette, palSize);
    }

    // ---- add the entry to the palette ----
    PALETTEENTRY newEntry;
    newEntry.peRed = LOBYTE(r);
    newEntry.peGreen = LOBYTE(g);
    newEntry.peBlue = LOBYTE(b);
    newEntry.peFlags = 0;
    SetPaletteEntries(palette, numEntries, 1, &newEntry);
    ++numEntries;

    // ---- add to the hash table ----
    table->insert(color, color);

    return color;
}

Boolean MColormap::find_entry(int r, int g, int b, COLORREF& value) {
    COLORREF color = PALETTERGB( LOBYTE(r), LOBYTE(g), LOBYTE(b));
    return table->find(value, color);
}

// Initially, the colormap is shared by all of the windows
static MColormap the_colormap;

// class Win32Screen

/*
 * To avoid a circularity, a screen does not hold a reference
 * to its display.  We could be safer by observing the display,
 * but for now we just assume that the display will live
 * at least as long as the screen.
 */

//+ XfWin32Screen(XfWin32Screen)
void* XfWin32Screen::_this_ptr() { return this; }
_Ix_extern__tid(XfWin32Screen);
void* XfWin32Screen::_narrow_tid(CORBA::TypeId t) {
    if (_Ix__tid(XfWin32Screen) == t) return this;
    void* p;
    if (p = XfScreen::_narrow_tid(t)) return p;
    return nil;
};
_Ix_extern__tid(XfWin32Screen);
_Ix_ImplementGetTid(XfWin32Screen,_Ix__tid(XfWin32Screen))
_Ix_ImplementNarrow(XfWin32Screen*,XfWin32Screen,_Ix_NullParameter,XfWin32Screen,0)
_Ix_ImplementDuplicate(XfWin32Screen*,XfWin32Screen,_Ix_NullParameter,_Ix__tid(XfWin32Screen),0)
//+

Win32Screen::Win32Screen(Win32Display* d, long n) {
    display_ = d;
    number_ = n;

    HWND root = GetDesktopWindow();
    HDC hdc = GetDC(root);
    MWassert(hdc);
    int result = SetMapMode(hdc, MM_TWIPS);
    MWassert(result);

/*
    //
    // set conversion factors, these are different between the X and Y
    // dimension generally (and definitely in MS-Windows), so this is
    // handled differently than the InterViews distribution.  The logical
    // to device point conversion function is used to convert an inch worth
    // of printer points from page space to device space to determine what
    // the scale factors are.
    POINT p;
    p.x = coord_to_twips(Coord(72));
    p.y = coord_to_twips(Coord(72));
    result = LPtoDP(hdc, &p, 1);
    MWassert(result);
    x_point_ = p.x / 72.0f;
    y_point_ = float(Math::abs( p.y / 72.0 ));
 */

    x_pixels_ = GetDeviceCaps(hdc, HORZRES);
    y_pixels_ = GetDeviceCaps(hdc, VERTRES);

    int x_size = GetDeviceCaps(hdc, HORZSIZE);
    int y_size = GetDeviceCaps(hdc, VERTSIZE);

    const float mm_per_point = 25.4f / 72.0f;

    // pixels per point
    x_point_ = (float(x_pixels_) / float(x_size)) * mm_per_point;
    y_point_ = (float(y_pixels_) / float(y_size)) * mm_per_point;

    // points per pixel
    x_pixel_ = 1.0f / x_point_;
    y_pixel_ = 1.0f / y_point_;

    ReleaseDC(root, hdc);
}

Win32Screen::~Win32Screen() {
}

OxExchange* Win32Screen::_exchange() { return &object_; }

//+ Win32Screen(XfFrescoObject::=object_.)
XfTag Win32Screen::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void Win32Screen::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void Win32Screen::disconnect() {
    object_.disconnect();
}
void Win32Screen::notify_observers() {
    object_.notify_observers();
}
void Win32Screen::update() {
    object_.update();
}
//+

//+ Win32Screen(XfScreen::dpi=c)
void Win32Screen::dpi(XfCoord c) {
    //
    // This request is not honered because it should be set in the
    // registry where it can be used by all applications and the 
    // MS-Windows mapping modes will use it (which is were we get
    // our conversion factors).
    //
}

//+ Win32Screen(XfScreen::dpi?)
XfCoord Win32Screen::dpi() {
    // probably should add axis argument
    return Coord(72.0) / x_pixel_;
}

//+ Win32Screen(XfScreen::display_ptr)
XfDisplay_ptr Win32Screen::display_ptr() {
    return Display::_duplicate(display_);
}

//+ Win32Screen(XfScreen::width)
XfCoord Win32Screen::width() {
    return to_coord(x_pixels_, X_axis);
}

//+ Win32Screen(XfScreen::height)
XfCoord Win32Screen::height() {
    return to_coord(y_pixels_, Y_axis);
}

//+ Win32Screen(XfScreen::to_pixels)
XfPixelCoord Win32Screen::to_pixels(XfCoord c, XfAxis a) {
    switch(a) {
        case X_axis:
            return PixelCoord(c * x_point_ + ((c > 0) ? 0.5 : -0.5));
        case Y_axis:
            return PixelCoord(c * y_point_ + ((c > 0) ? 0.5 : -0.5));
    }
    return PixelCoord(0);
}

//+ Win32Screen(XfScreen::to_coord)
XfCoord Win32Screen::to_coord(XfPixelCoord p, XfAxis a) {
    switch(a) {
        case X_axis:
            return Coord(p) * x_pixel_;
        case Y_axis:
            return Coord(p) * y_pixel_;
    }
    return Coord(0);
}

//+ Win32Screen(XfScreen::to_pixels_coord)
XfCoord Win32Screen::to_pixels_coord(XfCoord c, XfAxis a) {
    return to_coord(to_pixels(c, a), a);
}

//+ Win32Screen(XfScreen::move_pointer)
void Win32Screen::move_pointer(XfCoord x, XfCoord y) {
    SetCursorPos(
        int(to_pixels(x, X_axis)), int(y_pixels_ - to_pixels(y, Y_axis))
    );
}

//+ Win32Screen(XfScreen::application)
XfWindow_ptr Win32Screen::application(XfViewer_ptr v) {
    return new Win32ApplicationWindow(display_, this, v);
}

//+ Win32Screen(XfScreen::top_level)
XfWindow_ptr Win32Screen::top_level(XfViewer_ptr v, XfWindow_ptr group_leader) {
    return new Win32TopLevelWindow(display_, this, v, group_leader);
}

//+ Win32Screen(XfScreen::transient)
XfWindow_ptr Win32Screen::transient(XfViewer_ptr v, XfWindow_ptr transient_for) {
    return new Win32TransientWindow(display_, this, v, transient_for);
}

//+ Win32Screen(XfScreen::popup)
XfWindow_ptr Win32Screen::popup(XfViewer_ptr v) {
    return new Win32PopupWindow(display_, this, v);
}

//+ Win32Screen(XfScreen::icon)
XfWindow_ptr Win32Screen::icon(XfViewer_ptr v) {
    return new Win32IconWindow(display_, this, v);
}

HPALETTE Win32Screen::find_colormap(Window*) {
    return the_colormap.handle();
}

void Win32Screen::find_color(Window*, COLORREF& c) {
    int r = GetRValue(c);
    int g = GetGValue(c);
    int b = GetBValue(c);
    c = the_colormap.add_entry(r, g, b);
}
