/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1995 Silicon Graphics, Inc.
 * Copyright (c) 1995 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include <Fresco/defs/fresco.h>

#include <Fresco/defs_impl/action.h>
#include <Fresco/defs_impl/charstr.h>
#include <Fresco/defs_impl/displays.h>
#include <Fresco/defs_impl/list.h>
#include <Fresco/defs_impl/styles.h>
#include <Fresco/defs_impl/table.h>
#include <Fresco/defs_impl/threads.h>

#include <Fresco/Ox/schema.h>

// class DisplayBase

declarePtrList(DisplayDamage,Window)
implementPtrList(DisplayDamage,Window)

struct DisplayFilterInfo {
    Boolean general;
    Tag tag;
    EventFilter_ptr filter;
};

declarePtrList(DisplayFilters,DisplayFilterInfo)
implementPtrList(DisplayFilters,DisplayFilterInfo)

declareActionCallback(DisplayBase)
implementActionCallback(DisplayBase)

DisplayBase::DisplayBase() {
    ThreadKit_ptr t = ThreadLib::thread_kit_ptr();
    lock_ = t->mutex_ptr();

    reading_ = false;
    waiting_to_read_ = 0;
    wait_to_read_ = t->condition();

    requesting_ = false;
    waiting_to_request_ = 0;
    wait_to_request_ = t->condition();

    running_ = false;

    style_ = new StyleImpl;
    FrescoContext* fc = FrescoLib::current_context();
    style_->merge(StyleContext_var(fc->style()));

    double_click_ = Float(0);
    StyleValue_var a = style_->resolve(
        FrescoLib::tmp_string_ref("dblclick_threshold")
    );
    if (is_not_nil(a)) {
        double v;
        if (a->read_real(v)) {
	    double_click_ = Float(v);
	}
    }

    filter_lock_ = t->mutex_ptr();
    filters_ = new DisplayFilters;
    filter_tag_ = 0;

    redisplay_ = nil;
    damage_pending_ = t->condition();
    damaged_ = new DisplayDamage;
}

DisplayBase::~DisplayBase() {
    CORBA::release(style_);

    for (ListItr(DisplayFilters) f(*filters_); f.more(); f.next()) {
	DisplayFilterInfo* info = f.cur();
	CORBA::release(info->filter);
	delete info;
    }
    delete filters_;

    for (ListItr(DisplayDamage) d(*damaged_); d.more(); d.next()) {
        CORBA::release(d.cur());
    }
    delete damaged_;
}

OxExchange* DisplayBase::_exchange() { return &object_; }

//+ DisplayBase(XfFrescoObject::=object_.)
XfTag DisplayBase::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void DisplayBase::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void DisplayBase::disconnect() {
    object_.disconnect();
}
void DisplayBase::notify_observers() {
    object_.notify_observers();
}
void DisplayBase::update() {
    object_.update();
}
//+

//+ DisplayBase(XfDisplay::style)
XfStyleContext_ptr DisplayBase::style() {
    return StyleContext::_duplicate(style_);
}

/*
 * Before trying to read an event, we need to check to see if
 * another thread is waiting to use the display.  If so, we wait
 * until they are done.  Otherwise, we block until input is ready.
 *
 * If the display has been closed, then we don't try to read and
 * return false.  Otherwise, we return true.
 */

Boolean DisplayBase::prepare_to_read() {
    Boolean b = false;
    lock_->acquire();
    if (running_) {
        b = true;
        if (waiting_to_request_ > 0) {
            --waiting_to_request_;
            requesting_ = true;
            wait_to_request_->notify();
            while (requesting_) {
                ++waiting_to_read_;
                wait_to_read_->wait(lock_);
            }
        }
        if (peek() && (damaged_->count() != 0)) {
            notify_redisplay();
        }
        reading_ = true;
    }
    lock_->release();
    return b;
}

/*
 * After reading an event, we should re-acquire the display lock and
 * notify other threads that round-trip requests are now allowed.
 */

void DisplayBase::after_read() {
    lock_->acquire();
    reading_ = false;
    lock_->release();
}

/*
 * Main event dispatching loop for the display.
 *
 * Unless the "single_threaded" attribute is set, we spawn
 * a thread to handle redisplay asynchronously.
 */

//+ DisplayBase(XfDisplay::run)
void DisplayBase::run(Boolean b) {
    if (b) {
        running_ = true;

        if (!style_->is_on(FrescoLib::tmp_string_ref("single_threaded"))) {
            ThreadKit_ptr t = ThreadLib::thread_kit_ptr();
            redisplay_ = t->thread_ptr(
                new ActionCallback(DisplayBase)(
                    this, &DisplayBase::redisplay_thread
                )
            );
            if (is_not_nil(redisplay_)) {
                redisplay_->run();
            }
        }

        while (run_one_event_bool()) {
	}
    }

    if (running_) {
        lock_->acquire();
        running_ = false;
        if (is_not_nil(redisplay_)) {
            redisplay_->terminate();
        }
        lock_->release();
    }
}

//+ DisplayBase(XfDisplay::running)
Boolean DisplayBase::running() {
    return running_;
}

//+ DisplayBase(XfDisplay::run_one_event)
void DisplayBase::run_one_event() {
    run_one_event_bool();
}

// Check to see if any filters will handle this event.
Boolean DisplayBase::filtered(Event_ptr event) {
    Boolean b = false;
    filter_lock_->acquire();
    ListItr(DisplayFilters) i(*filters_);
    for (;;) {
	if (!i.more()) {
	    filter_lock_->release();
	    break;
	}
	DisplayFilterInfo* info = i.cur();
        Boolean general = info->general;
	EventFilter_ptr f = info->filter;
	filter_lock_->release();
        if ((general || event->type() != Event::other) && f->handle(event)) {
	    b = true;
	    break;
	}
	filter_lock_->acquire();
	i.next();
    }
    return b;
}

XfTag DisplayBase::add_filter_info(XfEventFilter_ptr f, Boolean general) {
    filter_lock_->acquire();
    DisplayFilterInfo* info = new DisplayFilterInfo;
    ++filter_tag_;
    info->tag = filter_tag_;
    info->general = general;
    info->filter = EventFilter::_duplicate(f);
    filters_->prepend(info);
    filter_lock_->release();
    return info->tag;
}

//+ DisplayBase(XfDisplay::add_filter)
XfTag DisplayBase::add_filter(XfEventFilter_ptr f) {
    return add_filter_info(f, false);
}

//+ DisplayBase(XfDisplay::add_general_filter)
XfTag DisplayBase::add_general_filter(XfEventFilter_ptr f) {
    return add_filter_info(f, true);
}

//+ DisplayBase(XfDisplay::remove_filter)
void DisplayBase::remove_filter(XfTag add_tag) {
    filter_lock_->acquire();
    for (ListUpdater(DisplayFilters) i(*filters_); i.more(); i.next()) {
        DisplayFilterInfo* info = i.cur();
        if (info->tag == add_tag) {
	    CORBA::release(info->filter);
	    i.remove_cur();
	    delete info;
	    break;
        }
    }
    filter_lock_->release();
}

/*
 * Repair all the windows on the damage list.
 */

void DisplayBase::do_repairs() {
    Window_ptr w;
    for (ListItr(DisplayDamage) i(*damaged_); i.more(); i.next()) {
        w = i.cur();
        w->repair();
        CORBA::release(w);
    }
    damaged_->remove_all();
}

/*
 * Main thread for controlling redraw.  This function
 * just loops, blocking when there is no damage to repair.
 *
 * The synchronization looks wrong because it acquires the lock and
 * never releases it explicitly.  However, the lock will be released
 * whenever redisplay is waiting for more damage.  We assume that
 * Display::close will terminate the redisplay thread while blocked
 * waiting for damage, so the releasing the lock will not be a concern.
 */

void DisplayBase::redisplay_thread() {
    lock_->acquire();
    for (;;) {
        while (damaged_->count() == 0) {
            damage_pending_->wait(lock_);
        }

        do_repairs();
        flush();
    }
}

/*
 * Notify the redisplay thread that any damaged windows should now
 * be repaired, or, in the single threaded case, go ahead and
 * do the repairs.
 */

void DisplayBase::notify_redisplay() {
    if (is_not_nil(redisplay_)) {
        damage_pending_->notify();
    } else {
        do_repairs();
    }
}

/*
 * Note that a window needs to be repaired by putting it
 * on the damage list.  This operation must synchronize
 * with the redisplay thread.
 *
 * If the event thread is reading input, then we
 * immediately wake up the redisplay thread.  Otherwise,
 * we will effectively batch multiple repair operations by
 * waiting until the next time around in the event loop
 * to check for damage.
 */

//+ DisplayBase(XfDisplay::need_repair)
void DisplayBase::need_repair(XfWindow_ptr w) {
    lock_->acquire();
    damaged_->append(Window::_duplicate(w));
    if (reading_) {
        damage_pending_->notify();
    }
    lock_->release();
}

//+ DisplayBase(XfDisplay::repair)
void DisplayBase::repair() {
    lock_->acquire();
    if (damaged_->count() != 0) {
        notify_redisplay();
    }
    lock_->release();
}

//+ DisplayBase(XfDisplay::close)
void DisplayBase::close() {
    run(false);
}

//+ DisplayBase(XfDisplay::auto_repeat=b)
void DisplayBase::auto_repeat(Boolean b) {
    // !!! unimplemented
}

//+ DisplayBase(XfDisplay::auto_repeat?)
Boolean DisplayBase::auto_repeat() {
    // !!! unimplemented
    return false;
}

//+ DisplayBase(XfDisplay::key_click_volume=v)
void DisplayBase::key_click_volume(Float v) {
    // !!! unimplemented
}

//+ DisplayBase(XfDisplay::key_click_volume?)
Float DisplayBase::key_click_volume() {
    // !!! unimplemented
    return Float(0.5);
}

//+ DisplayBase(XfDisplay::pointer_acceleration=a)
void DisplayBase::pointer_acceleration(Float a) {
    // !!! unimplemented
}

//+ DisplayBase(XfDisplay::pointer_acceleration?)
Float DisplayBase::pointer_acceleration() {
    // !!! unimplemented
    return Float(0.5);
}

//+ DisplayBase(XfDisplay::pointer_threshold=t)
void DisplayBase::pointer_threshold(Long t) {
    // !!! unimplemented
}

//+ DisplayBase(XfDisplay::pointer_threshold?)
Long DisplayBase::pointer_threshold() {
    // !!! unimplemented
    return 0;
}

//+ DisplayBase(XfDisplay::pointer_double_click_threshold=t)
void DisplayBase::pointer_double_click_threshold(Float t) {
    double_click_ = t;
}

//+ DisplayBase(XfDisplay::pointer_double_click_threshold?)
Float DisplayBase::pointer_double_click_threshold() {
    return double_click_;
}
