//
// $XConsortium$
//

//
// Copyright (c) 1992-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the name of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#include <Fresco/defs/display.h>
#include <Fresco/defs/drawing.h>
#include <Fresco/defs/fcontext.h>
#include <Fresco/defs/fresco.h>
#include <Fresco/defs/glyph.h>
#include <Fresco/defs_impl/charstr.h>
#include <Fresco/defs_impl/dload.h>
#include <Fresco/defs_impl/fobjects.h>
#include <Fresco/defs_impl/list.h>
#include <Fresco/defs_impl/names.h>
#include <Fresco/defs_impl/styles.h>
#include <Fresco/defs_impl/table.h>
#include <Fresco/defs_impl/threads.h>
#include <Fresco/defs_impl/viewers.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if defined(__hpux)
#include <cxxdl.h>
#endif

#if defined(__sgi)
#include <malloc.h>
#endif

#if defined(WIN32)
#include <windows.h>
#define strdup _strdup
#endif

//
// Standard command-line options.
//

static Option standard_options[] = {
    { "-background", "*background", Option::value },
    { "-bg", "*background", Option::value },
    { "-dbuf", "*double_buffered", Option::implicit, "on" },
    { "-display", "*display", Option::value },
    { "-dpi", "*dpi", Option::value },
    { "-fg", "*foreground", Option::value },
    { "-flat", "*flat", Option::value },
    { "-fn", "*font", Option::value },
    { "-font", "*font", Option::value },
    { "-foreground", "*foreground", Option::value },
    { "-geometry", "*geometry", Option::value },
    { "-iconic", "*iconic", Option::implicit, "on" },
    { "-monochrome", "*gui", Option::implicit, "monochrome" },
    { "-mthread", "*single_threaded", Option::implicit, "off" },
    { "-name", "*name", Option::value },
    { "-nodbuf", "*double_buffered", Option::implicit, "off" },
    { "-sthread", "*single_threaded", Option::implicit, "on" },
    { "-synchronous", "*synchronous", Option::implicit, "on" },
    { "+synchronous", "*synchronous", Option::implicit, "off" },
    { "-title", "*title", Option::value },
    { "-visual", "*visual", Option::value },
    { "-visual_id", "*visual_id", Option::value },
    { "-xrm", nil, Option::path_value },
#if defined(__sgi)
    { "-malloc", "*malloc_debug", Option::implicit, "on" },
#endif
    { nil }
};

//
// Default root style attributes.
//

static const char* standard_attributes[] = {
    "foreground", "#000000",
    "background", "#ffffff",
    "font", "fixed",
    "double_buffered", "on",
    "*PopupWindow*saveUnder", "on",
    "*TransientWindow*saveUnder", "on",
#if defined(__sgi)
    "*PopupWindow*overlay: true", "on",
#endif
    nil,
};

// class SharedFrescoObjectImpl

struct ObservableInfo {
    Tag tag;
    FrescoObject_ptr observer;
};

declareList(ObserverList,ObservableInfo)
implementList(ObserverList,ObservableInfo)

SharedFrescoObjectImpl::SharedFrescoObjectImpl() {
    observers_ = nil;
}

SharedFrescoObjectImpl::~SharedFrescoObjectImpl() {
    ObserverList* list = observers_;
    if (list != nil) {
	ListItr(ObserverList) i(*list);
	//
	// First element is just for tag counter
	//
	i.next();
	for (; i.more(); i.next()) {
	    i.cur().observer->disconnect();
	}
	delete list;
    }
}

//
// Observable operations.
//
// Note that the first "observer" on the list is really a dummy
// to hold the counter for the object.  This way we pay an extra
// word for observables, but one less word for objects that are
// never observed.
//

//+ SharedFrescoObjectImpl(XfFrescoObject::attach)
XfTag SharedFrescoObjectImpl::attach(XfFrescoObject_ptr observer) {
    ObservableInfo i;
    if (observers_ == nil) {
	observers_ = new ObserverList;
	i.tag = 0;
	i.observer = nil;
	observers_->append(i);
    }
    ObservableInfo& hdr = observers_->item_ref(0);
    hdr.tag += 1;
    i.tag = hdr.tag;
    i.observer = observer;
    observers_->append(i);
    return i.tag;
}

//
// Detach the observer corresponding to the given tag.
// Probably should raise an exception or otherwise
// return an error if the tag is not found, but for now
// we just silently ignore.
//

//+ SharedFrescoObjectImpl(XfFrescoObject::detach)
void SharedFrescoObjectImpl::detach(XfTag attach_tag) {
    if (observers_ != nil) {
	ListUpdater(ObserverList) i(*observers_);
	for (i.next(); i.more(); i.next()) {
	    if (i.cur().tag == attach_tag) {
		i.remove_cur();
		break;
	    }
	}
    }
}

//+ SharedFrescoObjectImpl(XfFrescoObject::notify_observers)
void SharedFrescoObjectImpl::notify_observers() {
    if (observers_ != nil) {
	ListUpdater(ObserverList) i(*observers_);
	for (i.next(); i.more(); i.next()) {
	    i.cur().observer->update();
	}
    }
}

// class LockedFrescoObjectImpl

LockedFrescoObjectImpl::LockedFrescoObjectImpl() { }
LockedFrescoObjectImpl::~LockedFrescoObjectImpl() { }

//+ LockedFrescoObjectImpl(XfFrescoObject::attach)
XfTag LockedFrescoObjectImpl::attach(XfFrescoObject_ptr observer) {
    lock_->acquire();
    Tag t = SharedFrescoObjectImpl::attach(observer);
    lock_->release();
    return t;
}

//+ LockedFrescoObjectImpl(XfFrescoObject::detach)
void LockedFrescoObjectImpl::detach(XfTag attach_tag) {
    lock_->acquire();
    SharedFrescoObjectImpl::detach(attach_tag);
    lock_->release();
}

//+ LockedFrescoObjectImpl(XfFrescoObject::notify_observers)
void LockedFrescoObjectImpl::notify_observers() {
    lock_->acquire();
    SharedFrescoObjectImpl::notify_observers();
    lock_->release();
}

// class FrescoContext

FrescoContext::FrescoContext() { }
FrescoContext::~FrescoContext() { }

// class FrescoContextImpl

struct FrescoContextEntry {
    CORBA::Object_ptr obj;
#if defined(SVR4) || defined(__alpha) || defined(linux) || defined(__sun)
    void* dlhandle;
#endif
#if defined(__hpux)
    shl_t dlhandle;
#endif
#if defined(WIN32)
    HANDLE dlhandle;
#endif
};

declareTable(FrescoContextTable,const char*,FrescoContextEntry)
implementTable(FrescoContextTable,const char*,FrescoContextEntry)

class FrescoContextImpl : public FrescoContext {
public:
    FrescoContextImpl();
    ~FrescoContextImpl();

    OxExchange* _exchange();

    //+ Naming::NamingContext::*
    /* NamingContext */
    void bind(const Naming::Name& n, CORBA::Object_ptr obj);
    void bind_context(const Naming::Name& n, Naming::NamingContext_ptr nc);
    void rebind(const Naming::Name& n, CORBA::Object_ptr obj);
    void unbind(const Naming::Name& n);
    CORBA::Object_ptr resolve(const Naming::Name& n);
    Naming::NamingContext_ptr new_context();
    Naming::NamingContext_ptr bind_new_context(const Naming::Name& n);
    Naming::BindingList* list_names(ULong start, ULong how_many);
    Naming::BindingList* list_all();
    void destroy();
    //+

    void init(const char* name, int& argc, char** argv, Option*);
    void init_display(const char* display_name);

    const char* name();
    int argc();
    char** argv();
    StyleContext_ptr style();
    DisplayContext_ptr display_context();
    Display_ptr display_ptr();
    void default_drawing_kit(DrawingKit_ptr);
    DrawingKit_ptr default_drawing_kit();
protected:
    SharedFrescoObjectImpl object_;
    CORBA::String_var name_;
    int argc_;
    char** argv_;
    StyleContext_ptr style_;
    Display_ptr display_;
    DrawingKit_ptr drawing_;
    FrescoContextTable* table_;

    void create_root_style(
	const char* name, int& argc, char** argv, Option* options
    );
    void find_name(int argc, char** argv);
    void parse_args(int& argc, char** argv, Option* options);
    void extract(Option*, int& i, int argc, char** argv);
    const char* next_arg(int& i, int argc, char** argv, const char* message);
    void bad_arg(const char* message);
};

FrescoContextImpl::FrescoContextImpl() {
    style_ = nil;
    argc_ = 0;
    argv_ = nil;
    display_ = nil;
    drawing_ = nil;
    table_ = new FrescoContextTable(100);
}

FrescoContextImpl::~FrescoContextImpl() {
    CORBA::release(style_);
    CORBA::release(drawing_);

    FrescoContextTable* t = table_;
    for (TableIterator(FrescoContextTable) i(*t); i.more(); i.next()) {
	CORBA::release(i.cur_value().obj);
    }
    delete table_;
}

OxExchange* FrescoContextImpl::_exchange() { return &object_; }

void FrescoContextImpl::init(
    const char* name, int& argc, char** argv, Option* options
) {
    create_root_style(name, argc, argv, options);
    name_ = new char[strlen(name) + 1];
    strcpy(name_, name);
    argc_ = argc;
    argv_ = argv;
#if defined(sgi)
    if (style_->is_on(FrescoLib::tmp_string_ref("malloc_debug"))) {
	mallopt(M_DEBUG, 1);
    }
#endif
}

void FrescoContextImpl::init_display(const char* display_name) {
    DisplayContext_ptr dc = display_context();
    assert(dc != nil);
    const char* dname;
    if (display_name == nil) {
	display_ = dc->open_default();
	dname = "default display";
    } else {
	display_ = dc->open_string(display_name);
	dname = display_name;
    }
    if (is_nil(display_)) {
	fprintf(stderr, "Can't open %s\n", dname);
	exit(1);
    }
    drawing_ = display_->drawing_kit();
}

//+ FrescoContextImpl(Naming::NamingContext::bind) !n,obj
void FrescoContextImpl::bind(const Naming::Name&, CORBA::Object_ptr) {
    // unimplemented
}

//+ FrescoContextImpl(Naming::NamingContext::bind_context) !n,nc
void FrescoContextImpl::bind_context(const Naming::Name&, Naming::NamingContext_ptr) {
    // unimplemented
}

//+ FrescoContextImpl(Naming::NamingContext::rebind) !n,obj
void FrescoContextImpl::rebind(const Naming::Name&, CORBA::Object_ptr) {
    // unimplemented
}

//+ FrescoContextImpl(Naming::NamingContext::unbind) !n
void FrescoContextImpl::unbind(const Naming::Name&) {
    // unimplemented
}

//
// Not sure what to do here if we don't have dlopen or something equivalent.
//

#if defined(SVR4) || defined(__alpha) || defined(linux)
#include <dlfcn.h>

#define LoadLibrary(name) dlopen(name, RTLD_LAZY)
#define GetProcAddress(h,name) dlsym(h, name)
#endif

#if defined(WIN32)
#include <windows.h>

#define dlerror() ""
#endif

#if defined(__hpux)
#define LoadLibrary(name) cxxshl_load(name, BIND_DEFERRED | DYNAMIC_PATH, 0L)
static void* GetProcAddress (shl_t h, const char* name) {
    void* addr;
    int status;

    status = shl_findsym (&h, name, TYPE_PROCEDURE, &addr);
    if (status < 0) {
        fprintf (stderr, "Can't find symbol %s in library\n", name);
        exit (1);
    }
    return addr;
}
#define dlerror() ""
#endif

#if defined(__sun) && !defined(SVR4)
#include <dlfcn.h>

extern void *LoadLibrary(const char *name);
extern void *GetProcAddress(const void *handle, const char *name);
#endif

//+ FrescoContextImpl(Naming::NamingContext::resolve)
CORBA::Object_ptr FrescoContextImpl::resolve(const Naming::Name& n) {
    FrescoContextEntry e;
    e.obj = nil;
    const char* name = NamingLib::to_string(n);
    if (!table_->find(e, name)) {
	const char* dlname = name;
#if defined(SVR4) || defined(__hpux) || defined(__alpha) || defined(linux) || defined(__sun)
	unsigned int len = strlen(name) + 3 + 1;
	char fixed_buf[128];
	char* buf = len <= sizeof(fixed_buf) ? fixed_buf : new char[len];
	const char* slash = strrchr(name, '/');
	if (slash != nil) {
	    ++slash;
	} else {
	    slash = name;
	}
#if defined(SVR4) || defined(__alpha) || defined(linux) || defined(__sun)
      sprintf(buf, "%.*slib%s.so", slash - name, name, slash);
#endif
#if defined(__hpux)
      sprintf(buf, "%.*slib%s.sl", slash - name, name, slash);
#endif
	dlname = buf;
#endif
#if defined(SVR4) || defined(WIN32) || defined(__hpux) || defined(__alpha) || defined(linux) || defined(__sun)
	e.dlhandle = LoadLibrary(dlname);
	if (e.dlhandle == nil) {
	    const char* msg = dlerror();
	    const char* sep = msg[0] == '\0' ? msg : ": ";
	    fprintf(stderr, "Unable to load %s%s%s\n", dlname, sep, msg);
	} else {
	    DLoadFunc* f = (DLoadFunc*)GetProcAddress(e.dlhandle, "load");
	    if (f == nil) {
		fprintf(stderr, "Unable to create object for %s\n", name);
	    } else {
		e.obj = (*f)();
		table_->insert(name, e);
	    }
	}
#endif
#if defined(SVR4) || defined(__hpux) || defined(__alpha) || defined(linux) || defined(__sun)
	if (buf != fixed_buf) {
	    delete buf;
	}
#endif
    }
    return CORBA::Object::_duplicate(e.obj);
}

//+ FrescoContextImpl(Naming::NamingContext::new_context) !n
Naming::NamingContext_ptr FrescoContextImpl::new_context() {
    // unimplemented
    return nil;
}

//+ FrescoContextImpl(Naming::NamingContext::bind_new_context) !n
Naming::NamingContext_ptr FrescoContextImpl::bind_new_context(const Naming::Name&) {
    // unimplemented
    return nil;
}

//+ FrescoContextImpl(Naming::NamingContext::list_names) !start,how_many,bl
Naming::BindingList* FrescoContextImpl::list_names(ULong, ULong) {
    // unimplemented
    return nil;
}

//+ FrescoContextImpl(Naming::NamingContext::list_all)
Naming::BindingList* FrescoContextImpl::list_all() {
    // unimplemented
    return nil;
}

//+ FrescoContextImpl(Naming::NamingContext::destroy)
void FrescoContextImpl::destroy() {
    // unimplemented
}

const char* FrescoContextImpl::name() { return name_; }
int FrescoContextImpl::argc() { return argc_; }
char** FrescoContextImpl::argv() { return argv_; }

StyleContext_ptr FrescoContextImpl::style() {
    return StyleContext::_duplicate(style_);
}

DisplayContext_ptr FrescoContextImpl::display_context() {
    return DisplayContext::_narrow(
	FrescoLib::resolve_string(this, "fdisplay")
    );
}

Display_ptr FrescoContextImpl::display_ptr() {
    return Display::_duplicate(display_);
}

void FrescoContextImpl::default_drawing_kit(DrawingKit_ptr d) {
    CORBA::release(drawing_);
    drawing_ = DrawingKit::_duplicate(d);
}

DrawingKit_ptr FrescoContextImpl::default_drawing_kit() {
    return DrawingKit::_duplicate(drawing_);
}

void FrescoContextImpl::create_root_style(
    const char* name, int& argc, char** argv, Option* options
) {
    style_ = new StyleImpl;

    style_->alias(FrescoLib::tmp_string_ref(name));
    find_name(argc, argv);

    for (const char** attr = standard_attributes; *attr != nil; attr += 2) {
	StyleValue_var a = style_->bind(FrescoLib::tmp_string_ref(attr[0]));
	a->write_string(FrescoLib::tmp_string_ref(attr[1]));
	a->priority(-10);
    }
    if (options != nil) {
	parse_args(argc, argv, options);
    }
    parse_args(argc, argv, standard_options);
}

//
// Figure out the name for the root style.  If -name is specified
// on the command line then we use it.  Otherwise, we have to check
// for the RESOURCE_NAME environment variable, or finally try
// to use argv[0].
//

void FrescoContextImpl::find_name(int argc, char** argv) {
    const char* value = nil;
    for (int i = 1; i < argc - 1; i++) {
	if (strcmp(argv[i], "-name") == 0) {
	    value = argv[i + 1];
	    break;
	}
    }
    if (value == nil) {
	value = getenv("RESOURCE_NAME");
	if (value == nil) {
	    if (argc > 0) {
		value = argv[0];
		const char* p = value + strlen(value) - 1;
		for (; p >= value; p--) {
		    if (*p == '/') {
			value = p + 1;
			break;
		    }
		}
	    } else {
		value = "Anon-7";
	    }
	}
    }
    style_->context_name(FrescoLib::tmp_string_ref(value));
}

//
// Parse an argument list, setting any attributes that are specified
// in the option list.  Matching arguments are removed (in-place)
// from the argument list.
//

void FrescoContextImpl::parse_args(int& argc, char** argv, Option* options) {
    int i;
    int newargc = 1;
    char* newargv[1024];
    newargv[0] = argv[0];
    for (i = 1; i < argc; i++) {
	Boolean matched = false;
	const char* s = argv[i];
	for (Option* o = options; o->name != nil; o++) {
	    if (strcmp(s, o->name) == 0) {
		extract(o, i, argc, argv);
		matched = true;
		break;
	    }
	}
	if (!matched) {
	    if (newargc != i) {
		argv[newargc] = argv[i];
	    }
	    ++newargc;
	}
    }
    if (newargc != argc) {
	argc = newargc;
	argv[argc] = nil;
    }
}

//
// Extract an attribute definition from an argument.
//

void FrescoContextImpl::extract(Option* o, int& i, int argc, char** argv) {
    const char* name;
    Long name_length;
    const char* value;
    Long value_length;
    const char* colon;
    switch (o->format) {
    case Option::path_value:
	value  = next_arg(i, argc, argv, "Missing attribute definition");
	value_length = strlen(value);
	colon = CharStringImpl::index(value, value_length, ':');
	if (colon == nil) {
	    bad_arg("Missing ':' in attribute definition");
	} else {
	    name = value;
	    name_length = colon - value;
	    value = colon + 1;
	    value_length -= (name_length + 1);
	}
	break;
    case Option::value:
	name = o->path;
	name_length = strlen(name);
	value = next_arg(i, argc, argv, "Missing value");
	value_length = strlen(value);
	break;
    case Option::implicit:
	name = o->path;
	name_length = strlen(name);
	value = o->implicit_value;
	value_length = strlen(value);
	break;
    }
    StyleValue_var a = style_->bind(
	CharString_var(new CharStringImpl(name, name_length))
    );
    a->write_string(CharString_var(new CharStringImpl(value, value_length)));
    a->priority(0);
}

const char* FrescoContextImpl::next_arg(
    int& i, int argc, char** argv, const char* message
) {
    ++i;
    if (i == argc) {
	bad_arg(message);
    }
    return argv[i];
}

void FrescoContextImpl::bad_arg(const char* message) {
    fflush(stdout);
    fprintf(stderr, "%s\n", message);
    exit(1);
}

// class FrescoLib

FrescoContext* FrescoLib::current_;

FrescoContext* FrescoLib::init_context(
    const char* name, int& argc, char** argv, Option* options
) {
    FrescoContextImpl* fc = new FrescoContextImpl;
    current_ = fc;
    fc->init(name, argc, argv, options);
    return fc;
}

FrescoContext* FrescoLib::init_default_context(
    const char* name, int& argc, char** argv, Option* options
) {
    FrescoContextImpl* fc = new FrescoContextImpl;
    current_ = fc;
    fc->init(name, argc, argv, options);
    fc->init_display(nil);
    return fc;
}

FrescoContext* FrescoLib::init_display_context(
    const char* display_name,
    const char* name, int& argc, char** argv, Option* options
) {
    FrescoContextImpl* fc = new FrescoContextImpl;
    current_ = fc;
    fc->init(name, argc, argv, options);
    fc->init_display(display_name);
    return fc;
}

void FrescoLib::current_context(FrescoContext* fc) { current_ = fc; }
FrescoContext* FrescoLib::current_context() { return current_; }

void FrescoLib::run(Viewer_ptr v, Glyph_ptr g) {
    FrescoContext* fc = FrescoLib::current_context();
    if (is_nil(fc)) {
	// should raise an exception
	fprintf(stderr, "No context\n");
	exit(1);
    }
    Display_var d = fc->display_ptr();
    if (is_nil(d)) {
	d = DisplayContext_var(fc->display_context())->open_default();
	if (is_nil(d)) {
	    // should raise an exception
	    fprintf(stderr, "Can't open display\n");
	    exit(1);
	} else {
	    fc->default_drawing_kit(DrawingKit_var(d->drawing_kit()));
	}
    }
    Screen_var s = d->default_screen();
    Viewer_ptr nv;
    if (is_nil(g)) {
	nv = v;
    } else {
	if (is_nil(v)) {
	    nv = new BgViewer;
	} else {
	    nv = new ViewerImpl(true);
	    nv->append_viewer(v);
	}
	nv->body(g);
    }
    Window_var w = s->application(nv);
    w->map();
    d->run(true);
}

CharString_ptr FrescoLib::string_ref(const char* ptr) {
    return CharStringImpl::create_static(ptr);
}

CharString_ptr FrescoLib::string_copy(const char* ptr) {
    return new CharStringImpl(ptr);
}

CharString_ptr FrescoLib::get_string(StyleContext_ptr s, const char* name) {
    CharString_ptr r = nil;
    StyleValue_var a = s->resolve(tmp_string_ref(name));
    if (is_not_nil(a)) {
	a->read_string(r);
    }
    return r;
}

CharString_ptr FrescoLib::get_fcontext_string(
    FrescoContext* fc, const char* name
) {
    StyleContext_ptr s = fc->style();
    return get_string(s, name);
}

CORBA::Object_ptr FrescoLib::resolve_string(
    FrescoContext* fc, const char* name
) {
    return fc->resolve(*NamingLib::to_name(name));
}

char* FrescoLib::keysym_to_string(Event::KeySym k) {
    return is_nil(current_) ?
	nil : Display_var(current_->display_ptr())->keysym_to_string(k);
}

Event::KeySym FrescoLib::string_to_keysym(const char* s) {
    return is_nil(current_) ?
	0 : Display_var(current_->display_ptr())->string_to_keysym(s);
}

#if defined(__powerc)

/* not implemented */
Boolean FrescoLib::delay(Float seconds) {
	return false;
}

#else

#if defined(__sgi)

extern "C" int sginap(long);

Boolean FrescoLib::delay(Float seconds) {
    long n = long(100.0 * seconds + 0.5);
    if (n <= 2) {
	n = 3;
    }
    return sginap(n) == 0;
}

#else

#if defined(__alpha)

//
// We can't use either poll or select under OSF/1 because they're both
// broken, though in different ways.  When called with no struct pollfd's
// and an nfds argument of zero to do a timeout, poll returns immediately
// instead of waiting until the timeout period has expired.  It also sets
// errno to EAGAIN sometimes, but not always.  Select seems to work for
// timeouts in simple cases, and even in some complex cases, but in some
// multithreaded programs FrescoLib::delay calls select and select never
// returns.
//
// The most reliable way to do a sleep in a multithreaded program under
// OSF/1 seems to be using the pthread_delay_np routine, which puts the
// current thread to sleep.  The following code assumes that you always
// build multithreaded Fresco libraries on the alpha, which matches the
// assumption in threads.cxx; if we switch to using the X11R6 threads
// macros, the use of pthread_delay_np should be conditionalized on
// the XTHREADS symbol being defined.
//

#include <pthread.h>

Boolean FrescoLib::delay(Float seconds) {
    struct timespec ts;
    ts.tv_sec = time_t(seconds);
    ts.tv_nsec = long(1000000000.0 * (seconds - float(ts.tv_sec)) + 0.5);
    return pthread_delay_np(&ts) == 0;
}

#else

#if defined(USE_POLL)

#include <sys/poll.h>

#if (defined(sun) || defined(__sun)) && !defined(SVR4) && !defined(__alpha)
extern "C" int poll(struct pollfds*, unsigned long, int);
#endif

Boolean FrescoLib::delay(Float seconds) {
    return poll(NULL, 0, int(1000.0* seconds + 0.5)) == 0;
}

#else

#include <sys/types.h>
#ifdef WIN32
#include <winsock.h>
#else
#include <sys/time.h>
#endif

Boolean FrescoLib::delay(Float seconds) {
    struct timeval tv;
    tv.tv_sec = time_t(seconds);
    tv.tv_usec = time_t(1000000.0 * (seconds - float(tv.tv_sec)) + 0.5);
    return select(0, nil, nil, nil, &tv) == 0;
}

// defined(USE_POLL)
#endif

// defined(__alpha)
#endif

// defined(__sgi)
#endif

// defined(__powerc)
#endif
