/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1993-95 Silicon Graphics, Inc.
 * Copyright (c) 1993-95 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include <assert.h>

#include <Fresco/defs/fresco.h>

#include <Fresco/defs_impl/list.h>
#include <Fresco/defs_impl/math.h>
#include <Fresco/defs_impl/memory.h>
#include <Fresco/defs_impl/painters.h>
#include <Fresco/defs_impl/region.h>
#include <Fresco/defs_impl/transform.h>

#include <Fresco/Ox/schema.h>
	  
// class PathImpl

PathImpl::PathImpl() {
    size_ = 8;
    buffer_ = new PathImpl::Point[size_];
    count_ = 0;
    is_closed_ = false;
}

PathImpl::~PathImpl() {
    delete buffer_;
}

void PathImpl::begin_path() {
    count_ = 0;
    is_closed_ = false;
}

void PathImpl::move_to(XfCoord x, XfCoord y) {
    expand();

    PathImpl::Point& p = buffer_[count_];
    p.type = PathImpl::MoveTo;
    p.x = x;
    p.y = y;

    ++count_;
}

void PathImpl::line_to(XfCoord x, XfCoord y) {
    expand();

    PathImpl::Point& p = buffer_[count_];
    p.type = PathImpl::LineTo;
    p.x = x;
    p.y = y;

    ++count_;
}

void PathImpl::curve_to(XfCoord x, XfCoord y, XfCoord x1, XfCoord y1, XfCoord x2, XfCoord y2) {
    expand();

    PathImpl::Point& p = buffer_[count_];
    p.type = PathImpl::CurveTo;
    p.x = x;
    p.y = y;
    PathImpl::Point& p1 = buffer_[count_ + 1];
    p1.type = PathImpl::CurveTo;
    p1.x = x1;
    p1.y = y1;
    PathImpl::Point& p2 = buffer_[count_ + 2];
    p2.type = PathImpl::CurveTo;
    p2.x = x2;
    p2.y = y2;

    count_ += 3;
}

void PathImpl::close_path() {
    is_closed_ = true;
}

void PathImpl::transform(Transform::Matrix m) {
    for (int i = 0; i < count_; ++i) {
        PathImpl::Point& p = buffer_[i];
        p.px = PixelCoord(p.x * m[0][0] + p.y * m[1][0] + m[2][0]);
        p.py = PixelCoord(p.x * m[0][1] + p.y * m[1][1] + m[2][1]);
    }
}

int PathImpl::count() {
    return count_;
}

PathImpl::Point& PathImpl::point(int i) {
    return buffer_[i];
}

Boolean PathImpl::is_closed() {
    return is_closed_;
}

void PathImpl::expand() {
    if ((count_ + 3) < size_) {
	return;
    }

    PathImpl::Point* old_buffer = buffer_;
    size_ = size_ + size_;
    buffer_ = new PathImpl::Point[size_];
    Memory::copy(old_buffer, buffer_, count_ * sizeof(PathImpl::Point));
    delete old_buffer;
}

// class PainterBase

static void rect_path(Painter* p, Coord l, Coord b, Coord r, Coord t) {
    p->begin_path();
    p->move_to(l, b);
    p->line_to(l, t);
    p->line_to(r, t);
    p->line_to(r, b);
    p->close_path();
}

/*
 * We only need this silly PainterMatrixStackElement because the compiler
 * (language) can't deal with functions that return arrays.
 */

struct PainterMatrixStackElement {
    Transform::Matrix matrix;
};

declareList(PainterMatrixStack,PainterMatrixStackElement)
implementList(PainterMatrixStack,PainterMatrixStackElement)

PainterBase::PainterBase() {
    matrix_ = new TransformImpl;
    transforms_ = new PainterMatrixStack;

    brush_ = nil;
    color_ = nil;
    font_ = nil;
    pattern_ = nil;
}

PainterBase::~PainterBase() {            
    CORBA::release(matrix_);
    delete transforms_;
}

OxExchange* PainterBase::_exchange() { return &object_; } 

void PainterBase::begin_paint() {
    path_.begin_path();
}

void PainterBase::end_paint() {
    CORBA::release(brush_);
    brush_ = nil;
    CORBA::release(color_);
    color_ = nil;
    CORBA::release(font_);
    font_ = nil;
    CORBA::release(pattern_);
    pattern_ = nil;
}

//+ PainterBase(XfFrescoObject::=object_.)
XfTag PainterBase::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void PainterBase::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void PainterBase::disconnect() {
    object_.disconnect();
}
void PainterBase::notify_observers() {
    object_.notify_observers();
}
void PainterBase::update() {
    object_.update();
}
//+

//+ PainterBase(XfPainter::begin_path)
void PainterBase::begin_path() {
    path_.begin_path();
}

//+ PainterBase(XfPainter::move_to)
void PainterBase::move_to(XfCoord x, XfCoord y) {
    path_.move_to(x, y);
}

//+ PainterBase(XfPainter::line_to)
void PainterBase::line_to(XfCoord x, XfCoord y) {
    path_.line_to(x, y);
}

//+ PainterBase(XfPainter::curve_to)
void PainterBase::curve_to(XfCoord x, XfCoord y, XfCoord x1, XfCoord y1, XfCoord x2, XfCoord y2) {
    path_.curve_to(x, y, x1, y1, x2, y2);
}

//+ PainterBase(XfPainter::close_path)
void PainterBase::close_path() {
    path_.close_path();
}

//+ PainterBase(XfPainter::current_brush=b)
void PainterBase::current_brush(XfBrush_ptr b) {
    if (brush_ == b) {
        return;
    }

    CORBA::release(brush_);
    brush_ = Brush::_duplicate(b);
}

//+ PainterBase(XfPainter::current_brush?)
XfBrush_ptr PainterBase::current_brush() {
    return Brush::_duplicate(brush_);
}

//+ PainterBase(XfPainter::current_color=c)
void PainterBase::current_color(XfColor_ptr c) {
    if (color_ == c) {
        return;
    }

    CORBA::release(color_);
    color_ = Color::_duplicate(c);
}

//+ PainterBase(XfPainter::current_color?)
XfColor_ptr PainterBase::current_color() {
    return Color::_duplicate(color_);
}

//+ PainterBase(XfPainter::current_font=f)
void PainterBase::current_font(XfFont_ptr f) {
    if (font_ == f) {
        return;
    }

    CORBA::release(font_);
    font_ = Font::_duplicate(f);
}

//+ PainterBase(XfPainter::current_font?)
XfFont_ptr PainterBase::current_font() {
    return Font::_duplicate(font_);
}

//+ PainterBase(XfPainter::current_pattern=p)
void PainterBase::current_pattern(XfPattern_ptr p) {
    if (pattern_ == p) {
        return;
    }

    CORBA::release(pattern_);
    pattern_ = Pattern::_duplicate(p);
}

//+ PainterBase(XfPainter::current_pattern?)
XfPattern_ptr PainterBase::current_pattern() {
    return Pattern::_duplicate(pattern_);
}

//+ PainterBase(XfPainter::line)
void PainterBase::line(XfCoord x0, XfCoord y0, XfCoord x1, XfCoord y1) {
    begin_path();
    move_to(x0, y0);
    line_to(x1, y1);
    stroke();
}

//+ PainterBase(XfPainter::rect)
void PainterBase::rect(XfCoord x0, XfCoord y0, XfCoord x1, XfCoord y1) {
    rect_path(this, x0, y0, x1, y1);
    stroke();
}

//+ PainterBase(XfPainter::fill_rect)
void PainterBase::fill_rect(XfCoord x0, XfCoord y0, XfCoord x1, XfCoord y1) {
    rect_path(this, x0, y0, x1, y1);
    fill();
}

//+ PainterBase(XfPainter::current_matrix)
void PainterBase::current_matrix(Transform_ptr t) {
    matrix_->copy(t);
}

//+ PainterBase(XfPainter::current_matrix?)
XfTransform_ptr PainterBase::current_matrix() {
    return Transform::_duplicate(matrix_);
}

//+ PainterBase(XfPainter::push_matrix)
void PainterBase::push_matrix() {
    PainterMatrixStackElement item;
    matrix_->store_matrix(item.matrix);
    transforms_->prepend(item);
}

//+ PainterBase(XfPainter::pop_matrix)
void PainterBase::pop_matrix() {
    assert(transforms_->count() > 0);
    PainterMatrixStackElement item = transforms_->item(0);
    matrix_->load_matrix(item.matrix);
    transforms_->remove(0);
}

//+ PainterBase(XfPainter::premultiply)
void PainterBase::premultiply(XfTransform_ptr t) {
    matrix_->premultiply(t);
}

//+ PainterBase(XfPainter::clip_rect)
void PainterBase::clip_rect(XfCoord x0, XfCoord y0, XfCoord x1, XfCoord y1) {
    rect_path(this, x0, y0, x1, y1);
    clip();
}

//+ PainterBase(XfPainter::comment)
void PainterBase::comment(XfCharString_ptr s) {
}

//+ PainterBase(XfPainter::page_number)
void PainterBase::page_number(XfCharString_ptr s) {
}
