/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1995 Silicon Graphics, Inc.
 * Copyright (c) 1995 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

/*
 * Wraster.cxx
 */

#include <Fresco/defs/display.h>

#include <Fresco/defs_impl/drawings.h>
#include <Fresco/defs_impl/memory.h>
#include <Fresco/defs_impl/rasters.h>

// class RasterImageImpl

RasterImageImpl::RasterImageImpl(const Raster::Info& info, Pixel* image) {
    scale_ = Coord(0);
    info_ = info;
    long pixels = info_.rows * info_.columns;
    image_ = new RasterImageImpl::Pixel[pixels];
    Memory::copy(image, image_, pixels * sizeof(RasterImageImpl::Pixel));
}

RasterImageImpl::~RasterImageImpl() {
    delete image_;
}

OxExchange* RasterImageImpl::_exchange() { return &object_; }
 
//+ RasterImageImpl(XfFrescoObject::=object_.)
XfTag RasterImageImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void RasterImageImpl::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void RasterImageImpl::disconnect() {
    object_.disconnect();
}
void RasterImageImpl::notify_observers() {
    object_.notify_observers();
}
void RasterImageImpl::update() {
    object_.update();
}
//+

Boolean RasterImageImpl::equal(XfRaster_ptr r) {
    return is_eq(this, r);
}

ULong RasterImageImpl::hash() {
    return (ULong) this;
}

void RasterImageImpl::raster_info(XfRaster::Info& i) {
    i = info_;
}

void RasterImageImpl::set_color(XfRaster::Index row, XfRaster::Index col, XfColor_ptr c) {
    Pixel& pixel = image_[(info_.rows - 1 - row) * info_.columns + col];
    c->rgb(pixel.r, pixel.g, pixel.b);
    pixel.a = Float(1);
    notify_observers();
}

void RasterImageImpl::set_color_blend(XfRaster::Index row, XfRaster::Index col, XfColor_ptr c, Float blend) {
    Pixel& pixel = image_[(info_.rows - 1 - row) * info_.columns + col];
    c->rgb(pixel.r, pixel.g, pixel.b);
    pixel.a = blend;
    notify_observers();
}

void RasterImageImpl::set_mask(XfRaster::Index row, XfRaster::Index col, Boolean on) {
    Pixel& pixel = image_[(info_.rows - 1 - row) * info_.columns + col];
    pixel.a = on ? Float(1) : Float(0);
    notify_observers();
}

XfColor_ptr RasterImageImpl::get_color(XfRaster::Index row, XfRaster::Index col) {
    Pixel& pixel = image_[(info_.rows - 1 - row) * info_.columns + col];
    return new ColorBase(pixel.r, pixel.g, pixel.b);
}

void RasterImageImpl::get_color_blend(XfRaster::Index row, XfRaster::Index col, XfColor_out c, Float& blend) {
    Pixel& pixel = image_[(info_.rows - 1 - row) * info_.columns + col];
    c = new ColorBase(pixel.r, pixel.g, pixel.b);
    blend = pixel.a;
}

Boolean RasterImageImpl::get_mask(XfRaster::Index row, XfRaster::Index col) {
    Pixel& pixel = image_[(info_.rows - 1 - row) * info_.columns + col];
    return pixel.a == Float(1);
}

XfCoord RasterImageImpl::scale() {
    return scale_;
}

void RasterImageImpl::scale(XfCoord scale) {
    scale_ = scale;
    notify_observers();
}

// class RasterBitmapImpl

RasterBitmapImpl::RasterBitmapImpl(const Raster::Info& info, unsigned char* image) {
    scale_ = Coord(0);
    info_ = info;
    bytes_per_row_ = (info_.columns + 7) / 8;
    long size = info_.rows * bytes_per_row_;
    image_ = new unsigned char[size];
    Memory::copy(image, image_, size);
}

RasterBitmapImpl::RasterBitmapImpl(
    const DrawingKit::Data8& data,
    Raster::Index rows, Raster::Index columns,
    Raster::Index origin_row, Raster::Index origin_column
) {
    scale_ = Coord(0);
    info_.rows = rows;
    info_.columns = columns;
    info_.origin_y = origin_row;
    info_.origin_x = origin_column;
    bytes_per_row_ = (info_.columns + 7) / 8;
    long size = info_.rows * bytes_per_row_;
    image_ = new unsigned char[size];
    Memory::copy(data._buffer, image_, size);
}

RasterBitmapImpl::~RasterBitmapImpl() {
    delete image_;
}

OxExchange* RasterBitmapImpl::_exchange() { return &object_; }
 
//+ RasterBitmapImpl(XfFrescoObject::=object_.)
XfTag RasterBitmapImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void RasterBitmapImpl::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void RasterBitmapImpl::disconnect() {
    object_.disconnect();
}
void RasterBitmapImpl::notify_observers() {
    object_.notify_observers();
}
void RasterBitmapImpl::update() {
    object_.update();
}
//+

Boolean RasterBitmapImpl::equal(XfRaster_ptr r) {
    return is_eq(this, r);
}

ULong RasterBitmapImpl::hash() {
    return (ULong) this;
}

void RasterBitmapImpl::raster_info(XfRaster::Info& i) {
    i = info_;
}

static inline unsigned char bit(XfRaster::Index col) {
    static unsigned char mask[] = {0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80};
    return mask[col % 8];
}

void RasterBitmapImpl::set_color(XfRaster::Index row, XfRaster::Index col, XfColor_ptr c) {
    // not supported
}

void RasterBitmapImpl::set_color_blend(XfRaster::Index row, XfRaster::Index col, XfColor_ptr c, Float blend) {
    // not supported
}

void RasterBitmapImpl::set_mask(XfRaster::Index row, XfRaster::Index col, Boolean on) {
    unsigned char& byte = image_[(info_.rows - 1 - row) * bytes_per_row_ + col / 8];
    if (on) {
        byte |= bit(col);
    } else {
        byte &= ~bit(col);
    }
    notify_observers();
}

XfColor_ptr RasterBitmapImpl::get_color(XfRaster::Index row, XfRaster::Index col) {
    unsigned char& byte = image_[(info_.rows - 1 - row) * bytes_per_row_ + col / 8];
    Color::Intensity v = byte & bit(col) ? Color::Intensity(1) : Color::Intensity(0);
    return new ColorBase(v, v, v);
}

void RasterBitmapImpl::get_color_blend(XfRaster::Index row, XfRaster::Index col, XfColor_out c, Float& blend) {
    unsigned char& byte = image_[(info_.rows - 1 - row) * bytes_per_row_ + col / 8];
    Color::Intensity v = byte & bit(col) ? Color::Intensity(1) : Color::Intensity(0);
    c = new ColorBase(v, v, v);
    blend = v;
}

Boolean RasterBitmapImpl::get_mask(XfRaster::Index row, XfRaster::Index col) {
    unsigned char& byte = image_[(info_.rows - 1 - row) * bytes_per_row_ + col / 8];
    return byte & bit(col);
}

XfCoord RasterBitmapImpl::scale() {
    return scale_;
}

void RasterBitmapImpl::scale(XfCoord scale) {
    scale_ = scale;
    notify_observers();
}
