//
// $XConsortium$
//

//
// Copyright (c) 1992-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names
// of Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#include <Fresco/defs_impl/action.h>
#include <Fresco/defs_impl/charstr.h>
#include <Fresco/defs_impl/fobjects.h>
#include <Fresco/defs_impl/math.h>
#include <Fresco/defs_impl/region.h>
#include <Fresco/defs_impl/telltales.h>
#include <Fresco/defs_impl/transform.h>
#include <assert.h>
#include <string.h>
#include <ctype.h>

//
// Sony NEWS-OS 6.0 has conflicting prototypes for abs() in math.h
// and stdlib.h, so you cannot include both headers in any C++ source
// file.  Until that bug is fixed (in NEWS-OS 6.0.1), we'll have to
// explicitly define the symbols we need from math.h on the Sony.
// On other platforms, we can just include math.h.
//

#if defined(sony) && defined(SVR4) || defined(__powerc)
extern "C" {
    double cos(double);
    double sin(double);
}
#else
#include <math.h>
#endif

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

CharStringImpl::CharStringImpl() {
    data_ = nil;
    length_ = 0;
    dynamic_ = false;
}

CharStringImpl::CharStringImpl(const char* s) {
    set(s, -1);
}

CharStringImpl::CharStringImpl(const char* s, Long length) {
    set(s, length);
}

CharStringImpl::CharStringImpl(CharString_ptr s) {
    data_ = nil;
    dynamic_ = false;
    copy(s);
}

CharStringImpl::CharStringImpl(const CharStringImpl& s) {
    set(s.data_, s.length_);
}

CharStringImpl::~CharStringImpl() {
    free();
}

OxExchange* CharStringImpl::_exchange() { return &object_; }

CharStringImpl* CharStringImpl::create_static(const char* str) {
    CharStringImpl* c = new CharStringImpl;
    c->data_ = (char*)str;
    c->length_ = strlen(str);
    c->dynamic_ = false;
    return c;
}

void CharStringImpl::set(const char* str, Long n) {
    length_ = (n == -1) ? strlen(str) : n;
    data_ = new char[length_ + 1];
    strncpy(data_, str, int(length_));
    data_[length_] = '\0';
    dynamic_ = true;
}

void CharStringImpl::free() {
    if (dynamic_) {
	delete data_;
    }
}

//+ CharStringImpl(XfFrescoObject::=object_.)
XfTag CharStringImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void CharStringImpl::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void CharStringImpl::disconnect() {
    object_.disconnect();
}
void CharStringImpl::notify_observers() {
    object_.notify_observers();
}
void CharStringImpl::update() {
    object_.update();
}
//+

//+ CharStringImpl(XfCharString::hash)
XfCharString::HashValue CharStringImpl::hash() {
    ULong v = 0;
    const char* q = &data_[length_];
    for (const char* p = data_; p < q; p++) {
	v = (v << 1) ^ (*p);
    }
    ULong t = v >> 10;
    t ^= (t >> 10);
    return v ^ t;
}

//+ CharStringImpl(XfCharString::count)
ULong CharStringImpl::count() {
    return length_;
}

//+ CharStringImpl(XfCharString::copy)
void CharStringImpl::copy(XfCharString_ptr s) {
    free();
    length_ = s->count();
    data_ = new char[length_ + 1];
    CharString::CharData buf(length_ + 1, 0, data_);
    s->get_char_data(buf);
    data_[length_] = '\0';
    dynamic_ = true;
    // avoid deallocating the newly-copied data
    buf._buffer = nil;
}

//+ CharStringImpl(XfCharString::equal)
Boolean CharStringImpl::equal(XfCharString_ptr s) {
    if (length_ == s->count()) {
	CharStringBuffer buf(s);
	Boolean b = strncmp(data_, buf.string(), int(length_)) == 0;
	return b;
    }
    return false;
}

//+ CharStringImpl(XfCharString::case_insensitive_equal)
Boolean CharStringImpl::case_insensitive_equal(XfCharString_ptr s) {
    if (length_ == s->count()) {
	CharStringBuffer buf(s);
	const char* str = buf.string();
	for (unsigned long i = 0; i < length_ && data_[i] != '\0'; i++) {
	    if (tolower(str[i]) != tolower(data_[i])) {
		return false;
		break;
	    }
	}
	return true;
    }
    return false;
}

//+ CharStringImpl(XfCharString::get_data)
void CharStringImpl::get_data(XfCharString::Data& d) {
    ULong n = (length_ > d._maximum) ? d._maximum : length_;
    for (ULong i = 0; i < n; i++) {
	d._buffer[i] = data_[i];
    }
}

//+ CharStringImpl(XfCharString::get_char_data)
void CharStringImpl::get_char_data(XfCharString::CharData& d) {
    ULong n;
    if (d._maximum == 0) {
	n = length_;
    } else {
	ULong m = d._maximum - 1;
	n = (length_ > m) ? m : length_;
    }
    for (ULong i = 0; i < n; i++) {
	d._buffer[i] = data_[i];
    }
    d._buffer[n] = '\0';
    d._length = n;
}

//+ CharStringImpl(XfCharString::put_data)
void CharStringImpl::put_data(const XfCharString::Data& d) {
    free();
    length_ = d._length;
    data_ = new char[length_];
    dynamic_ = true;
    for (ULong i = 0; i < length_; i++) {
	data_[i] = (char)d._buffer[i];
    }
}

//+ CharStringImpl(XfCharString::put_char_data)
void CharStringImpl::put_char_data(const XfCharString::CharData& d) {
    free();
    length_ = d._length;
    data_ = new char[length_];
    dynamic_ = true;
    for (ULong i = 0; i < length_; i++) {
	data_[i] = d._buffer[i];
    }
}

//+ CharStringImpl(XfCharString::string_ptr)
char* CharStringImpl::string_ptr() {
    char* s = new char[length_ + 1];  // Should use CORBA::string_alloc
    CharStringBuffer buf(this);
    strcpy(s, buf.string());
    return s;
}

const char* CharStringImpl::index(const char* s, Long length, char c) {
    const char* p = s;
    const char* q = s + length;
    for (; p < q; p++) {
	if (*p == c) {
	    return p;
	}
    }
    return nil;
}

void CharStringBuffer::init(CharString_ptr s) {
    Long n = s->count();
    if (n > sizeof(fixed_)) {
	buf_._buffer = new char[n + 1];
    } else {
	n = sizeof(fixed_);
	buf_._buffer = fixed_;
    }
    buf_._maximum = n;
    buf_._length = 0;
    s->get_char_data(buf_);
}

CharStringBuffer::~CharStringBuffer() {
    if (buf_._buffer == fixed_) {
	buf_._buffer = nil;
    }
}

Boolean CharStringBuffer::operator ==(const char* str) const {
    return strcmp(buf_._buffer, str) == 0;
}

Boolean CharStringBuffer::operator ==(const CharStringBuffer& b) const {
    return strcmp(buf_._buffer, b.buf_._buffer) == 0;
}

Boolean CharStringBuffer::operator !=(const char* str) const {
    return strcmp(buf_._buffer, str) != 0;
}

Boolean CharStringBuffer::operator !=(const CharStringBuffer& b) const {
    return strcmp(buf_._buffer, b.buf_._buffer) != 0;
}

ActionImpl::ActionImpl() { }
ActionImpl::~ActionImpl() { }

OxExchange* ActionImpl::_exchange() { return &object_; }

//+ ActionImpl(XfFrescoObject::=object_.)
XfTag ActionImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void ActionImpl::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void ActionImpl::disconnect() {
    object_.disconnect();
}
void ActionImpl::notify_observers() {
    object_.notify_observers();
}
void ActionImpl::update() {
    object_.update();
}
//+

//+ ActionImpl(XfAction::execute)
void ActionImpl::execute() { }

//+ ActionImpl(XfAction::reversible)
Boolean ActionImpl::reversible() { return false; }

//+ ActionImpl(XfAction::unexecute)
void ActionImpl::unexecute() { }

RegionImpl::RegionImpl() {
    Coord zero = Coord(0.0);
    defined_ = false;
    lower_.x = lower_.y = zero;
    upper_.x = upper_.y = zero;
    xalign_ = yalign_ = zalign_ = zero;
    lower_.z = Coord(-1e6);
    upper_.z = Coord(1e6);
}

RegionImpl::RegionImpl(Region_ptr a, Transform_ptr t1, Transform_ptr t2) {
    TransformImpl cumulative;
    Boolean xf = false;
    if (is_not_nil(t1) && !t1->identity()) {
	cumulative.copy(t1);
	xf = true;
    }
    if (is_not_nil(t2) && !t2->identity()) {
	cumulative.premultiply(t2);
	xf = true;
    }
    RegionImpl::copy(a);
    if (xf) {
	RegionImpl::apply_transform(&cumulative);
    }
}

RegionImpl::~RegionImpl() { }

OxExchange* RegionImpl::_exchange() { return &object_; }

//+ RegionImpl(XfFrescoObject::=object_.)
XfTag RegionImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void RegionImpl::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void RegionImpl::disconnect() {
    object_.disconnect();
}
void RegionImpl::notify_observers() {
    object_.notify_observers();
}
void RegionImpl::update() {
    object_.update();
}
//+

//+ RegionImpl(XfRegion::contains)
Boolean RegionImpl::contains(const XfVertex& v) {
    return (
	defined_ &&
	v.x >= lower_.x && v.x <= upper_.x &&
	v.y >= lower_.y && v.y <= upper_.y &&
	v.z >= lower_.z && v.z <= upper_.z
    );
}

//+ RegionImpl(XfRegion::contains_plane)
Boolean RegionImpl::contains_plane(const XfVertex& v, XfAxis a) {
    Boolean b = false;
    if (defined_) {
	switch (a) {
	case X_axis:
	    b = (
		v.y >= lower_.y && v.y <= upper_.y &&
		v.z >= lower_.z && v.z <= upper_.z
	    );
	    break;
	case Y_axis:
	    b = (
		v.x >= lower_.x && v.x <= upper_.x &&
		v.z >= lower_.z && v.z <= upper_.z
	    );
	    break;
	case Z_axis:
	    b = (
		v.x >= lower_.x && v.x <= upper_.x &&
		v.y >= lower_.y && v.y <= upper_.y
	    );
	    break;
	}
    }
    return b;
}

//+ RegionImpl(XfRegion::intersects)
Boolean RegionImpl::intersects(XfRegion_ptr r) {
    Boolean b = false;
    if (defined_) {
	Vertex lower, upper;
	r->bounds(lower, upper);
	b = (
	    upper.x >= lower_.x && lower.x <= upper_.x &&
	    upper.y >= lower_.y && lower.y <= upper_.y &&
	    upper.z >= lower_.z && lower.z <= upper_.z
	);
    }
    return b;
}

//+ RegionImpl(XfRegion::copy)
void RegionImpl::copy(XfRegion_ptr r) {
    Region::BoundingSpan x, y, z;
    r->span(X_axis, x);
    r->span(Y_axis, y);
    r->span(Z_axis, z);
    defined_ = true;
    lower_.x = x.begin;
    lower_.y = y.begin;
    lower_.z = z.begin;
    upper_.x = x.end;
    upper_.y = y.end;
    upper_.z = z.end;
    xalign_ = x.align;
    yalign_ = y.align;
    zalign_ = z.align;
    notify();
}

//+ RegionImpl(XfRegion::merge_intersect)
void RegionImpl::merge_intersect(XfRegion_ptr r) {
    if (defined_) {
	Vertex lower, upper;
	r->bounds(lower, upper);
	merge_max(lower_, lower);
	merge_min(upper_, upper);
    } else {
	copy(r);
    }
    notify();
}

//+ RegionImpl(XfRegion::merge_union)
void RegionImpl::merge_union(XfRegion_ptr r) {
    if (defined_) {
	Vertex lower, upper;
	r->bounds(lower, upper);
	merge_min(lower_, lower);
	merge_max(upper_, upper);
    } else {
	copy(r);
    }
    notify();
}

//+ RegionImpl(XfRegion::subtract) !r
void RegionImpl::subtract(XfRegion_ptr) {
    // not implemented
    notify();
}

//
// Transform a region.  This algorithm takes advantage of some
// interesting properties of affine transformations: i.e., opposite sides
// are always parallel and same in length.  Another property of
// affine transformation is that the transformed center of a box
// is equal to the center of the transformed box.  Realizing these
// properties, finding the transformed box can be broken down as finding
// the effective width, height, depth, and center.  The last is easy.
// Each of the other three values is found by first breaking down
// each of the transformed axis vectors into x, y,  and z vectors.
// Joining all the absolute vectors on a certain axis gives you
// the size of the box on that axis.  Width, for example,
// forms a (w, 0, 0) vector.  After transformation, it becomes
// (Xw, Yw, Zw).  Adding all the Xw's gives you the new width.
//

//+ RegionImpl(XfRegion::apply_transform)
void RegionImpl::apply_transform(XfTransform_ptr t) {
    if (defined_) {
	Vertex o;

	origin(o);
	t->transform_vertex(o);
	Transform::Matrix m;
	t->store_matrix(m);

        Coord x = upper_.x - lower_.x;
        Coord y = upper_.y - lower_.y;

        Vertex center;
        center.x = Coord((upper_.x + lower_.x) * 0.5);
        center.y = Coord((upper_.y + lower_.y) * 0.5);

	// transform the center

        t->transform_vertex(center);

	// optimized computation of new width and height
        Coord nx = Coord(Math::abs(x * m[0][0]) + Math::abs(y * m[1][0]));
        Coord ny = Coord(Math::abs(x * m[0][1]) + Math::abs(y * m[1][1]));

	// form the new box
        lower_.x = Coord(center.x - nx * 0.5);
	upper_.x = Coord(center.x + nx * 0.5);
        lower_.y = Coord(center.y - ny * 0.5);
	upper_.y = Coord(center.y + ny * 0.5);

        if (!Math::equal(nx, Coord(0), float(1e-4))) {
            xalign_ = (o.x - lower_.x) / nx;
        }
        if (!Math::equal(ny, Coord(0), float(1e-4))) {
            yalign_ = (o.y - lower_.y) / ny;
        }
	notify();
    }
}

Coord RegionImpl::span_align(Coord lower, Coord upper, Coord origin) {
    Coord s;
    if (Math::equal(lower, upper, float(1e-4))) {
	s = Coord(0.0);
    } else {
	s = Coord((origin - lower) / (upper - lower));
    }
    return s;
}

//+ RegionImpl(XfRegion::bounds)
void RegionImpl::bounds(XfVertex& lower, XfVertex& upper) {
    assert(defined_);
    lower = lower_;
    upper = upper_;
}

//+ RegionImpl(XfRegion::origin)
void RegionImpl::origin(XfVertex& v) {
    assert(defined_);
    v.x = span_origin(lower_.x, upper_.x, xalign_);
    v.y = span_origin(lower_.y, upper_.y, yalign_);
    v.z = span_origin(lower_.z, upper_.z, zalign_);
}

Coord RegionImpl::span_origin(Coord lower, Coord upper, Coord align) {
    Coord orig;
    if (Math::equal(lower, upper, float(1e-4))) {
	orig = Coord(0.0);
    } else {
	orig = lower + align * (upper - lower);
    }
    return orig;
}

//+ RegionImpl(XfRegion::span)
void RegionImpl::span(XfAxis a, XfRegion::BoundingSpan& s) {
    assert(defined_);
    switch (a) {
    case X_axis:
	s.begin = lower_.x;
	s.end = upper_.x;
	s.align = xalign_;
	break;
    case Y_axis:
	s.begin = lower_.y;
	s.end = upper_.y;
	s.align = yalign_;
	break;
    case Z_axis:
	s.begin = lower_.z;
	s.end = upper_.z;
	s.align = zalign_;
	break;
    }
    s.length = s.end - s.begin;
    s.origin = s.begin + s.align * s.length;
}

void RegionImpl::notify() { }

void RegionImpl::merge_min(Vertex& v0, const Vertex& v) {
    v0.x = Math::min(v0.x, v.x);
    v0.y = Math::min(v0.y, v.y);
    v0.z = Math::min(v0.z, v.z);
}

void RegionImpl::merge_max(Vertex& v0, const Vertex& v) {
    v0.x = Math::max(v0.x, v.x);
    v0.y = Math::max(v0.y, v.y);
    v0.z = Math::max(v0.z, v.z);
}

static const double radians_per_degree = M_PI / 180;
static const float tolerance = float(1e-4);

TransformImpl::TransformImpl() { init(); }

TransformImpl::TransformImpl(Transform::Matrix m) {
    load_matrix(m);
    identity_ = false;
    translate_only_ = false;
    xy_ = true;
    modified_ = true;
}

TransformImpl::~TransformImpl() { }

OxExchange* TransformImpl::_exchange() { return &object_; }

//+ TransformImpl(XfFrescoObject::=object_.)
XfTag TransformImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void TransformImpl::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void TransformImpl::disconnect() {
    object_.disconnect();
}
void TransformImpl::notify_observers() {
    object_.notify_observers();
}
void TransformImpl::update() {
    object_.update();
}
//+

void TransformImpl::init() {
    Coord zero = Coord(0), one = Coord(1);
    mat_[0][0] = mat_[1][1] = one;
    mat_[0][1] = mat_[1][0] = zero;
    mat_[0][2] = mat_[1][2] = zero;
    mat_[0][3] = mat_[1][3] = zero;
    mat_[2][0] = mat_[2][1] = zero;
    mat_[2][2] = mat_[2][3] = zero;
    mat_[3][0] = mat_[3][1] = zero;
    mat_[3][2] = mat_[3][3] = zero;
    identity_ = true;
    translate_only_ = true;
    xy_ = true;
    modified_ = false;
}

void TransformImpl::modified() {
    modified_ = true;
    notify_observers();
}

void TransformImpl::recompute() {
    Coord m00 = mat_[0][0];
    Coord m01 = mat_[0][1];
    Coord m10 = mat_[1][0];
    Coord m11 = mat_[1][1];
    Coord m20 = mat_[2][0];
    Coord m21 = mat_[2][1];
    Coord one_plus = 1 + tolerance, one_minus = 1 - tolerance;
    Coord zero_plus = 0 + tolerance, zero_minus = 0 - tolerance;
    identity_ = (
	m00 > one_minus && m00 < one_plus &&
	m11 > one_minus && m11 < one_plus &&
	m01 > zero_minus && m01 < zero_plus &&
	m10 > zero_minus && m10 < zero_plus &&
	m20 > zero_minus && m20 < zero_plus &&
	m21 > zero_minus && m21 < zero_plus
    );
    modified_ = false;
}

Coord TransformImpl::det() {
    return mat_[0][0] * mat_[1][1] - mat_[0][1] * mat_[1][0];
}

//+ TransformImpl(XfTransform::copy)
void TransformImpl::copy(XfTransform_ptr t) {
    if (is_nil(t)) {
	init();
    } else {
	Transform::Matrix m;
	t->store_matrix(m);
	load_matrix(m);
    }
}

//+ TransformImpl(XfTransform::load_matrix)
void TransformImpl::load_matrix(Matrix m) {
    mat_[0][0] = m[0][0]; mat_[0][1] = m[0][1];
    mat_[0][2] = m[0][2]; mat_[0][3] = m[0][3];
    mat_[1][0] = m[1][0]; mat_[1][1] = m[1][1];
    mat_[1][2] = m[1][2]; mat_[1][3] = m[1][3];
    mat_[2][0] = m[2][0]; mat_[2][1] = m[2][1];
    mat_[2][2] = m[2][2]; mat_[2][3] = m[2][3];
    mat_[3][0] = m[3][0]; mat_[3][1] = m[3][1];
    mat_[3][2] = m[3][2]; mat_[3][3] = m[3][3];
    modified();
}

//+ TransformImpl(XfTransform::load_identity)
void TransformImpl::load_identity() { init(); }

//+ TransformImpl(XfTransform::store_matrix)
void TransformImpl::store_matrix(Matrix m) {
    m[0][0] = mat_[0][0]; m[0][1] = mat_[0][1];
    m[0][2] = mat_[0][2]; m[0][3] = mat_[0][3];
    m[1][0] = mat_[1][0]; m[1][1] = mat_[1][1];
    m[1][2] = mat_[1][2]; m[1][3] = mat_[1][3];
    m[2][0] = mat_[2][0]; m[2][1] = mat_[2][1];
    m[2][2] = mat_[2][2]; m[2][3] = mat_[2][3];
    m[3][0] = mat_[3][0]; m[3][1] = mat_[3][1];
    m[3][2] = mat_[3][2]; m[3][3] = mat_[3][3];
}

//+ TransformImpl(XfTransform::equal)
Boolean TransformImpl::equal(XfTransform_ptr t) {
    if (modified_) {
	recompute();
    }
    if (identity_) {
	return is_nil(t) || t->identity();
    }
    if (is_nil(t) || t->identity()) {
	return false;
    }
    Transform::Matrix m;
    t->store_matrix(m);
    Coord m00 = mat_[0][0];
    Coord m01 = mat_[0][1];
    Coord m10 = mat_[1][0];
    Coord m11 = mat_[1][1];
    Coord m20 = mat_[2][0];
    Coord m21 = mat_[2][1];
    return (
	xy_ &&
	m00 < m[0][0] + tolerance && m00 > m[0][0] - tolerance &&
	m01 < m[0][1] + tolerance && m01 > m[0][1] - tolerance &&
	m10 < m[1][0] + tolerance && m10 > m[1][0] - tolerance &&
	m11 < m[1][1] + tolerance && m11 > m[1][1] - tolerance &&
	m20 < m[2][0] + tolerance && m20 > m[2][0] - tolerance &&
	m21 < m[2][1] + tolerance && m21 > m[2][1] - tolerance
    );
}

//+ TransformImpl(XfTransform::identity)
Boolean TransformImpl::identity() {
    if (modified_) {
	recompute();
    }
    return identity_;
}

//+ TransformImpl(XfTransform::det_is_zero)
Boolean TransformImpl::det_is_zero() {
    Coord d = det();
    return d < tolerance && d > -tolerance;
}

//+ TransformImpl(XfTransform::scale)
void TransformImpl::scale(const XfVertex& v) {
    mat_[0][0] *= v.x;
    mat_[0][1] *= v.y;

    mat_[1][0] *= v.x;
    mat_[1][1] *= v.y;
    mat_[2][0] *= v.x;
    mat_[2][1] *= v.y;
    translate_only_ = false;
    modified();
}

//+ TransformImpl(XfTransform::rotate)
void TransformImpl::rotate(Float angle, XfAxis a) {
    if (a != Z_axis) {
	return;
    }
    Transform::Matrix m;
    Coord r_angle = Coord(angle * radians_per_degree);
    Coord tmp1 = Coord(cos(r_angle));
    Coord tmp2 = Coord(sin(r_angle));
    
    m[0][0] = tmp1 * mat_[0][0];
    m[0][1] = tmp2 * mat_[0][1];
    m[1][0] = tmp1 * mat_[1][0];
    m[1][1] = tmp2 * mat_[1][1];
    m[2][0] = tmp1 * mat_[2][0];
    m[2][1] = tmp2 * mat_[2][1];

    mat_[0][1] = mat_[0][0] * tmp2 + mat_[0][1] * tmp1;
    mat_[1][1] = mat_[1][0] * tmp2 + mat_[1][1] * tmp1;
    mat_[2][1] = mat_[2][0] * tmp2 + mat_[2][1] * tmp1;
    mat_[0][0] = m[0][0] - m[0][1];
    mat_[1][0] = m[1][0] - m[1][1];
    mat_[2][0] = m[2][0] - m[2][1];
    translate_only_ = false;
    modified();
}

//+ TransformImpl(XfTransform::translate)
void TransformImpl::translate(const XfVertex& v) {
    mat_[2][0] += v.x;
    mat_[2][1] += v.y;
    modified();
}

//+ TransformImpl(XfTransform::premultiply)
void TransformImpl::premultiply(XfTransform_ptr t) {
    if (is_not_nil(t) && !t->identity()) {
	Transform::Matrix m;
	t->store_matrix(m);

	Coord tmp1 = mat_[0][0];
	Coord tmp2 = mat_[1][0];
	mat_[0][0]  = m[0][0] * tmp1 + m[0][1] * tmp2;
	mat_[1][0]  = m[1][0] * tmp1 + m[1][1] * tmp2;
	mat_[2][0] += m[2][0] * tmp1 + m[2][1] * tmp2;
     
	tmp1 = mat_[0][1];
	tmp2 = mat_[1][1];
     
	mat_[0][1]  = m[0][0] * tmp1 + m[0][1] * tmp2;
	mat_[1][1]  = m[1][0] * tmp1 + m[1][1] * tmp2;
	mat_[2][1] += m[2][0] * tmp1 + m[2][1] * tmp2;
	modified();
    }
}    

//+ TransformImpl(XfTransform::postmultiply)
void TransformImpl::postmultiply(XfTransform_ptr t) {
    if (is_not_nil(t) && !t->identity()) {
	Transform::Matrix m;
	t->store_matrix(m);

	Coord tmp = mat_[0][0] * m[0][1] + mat_[0][1] * m[1][1];
	mat_[0][0] = mat_[0][0] * m[0][0] + mat_[0][1] * m[1][0];
	mat_[0][1] = tmp;

	tmp = mat_[1][0] * m[0][1] + mat_[1][1] * m[1][1];
	mat_[1][0] = mat_[1][0] * m[0][0] + mat_[1][1] * m[1][0];
	mat_[1][1] = tmp;

	tmp = mat_[2][0] * m[0][1] + mat_[2][1] * m[1][1];
	mat_[2][0] = mat_[2][0] * m[0][0] + mat_[2][1] * m[1][0];
	mat_[2][1] = tmp;

	mat_[2][0] += m[2][0];
	mat_[2][1] += m[2][1];
	modified();
    }
}

//+ TransformImpl(XfTransform::invert)
void TransformImpl::invert() {
    Coord d = det();
    if (Math::equal(d, float(0), tolerance)) {
	return;
    }
    Coord t00 = mat_[0][0];
    Coord t20 = mat_[2][0];

    mat_[2][0] = (mat_[1][0] * mat_[2][1] - mat_[1][1] * mat_[2][0]) / d;
    mat_[2][1] = (mat_[0][1] * t20 - mat_[0][0] * mat_[2][1]) / d;
    mat_[0][0] = mat_[1][1]/d;
    mat_[1][1] = t00/d;
    mat_[1][0] = -mat_[1][0]/d;
    mat_[0][1] = -mat_[0][1]/d;
    modified();
}

//+ TransformImpl(XfTransform::transform_vertex)
void TransformImpl::transform_vertex(XfVertex& v) {
    Coord tx = v.x;
    v.x = tx * mat_[0][0] + v.y * mat_[1][0] + mat_[2][0];
    v.y = tx * mat_[0][1] + v.y * mat_[1][1] + mat_[2][1];
}

//+ TransformImpl(XfTransform::inverse_transform_vertex)
void TransformImpl::inverse_transform_vertex(XfVertex& v) {
    Coord d = det();
    Coord a = (v.x - mat_[2][0]) / d;
    Coord b = (v.y - mat_[2][1]) / d;
    v.x = a * mat_[1][1] - b * mat_[1][0];
    v.y = b * mat_[0][0] - a * mat_[0][1];
}

// class TelltaleImpl

TelltaleImpl::TelltaleImpl(TelltaleImpl::FlagSet f) {
    flags_ = f;
    current_ = nil;
}

TelltaleImpl::~TelltaleImpl() { }

OxExchange* TelltaleImpl::_exchange() { return &object_; }

//+ TelltaleImpl(XfFrescoObject::=object_.)
XfTag TelltaleImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void TelltaleImpl::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void TelltaleImpl::disconnect() {
    object_.disconnect();
}
void TelltaleImpl::notify_observers() {
    object_.notify_observers();
}
void TelltaleImpl::update() {
    object_.update();
}
//+

//+ TelltaleImpl(XfTelltale::set)
void TelltaleImpl::set(XfTelltale::Flag f) { modify(f, true); }

//+ TelltaleImpl(XfTelltale::clear)
void TelltaleImpl::clear(XfTelltale::Flag f) { modify(f, false); }

//+ TelltaleImpl(XfTelltale::test)
Boolean TelltaleImpl::test(XfTelltale::Flag f) {
    return (flags_ & (1 << f)) != 0;
}

//+ TelltaleImpl(XfTelltale::current=g)
void TelltaleImpl::current(XfTelltale_ptr g) {
    CORBA::release(current_);
    current_ = Telltale::_duplicate(g);
}

//+ TelltaleImpl(XfTelltale::current?)
XfTelltale_ptr TelltaleImpl::current() {
    return Telltale::_duplicate(current_);
}

void TelltaleImpl::modify(Telltale::Flag f, Boolean on) {
    TelltaleImpl::FlagSet fs = (1 << f);
    TelltaleImpl::FlagSet newflags = on ? (flags_ | fs) : (flags_ & ~fs);
    if (newflags != flags_) {
	if (on && f == Telltale::chosen && is_not_nil(current_)) {
	    Telltale_var t = current_->current();
	    if (is_not_nil(t)) {
		t->clear(Telltale::chosen);
	    }
	    current_->current(this);
	}
	flags_ = newflags;
	notify_observers();
    }
}

TelltaleImpl::FlagSet TelltaleImpl::flags_of(Telltale_ptr t) {
    return
        (t->test(Telltale::enabled) ? TelltaleImpl::enabled_bit : 0) |
        (t->test(Telltale::visible) ? TelltaleImpl::visible_bit : 0) |
        (t->test(Telltale::active) ? TelltaleImpl::active_bit : 0) |
        (t->test(Telltale::chosen) ? TelltaleImpl::chosen_bit : 0);
}
