/*
 * Copyright (c) 1995 Firefly Design
 * Copyright (c) 1993-95 Silicon Graphics, Inc.
 * Copyright (c) 1993-95 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include <assert.h>

#include <Fresco/defs/display.h>
#include <Fresco/defs/fresco.h>

#include <Fresco/defs_impl/charstr.h>
#include <Fresco/defs_impl/region.h>
#include <Fresco/defs_impl/styles.h>
#include <Fresco/defs_impl/threads.h>
#include <Fresco/defs_impl/traversal.h>
#include <Fresco/defs_impl/viewers.h>
#include <Fresco/defs_impl/windows.h>
	  
#include <Fresco/Ox/schema.h>

// class DamageImpl

DamageImpl::DamageImpl(WindowBase* window) {
    window_ = window;
}

OxExchange* DamageImpl::_exchange() { return &object_; } 

//+ DamageImpl(XfFrescoObject::=object_.)
XfTag DamageImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void DamageImpl::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void DamageImpl::disconnect() {
    object_.disconnect();
}
void DamageImpl::notify_observers() {
    object_.notify_observers();
}
void DamageImpl::update() {
    object_.update();
}
//+

void DamageImpl::incur() {
    window_->incur();
}

void DamageImpl::extend(XfRegion_ptr r) {
    window_->extend(r);
}

XfRegion_ptr DamageImpl::current() {
    return window_->current();
}

void DamageImpl::begin_update() {
    window_->begin_update();
}

void DamageImpl::end_update() {
    window_->end_update();
}

XfWindow_ptr DamageImpl::window_ptr() {
    return Window::_duplicate(window_);
}

// class WindowBase

/*
 * The relationship with the main viewer is a bit tricky to avoid
 * a circularity that would prevent either from being destroyed.
 * We assume that the "external" reference is to the window, and
 * that when it is destroyed it will in turn destroy the main viewer.
 * So, the main viewer doesn't hold a reference to the window.
 */

WindowBase::WindowBase(Display_ptr d, Viewer_ptr v) {
    style_ = new StyleImpl;
    StyleContext_var ds = d->style();
    style_->link_parent(ds);

    viewer_ = new MainViewer;
    viewer_->append_viewer(v);
    viewer_->body(v);
    viewer_->window_ = this;
    viewer_->allocation_ = new RegionImpl;
    viewer_->style(ds);

    ThreadKit_ptr t = ThreadLib::thread_kit_ptr();
    lock_ = t->mutex_ptr();

    wdamage_ = new DamageImpl(this);
    damage_lock_ = t->mutex_ptr();
    damage_ = new RegionImpl;

    pick_ = new PickTraversal(GlyphTraversal::pick_top, this, wdamage_);

    configured_ = false;
    left_ = 0;
    top_ = 0;
    width_ = 0;
    height_ = 0;
    x_align_ = Alignment(0);
    y_align_ = Alignment(0);

    cursor_ = nil;
    cursor_foreground_ = nil;
    cursor_background_ = nil;

    icon_ = nil;

    double_buffer_ = double_buffered();
    update_nesting_ = 0;
    draw_ = new GlyphTraversalImpl(GlyphTraversal::draw, this, wdamage_);
}

WindowBase::~WindowBase() {
    CORBA::release(style_);

    CORBA::release(viewer_->allocation_);
    CORBA::release(viewer_);

    CORBA::release(wdamage_);
    CORBA::release(damage_);

    CORBA::release(pick_);

    CORBA::release(cursor_);
    CORBA::release(cursor_foreground_);
    CORBA::release(cursor_background_);

    CORBA::release(icon_);

    CORBA::release(draw_);
}

Boolean WindowBase::configured() {
    return configured_;
}

void WindowBase::get_client_geometry(
    PixelCoord& x, PixelCoord& y, PixelCoord& w, PixelCoord& h
) {
    if (configured_) {
        x = left_;
        y = top_;
        w = width_;
        h = height_;
    } else {
        Glyph::Requisition r;
        GlyphImpl::init_requisition(r);
        viewer_->request(r);
        assert(r.x.defined && r.y.defined);
	Screen_var screen = screen_ptr();
        w = screen->to_pixels(r.x.natural, X_axis);
        h = screen->to_pixels(r.y.natural, Y_axis);
    }
}

OxExchange* WindowBase::_exchange() { return &object_; } 

//+ WindowBase(XfFrescoObject::=object_.)
XfTag WindowBase::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void WindowBase::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void WindowBase::disconnect() {
    object_.disconnect();
}
void WindowBase::notify_observers() {
    object_.notify_observers();
}
void WindowBase::update() {
    object_.update();
}
//+

//+ WindowBase(XfWindow::double_buffered=b)
void WindowBase::double_buffered(Boolean b) {
    set_attribute(
        "doubleBuffered", FrescoLib::tmp_string_ref(b ? "true" : "false")
    );
    double_buffer_ = b;
}

//+ WindowBase(XfWindow::double_buffered?)
Boolean WindowBase::double_buffered() {
    StyleValue_var value = style_->resolve(FrescoLib::tmp_string_ref("doubleBuffered"));
    return is_nil(value) || value->is_on();
}

//+ WindowBase(XfWindow::default_cursor=c)
void WindowBase::default_cursor(XfCursor_ptr c) {
    CORBA::release(cursor_);
    cursor_ = Cursor::_duplicate(c);
}

//+ WindowBase(XfWindow::default_cursor?)
XfCursor_ptr WindowBase::default_cursor() {
    return Cursor::_duplicate(cursor_);
}

//+ WindowBase(XfWindow::cursor_foreground=c)
void WindowBase::cursor_foreground(XfColor_ptr c) {
    cursor_foreground_ = Color::_duplicate(c);
}

//+ WindowBase(XfWindow::cursor_foreground?)
XfColor_ptr WindowBase::cursor_foreground() {
    return Color::_duplicate(cursor_foreground_);
}

//+ WindowBase(XfWindow::cursor_background=c)
void WindowBase::cursor_background(XfColor_ptr c) {
    cursor_background_ = Color::_duplicate(c);
}

//+ WindowBase(XfWindow::cursor_background?)
XfColor_ptr WindowBase::cursor_background() {
    return Color::_duplicate(cursor_background_);
}

//+ WindowBase(XfWindow::geometry=g)
void WindowBase::geometry(XfCharString_ptr g) {
    set_attribute("geometry", g);
}

//+ WindowBase(XfWindow::geometry?)
XfCharString_ptr WindowBase::geometry() {
    return get_attribute("geometry");
}

//+ WindowBase(XfWindow::icon=w)
void WindowBase::icon(XfWindow_ptr w) {
    CORBA::release(icon_);
    icon_ = Window::_duplicate(w);
}

//+ WindowBase(XfWindow::icon?)
XfWindow_ptr WindowBase::icon() {
    return Window::_duplicate(icon_);
}

//+ WindowBase(XfWindow::iconic=b)
void WindowBase::iconic(Boolean b) {
    set_attribute("iconic", FrescoLib::tmp_string_ref(b ? "true" : "false"));
}

//+ WindowBase(XfWindow::iconic?)
Boolean WindowBase::iconic() {
    return style_->is_on(FrescoLib::tmp_string_ref("iconic"));
}

//+ WindowBase(XfWindow::title=t)
void WindowBase::title(XfCharString_ptr t) {
    set_attribute("title", t);
}

//+ WindowBase(XfWindow::title?)
XfCharString_ptr WindowBase::title() {
    return get_attribute("title");
}

//+ WindowBase(XfWindow::xor_pixel=p)
void WindowBase::xor_pixel(Long p) {
    StyleValue_var a = style_->bind(FrescoLib::tmp_string_ref("xor_pixel"));
    a->write_integer(p);
}

//+ WindowBase(XfWindow::xor_pixel?)
Long WindowBase::xor_pixel() {
    long p = 0;
    StyleValue_var a = style_->resolve(FrescoLib::tmp_string_ref("xor_pixel"));
    if (is_not_nil(a)) {
        a->read_integer(p);
    }
    return p;
}

//+ WindowBase(XfWindow::style)
XfStyleContext_ptr WindowBase::style() {
    return StyleContext::_duplicate(style_);
}

//+ WindowBase(XfWindow::main_viewer)
XfViewer_ptr WindowBase::main_viewer() {
    return Viewer::_duplicate(viewer_);
}

//+ WindowBase(XfWindow::configure)
void WindowBase::configure(const XfWindow::Placement& p) {
    Screen_var screen = screen_ptr();
    width_ = screen->to_pixels(p.width, X_axis);
    height_ = screen->to_pixels(p.height, Y_axis);
    left_ = screen->to_pixels(p.x - p.align_x * p.width, X_axis);
    top_ = screen->to_pixels(screen->height() -
        (p.y + (Alignment(1) - p.align_y) * p.height), Y_axis);
    x_align_ = p.align_x;
    y_align_ = p.align_y;
    configured_ = true;
}

//+ WindowBase(XfWindow::get_configuration)
void WindowBase::get_configuration(Boolean position, XfWindow::Placement& p) {
    lock_->acquire();

    Screen_var screen = screen_ptr();
    p.width = screen->to_coord(width_, X_axis);
    p.height = screen->to_coord(height_, Y_axis);
    if (position) {
        p.x = screen->to_coord(left_, X_axis) + (x_align_ * p.width);
        p.y = screen->height() - (screen->to_coord(top_, Y_axis) +
            (Alignment(1) - y_align_) * screen->to_coord(height_, Y_axis));
        p.align_x = x_align_;
        p.align_y = y_align_;
    }

    lock_->release();
}

//+ WindowBase(XfWindow::configure_notify)
void WindowBase::configure_notify(XfCoord width, XfCoord height) {
}

//+ WindowBase(XfWindow::move_notify)
void WindowBase::move_notify(XfCoord left, XfCoord bottom) {
}

//+ WindowBase(XfWindow::map_notify)
void WindowBase::map_notify() {
}

//+ WindowBase(XfWindow::unmap_notify)
void WindowBase::unmap_notify() {
}

//+ WindowBase(XfWindow::redraw)
void WindowBase::redraw(XfCoord left, XfCoord bottom, XfCoord width, XfCoord height) {
    lock_->acquire();
    RegionImpl r;
    r.defined_ = true;
    r.lower_.x = left;
    r.lower_.y = bottom;
    r.upper_.x = left + width;
    r.upper_.y = bottom + height;
    extend(&r);
    lock_->release();
}

void WindowBase::render() {
    draw_->push(viewer_, viewer_, nil, viewer_->allocation_, nil);
    viewer_->traverse(draw_);
    draw_->pop();
}

//+ WindowBase(XfWindow::hits)
Boolean WindowBase::hits(XfEvent_ptr e) {
    return e->hits(viewer_);
}

//+ WindowBase(XfWindow::handle_event)
void WindowBase::handle_event(XfEvent_ptr e) {
    if (!e->positional()) {
        pick_->push(viewer_, viewer_, nil, viewer_->allocation_, nil);
        pick_->begin_viewer(viewer_);
        viewer_->handle(pick_, e);
        pick_->end_viewer();
        pick_->pop();
    } else {
        //
        // The pick traversal uses clipping to determine visibility so 
        // we clip down around the point before doing the traversal.
        //
        pick_->clear();
        Coord x = e->pointer_x(), y = e->pointer_y();
        Painter_var painter = pick_->current_painter();
        painter->push_clipping();
        painter->clip_rect(x, y, x, y);
        pick_->push(viewer_, viewer_, nil, viewer_->allocation_, nil);
        viewer_->traverse(pick_);
        pick_->pop();
        painter->pop_clipping();

        GlyphTraversal_var p = pick_->picked();
        if (is_not_nil(p)) {
            Vertex ev;
            do {
                Glyph_var g = p->current_glyph();
                Viewer_var v = p->current_viewer();
                if (is_nil(g) && is_not_nil(v)) {
                    ev.x = x; ev.y = y; ev.z = Coord(0);
                    Transform_var t = p->current_transform();
                    if (is_not_nil(t)) {
                        t->inverse_transform_vertex(ev);
                    }
                    painter->push_clipping();
                    painter->clip_rect(ev.x, ev.y, ev.x, ev.y);
                    Boolean ok = v->handle(p, e);
                    painter->pop_clipping();
                    if (ok) {
                        break;
                    }
                }
            } while (p->backward());
        }
   }
}

//+ WindowBase(XfDamage::incur)
void WindowBase::incur() {
    Screen_var screen = screen_ptr();
    RegionImpl r;
    r.defined_ = true;
    r.lower_.x = Coord(0);
    r.lower_.y = Coord(0);
    r.upper_.x = screen->to_coord(width_, X_axis);
    r.upper_.y = screen->to_coord(height_, Y_axis);
    extend(&r);
}

//+ WindowBase(XfDamage::extend)
void WindowBase::extend(XfRegion_ptr r) {
    damage_lock_->acquire();
    if (damage_->defined_) {
        damage_->merge_union(r);
    } else {
        damage_->copy(r);
        if (update_nesting_ == 0) {
            Screen_var screen = screen_ptr();
	    Display_var display = screen->display_ptr();
            display->need_repair(this);
        }
    }
    damage_lock_->release();
}

//+ WindowBase(XfDamage::current)
XfRegion_ptr WindowBase::current() {
    damage_lock_->acquire();
    Region_ptr r = Region::_duplicate(damage_);
    damage_lock_->release();
    return r;
}

//+ WindowBase(XfDamage::begin_update)
void WindowBase::begin_update() {
    damage_lock_->acquire();
    ++update_nesting_;
    damage_lock_->release();
}

//+ WindowBase(XfDamage::end_update)
void WindowBase::end_update() {
    damage_lock_->acquire();
    Long n = --update_nesting_;
    if (n == 0 && damage_->defined_) {
        Screen_var screen = screen_ptr();
	Display_var display = screen->display_ptr();
        display->need_repair(this);
    }
    damage_lock_->release();
}

//+ WindowBase(XfDamage::window_ptr)
XfWindow_ptr WindowBase::window_ptr() {
    return Window::_duplicate(this);
}

void WindowBase::set_attribute(const char* name, CharString_ptr value) {
    StyleValue_var a = style_->bind(FrescoLib::tmp_string_ref(name));
    a->write_string(value);
}

CharString_ptr WindowBase::get_attribute(const char* name) {
    return FrescoLib::get_string(style_, name);
}
