#include <Fresco/defs/fresco.h>
#include <Fresco/defs/drawing.h>
#include <Fresco/figures/figures.h>
#include <Fresco/layouts/layouts.h>
#include <Fresco/widgets/widgets.h>
#include <Fresco/defs_impl/viewers.h>
#include <Fresco/defs_impl/transform.h>
#include <Fresco/defs_impl/region.h>
#include "ftutorial.h"

class FiggyViewer : public ViewerImpl {
public:
    FiggyViewer(Coord w=300, Coord h=300);
    ~FiggyViewer();
    void need_resize();
    Boolean press(GlyphTraversal_ptr, Event_ptr);
    Boolean drag(GlyphTraversal_ptr, Event_ptr);
    Boolean release(GlyphTraversal_ptr, Event_ptr);
    void add(Coord x, Coord y);
protected:
    FigureKit_var figures_;
    FigureStyle_var style_;
    Glyph_var root_polyfig_;
    Vertex start_;
    Glyph_var target_;
    Transform_var target_tx_;
    enum {
	create_tool, move_tool, move_root_tool,
	scale_tool, scale_root_tool
    } curtool_;
};

FiggyViewer::FiggyViewer(
    Coord w, Coord h
) : ViewerImpl(false /* opaque */) {
    figures_ = Fresco_resolve_object(FigureKit, "figures_impl");
    style_ = figures_->default_style();
    target_tx_ = new TransformImpl;
    root_polyfig_ = figures_->group();
    Glyph_var g = new Requestor(root_polyfig_, w,h);
    body(g);
}

FiggyViewer::~FiggyViewer() { }

void FiggyViewer::need_resize() { }

void FiggyViewer::add(Coord x, Coord y) {
    static int toggle = 0;
    Glyph_var fig;
    if (++toggle%2) {
	const float r = 30.0f;
 	fig = figures_->circle(FigureKit::fill, style_, x, y, r);
    } else {
	const Coord hh = Coord(25), hw = Coord(25);  // half-height,width
	fig = figures_->rectangle(
	    FigureKit::fill_stroke, style_, x-hw,y-hh, x+hw,y+hh
	);
    }
    root_polyfig_->append(fig);
    Transform_var tx = fig->transformation();
    tx->copy(target_tx_);
    tx->invert();
    fig->need_redraw();
    fig->need_resize();
}

Boolean FiggyViewer::press(GlyphTraversal_ptr t, Event_ptr e) {
    start_.x = e->pointer_x();
    start_.y = e->pointer_y();
    Boolean picked = t->forward();
    if (picked) {
	while (t->forward());
	target_ = t->current_glyph();
	target_tx_->copy(Transform_var(t->current_transform()));
	if (e->modifier_is_down(Event::shift)) {
	    curtool_ = scale_tool;
	} else {
	    curtool_ = move_tool;
	}
    } else {
	target_ = root_polyfig_;
	target_tx_->copy(Transform_var(t->current_transform()));
	target_tx_->premultiply(
	    Transform_var(root_polyfig_->transformation())
	);
	if (e->modifier_is_down(Event::shift)) {
	    curtool_ = scale_root_tool;
	} else if (e->modifier_is_down(Event::control)) {
	    curtool_ = move_root_tool;
	} else {
	    curtool_ = create_tool;
	}
    }
    return true;
}

Boolean FiggyViewer::drag(GlyphTraversal_ptr, Event_ptr e) {    
    if (curtool_ == move_tool || curtool_ == move_root_tool) {
	Transform_var tx = target_->transformation();
	if (is_not_nil(tx)) {
	    Vertex start, end;
	    start.x = start_.x;
	    start.y = start_.y;
	    end.x = e->pointer_x();
	    end.y = e->pointer_y();
	    if (curtool_ == move_tool) {
		target_tx_->inverse_transform_vertex(start);
		target_tx_->inverse_transform_vertex(end);
	    }
	    Vertex v;
	    v.x = end.x-start.x;
	    v.y = end.y-start.y;
	    target_->need_redraw();
	    tx->translate(v);
	    target_->need_redraw();
	    target_->need_resize();
	    start_.x = e->pointer_x();
	    start_.y = e->pointer_y();
	}
    } else if (curtool_ == scale_tool || curtool_ == scale_root_tool) {
	Transform_var tx = target_->transformation();
	if (is_not_nil(tx)) {
	    Coord dx = Coord(0);
	    Coord dy = e->pointer_y() - start_.y;
	    if (dy != Coord(0))	{
		const float scale_factor = 400.0f;
		float factor;
		if (dy > 0.0) {
		    factor = scale_factor/(scale_factor+dy);
		} else {
		    factor = (scale_factor-dy)/scale_factor;
		}
		Vertex zoom, trans;
		zoom.x = factor; zoom.y = factor; zoom.z = Coord(0);
		trans.x = Coord(0); trans.y = Coord(0); trans.z = Coord(0);
		target_->need_redraw();
		if (curtool_ == scale_tool) {
		    RegionImpl ext;
		    Glyph::AllocationInfo info;
		    target_->extension(info, &ext);
		    Region::BoundingSpan xspan, yspan;
		    ext.span(X_axis, xspan);
		    ext.span(Y_axis, yspan);
		    trans.x = -(xspan.begin + xspan.end)*0.5f;
		    trans.y = -(yspan.begin + yspan.end)*0.5f;
		    tx->translate(trans);
		}
		tx->scale(zoom);
		if (curtool_ == scale_tool) {
		    trans.x = -trans.x;
		    trans.y = -trans.y;
		    tx->translate(trans);
		}
		target_->need_redraw();
		target_->need_resize();
	    }
	}
    }
    return true;
}
 
Boolean FiggyViewer::release(GlyphTraversal_ptr, Event_ptr) {    
    if (curtool_ == create_tool) {
	add(start_.x, start_.y);
    }
    return true;
}

int main(int argc, char** argv) {
    Coord w = Coord(300), h = Coord(300);
    FrescoContext* fc = FrescoLib::init_default_context(
	"FTutorial", argc, argv
    );
    WidgetKit_var widgets = Fresco_resolve_object(WidgetKit, "widgets_impl");
    LayoutKit_var layouts = Fresco_resolve_object(LayoutKit, "layouts_impl");
    Viewer_var fv = new FiggyViewer(w,h);
    DrawingKit_var drawing = fc->default_drawing_kit();
    Glyph_var glyph = layouts->back(
	Glyph_var(layouts->flexible(fv, layouts->fil(), Coord(0))),
	Glyph_var(widgets->color_filler(
	    Color_var(drawing->background(StyleContext_var(drawing->style())))
	))
    );
    FrescoLib::run(fv, glyph);
    CORBA::release(fc);
    return 0;
}
