//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names
// of Stanford, Silicon Graphics, and Fujitsu may not be used in any
// advertising or publicity relating to the software without the specific,
// prior written permission of Stanford, Silicon Graphics, and Fujitsu.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD, SILICON GRAPHICS, OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#include "box.h"
#include "layout_mgr.h"
#include "place.h"
#include <Fresco/defs/fresco.h>
#include <Fresco/defs_impl/transform.h>

Fresco_file_id(Fresco_layouts_impl_box_cxx)

const int Box::static_size_ = 10;

Box::Box(LayoutManager* layout) {
    layout_ = layout;
    requested_ = false;
}

Box::~Box() {
    delete layout_;
}

//+ Box(XfGlyph::request)
void Box::request(XfGlyph::Requisition& r) {
    if (!requested_) {
	GlyphImpl::init_requisition(requisition_);
	long n = children_.count();
	if (n > 0) {
	    Glyph::Requisition req[static_size_];
	    Glyph::Requisition* r = children_requests(req, static_size_);
	    layout_->request(n, r, requisition_);
	    if (r != req) {
		delete [] r;
	    }
	}
	requested_ = true;
    }
    r = requisition_;
}

//+ Box(XfGlyph::extension)
void Box::extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r) {
    Long n = children_.count();
    if (n > 0) {
	Glyph::AllocationInfo child;
	Vertex prev_o, o, v;
	prev_o.x = Coord(0); prev_o.y = Coord(0); prev_o.z = Coord(0);
	TransformImpl child_tx, tmp_tx;
	child.transformation.ptr_ = &child_tx;
	child.transformation->copy(a.transformation);
	child.damaged = a.damaged;
	RegionImpl** result = children_allocations(a.allocation);
	for (Long i = 0; i < n; i++) {
	    Placement::normal_origin(result[i], o);
	    v.x = o.x - prev_o.x;
	    v.y = o.y - prev_o.y;
	    v.z = o.z - prev_o.z;
	    tmp_tx.load_identity();
	    tmp_tx.translate(v);
	    child.allocation = result[i];
	    child.transformation->premultiply(&tmp_tx);
	    children_.item(i)->child_->extension(child, r);
	    prev_o = o;
	}
	child.transformation.ptr_ = nil;
	delete [] result;
    }
}

//
// The cull test below (allocation_is_visible) is not quite ideal.
// If the box contains something that draws outside the box's allocation,
// then the box might not redraw correctly.  However, the alternative,
// using extensions, can be pretty expensive.  In practice, using
// allocations works in every case we have seen.
//

//+ Box(XfGlyph::traverse)
void Box::traverse(XfGlyphTraversal_ptr t) {
    Region_var given = t->allocation();
    if (is_not_nil(given)) {
	if (t->allocation_is_visible()) {
	    traverse_with_allocation(t, given);
	}
    } else {
	traverse_without_allocation(t);
    }
}

//
// Both traverse and child_allocate could be made more efficient
// by avoiding memory allocation using a fixed size array of regions
// when the number of children is below a reasonable amount.
//

RegionImpl** Box::children_allocations(Region_ptr a) {
    Long n = children_.count();
    Glyph::Requisition req[static_size_];
    Glyph::Requisition* r = children_requests(req, static_size_);
    if (!requested_) {
	GlyphImpl::init_requisition(requisition_);
	layout_->request(n, r, requisition_);
	requested_ = true;
    }
    RegionImpl** result = new RegionImpl*[n];
    for (Long i = 0; i < n; i++) {
	RegionImpl* region = new RegionImpl;
	region->defined_ = true;
	result[i] = region;
    }
    layout_->allocate(n, r, a, result);
    if (r != req) {
	delete [] r;
    }
    return result;
}

void Box::traverse_with_allocation(GlyphTraversal_ptr t, Region_ptr a) {
    RegionImpl** result = children_allocations(a);
    Vertex prev_o;
    prev_o.x = Coord(0); prev_o.y = Coord(0); prev_o.z = Coord(0);
    Long i_start, i_stop, i_incr;
    Boolean i_picking;
    GlyphTraversal::Operation op = t->op();
    switch (op) {
    case GlyphTraversal::pick_top:
    case GlyphTraversal::pick_any:
	i_picking = true;
	i_start = children_.count() - 1;
	i_stop = -1;
	i_incr = -1;
	break;
    default:
	i_picking = false;
	i_start = 0;
	i_stop = children_.count();
	i_incr = +1;
	break;
    }
    Vertex o, v;
    TransformImpl tx;
    Painter_var p = t->current_painter();
    p->push_matrix();
    for (Long i = i_start; i != i_stop; i += i_incr) {
	Placement::normal_origin(result[i], o);
	v.x = o.x - prev_o.x;
	v.y = o.y - prev_o.y;
	v.z = o.z - prev_o.z;
	tx.load_identity();
	tx.translate(v);
	p->premultiply(&tx);
	t->traverse_child(children_.item(i), result[i]);
	if (i_picking && is_not_nil(GlyphTraversal_var(t->picked()))) {
	    break;
	}
	prev_o = o;
    }
    p->pop_matrix();
    Long n = children_.count();
    for (i = 0; i < n; i++) {
        CORBA::release(result[i]);
    }
    delete [] result;
}

void Box::traverse_without_allocation(GlyphTraversal_ptr t) {
    GlyphTraversal::Operation op = t->op();
    if (op == GlyphTraversal::pick_top || op == GlyphTraversal::pick_any) {
	for (Long i = children_.count() - 1; i >= 0; i--) {
	    t->traverse_child(children_.item(i), nil);
	    if (is_not_nil(GlyphTraversal_var(t->picked()))) {
		break;
	    }
	}
    } else {
	for (ListItr(PolyGlyphOffsetList) i(children_); i.more(); i.next()) {
	    t->traverse_child(i.cur(), nil);
	}
    }
}

void Box::child_allocate(Long index, Glyph::AllocationInfo& a) {
    Region_ptr given = a.allocation;
    RegionImpl** result = children_allocations(given);
    TransformImpl tx;
    Placement::normal_transform(result[index], &tx);
    if (is_nil(a.transformation)) {
	a.transformation = new TransformImpl;
    }
    a.transformation->premultiply(&tx);
    given->copy(result[index]);
    Long n = children_.count();
    for (Long i = 0; i < n; i++) {
	CORBA::release(result[i]);
    }
    delete [] result;
}

//+ Box(XfGlyph::need_resize)
void Box::need_resize() {
    modified();
    PolyGlyph::need_resize();
}

void Box::modified() {
    requested_ = false;
}

// class BoxAlignElements
 
BoxAlignElements::BoxAlignElements(
    LayoutManager* layout, Axis axis, Alignment alignment
) : Box(layout) {
    axis_ = axis;
    alignment_ = alignment;
}
 
BoxAlignElements::~BoxAlignElements() { }
 
//+ BoxAlignElements(XfGlyph::append)
void BoxAlignElements::append(XfGlyph_ptr g) {
    Box::append(
        Glyph_var(new Placement(g, new LayoutCenter(axis_, alignment_)))
    );
}
 
//+ BoxAlignElements(XfGlyph::prepend)
void BoxAlignElements::prepend(XfGlyph_ptr g) {
    Box::prepend(
        Glyph_var(new Placement(g, new LayoutCenter(axis_, alignment_)))
    );
}
