//
// Copyright (c) 1994-95 Firefly Design
// Copyright (c) 1991-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names
// of Stanford, Silicon Graphics, and Fujitsu may not be used in any
// advertising or publicity relating to the software without the specific,
// prior written permission of Stanford, Silicon Graphics, and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL STANFORD, SILICON GRAPHICS, OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
// OF THIS SOFTWARE.
//
 
#include "grid.h"
#include "layout_mgr.h"
#include <Fresco/defs/drawing.h>
#include <Fresco/defs_impl/region.h>
#include <Fresco/defs_impl/math.h>
#include <Fresco/defs_impl/memory.h>
#include <Fresco/layouts/layouts.h>
#include <assert.h>
#include <string.h>

// Span

static void set_span(
    Span& s, Coord origin, Coord length, Alignment align
) {
    Coord begin = origin - length * align;
    Coord end = begin + length;
    s.lower_ = begin;
    s.upper_ = end;
    s.align_ = align;
}

static void spans_to_region(Span& x_span, Span& y_span, RegionImpl* r) {
    r->defined_ = true;
    r->lower_.x = x_span.lower_;
    r->upper_.x = x_span.upper_;
    r->xalign_ = x_span.align_;
    r->lower_.y = y_span.lower_;
    r->upper_.y = y_span.upper_;
    r->yalign_ = y_span.align_;
}

static void offset_region(RegionImpl* r, Coord dx, Coord dy) {
    r->lower_.x += dx;
    r->upper_.x += dx;
    r->lower_.y += dy;
    r->upper_.y += dy;
}

// LayoutAlignRequest

class LayoutAlignRequest {
public:
    LayoutAlignRequest();
 
    void margin(Coord margin);
 
    void accumulate(const Glyph::Requirement& r);
    void requirement(Glyph::Requirement& r) const;
 
protected:
    Coord natural_lead;
    Coord min_lead;
    Coord max_lead;
    Coord natural_trail;
    Coord min_trail;
    Coord max_trail;
};
 
LayoutAlignRequest::LayoutAlignRequest() {
    natural_lead = Coord(0);
    min_lead = -infinite_coord;
    max_lead = infinite_coord;
    natural_trail = Coord(0);
    min_trail = -infinite_coord;
    max_trail = infinite_coord;
}
 
void LayoutAlignRequest::margin(Coord margin) {
    natural_lead += margin;
    min_lead -= margin;
    max_lead += margin;
    natural_trail += margin;
    min_trail -= margin;
    max_trail += margin;
}
 
void LayoutAlignRequest::accumulate(const Glyph::Requirement& r) {
    if (r.defined) {
        Coord r_nat = r.natural;
        Coord r_max = r.maximum;
        Coord r_min = r.minimum;
        Coord r_align = r.align;
        Coord r_inv_align = Coord(1) - r_align;
        natural_lead = Math::max(natural_lead, Coord(r_nat * r_align));
        max_lead = Math::min(max_lead, Coord(r_max * r_align));
        min_lead = Math::max(min_lead, Coord(r_min * r_align));
        natural_trail = Math::max(
            natural_trail, Coord(r_nat * r_inv_align)
        );
        max_trail = Math::min(max_trail, Coord(r_max * r_inv_align));
        min_trail = Math::max(min_trail, Coord(r_min * r_inv_align));
    }
}
 
void LayoutAlignRequest::requirement(Glyph::Requirement& r) const {
    GlyphImpl::require_lead_trail(
        r, natural_lead, max_lead, min_lead, natural_trail, max_trail, min_trail
    );
}
 
// class LayoutTileRequest

class LayoutTileRequest {
public:
    LayoutTileRequest();
 
    void space(Coord space);
    void flexible();
 
    void accumulate(const Glyph::Requirement& r);
    void requirement(Glyph::Requirement& r) const;
 
    Coord natural;
    Coord min_size;
    Coord max_size;
};
 
LayoutTileRequest::LayoutTileRequest() {
    natural = Coord(0);
    min_size = Coord(0);
    max_size = Coord(0);
}
 
void LayoutTileRequest::space(Coord space) {
    natural += space;
    max_size += space;
    min_size += space;
}
 
void LayoutTileRequest::flexible() {
    max_size = infinite_coord;
}
 
void LayoutTileRequest::accumulate(const Glyph::Requirement& r) {
    if (r.defined) {
        natural += r.natural;
        max_size += r.maximum;
        min_size += r.minimum;
    }
}
 
void LayoutTileRequest::requirement(Glyph::Requirement& r) const {
    r.defined = true;
    r.natural = natural;
    r.maximum = max_size;
    r.minimum = min_size;
    r.align = Alignment(0);
}
 
// class LayoutTileAllocate

class LayoutTileAllocate {
public:
    LayoutTileAllocate(
        Axis a, Glyph::Requisition& total, Boolean first_aligned,
        Region_ptr given
    );

    static Coord compute_length(
        const Glyph::Requirement& r, const Region::BoundingSpan& s
    );
    static float compute_squeeze(
       const Glyph::Requirement& r, Coord length
    );

    void next_span(const Glyph::Requirement& r, Span& s);

private:
    Boolean first_aligned_;
    Boolean growing_;
    Boolean shrinking_;
    Coord f_;
    Coord p_;
    long i_;
};

LayoutTileAllocate::LayoutTileAllocate(
    Axis a, Glyph::Requisition& total, Boolean first_aligned, Region_ptr given
) {
    first_aligned_ = first_aligned;
    Glyph::Requirement* r;
    Region::BoundingSpan s;
    r = GlyphImpl::requirement(total, a);
    given->span(a, s);
    Coord length = compute_length(*r, s);
    growing_ = length > r->natural;
    shrinking_ = length < r->natural;
    f_ = compute_squeeze(*r, length);
    p_ = s.origin;
    i_ = 0;
}

Coord LayoutTileAllocate::compute_length(
    const Glyph::Requirement& r, const Region::BoundingSpan& s
) {
    Coord length = s.length;
    Coord s_a = s.align;
    Coord r_a = r.align;
    if (r_a == 0) {
        length *= (1 - s_a);
    } else if (r_a == 1) {
        length *= s_a;
    } else {
        length *= Math::min(s_a / r_a, (1 - s_a) / (1 - r_a));
    }
    return length;
}
 
float LayoutTileAllocate::compute_squeeze(
    const Glyph::Requirement& r, Coord length
) {
    float f;
    Coord nat = r.natural;
    if (length > nat && r.maximum > nat) {
        f = (length - nat) / (r.maximum - nat);
    } else if (length < nat && r.minimum < nat) {
        f = (nat - length) / (nat - r.minimum);
    } else {
        f = 0.0f;
    }
    return f;
}

void LayoutTileAllocate::next_span(const Glyph::Requirement& r, Span& s) {
    if (r.defined) {
        Coord cspan = r.natural;
        if (growing_) {
            cspan += f_ * (r.maximum - r.natural);
        } else if (shrinking_) {
            cspan -= f_ * (r.natural - r.minimum);
        }
        if (first_aligned_ && (i_ == 0)) {
            p_ -= r.align * cspan;
        }
        set_span(s, p_ + cspan * r.align, cspan, r.align);
        p_ += cspan;
    } else {
        set_span(s, p_, Coord(0), Alignment(0));
    }
    ++i_;
}

// class GridImpl

GridImpl::GridImpl(const Grid::Index& upper) {
    dimensions_[X_axis].initialize(upper.col);
    dimensions_[Y_axis].initialize(upper.row);

    cursor_.col = cursor_.row = 0;

    requested_ = false;
    requisition_.z.defined = false;
}

GridImpl::~GridImpl() {
}

OxExchange* GridImpl::_exchange() {
    return &object_;
}

//+ GridImpl(XfFrescoObject::=object_.)
XfTag GridImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void GridImpl::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void GridImpl::disconnect() {
    object_.disconnect();
}
void GridImpl::notify_observers() {
    object_.notify_observers();
}
void GridImpl::update() {
    object_.update();
}
//+

XfGlyph_ptr GridImpl::clone_glyph() {
    return nil;
}
XfStyleContext_ptr GridImpl::style() {
    return nil;
}
void GridImpl::style(XfStyleContext_ptr) {
}
XfTransform_ptr GridImpl::transformation() {
    return nil;
}
void GridImpl::extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r) {
    GlyphImpl::default_extension(a, r);
}
void GridImpl::shape(XfRegion_ptr) {
}
void GridImpl::draw(XfGlyphTraversal_ptr) {
}
void GridImpl::pick(XfGlyphTraversal_ptr) {
}
XfGlyph_ptr GridImpl::body() {
    return nil;
}
void GridImpl::body(XfGlyph_ptr) {
}
Tag GridImpl::add_parent(XfGlyphOffset_ptr parent_offset) {
    Long n = parents_.count();
    parents_.append(parent_offset);
    return n;
}
void GridImpl::remove_parent(Tag add_tag) {
    Long n = Long(add_tag);
    if (n >= 0 && n < parents_.count()) {
        parents_.remove(n);
    }
}
XfGlyphOffset_ptr GridImpl::first_child_offset() {
    return nil;
}
XfGlyphOffset_ptr GridImpl::last_child_offset() {
    return nil;
}
void GridImpl::parent_offsets(XfGlyph::OffsetSeq& parents) {
    GlyphImpl::get_parent_offsets(parents, parents_);
}
void GridImpl::allocations(XfGlyph::AllocationInfoSeq& a) {
    for (ListItr(GlyphOffsetList) i(parents_); i.more(); i.next()) {
        i.cur()->allocations(a);
    }
}
void GridImpl::need_redraw() {
    Glyph::AllocationInfoSeq alist;
    allocations(alist);
    for (ULong i = 0; i < alist->length(); i++) {
        Glyph::AllocationInfo& a = alist[i];
        if (is_not_nil(a.damaged)) {
            RegionImpl r;
            extension(a, &r);
            if (r.defined_) {
                a.damaged->extend(&r);
            }
        }
    }
}
void GridImpl::need_redraw_region(XfRegion_ptr r) {
    Glyph::AllocationInfoSeq alist;
    allocations(alist);
    RegionImpl dr;
    for (ULong i = 0; i < alist->length(); i++) {
        Glyph::AllocationInfo& a = alist[i];
        if (is_not_nil(a.damaged)) {
            dr.copy(r);
            dr.apply_transform(a.transformation);
            a.damaged->extend(&dr);
        }
    }
}
Boolean GridImpl::restore_trail(XfGlyphTraversal_ptr t) {
    return false;
}

void GridImpl::replace(XfGlyph_ptr g, const Grid::Index& index) {
    assert((0 <= index.col) && (index.col < dimensions_[int(X_axis)].count_));
    assert((0 <= index.row) && (index.row < dimensions_[int(Y_axis)].count_));

    GridOffset* offset = g ? new GridOffset(this, g, index) : nil;

    GridOffset** o;
    o = &dimensions_[int(X_axis)].offsets_[index.col];
    for (; *o && ((*o)->index_.row <= index.row); o = &(*o)->next(Y_axis)) {
    }
    GridOffset* r = *o && ((*o)->index_.row == index.row) && ((*o)->index_.col == index.col) ?
        *o : nil;
    if (offset != nil) {
        if (r != nil) {
            offset->next(Y_axis) = r->next(Y_axis);
        } else {
            offset->next(Y_axis) = *o;
        }
    }
    *o = offset;

    o = &dimensions_[int(Y_axis)].offsets_[index.row];
    for (; *o && ((*o)->index_.col <= index.col); o = &(*o)->next(X_axis)) {
    }
    if (offset != nil) {
        if (r != nil) {
            assert(r == *o);
            offset->next(X_axis) = r->next(X_axis);
        } else {
            offset->next(X_axis) = *o;
        }
    }
    *o = offset;

    CORBA::release(r);
}

void GridImpl::append(XfGlyph_ptr g) {
    replace(g, cursor_);

    if (++cursor_.col >= dimensions_[int(X_axis)].count_) {
        Long count = dimensions_[int(Y_axis)].count_;
        cursor_.row = (cursor_.row + 1) % count;
        cursor_.col = 0;
    }
}

void GridImpl::prepend(XfGlyph_ptr g) {
    if (--cursor_.col < 0) {
        Long count = dimensions_[int(Y_axis)].count_;
        cursor_.row = (cursor_.row - 1 + count) % count;
        cursor_.col = dimensions_[int(X_axis)].count_ - 1;
    }
    // !!! this should do insert the glyph -denis
}

void GridImpl::request_range(
    XfGlyph::Requisition& r, const Grid::Range& range
) {
    cache_request();

    partial_request(X_axis, range.lower.col, range.upper.col, r.x);
    partial_request(Y_axis, range.lower.row, range.upper.row, r.y);
}

void GridImpl::request(XfGlyph::Requisition& r) {
    cache_request();

    r = requisition_;
}


void GridImpl::cache_request() {
    if (! requested_) {
        full_request(X_axis, Y_axis);
        full_request(Y_axis, X_axis);
        requested_ = true;
    }
}

void GridImpl::partial_request(
    Axis axis, Long begin, Long end, XfGlyph::Requirement& r
) {
    GridDimension& d = dimensions_[int(axis)];
    LayoutTileRequest tile;
    for (Long i = begin; i < end; ++i) {
        tile.accumulate(d.requirements_[i]);
    }
    tile.requirement(r);
}

void GridImpl::full_request(Axis axis, Axis direction) {
    GridDimension& d = dimensions_[int(axis)];
    LayoutTileRequest tile;
    for (int i = 0; i < d.count_; ++i) {
        LayoutAlignRequest align;
        for (GridOffset* o = d.offsets_[i]; o; o = o->next(direction)) {
            align.accumulate(o->requirement(axis));
        }
        XfGlyph::Requirement& r = d.requirements_[i];
        align.requirement(r);
        tile.accumulate(r);
    }
    XfGlyph::Requirement& r = *GlyphImpl::requirement(requisition_, axis);
    tile.requirement(r);
}

void GridImpl::traverse_range(GlyphTraversal_ptr t, const Grid::Range& range) {
    Region_var given = t->allocation();
    if (is_not_nil(given)) {
        if (t->allocation_is_visible()) {
            traverse_with_allocation(t, given, range);
        }
    } else {
        traverse_without_allocation(t, range);
    }
}

void GridImpl::traverse(XfGlyphTraversal_ptr t) {
    Grid::Range range;
    range.lower.col = 0;
    range.upper.col = dimensions_[int(X_axis)].count_;
    range.lower.row = 0;
    range.upper.row = dimensions_[int(Y_axis)].count_;

    traverse_range(t, range);
}

Span* GridImpl::full_allocate(Axis axis, Region_ptr given) {
    GridDimension& d = dimensions_[int(axis)];
    Span* spans = new Span[d.count_];
    LayoutTileAllocate allocate(axis, requisition_, false, given);
    for (int i = 0; i < d.count_; ++i) {
        allocate.next_span(d.requirements_[i], spans[i]);
    }
    return spans;
}

Grid::Index GridImpl::find(GlyphTraversal_ptr t) {
    Grid::Range range;
    range.lower.col = 0;
    range.upper.col = dimensions_[int(X_axis)].count_;
    range.lower.row = 0;
    range.upper.row = dimensions_[int(Y_axis)].count_;

    return find_range(t, range);
}

Grid::Index GridImpl::find_range(
    GlyphTraversal_ptr t, const Grid::Range& range
) {
    Region_var given = t->allocation();
    Span* x_spans = full_allocate(X_axis, given);
    Span* y_spans = full_allocate(Y_axis, given);
    Painter_var p = t->current_painter();
    Region_var e = p->visible();
    Vertex lower, upper;
    e->bounds(lower, upper);
    Coord x = lower.x;
    Coord y = lower.y;
    // If the point is outside the range find the outermost cell.
    for (Long c = range.lower.col; c < (range.upper.col - 1); ++c) {
        if (x <= x_spans[c].upper_) {
            break;
        }
    }
    for (Long r = range.lower.row; r < (range.upper.row - 1); ++r) {
        if (y <= y_spans[r].upper_) {
            break;
        }
    }
    Grid::Index index;
    index.col = c;
    index.row = r;
    return index;
}

Grid::Index GridImpl::upper() {
    Grid::Index upper;
    upper.col = dimensions_[int(X_axis)].count_;
    upper.row = dimensions_[int(Y_axis)].count_;
    return upper;
}

void GridImpl::traverse_with_allocation(
    GlyphTraversal_ptr t, Region_ptr given, const Grid::Range& range
) {
    Span* x_spans = full_allocate(X_axis, given);
    Span* y_spans = full_allocate(Y_axis, given);
    Coord dx = x_spans[0].lower_ - x_spans[range.lower.col].lower_;
    Coord dy = y_spans[0].lower_ - y_spans[range.lower.row].lower_;
    GridDimension& d = dimensions_[int(Y_axis)];
    for (Long i = range.lower.row; i < range.upper.row; ++i) {
        for (GridOffset* o = d.offsets_[i]; o; o = o->next(X_axis)) {
            if (
                (range.lower.col <= o->index_.col) &&
                (o->index_.col <= range.upper.col)
            ) {
                RegionImpl* r = new RegionImpl;
                spans_to_region(x_spans[o->index_.col], y_spans[o->index_.row], r);
                offset_region(r, dx, dy);
                t->traverse_child(o, r);
                CORBA::release(r);
            }
        }
    }
    delete x_spans;
    delete y_spans;
}

void GridImpl::traverse_without_allocation(
    GlyphTraversal_ptr t, const Grid::Range& range
) {
    GridDimension& d = dimensions_[int(Y_axis)];
    for (Long i = range.lower.row; i < range.upper.row; ++i) {
        for (GridOffset* o = d.offsets_[i]; o; o = o->next(X_axis)) {
            if (
                (range.lower.col <= o->index_.col) &&
                (o->index_.col <= range.upper.col)
            ) {
                t->traverse_child(o, nil);
            }
        }
    }
}

void GridImpl::allocate_cell(
    XfRegion_ptr given, const Grid::Index& index, XfRegion_ptr a
) {
    Span* x_spans = full_allocate(X_axis, given);
    Span* y_spans = full_allocate(Y_axis, given);
    RegionImpl r;
    spans_to_region(x_spans[index.col], y_spans[index.row], &r);
    a->copy(&r);
    delete x_spans;
    delete y_spans;
}

void GridImpl::need_resize() {
    requested_ = false;

    for (ListItr(GlyphOffsetList) i(parents_); i.more(); i.next()) {
        i.cur()->notify();
    }
}

GridOffset::GridOffset(
    GridImpl* parent, Glyph_ptr child, Grid::Index index
) {
    parent_ = parent;
    child_ = Glyph::_duplicate(child);
    index_ = index;
    remove_tag_ = child_->add_parent(this);
}

GridOffset::~GridOffset() {
    child_->remove_parent(remove_tag_);
    CORBA::release(child_);
}
 
Glyph::Requirement GridOffset::requirement(Axis axis) {
    Glyph::Requisition r;
    child_->request(r);
    return axis == X_axis ? r.x : r.y;
}
 
OxExchange* GridOffset::_exchange() {
    return &object_;
}
 
//+ GridOffset(XfFrescoObject::=object_.)
XfTag GridOffset::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void GridOffset::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void GridOffset::disconnect() {
    object_.disconnect();
}
void GridOffset::notify_observers() {
    object_.notify_observers();
}
void GridOffset::update() {
    object_.update();
}
//+

//+ GridOffset(XfGlyphOffset::next_child)
XfGlyphOffset_ptr GridOffset::next_child() {
    return nil;
}
//+ GridOffset(XfGlyphOffset::prev_child)
XfGlyphOffset_ptr GridOffset::prev_child() {
    return nil;
}
//+ GridOffset(XfGlyphOffset::insert) !g
void GridOffset::insert(XfGlyph_ptr) {
}
//+ GridOffset(XfGlyphOffset::replace) !g
void GridOffset::replace(XfGlyph_ptr) {
}
//+ GridOffset(XfGlyphOffset::remove)
void GridOffset::remove() {
}
//+ GridOffset(XfGlyphOffset::visit_trail) !t
void GridOffset::visit_trail(XfGlyphTraversal_ptr) {
}

//+ GridOffset(XfGlyphOffset::parent)
XfGlyph_ptr GridOffset::parent() {
    return Grid::_duplicate(parent_);
}

//+ GridOffset(XfGlyphOffset::child)
XfGlyph_ptr GridOffset::child() {
    return Glyph::_duplicate(child_);
}

//+ GridOffset(XfGlyphOffset::allocations)
void GridOffset::allocations(XfGlyph::AllocationInfoSeq& a) {
    Long start = a.length();
    parent_->allocations(a);
    for (ULong i = start; i < a.length(); i++) {
        parent_->allocate_cell(a[i].allocation, index_, a[i].allocation);
    }
}

//+ GridOffset(XfGlyphOffset::notify)
void GridOffset::notify() {
    parent_->need_resize();
}

//+ GridOffset(XfGlyphOffset::traverse)
void GridOffset::traverse(XfGlyphTraversal_ptr t) {
    child_->traverse(t);
}

//+ GridOffset(XfGlyphOffset::child_allocate)
void GridOffset::child_allocate(XfGlyph::AllocationInfo& a) {
    parent_->allocate_cell(a.allocation, index_, a.allocation);
}

// class GridDimension

GridDimension::GridDimension() {
    count_ = 0;
    offsets_ = nil;
    requirements_ = nil;
}

GridDimension::~GridDimension() {
    for (int i = 0; i < count_; ++i) {
        CORBA::release(offsets_[i]);
    }
    delete offsets_;
    delete requirements_;
}

void GridDimension::initialize(Long count) {
    count_ = count;
    offsets_ = new GridOffset*[count];
    Memory::zero(offsets_, sizeof(GridOffset*) * count);
    requirements_ = new Glyph::Requirement[count];
}

// class SubGridImpl

SubGridImpl::SubGridImpl(
    Grid_ptr grid, const Grid::Range& range
) : offset_(this, grid, range) {
}

OxExchange* SubGridImpl::_exchange() {
    return GlyphImpl::_exchange();
}

//+ SubGridImpl(XfFrescoObject::=GlyphImpl::)
XfTag SubGridImpl::attach(XfFrescoObject_ptr observer) {
    return GlyphImpl::attach(observer);
}
void SubGridImpl::detach(XfTag attach_tag) {
    GlyphImpl::detach(attach_tag);
}
void SubGridImpl::disconnect() {
    GlyphImpl::disconnect();
}
void SubGridImpl::notify_observers() {
    GlyphImpl::notify_observers();
}
void SubGridImpl::update() {
    GlyphImpl::update();
}
//+

XfGlyph_ptr SubGridImpl::clone_glyph() {
    return GlyphImpl::clone_glyph();
}
XfStyleContext_ptr SubGridImpl::style() {
    return GlyphImpl::style();
}
void SubGridImpl::style(XfStyleContext_ptr _p) {
    GlyphImpl::style(_p);
}
XfTransform_ptr SubGridImpl::transformation() {
    return GlyphImpl::transformation();
}
void SubGridImpl::extension(const Glyph::AllocationInfo& a, XfRegion_ptr r) {
    GlyphImpl::extension(a, r);
}
void SubGridImpl::shape(XfRegion_ptr r) {
    GlyphImpl::shape(r);
}
void SubGridImpl::draw(XfGlyphTraversal_ptr t) {
    GlyphImpl::draw(t);
}
void SubGridImpl::pick(XfGlyphTraversal_ptr t) {
    GlyphImpl::pick(t);
}
XfGlyph_ptr SubGridImpl::body() {
    return GlyphImpl::body();
}
void SubGridImpl::body(XfGlyph_ptr _p) {
    GlyphImpl::body(_p);
}
void SubGridImpl::append(XfGlyph_ptr g) {
    GlyphImpl::append(g);
}
void SubGridImpl::prepend(XfGlyph_ptr g) {
    GlyphImpl::prepend(g);
}
Tag SubGridImpl::add_parent(XfGlyphOffset_ptr parent_offset) {
    return GlyphImpl::add_parent(parent_offset);
}
void SubGridImpl::remove_parent(Tag add_tag) {
    GlyphImpl::remove_parent(add_tag);
}
XfGlyphOffset_ptr SubGridImpl::first_child_offset() {
    return GlyphImpl::first_child_offset();
}
XfGlyphOffset_ptr SubGridImpl::last_child_offset() {
    return GlyphImpl::last_child_offset();
}
void SubGridImpl::parent_offsets(Glyph::OffsetSeq& parents) {
    GlyphImpl::parent_offsets(parents);
}
void SubGridImpl::allocations(Glyph::AllocationInfoSeq& a) {
    GlyphImpl::allocations(a);
}
void SubGridImpl::need_redraw() {
    GlyphImpl::need_redraw();
}
void SubGridImpl::need_redraw_region(XfRegion_ptr r) {
    GlyphImpl::need_redraw_region(r);
}
void SubGridImpl::need_resize() {
    GlyphImpl::need_resize();
}
Boolean SubGridImpl::restore_trail(XfGlyphTraversal_ptr t) {
    return GlyphImpl::restore_trail(t);
}

void SubGridImpl::request(Glyph::Requisition& r) {
    offset_.request(r);
}

void SubGridImpl::traverse(XfGlyphTraversal_ptr t) {
    t->traverse_child(&offset_, Region_var(t->allocation()));
}

void SubGridImpl::visit_trail(GlyphTraversal_ptr t) {
    Glyph::AllocationInfo a;
    a.allocation = t->allocation();
    a.transformation = Painter_var(t->current_painter())->current_matrix();
    child_allocate(a);
}
 
void SubGridImpl::child_allocate(Glyph::AllocationInfo&) { }
 
// class SubGridOffset

SubGridOffset::SubGridOffset(
    SubGridImpl* parent, Grid_ptr child, const Grid::Range& range
) : range_(range) {
    parent_ = parent;
    child_ = Grid::_duplicate(child);
    remove_tag_ = child_->add_parent(this);
}

SubGridOffset::~SubGridOffset() {
    child_->remove_parent(remove_tag_);
}

OxExchange* SubGridOffset::_exchange() {
    return &object_;
}

//+ SubGridOffset(XfFrescoObject::=object_.)
XfTag SubGridOffset::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void SubGridOffset::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void SubGridOffset::disconnect() {
    object_.disconnect();
}
void SubGridOffset::notify_observers() {
    object_.notify_observers();
}
void SubGridOffset::update() {
    object_.update();
}
//+

//+ SubGridOffset(XfGlyphOffset::parent)
XfGlyph_ptr SubGridOffset::parent() {
    return XfGlyph::_duplicate(parent_);
}

//+ SubGridOffset(XfGlyphOffset::child)
XfGlyph_ptr SubGridOffset::child() {
    return Grid::_duplicate(child_);
}

//+ SubGridOffset(XfGlyphOffset::next_child)
XfGlyphOffset_ptr SubGridOffset::next_child() {
    return nil;
}

//+ SubGridOffset(XfGlyphOffset::prev_child)
XfGlyphOffset_ptr SubGridOffset::prev_child() {
    return nil;
}

//+ SubGridOffset(XfGlyphOffset::allocations)
void SubGridOffset::allocations(XfGlyph::AllocationInfoSeq& a) {
    Long start = a.length();
    parent_->allocations(a);
    for (ULong i = start; i < a.length(); i++) {
        parent_->child_allocate(a[i]);
    }
}

//+ SubGridOffset(XfGlyphOffset::insert) !g
void SubGridOffset::insert(XfGlyph_ptr) {
}
//+ SubGridOffset(XfGlyphOffset::replace) !g
void SubGridOffset::replace(XfGlyph_ptr) {
}
//+ SubGridOffset(XfGlyphOffset::remove)
void SubGridOffset::remove() {
}

//+ SubGridOffset(XfGlyphOffset::notify)
void SubGridOffset::notify() {
    parent_->need_resize();
}

//+ SubGridOffset(XfGlyphOffset::visit_trail)
void SubGridOffset::visit_trail(XfGlyphTraversal_ptr t) {
    parent_->visit_trail(t);
}
//+ SubGridOffset(XfGlyphOffset::child_allocate)
void SubGridOffset::child_allocate(XfGlyph::AllocationInfo& a) {
    parent_->child_allocate(a);
}

//+ SubGridOffset(XfGlyphOffset::traverse)
void SubGridOffset::traverse(XfGlyphTraversal_ptr t) {
    child_->traverse_range(t, range_);
}

void SubGridOffset::request(Glyph::Requisition& r) {
    child_->request_range(r, range_);
}
