//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names
// of Stanford, Silicon Graphics, and Fujitsu may not be used in any
// advertising or publicity relating to the software without the specific,
// prior written permission of Stanford, Silicon Graphics, and Fujitsu.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD, SILICON GRAPHICS, OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#include "layout_mgr.h"
#include <Fresco/defs_impl/glyphs.h>
#include <Fresco/defs_impl/math.h>
#include <Fresco/defs_impl/region.h>

// class LayoutManager

LayoutManager::LayoutManager() { }
LayoutManager::~LayoutManager() { }

void LayoutManager::set_span(
    RegionImpl* r, Axis a, Coord origin, Coord length, Alignment align
) {
    Coord begin = origin - length * align;
    Coord end = begin + length;
    switch (a) {
    case X_axis:
	r->lower_.x = begin;
	r->upper_.x = end;
	r->xalign_ = align;
	break;
    case Y_axis:
	r->lower_.y = begin;
	r->upper_.y = end;
	r->yalign_ = align;
	break;
    case Z_axis:
	r->lower_.z = begin;
	r->upper_.z = end;
	r->zalign_ = align;
	break;
    }
}

// class LayoutAlign

LayoutAlign::LayoutAlign(Axis a) {
    axis_ = a;
}

LayoutAlign::~LayoutAlign() { }

void LayoutAlign::request(
    long n, Glyph::Requisition* requests, Glyph::Requisition& result
) {
    Glyph::Requirement* r;
    Coord natural_lead = Coord(0);
    Coord min_lead = -infinite_coord;
    Coord max_lead = infinite_coord;
    Coord natural_trail = Coord(0);
    Coord min_trail = -infinite_coord;
    Coord max_trail = infinite_coord;
    for (long i = 0; i < n; i++) {
	r = GlyphImpl::requirement(requests[i], axis_);
	if (r->defined) {
	    Coord r_nat = r->natural;
	    Coord r_max = r->maximum;
	    Coord r_min = r->minimum;
	    Coord r_align = r->align;
	    Coord r_inv_align = Coord(1) - r_align;
            natural_lead = Math::max(natural_lead, Coord(r_nat * r_align));
            max_lead = Math::min(max_lead, Coord(r_max * r_align));
            min_lead = Math::max(min_lead, Coord(r_min * r_align));
            natural_trail = Math::max(
                natural_trail, Coord(r_nat * r_inv_align)
            );
            max_trail = Math::min(max_trail, Coord(r_max * r_inv_align));
            min_trail = Math::max(min_trail, Coord(r_min * r_inv_align));
        }
    }
    r = GlyphImpl::requirement(result, axis_);
    GlyphImpl::require_lead_trail(
	*r, natural_lead, max_lead, min_lead,
	natural_trail, max_trail, min_trail
    );
}

void LayoutAlign::allocate(
    long n, Glyph::Requisition* requests, Region_ptr given,
    LayoutManager::Allocations result
) {
    Glyph::Requirement* r;
    Region::BoundingSpan s;
    given->span(axis_, s);
    for (long i = 0; i < n; i++) {
        r = GlyphImpl::requirement(requests[i], axis_);
	if (r->defined) {
	    Coord length = Math::max(
		Math::min(LayoutTile::compute_length(*r, s), r->maximum),
		r->minimum
	    );
	    set_span(result[i], axis_, s.origin, length, r->align);
	} else {
	    set_span(result[i], axis_, s.origin, s.length, s.align);
	}
    }
}

// class LayoutCenter

LayoutCenter::LayoutCenter(Axis a, Alignment align) {
    axis_ = a;
    alignment_ = align;
}

LayoutCenter::~LayoutCenter() { }

void LayoutCenter::request(
    long, Glyph::Requisition*, Glyph::Requisition& result
) {
    Glyph::Requirement* r = GlyphImpl::requirement(result, axis_);
    r->align = alignment_;
}

void LayoutCenter::allocate(
    long, Glyph::Requisition* requests, Region_ptr,
    LayoutManager::Allocations result
) {
    Region::BoundingSpan s;
    result[0]->span(axis_, s);
    Glyph::Requirement* r = GlyphImpl::requirement(requests[0], axis_);
    if (r->defined) {
        Alignment a = r->align;
        Coord n = Math::min(r->maximum, Math::max(r->minimum, s.length));
        set_span(result[0], axis_, s.origin + (a - s.align) * n, s.length, a);
    }
}

// class LayoutFixed

LayoutFixed::LayoutFixed(Axis a, Coord size) {
    axis_ = a;
    size_ = size;
}

LayoutFixed::~LayoutFixed() { }

void LayoutFixed::request(
    long, Glyph::Requisition*, Glyph::Requisition& result
) {
    Glyph::Requirement* r = GlyphImpl::requirement(result, axis_);
    r->natural = size_;
    r->maximum = size_;
    r->minimum = size_;
    if (!r->defined) {
	r->defined = true;
	r->align = Alignment(0);
    }
}

void LayoutFixed::allocate(
    long, Glyph::Requisition* , Region_ptr, LayoutManager::Allocations result
) {
    Region::BoundingSpan s;
    result[0]->span(axis_, s);
    set_span(result[0], axis_, s.origin, size_, s.align);
}

// class LayoutMargin

LayoutMargin::LayoutMargin(Coord margin) {
    lnatural_ = margin; lstretch_ = Coord(0); lshrink_ = Coord(0);
    rnatural_ = margin; rstretch_ = Coord(0); rshrink_ = Coord(0);
    bnatural_ = margin; bstretch_ = Coord(0); bshrink_ = Coord(0);
    tnatural_ = margin; tstretch_ = Coord(0); tshrink_ = Coord(0);
}

LayoutMargin::LayoutMargin(Coord hmargin, Coord vmargin) {
    lnatural_ = hmargin; lstretch_ = Coord(0); lshrink_ = Coord(0);
    rnatural_ = hmargin; rstretch_ = Coord(0); rshrink_ = Coord(0);
    bnatural_ = vmargin; bstretch_ = Coord(0); bshrink_ = Coord(0);
    tnatural_ = vmargin; tstretch_ = Coord(0); tshrink_ = Coord(0);
}

LayoutMargin::LayoutMargin(
    Coord lmargin, Coord rmargin, Coord bmargin, Coord tmargin
) {
    lnatural_ = lmargin; lstretch_ = Coord(0); lshrink_ = Coord(0);
    rnatural_ = rmargin; rstretch_ = Coord(0); rshrink_ = Coord(0);
    bnatural_ = bmargin; bstretch_ = Coord(0); bshrink_ = Coord(0);
    tnatural_ = tmargin; tstretch_ = Coord(0); tshrink_ = Coord(0);
}

LayoutMargin::LayoutMargin(
    Coord lmargin, Coord lstretch, Coord lshrink,
    Coord rmargin, Coord rstretch, Coord rshrink,
    Coord bmargin, Coord bstretch, Coord bshrink,
    Coord tmargin, Coord tstretch, Coord tshrink
) {
    lnatural_ = lmargin; lstretch_ = lstretch; lshrink_ = lshrink;
    rnatural_ = rmargin; rstretch_ = rstretch; rshrink_ = rshrink;
    bnatural_ = bmargin; bstretch_ = bstretch; bshrink_ = bshrink;
    tnatural_ = tmargin; tstretch_ = tstretch; tshrink_ = tshrink;
}

LayoutMargin::~LayoutMargin() { }

void LayoutMargin::request(
    long, Glyph::Requisition*, Glyph::Requisition& result
) {
    Glyph::Requirement& rx = result.x;
    if (rx.defined) {
	Coord dx = lnatural_ + rnatural_;
	rx.natural += dx;
	rx.maximum += dx + (lstretch_ + rstretch_);
	rx.minimum += dx - (lshrink_ + rshrink_);
    }
    Glyph::Requirement& ry = result.y;
    if (ry.defined) {
	Coord dy = bnatural_ + tnatural_;
	ry.natural += dy;
	ry.maximum += dy + (bstretch_ + tstretch_);
	ry.minimum += dy - (bshrink_ + tshrink_);
    }
    requisition_ = result;
}

void LayoutMargin::allocate(
    long, Glyph::Requisition*, Region_ptr, LayoutManager::Allocations result
) {
    allocate_axis(
	X_axis,
	lnatural_, lstretch_, lshrink_, rnatural_, rstretch_, rshrink_,
	result
    );
    allocate_axis(
	Y_axis,
	bnatural_, bstretch_, bshrink_, tnatural_, tstretch_, tshrink_,
	result
    );
}

void LayoutMargin::allocate_axis(
    Axis a,
    Coord natural_lead, Coord stretch_lead, Coord shrink_lead,
    Coord natural_trail, Coord stretch_trail, Coord shrink_trail,
    LayoutManager::Allocations result
) {
    Region::BoundingSpan s;
    result[0]->span(a, s);
    Glyph::Requirement* r = GlyphImpl::requirement(requisition_, a);
    Coord lead = span(s.length, *r, natural_lead, stretch_lead, shrink_lead);
    Coord trail = span(
	s.length, *r, natural_trail, stretch_trail, shrink_trail
    );
    s.length -= (lead + trail);
    s.origin += ((1 - r->align) * lead - r->align * trail);
    set_span(result[0], a, s.origin, s.length, s.align);
}

Coord LayoutMargin::span(
    Coord span, Glyph::Requirement& total,
    Coord natural, Coord stretch, Coord shrink
) {
    Coord extra = span - total.natural;
    Coord result = natural;
    float ss = 0.0f;
    Coord total_stretch = total.maximum - total.natural;
    Coord total_shrink = total.natural - total.minimum;
    if (extra > 0 && total_stretch > 0) {
        ss = stretch / total_stretch;
    } else if (extra < 0 && total_shrink > 0) {
        ss = shrink / total_shrink;
    }
    return result + ss * extra;
}

// class LayoutNatural

LayoutNatural::LayoutNatural(Axis a, Coord natural) {
    axis_ = a;
    natural_ = natural;
}

LayoutNatural::~LayoutNatural() { }

void LayoutNatural::request(
    long, Glyph::Requisition*, Glyph::Requisition& result
) {
    Glyph::Requirement* r = GlyphImpl::requirement(result, axis_);
    r->defined = true;
    r->natural = natural_;
}

void LayoutNatural::allocate(
    long, Glyph::Requisition*, Region_ptr, LayoutManager::Allocations
) {
    // leave it as is
}

// class LayoutSuperpose

LayoutSuperpose::LayoutSuperpose(
    LayoutManager* first, LayoutManager* second
) {
    first_ = first;
    second_ = second;
    third_ = nil;
}

LayoutSuperpose::LayoutSuperpose(
    LayoutManager* first, LayoutManager* second, LayoutManager* third
) {
    first_ = first;
    second_ = second;
    third_ = third;
}

LayoutSuperpose::~LayoutSuperpose() {
    delete first_;
    delete second_;
    delete third_;
}

void LayoutSuperpose::request(
    long n, Glyph::Requisition* requests, Glyph::Requisition& result
) {
    if (first_ != nil) {
	first_->request(n, requests, result);
    }
    if (second_ != nil) {
	second_->request(n, requests, result);
    }
    if (third_ != nil) {
	third_->request(n, requests, result);
    }
}

void LayoutSuperpose::allocate(
    long n, Glyph::Requisition* requests, Region_ptr given,
    LayoutManager::Allocations result
) {
    if (first_ != nil) {
	first_->allocate(n, requests, given, result);
    }
    if (second_ != nil) {
	second_->allocate(n, requests, given, result);
    }
    if (third_ != nil) {
	third_->allocate(n, requests, given, result);
    }
}

// class LayoutTile

LayoutTile::LayoutTile(Axis a) { axis_ = a; }
LayoutTile::~LayoutTile() { }

void LayoutTile::request(
    long n, Glyph::Requisition* requests, Glyph::Requisition& result
) {
    compute_request(axis_, Alignment(0), n, requests, result);
    requisition_ = result;
}

void LayoutTile::allocate(
    long n, Glyph::Requisition* requests, Region_ptr given,
    LayoutManager::Allocations result
) {
    compute_allocations(
	axis_, requisition_, false, n, requests, given, result
    );
}

void LayoutTile::compute_request(
    Axis a, Alignment align,
    long n, Glyph::Requisition* requests, Glyph::Requisition& result
) {
    Glyph::Requirement* r;
    Coord natural = Coord(0), min_size = Coord(0), max_size = Coord(0);
    for (long i = 0; i < n; i++) {
        r = GlyphImpl::requirement(requests[i], a);
        if (r->defined) {
	    Coord n = r->natural;
	    natural += n;
	    max_size += r->maximum;
	    min_size += r->minimum;
	}
    }
    r = GlyphImpl::requirement(result, a);
    r->defined = true;
    r->natural = natural;
    r->maximum = max_size;
    r->minimum = min_size;
    r->align = align;
}

void LayoutTile::compute_allocations(
    Axis a, Glyph::Requisition& total, Boolean first_aligned,
    long n, Glyph::Requisition* requests, Region_ptr given,
    LayoutManager::Allocations result
) {
    Glyph::Requirement* r;
    Region::BoundingSpan s;
    r = GlyphImpl::requirement(total, a);
    given->span(a, s);
    Coord length = compute_length(*r, s);
    Boolean growing = length > r->natural;
    Boolean shrinking = length < r->natural;
    float f = compute_squeeze(*r, length);
    Coord p = s.origin;
    for (long i = 0; i < n; i++) {
        r = GlyphImpl::requirement(requests[i], a);
        if (r->defined) {
            Coord cspan = r->natural;
            if (growing) {
                cspan += f * (r->maximum - r->natural);
            } else if (shrinking) {
                cspan -= f * (r->natural - r->minimum);
            }
            if (first_aligned && i == 0) {
                p -= r->align * cspan;
            }
	    set_span(result[i], a, p + cspan * r->align, cspan, r->align);
            p += cspan;
        } else {
	    set_span(result[i], a, p, Coord(0), Coord(0));
        }
    }
}

Coord LayoutTile::compute_length(
    const Glyph::Requirement& r, const Region::BoundingSpan& s
) {
    Coord length = s.length;
    Coord s_a = s.align;
    Coord r_a = r.align;
    if (r_a == 0) {
	length *= (1 - s_a);
    } else if (r_a == 1) {
	length *= s_a;
    } else {
	length *= Math::min(s_a / r_a, (1 - s_a) / (1 - r_a));
    }
    return length;
}

float LayoutTile::compute_squeeze(const Glyph::Requirement& r, Coord length) {
    float f;
    Coord nat = r.natural;
    if (length > nat && r.maximum > nat) {
        f = (length - nat) / (r.maximum - nat);
    } else if (length < nat && r.minimum < nat) {
        f = (nat - length) / (nat - r.minimum);
    } else {
        f = 0.0f;
    }
    return f;
}

// class LayoutTileReversed

LayoutTileReversed::LayoutTileReversed(Axis a) { axis_ = a; }
LayoutTileReversed::~LayoutTileReversed() { }

void LayoutTileReversed::request(
    long n, Glyph::Requisition* requests, Glyph::Requisition& result
) {
    LayoutTile::compute_request(axis_, Alignment(1), n, requests, result);
    requisition_ = result;
}

void LayoutTileReversed::allocate(
    long n, Glyph::Requisition* requests, Region_ptr given,
    LayoutManager::Allocations result
) {
    compute_reversed_allocations(
	axis_, requisition_, false, n, requests, given, result
    );
}

void LayoutTileReversed::compute_reversed_allocations(
    Axis a, Glyph::Requisition& total, Boolean first_aligned,
    long n, Glyph::Requisition* requests, Region_ptr given,
    LayoutManager::Allocations result
) {
    Glyph::Requirement* r;
    Region::BoundingSpan s;
    r = GlyphImpl::requirement(total, a);
    given->span(a, s);
    Coord length = LayoutTile::compute_length(*r, s);
    Boolean growing = length > r->natural;
    Boolean shrinking = length < r->natural;
    float f = LayoutTile::compute_squeeze(*r, length);
    Coord p = s.origin;
    for (long i = 0; i < n; i++) {
        r = GlyphImpl::requirement(requests[i], a);
        if (r->defined) {
            Coord cspan = r->natural;
            if (growing) {
                cspan += f * (r->maximum - r->natural);
            } else if (shrinking) {
                cspan -= f * (r->natural - r->minimum);
            }
            if (first_aligned && i == 0) {
                p += (1 - r->align) * cspan;
            }
            p -= cspan;
	    set_span(result[i], a, p + r->align * cspan, cspan, r->align);
        } else {
	    set_span(result[i], a, p, Coord(0), Coord(0));
        }
    }
}

// class LayoutTileFirstAligned

LayoutTileFirstAligned::LayoutTileFirstAligned(Axis a) { axis_ = a; }
LayoutTileFirstAligned::~LayoutTileFirstAligned() { }

void LayoutTileFirstAligned::request(
    long n, Glyph::Requisition* requests, Glyph::Requisition& result
) {
    compute_request_first_aligned(axis_, n, requests, result);
    requisition_ = result;
}

void LayoutTileFirstAligned::allocate(
    long n, Glyph::Requisition* requests, Region_ptr given,
    LayoutManager::Allocations result
) {
    LayoutTile::compute_allocations(
	axis_, requisition_, true, n, requests, given, result
    );
}

void LayoutTileFirstAligned::compute_request_first_aligned(
    Axis a,
    long n, Glyph::Requisition* requests, Glyph::Requisition& result
) {
    Glyph::Requirement* r;
    Coord natural_lead = Coord(0);
    Coord min_lead = Coord(0);
    Coord max_lead = Coord(0);
    Coord natural_trail = Coord(0);
    Coord min_trail = Coord(0);
    Coord max_trail = Coord(0);
    for (long i = 0; i < n; i++) {
        r = GlyphImpl::requirement(requests[i], a);
        if (r->defined) {
            if (i == 0) {
		Alignment a = r->align;
		Alignment aa = 1 - a;
                natural_lead = a * r->natural;
		max_lead = a * r->maximum;
                min_lead = a * r->minimum;
                natural_trail = aa * r->natural;
                max_trail = aa * r->maximum;
                min_trail = aa * r->minimum;
            } else {
                natural_trail += r->natural;
                max_trail += r->maximum;
                min_trail += r->minimum;
            }
        }
    }
    r = GlyphImpl::requirement(result, a);
    GlyphImpl::require_lead_trail(
	*r, natural_lead, max_lead, min_lead,
	natural_trail, max_trail, min_trail
    );
}

// class LayoutTileReversedFirstAligned

LayoutTileReversedFirstAligned::LayoutTileReversedFirstAligned(Axis a) {
    axis_ = a;
}

LayoutTileReversedFirstAligned::~LayoutTileReversedFirstAligned() { }

void LayoutTileReversedFirstAligned::request(
    long n, Glyph::Requisition* requests, Glyph::Requisition& result
) {
    LayoutTileFirstAligned::compute_request_first_aligned(
	axis_, n, requests, result
    );
    requisition_ = result;
}

void LayoutTileReversedFirstAligned::allocate(
    long n, Glyph::Requisition* requests, Region_ptr given,
    LayoutManager::Allocations result
) {
    LayoutTileReversed::compute_reversed_allocations(
	axis_, requisition_, true, n, requests, given, result
    );
}

// class LayoutVariable

LayoutVariable::LayoutVariable(Axis a, Coord stretch, Coord shrink) {
    axis_ = a;
    stretch_ = stretch;
    shrink_ = shrink;
}

LayoutVariable::~LayoutVariable() { }

void LayoutVariable::request(
    long, Glyph::Requisition*, Glyph::Requisition& result
) {
    Glyph::Requirement* r = GlyphImpl::requirement(result, axis_);
    r->defined = true;
    r->maximum = r->natural + stretch_;
    r->minimum = r->natural - shrink_;
}

void LayoutVariable::allocate(
    long, Glyph::Requisition*, Region_ptr, LayoutManager::Allocations
) {
    // leave it as is
}
