//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-93 Silicon Graphics, Inc.
// Copyright (c) 1993 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names
// of Stanford, Silicon Graphics, and Fujitsu may not be used in any
// advertising or publicity relating to the software without the specific,
// prior written permission of Stanford, Silicon Graphics, and Fujitsu.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD, SILICON GRAPHICS, OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#ifndef Fresco_layouts_impl_layout_mgr_h
#define Fresco_layouts_impl_layout_mgr_h

#include <Fresco/defs/glyph.h>

class RegionImpl;

//
// You would think this should be a const float, but C++ compilers
// don't typically do the right thing with such a declaration.
//
#define infinite_coord Coord(10e6)

//
// LayoutManager -- determine appropriate allocations for a set of requests
//
// Note LayoutManagers are _not_ general objects, and therefore cannot
// be shared or observed.
//

class LayoutManager {
public:
    typedef RegionImpl** Allocations;

    LayoutManager();
    virtual ~LayoutManager();

    virtual void request(
	long n, Glyph::Requisition* requests, Glyph::Requisition& result
    ) = 0;
    virtual void allocate(
	long n, Glyph::Requisition* requests, Region_ptr given,
	LayoutManager::Allocations result
    ) = 0;

    static void set_span(
	RegionImpl* r, Axis a, Coord origin, Coord length, Alignment align
    );
};

//
// LayoutAlign -- align positions along an axis
//

class LayoutAlign : public LayoutManager {
public:
    LayoutAlign(Axis);
    virtual ~LayoutAlign();

    virtual void request(
	long n, Glyph::Requisition* requests, Glyph::Requisition& result
    );
    virtual void allocate(
	long n, Glyph::Requisition* requests, Region_ptr given,
	LayoutManager::Allocations result
    );
private:
    Axis axis_;
};

//
// LayoutCenter -- center positions along an axis
//

class LayoutCenter : public LayoutManager {
public:
    LayoutCenter(Axis, Alignment a);
    virtual ~LayoutCenter();

    virtual void request(
	long n, Glyph::Requisition* requests, Glyph::Requisition& result
    );
    virtual void allocate(
	long n, Glyph::Requisition* requests, Region_ptr given,
	LayoutManager::Allocations result
    );
private:
    Axis axis_;
    Alignment alignment_;
};

//
// LayoutFixed -- set size along an axis
//

class LayoutFixed : public LayoutManager {
public:
    LayoutFixed(Axis, Coord size);
    virtual ~LayoutFixed();

    virtual void request(
	long n, Glyph::Requisition* requests, Glyph::Requisition& result
    );
    virtual void allocate(
	long n, Glyph::Requisition* requests, Region_ptr given,
	LayoutManager::Allocations result
    );
private:
    Axis axis_;
    Coord size_;
};

//
// LayoutVariable -- allow flexibility along an axis
//

class LayoutVariable : public LayoutManager {
public:
    LayoutVariable(Axis, Coord stretch, Coord shrink);
    virtual ~LayoutVariable();

    virtual void request(
	long n, Glyph::Requisition* requests, Glyph::Requisition& result
    );
    virtual void allocate(
	long n, Glyph::Requisition* requests, Region_ptr given,
	LayoutManager::Allocations result
    );
private:
    Axis axis_;
    Coord stretch_;
    Coord shrink_;
};

//
// LayoutNatural -- set the natural size along an axis
//

class LayoutNatural : public LayoutManager {
public:
    LayoutNatural(Axis, Coord natural);
    virtual ~LayoutNatural();

    virtual void request(
	long n, Glyph::Requisition* requests, Glyph::Requisition& result
    );
    virtual void allocate(
	long n, Glyph::Requisition* requests, Region_ptr given,
	LayoutManager::Allocations result
    );
private:
    Axis axis_;
    Coord natural_;
};

//
// LayoutMargin -- leave a margin around the sides
//

class LayoutMargin : public LayoutManager {
public:
    LayoutMargin(Coord margin);
    LayoutMargin(Coord hmargin, Coord vmargin);
    LayoutMargin(Coord lmargin, Coord rmargin, Coord bmargin, Coord tmargin);
    LayoutMargin(
        Coord lmargin, Coord lstretch, Coord lshrink,
        Coord rmargin, Coord rstretch, Coord rshrink,
        Coord bmargin, Coord bstretch, Coord bshrink,
        Coord tmargin, Coord tstretch, Coord tshrink
    );
    virtual ~LayoutMargin();

    virtual void request(
	long n, Glyph::Requisition* requests, Glyph::Requisition& result
    );
    virtual void allocate(
	long n, Glyph::Requisition* requests, Region_ptr given,
	LayoutManager::Allocations result
    );
private:
    Coord lnatural_, lstretch_, lshrink_;
    Coord rnatural_, rstretch_, rshrink_;
    Coord bnatural_, bstretch_, bshrink_;
    Coord tnatural_, tstretch_, tshrink_;
    Glyph::Requisition requisition_;

    void allocate_axis(
	Axis,
	Coord natural_lead, Coord stretch_lead, Coord shrink_lead,
	Coord natural_trail, Coord stretch_trail, Coord shrink_trail,
	LayoutManager::Allocations result
    );
    static Coord span(
	Coord span, Glyph::Requirement& total,
	Coord natural, Coord stretch, Coord shrink
    );
};

//
// LayoutSuperpose - composite layout manager
//

class LayoutSuperpose : public LayoutManager {
public:
    LayoutSuperpose(LayoutManager*, LayoutManager*);
    LayoutSuperpose(LayoutManager*, LayoutManager*, LayoutManager*);
    virtual ~LayoutSuperpose();

    virtual void request(
	long n, Glyph::Requisition* requests, Glyph::Requisition& result
    );
    virtual void allocate(
	long n, Glyph::Requisition* requests, Region_ptr given,
	LayoutManager::Allocations result
    );
private:
    LayoutManager* first_;
    LayoutManager* second_;
    LayoutManager* third_;
};

//
// LayoutTile -- side-by-side, first-to-last along an axis
//

class LayoutTile : public LayoutManager {
public:
    LayoutTile(Axis);
    virtual ~LayoutTile();

    virtual void request(
	long n, Glyph::Requisition* requests, Glyph::Requisition& result
    );
    virtual void allocate(
	long n, Glyph::Requisition* requests, Region_ptr given,
	LayoutManager::Allocations result
    );

    static void compute_request(
	Axis a, Alignment align,
	long n, Glyph::Requisition* requests, Glyph::Requisition& result
    );
    static void compute_allocations(
	Axis a, Glyph::Requisition& total, Boolean first_aligned,
	long n, Glyph::Requisition* requests, Region_ptr given,
	LayoutManager::Allocations result
    );
    static Coord compute_length(
	const Glyph::Requirement& r, const Region::BoundingSpan& span
    );
    static float compute_squeeze(const Glyph::Requirement& r, Coord length);
private:
    Axis axis_;
    Glyph::Requisition requisition_;
};

//
// LayoutTileReversed -- side-by-side, last-to-first
//

class LayoutTileReversed : public LayoutManager {
public:
    LayoutTileReversed(Axis);
    virtual ~LayoutTileReversed();

    virtual void request(
	long n, Glyph::Requisition* requests, Glyph::Requisition& result
    );
    virtual void allocate(
	long n, Glyph::Requisition* requests, Region_ptr given,
	LayoutManager::Allocations result
    );
    static void compute_reversed_allocations(
	Axis a, Glyph::Requisition& total, Boolean first_aligned,
	long n, Glyph::Requisition* requests, Region_ptr given,
	LayoutManager::Allocations result
    );
private:
    Axis axis_;
    Glyph::Requisition requisition_;
};

//
// LayoutTileFirstAligned -- like Tile but use first element's origin
//

class LayoutTileFirstAligned : public LayoutManager {
public:
    LayoutTileFirstAligned(Axis);
    virtual ~LayoutTileFirstAligned();

    virtual void request(
	long n, Glyph::Requisition* requests, Glyph::Requisition& result
    );
    virtual void allocate(
	long n, Glyph::Requisition* requests, Region_ptr given,
	LayoutManager::Allocations result
    );
    static void compute_request_first_aligned(
	Axis a,
	long n, Glyph::Requisition* requests, Glyph::Requisition& result
    );
private:
    Axis axis_;
    Glyph::Requisition requisition_;
};

//
// LayoutTileReversedFirstAligned -- like TileReversed
//	but use first element's origin
//

class LayoutTileReversedFirstAligned : public LayoutManager {
public:
    LayoutTileReversedFirstAligned(Axis);
    virtual ~LayoutTileReversedFirstAligned();

    virtual void request(
	long n, Glyph::Requisition* requests, Glyph::Requisition& result
    );
    virtual void allocate(
	long n, Glyph::Requisition* requests, Region_ptr given,
	LayoutManager::Allocations result
    );
private:
    Axis axis_;
    Glyph::Requisition requisition_;
};

#endif
