//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-93 Silicon Graphics, Inc.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Stanford and Silicon Graphics may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Stanford and Silicon Graphics.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD OR SILICON GRAPHICS BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#include <math.h>
#include <Fresco/defs/display.h>
#include <Fresco/defs/drawing.h>
#include <Fresco/defs_impl/transform.h>
#include "bevel.h"

Beveler::Beveler(
    Coord thickness, Alignment x, Alignment y, Boolean h, Boolean v
) {
    thickness_ = thickness;
    xalign_ = x;
    yalign_ = y;
    hmargin_ = h;
    vmargin_ = v;
    allocation_ = new RegionImpl;
}

Beveler::~Beveler() {
    CORBA::release(allocation_);
}

//+ Beveler(XfGlyph::request)
void Beveler::request(Glyph::Requisition& r) {
    MonoGlyph::request(r);
    if (hmargin_ || vmargin_) {
	Coord t = thickness_ + thickness_;
	if (hmargin_ && r.x.defined) {
	    r.x.natural += t;
	    r.x.maximum += t;
	    r.x.minimum += t;
	}
	if (vmargin_ && r.y.defined) {
	    r.y.natural += t;
	    r.y.maximum += t;
	    r.y.minimum += t;
	}
    }
}

//+ Beveler(XfGlyph::traverse)
void Beveler::traverse(XfGlyphTraversal_ptr t) {
    switch (t->op()) {
    case GlyphTraversal::draw:
	draw(t);
	break;
    case GlyphTraversal::pick_top:
    case GlyphTraversal::pick_any:
    case GlyphTraversal::pick_all:
	pick(t);
	break;
    default:
	break;
    }
    if (is_not_nil(offset_.child_)) {
	if (hmargin_ || vmargin_) {
	    Painter_var p = t->current_painter();
	    p->push_matrix();
	    Glyph::AllocationInfo a;
	    a.allocation = t->allocation();
	    child_allocate(a);
	    p->premultiply(a.transformation);
	    t->traverse_child(&offset_, a.allocation);
	    p->pop_matrix();
	} else {
	    MonoGlyph::traverse(t);
	}
    }
}

//+ Beveler(Glyph::extension)
void Beveler::extension(const Glyph::AllocationInfo& a, Region_ptr r) {
    if (is_not_nil(a.allocation)) {
        default_extension(a, r);
    } else {
        MonoGlyph::extension(a, r);
    }
}

void Beveler::child_allocate(Glyph::AllocationInfo& info) {
    Glyph::Requisition req;
    GlyphImpl::init_requisition(req);
    MonoGlyph::request(req);
    RegionImpl* a = allocation_;
    a->defined_ = true;
    Region::BoundingSpan s;

    if (is_nil(info.transformation)) {
	info.transformation = new TransformImpl;
    }
    if (is_nil(info.allocation)) {
	info.allocation = Region::_duplicate(a);
    } else {
	Vertex o;
	RegionImpl* r = new RegionImpl;
	r->copy(info.allocation);
	r->origin(o);
	r->lower_.x -= o.x; r->upper_.x -= o.x;
	r->lower_.y -= o.y; r->upper_.y -= o.y;
	r->lower_.z -= o.z; r->upper_.z -= o.z;
	info.transformation->translate(o);
	info.allocation = r;
    }
    Vertex delta;
    info.allocation->span(X_axis, s);
    allocate_span(req.x, s, hmargin_ ? thickness_ : 0, xalign_);
    a->lower_.x = -s.length * s.align;
    a->upper_.x = a->lower_.x + s.length;
    a->xalign_ = s.align;
    delta.x = s.begin - a->lower_.x;

    info.allocation->span(Y_axis, s);
    allocate_span(req.y, s, vmargin_ ? thickness_ : 0, yalign_);
    a->lower_.y = -s.length * s.align;
    a->upper_.y = a->lower_.y + s.length;
    a->yalign_ = s.align;
    delta.y = s.begin - a->lower_.y;
    delta.z = Coord(0);

    info.allocation->copy(a);
    info.transformation->translate(delta);
}

void Beveler::allocate_span(
    Glyph::Requirement& r, Region::BoundingSpan& s, Coord margin, Alignment a
) {
    s.length -= (margin + margin);
    Coord offset = margin;
    if (r.defined) {
	if (s.length > r.maximum) {
	    offset += a * (s.length - r.maximum);
	    s.length = r.maximum;
	}
    }
    s.origin += offset * (1 - s.align - s.align);
    s.begin = s.origin - s.align * s.length;
    s.end = s.begin + s.length;
}

void Beveler::set_color(Painter_ptr p, Color_ptr c) {
    p->current_color(c);
}

void Beveler::fill_path(Painter_ptr p, Vertex* v, long n) {
    p->begin_path();
    p->move_to(v[0].x, v[0].y);
    for (long i = 1; i < n; i++) {
	p->line_to(v[i].x, v[i].y);
    }
    p->close_path();
    p->fill();
}

void Beveler::rect(
    Painter_ptr p, Coord thickness,
    Color_ptr light, Color_ptr medium, Color_ptr dark,
    Coord left, Coord bottom, Coord right, Coord top
) {
    if (is_not_nil(medium)) {
	set_color(p, medium);
	p->fill_rect(left, bottom, right, top);
    }

    Coord left_inside = left + thickness;
    Coord bottom_inside = bottom + thickness;
    Coord right_inside = right - thickness;
    Coord top_inside = top - thickness;

    Vertex v[10];

    // left edge
    set_color(p, light);
    v[0].x = left; v[0].y = bottom;
    v[1].x = left; v[1].y = top;
    v[2].x = left_inside; v[2].y = top;
    v[3].x = left_inside; v[3].y = bottom;
    fill_path(p, v, 4);

    // top edge
    v[0].x = left_inside; v[0].y = top_inside;
    v[1].x = left_inside; v[1].y = top;
    v[2].x = right; v[2].y = top;
    v[3].x = right; v[3].y = top_inside;
    fill_path(p, v, 4);

    // right edge
    set_color(p, dark);
    v[0].x = right_inside; v[0].y = bottom_inside;
    v[1].x = right_inside; v[1].y = top_inside;
    v[2].x = right; v[2].y = top;
    v[3].x = right; v[3].y = bottom;

    // bottom edge (as part of same path)
    v[4].x = left; v[4].y = bottom;
    v[5].x = left_inside; v[5].y = bottom_inside;
    v[6].x = right_inside; v[6].y = bottom_inside;
    fill_path(p, v, 7);
}

//
// A bit of a misnomer to call these arrows; they are really beveled
// triangles.  The only tricky part is dropping the bevel down and to the
// right of the original triangle, which requires finding delta x,y for
// a given thickness and matching the slope of the triangle.
//

void Beveler::left_arrow(
    Painter_ptr p, Coord thickness,
    Color_ptr light, Color_ptr medium, Color_ptr dark,
    Coord left, Coord bottom, Coord right, Coord top
) {
    Coord center_y = (bottom + top) * Coord(0.5);
    float slope = ((top - bottom) / (right - left)) * Coord(0.5);
    float delta_x = float(thickness / sqrt(slope * slope + 1.0f));
    float delta_y = slope * delta_x;

    Vertex v[10];

    set_color(p, medium);
    v[0].x = left; v[0].y = center_y;
    v[1].x = right; v[1].y = top;
    v[2].x = right; v[2].y = bottom;
    fill_path(p, v, 3);

    set_color(p, dark);
    v[0].x = right - thickness; v[0].y = bottom + thickness;
    v[1].x = v[0].x; v[1].y = top - thickness;
    v[2].x = right; v[2].y = top;
    v[3].x = right; v[3].y = bottom;
    fill_path(p, v, 4);

    v[0].x = left; v[1].y = center_y;
    v[1].x = left + thickness; v[1].y = center_y;
    v[2].x = right - thickness; v[2].y = bottom + thickness;
    v[3].x = right; v[3].y = bottom;
    fill_path(p, v, 4);

    set_color(p, light);
    v[0].x = left + delta_x; v[0].y = center_y - delta_y;
    v[1].x = left; v[1].y = center_y;
    v[2].x = right; v[2].y = top;
    v[3].x = right; v[3].y = top - thickness;
    fill_path(p, v, 4);
}

void Beveler::right_arrow(
    Painter_ptr p, Coord thickness,
    Color_ptr light, Color_ptr medium, Color_ptr dark,
    Coord left, Coord bottom, Coord right, Coord top
) {
    Coord center_y = (bottom + top) * Coord(0.5);
    float slope = ((top - bottom) / (right - left)) * Coord(0.5);
    float delta_x = float(thickness / sqrt(slope * slope + 1.0f));
    float delta_y = slope * delta_x;

    Vertex v[10];

    set_color(p, medium);
    v[0].x = left; v[0].y = bottom;
    v[1].x = left; v[1].y = top;
    v[2].x = right; v[2].y = center_y;
    fill_path(p, v, 3);

    set_color(p, dark);
    v[0].x = left; v[0].y = bottom;
    v[1].x = left + thickness; v[1].y = bottom + thickness;
    v[2].x = right - thickness; v[2].y = center_y;
    v[3].x = right; v[3].y = center_y;
    fill_path(p, v, 4);

    set_color(p, light);
    v[0].x = left; v[0].y = bottom;
    v[1].x = left; v[1].y = top;
    v[2].x = left + thickness; v[2].y = top - thickness;
    v[3].x = left + delta_x; v[3].y = bottom + delta_y;
    fill_path(p, v, 4);

    v[0].x = left + thickness; v[0].y = top - thickness;
    v[1].x = left; v[1].y = top;
    v[2].x = right; v[2].y = center_y;
    v[3].x = right - delta_x; v[3].y = center_y - delta_y;
    fill_path(p, v, 4);
}

void Beveler::up_arrow(
    Painter_ptr p, Coord thickness,
    Color_ptr light, Color_ptr medium, Color_ptr dark,
    Coord left, Coord bottom, Coord right, Coord top
) {
    Coord center_x = (left + right) * Coord(0.5);
    float slope = Coord(2) * ((top - bottom) / (right - left));
    float delta_x = float(thickness / sqrt(slope * slope + 1.0f));
    float delta_y = slope * delta_x;

    Vertex v[10];

    set_color(p, medium);
    v[0].x = left; v[0].y = bottom;
    v[1].x = center_x; v[1].y = top;
    v[2].x = right; v[2].y = bottom;
    fill_path(p, v, 3);

    set_color(p, dark);
    v[0].x = left; v[0].y = bottom;
    v[1].x = right; v[1].y = bottom;
    v[2].x = right - thickness; v[2].y = bottom + thickness;
    v[3].x = left + thickness; v[3].y = v[2].y;
    fill_path(p, v, 4);

    v[0].x = center_x; v[0].y = top - thickness;
    v[1].x = center_x; v[1].y = top;
    v[2].x = right; v[2].y = bottom;
    v[3].x = right - thickness; v[3].y = bottom + thickness;
    fill_path(p, v, 4);

    set_color(p, light);
    v[0].x = left; v[0].y = bottom;
    v[1].x = left + delta_x; v[1].y = bottom;
    v[2].x = center_x + delta_x; v[2].y = top - delta_y;
    v[3].x = center_x; v[3].y = top;
    fill_path(p, v, 4);
}

void Beveler::down_arrow(
    Painter_ptr p, Coord thickness,
    Color_ptr light, Color_ptr medium, Color_ptr dark,
    Coord left, Coord bottom, Coord right, Coord top
) {
    Coord center_x = (left + right) * Coord(0.5);
    float slope = Coord(2) * ((top - bottom) / (right - left));
    float delta_x = float(thickness / sqrt(slope * slope + 1.0f));
    float delta_y = slope * delta_x;

    Vertex v[10];

    set_color(p, medium);
    v[0].x = left; v[0].y = top;
    v[1].x = right; v[1].y = top;
    v[2].x = center_x; v[2].y = bottom;
    fill_path(p, v, 3);

    set_color(p, dark);
    v[0].x = center_x; v[0].y = bottom;
    v[1].x = center_x; v[1].y = bottom + thickness;
    v[2].x = right - thickness; v[2].y = top - thickness;
    v[3].x = right; v[3].y = top;
    fill_path(p, v, 4);

    set_color(p, light);
    v[0].x = left; v[0].y = top;
    v[1].x = center_x; v[1].y = bottom;
    v[2].x = center_x + delta_x; v[2].y = bottom + delta_y;
    v[3].x = left + thickness; v[3].y = top - thickness;
    fill_path(p, v, 4);

    v[0].x = left; v[0].y = top;
    v[1].x = right; v[1].y = top;
    v[2].x = right - delta_x; v[2].y = top - delta_y;
    v[3].x = left + thickness; v[3].y = top - thickness;
    fill_path(p, v, 4);
}

void Beveler::diamond(
    Painter_ptr p, Coord thickness,
    Color_ptr light, Color_ptr medium, Color_ptr dark,
    Coord left, Coord bottom, Coord right, Coord top
) {
    Coord x_mid = (left + right) * Coord(0.5);
    Coord y_mid = (bottom + top) * Coord(0.5);
    Coord left_inside = left + thickness;
    Coord top_inside = top - thickness;
    Coord right_inside = right - thickness;
    Coord bottom_inside = bottom + thickness;

    Vertex v[10];

    // interior of diamond
    set_color(p, medium);
    v[0].x = left; v[0].y = y_mid;
    v[1].x = x_mid; v[1].y = top;
    v[2].x = right; v[2].y = y_mid;
    v[3].x = x_mid; v[3].y = bottom;
    fill_path(p, v, 4);

    // lower half
    set_color(p, dark);
    v[0].x = left; v[0].y = y_mid;
    v[1].x = x_mid; v[1].y = bottom;
    v[2].x = right; v[2].y = y_mid;
    v[3].x = right_inside; v[3].y = y_mid;
    v[4].x = x_mid; v[4].y = bottom_inside;
    v[5].x = left_inside; v[5].y = y_mid;
    fill_path(p, v, 6);

    // upper half
    set_color(p, light);
    v[0].x = left; v[0].y = y_mid;
    v[1].x = x_mid; v[1].y = top;
    v[2].x = right; v[2].y = y_mid;
    v[3].x = right_inside; v[3].y = y_mid;
    v[4].x = x_mid; v[4].y = top_inside;
    v[5].x = left_inside; v[5].y = y_mid;
    fill_path(p, v, 6);
}
