/*
 * Copyright (c) 1994-95 Firefly Design
 * Copyright (c) 1994 Silicon Graphics, Inc.
 * Copyright (c) 1994 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THIS SOFTWARE.
 */

#include "box.h"
#include <Fresco/defs/fresco.h>
#include <Fresco/defs_impl/glyphs.h>
#include <Fresco/layouts/layouts.h>

Fresco_file_id(Fresco_widgets_impl_box_cxx)

// class MKPicker

class MKPicker : public MonoGlyph {
public:
    MKPicker(XfGlyph_ptr g, Long info);

    void traverse(XfGlyphTraversal_ptr t); //+ XfGlyph::traverse

protected:
    Long info_;
};

MKPicker::MKPicker(XfGlyph_ptr g, Long info) {
    body(g);
    info_ = info;
}

void MKPicker::traverse(XfGlyphTraversal_ptr t) {
    if (t->op() == GlyphTraversal::pick_top) {
        Painter_var painter = t->current_painter();
        if (painter->is_visible(Region_var(t->allocation()))) {
            t->hit_info(info_);
            t->hit();
        }
    } else {
        MonoGlyph::traverse(t);
    }
}

// class MKBoxInfo

MKBoxInfo::MKBoxInfo(MenuItem_ptr item, Glyph_ptr glyph, Telltale_ptr telltale) {
    item_ = MenuItem::_duplicate(item);
    glyph_ = Glyph::_duplicate(glyph);
    telltale_ = Telltale::_duplicate(telltale);
    tag_ = item_->attach(this);
}

MKBoxInfo::~MKBoxInfo() {
    item_->detach(tag_);
}

OxExchange* MKBoxInfo::_exchange() { return &object_; }

//+ MKBoxInfo(XfFrescoObject::attach)
XfTag MKBoxInfo::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}

//+ MKBoxInfo(XfFrescoObject::detach)
void MKBoxInfo::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}

//+ MKBoxInfo(XfFrescoObject::disconnect)
void MKBoxInfo::disconnect() {
    object_.disconnect();
}

//+ MKBoxInfo(XfFrescoObject::notify_observers)
void MKBoxInfo::notify_observers() {
    object_.notify_observers();
}

void MKBoxInfo::update() {
    if (item_->enabled()) {
        telltale_->set(Telltale::enabled);
    } else {
        telltale_->clear(Telltale::enabled);
    }
    // !!! the hilite glyph should do this -denis
    glyph_->need_redraw();
}

// class MKBox

implementPtrList(MKBoxInfoList,MKBoxInfo)

MKBox::MKBox(MenuKitImpl::Info* info) {
    info_ = info;
    selection_ = nil;
    manager_ = nil;
    grab_tag_ = 0;
    grab_display_ = nil;
}

MKBox::~MKBox() {
    for (ListItr(MKBoxInfoList) g(children_); g.more(); g.next()) {
        CORBA::release(g.cur());
    }
}

//+ MKBox(XfViewer::handle)
Boolean MKBox::handle(XfGlyphTraversal_ptr t, XfEvent_ptr e) {
    Boolean b = (e->type() == Event::down) || e->button_is_down(1);
    if (b) {
        GlyphTraversal_var picker = t->trail();
        picker->clear();
        picker->traverse_child(&offset_, Region_var(picker->allocation()));
        GlyphTraversal_var picked = picker->picked();
        if (is_not_nil(picked)) {
            Long index = picker->hit_info();
            grab(t);
            MenuItem_ptr item = children_.item(index)->item_;
            if (e->type() == Event::up) {
                ungrab();
                if (item->type() == MenuItem::submenu) {
                    manager_->select(MenuKitImpl::first_item(item), false);
                } else {
                    manager_->select(item, false);
                    manager_->activate();
                }
            } else {
                manager_->select(item, true);
            }
        } else {
            if (grab_tag_ != 0) {
                pop_selection();
                ungrab();
                b = false;
            }
        }
    }
    return b;
}

void MKBox::grab(GlyphTraversal_ptr t) {
    if (grab_tag_ == 0) {
        grab_display_ = t->current_display();
        grab_tag_ = grab_display_->add_filter(
            EventFilter_var(t->event_filter())
        );
        manager_->open_menus(Window_var(t->current_window()));
    }
}

void MKBox::ungrab() {
    if (grab_tag_ != 0) {
        grab_display_->remove_filter(grab_tag_);
        CORBA::release(grab_display_);
        grab_tag_ = 0;
        grab_display_ = nil;
    }
}

void MKBox::pop_selection() {
    MenuItem_ptr selection = manager_->selection();
    if (selection->type() == MenuItem::leaf) {
        MenuItem_ptr root = manager_->root();
        MenuItem_ptr parent = selection->parent();
        if (parent != root) {
            manager_->select(parent, true);
        } else {
            manager_->unselect();
        }
        CORBA::release(root);
    }
    CORBA::release(selection);
}

void MKBox::menu_manager(MenuManager_ptr m) {
    manager_ = m;
}

const MKBoxInfoList& MKBox::children() {
    return children_;
}

void MKBox::select(const MenuItem::Path& path, Boolean flap) {
    if (path.length() > 0) {
        MenuItem_ptr item = path[0];

        if (is_not_nil(selection_) && (selection_->item_ != item)) {
            unselect_item();
        }

        if (is_nil(selection_)) {
            select_item(item);
        }

        if ((path.length() > 1) || flap) {
            Long length = path.length() - 1;
            MenuItem::Path p(length, length);
            for (unsigned long i = 1; i < path.length(); ++i) {
                p[i - 1] = MenuItem::_duplicate(path[i]);
            }
            open_selection(p, flap);
        } else {
            close_selection();
        }
    } else {
        unselect_item();
    }
}

void MKBox::selection(MenuItem::Path& path) {
    if (is_not_nil(selection_)) {
        MenuKitImpl::path_append(path, selection_->item_);
        if (is_not_nil(selection_->menu_)) {
            selection_->menu_->selection(path);
        }
    }
}

void MKBox::unselect_item() {
    if (is_not_nil(selection_)) {
        close_selection();
        selection_->telltale_->clear(Telltale::visible);
        selection_->glyph_->need_redraw(); // !!! shouldn't need this -denis
        selection_ = nil;
        ungrab();
    }
}

void MKBox::select_item(MenuItem_ptr item) {
    for (ListItr(MKBoxInfoList) g(children_); g.more(); g.next()) {
        if (g.cur()->item_ == item) {
            break;
        }
    }

    selection_ = g.cur();
    selection_->telltale_->set(Telltale::visible);
    selection_->glyph_->need_redraw(); // !!! shouldn't need this -denis
}

void MKBox::open_selection(const MenuItem::Path& p, Boolean flap) {
    if (selection_->item_->type() == MenuItem::submenu) {
        if (is_nil(selection_->menu_)) {
            selection_->menu_ = create_popup_menu(selection_->item_);
            selection_->menu_->menu_manager(manager_);
        }
        RegionImpl a;
        selection_allocation(a);
        selection_->menu_->show(&a);
        selection_->menu_->select(p, flap);
    }
}

void MKBox::close_selection() {
    if (is_not_nil(selection_->menu_)) {
        MenuItem::Path p;
        selection_->menu_->select(p, false);
        selection_->menu_->hide();
    }
}

void MKBox::selection_allocation(RegionImpl& a) {
    Glyph::AllocationInfoSeq s;
    selection_->glyph_->allocations(s);
    // !!! assumes the box is at only one point in the glyph DAG -denis
    if (s.length() > 0) {
        Glyph::AllocationInfo& i = s[0];
        a.copy(i.allocation);
        a.apply_transform(i.transformation);
        Window::Placement p;
        Window_var window = i.damaged->window_ptr();
        window->get_configuration(true, p);
        Coord dx = p.x - p.width * p.align_x;
        Coord dy = p.y - p.height * p.align_y;
        a.lower_.x += dx;
        a.upper_.x += dx;
        a.lower_.y += dy;
        a.upper_.y += dy;
    } else {
        Screen_var screen = manager_->filter_screen();
        a.defined_ = true;
        a.lower_.x = a.upper_.x = screen->width() / Coord(2);
        a.lower_.y = a.upper_.y = screen->height() / Coord(2);
        a.xalign_ = a.yalign_ = Alignment(0);
    }
}

// class MKHBox
 
MKHBox::MKHBox(MenuKitImpl::Info* info, MenuItem_ptr menu) : MKBox(info) {
    LayoutKit_ptr layout = info_->layout;
    WidgetKit_ptr widget = info_->widget;
 
    Glyph_var hbox = layout->hbox();
    Long count = menu->count();
    for (Long i = 0; i < count; ++i) {
        MenuItem_ptr item = menu->item(i);

        switch (item->type()) {
        case MenuItem::submenu:
        case MenuItem::leaf: {
            Telltale_var telltale = widget->telltale_group();
            if (item->enabled()) {
                telltale->set(Telltale::enabled);
            }
            Glyph_var line = layout->hbox();
            Glyph_var mark = item->zone_glyph(MenuItem::mark);
            if (is_not_nil(mark)) {
                line->append(mark);
            }
            Glyph_var label = item->zone_glyph(MenuItem::label);
            if (is_not_nil(label)) {
                line->append(label);
            }
            Glyph_var shortcut = item->zone_glyph(MenuItem::shortcut);
            if (is_not_nil(shortcut)) {
                line->append(shortcut);
            }
            Glyph_var glyph = new MKPicker(
                Glyph_var(widget->hilite_frame(line, telltale)), children_.count()
            );
            children_.append(new MKBoxInfo(item, glyph, telltale));
            hbox->append(glyph);
        } break;

        case MenuItem::separator:
            hbox->append(info_->menu_glue);
        }
    }
    body(hbox);
}
 
PopupMenu_ptr MKHBox::create_popup_menu(MenuItem_ptr item) {
    return item->pulldown_menu();
}
 
// class MKVBox
 
MKVBox::MKVBox(MenuKitImpl::Info* info, MenuItem_ptr menu) : MKBox(info) {
    LayoutKit_ptr layout = info_->layout;
    WidgetKit_ptr widget = info_->widget;
 
    Glyph_var vbox = layout->vbox();
    Long count = menu->count();
    Grid::Index upper;
    upper.col = 3;
    upper.row = count;
    Grid_var grid = layout->fixed_grid(upper);
    for (Long i = 0; i < count; ++i) {
        MenuItem_ptr item = menu->item(i);
 
        switch (item->type()) {
        case MenuItem::submenu:
        case MenuItem::leaf: {
            Grid::Index index;
            index.row = i;
            Telltale_var telltale = widget->telltale_group();
            if (item->enabled()) {
                telltale->set(Telltale::enabled);
            }
            index.col = 0;
            grid->replace(Glyph_var(item->zone_glyph(MenuItem::mark)), index);
            index.col = 1;
            grid->replace(Glyph_var(item->zone_glyph(MenuItem::label)), index);
            Glyph_var shortcut;
            if (item->type() == MenuItem::submenu) {
                shortcut = Glyph::_duplicate(info_->menu_shortcut);
            } else {
                shortcut = item->zone_glyph(MenuItem::shortcut);
            }
            index.col = 2;
            grid->replace(shortcut, index);
            Grid::Range range;
            range.lower.col = 0;
            range.upper.col = 3;
            range.lower.row = i;
            range.upper.row = i + 1;
            Glyph_var line = layout->fixed_range(grid, range);
            Glyph_var glyph = new MKPicker(
                Glyph_var(widget->hilite_frame(line, telltale)), children_.count()
            );
            children_.append(new MKBoxInfo(item, glyph, telltale));
            vbox->append(glyph);
        } break;

        case MenuItem::separator:
            vbox->append(info_->menu_separator);
        }
    }
    body(vbox);
}
 
PopupMenu_ptr MKVBox::create_popup_menu(MenuItem_ptr item) {
    return item->pullright_menu();
}
