//
// Copyright (c) 1994 Silicon Graphics, Inc.
// Copyright (c) 1994 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Silicon Graphics and Fujitsu may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Silicon Graphics and Fujitsu.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#include <stdlib.h>
#include <assert.h>

#include <ctype.h>
#include <string.h>
#include <X11/keysym.h>
#include <Fresco/defs/display.h>
#include <Fresco/defs/drawing.h>
#include <Fresco/defs/fresco.h>
#include <Fresco/defs_impl/action.h>
#include <Fresco/defs_impl/polyglyph.h>
#include <Fresco/defs_impl/charstr.h>
#include <Fresco/defs_impl/math.h>
#include <Fresco/defs_impl/region.h>
#include <Fresco/defs_impl/threads.h>
#include <Fresco/defs_impl/transform.h>
#include <Fresco/defs_impl/names.h>
#include <Fresco/widgets/menus.h>
#include "bfloat.h"
#include "field.h"
#include "kit.h"

// static const char* SEDone = "\r\t\007\033";
static const CharCode SEBeginningOfLine = '\001';
static const CharCode SEEndOfLine = '\005';
static const CharCode SESelectAll = '\025';
static const CharCode SESelectWord = '\027';
static const CharCode SEPreviousCharacter = '\002';
static const CharCode SENextCharacter = '\006';
static const CharCode SEDeleteNextCharacter = '\004';
static const CharCode SEDeletePreviousCharacter = '\177';
static const CharCode SEDeletePreviousCharacterAlt = '\010';
static const CharCode SEESCCode = '\033';
//static const CharCode SERETURN  = '\n';
static const CharCode SERETURN  = '\015';


// Temp actions 

class CopyAction : public ActionImpl {
public:
    CopyAction() {}
    void execute() { /* unimplemented */ }
};

class CutAction : public ActionImpl {
public:
    CutAction() {}
    void execute() { /* unimplemented */ }
};

class PasteAction : public ActionImpl {
public:
    PasteAction() {}
    void execute() { /* unimplemented */ }
};





// class UpdateCallback(T)

/* Since a given observer cannot attach itself to more than one object,
 * a general mechanism is needed to define a "proxy" object that
 * will attach itself on behalf of the observer. When update() is
 * called on the proxy, it makes a callback to the given observer.
 */
  
#if !defined(UNIXCPP)
#define __UpdateCallback(T) T##_UpdateCallback
#define UpdateCallback(T) __UpdateCallback(T)
#else
#define __UpdateCallback(T) T/**/_UpdateCallback
#define UpdateCallback(T) __UpdateCallback(T)
#endif

#define declareUpdateCallback(T) \
class UpdateCallback(T) : public ActionCallback(T) { \
public: \
    UpdateCallback(T)( \
        T*, ActionMemberFunction(T), FrescoObject_ptr attach_to \
    ); \
    virtual ~UpdateCallback(T)(); \
\
    void update(); \
private: \
    FrescoObject_ptr attached_to_; \
    Tag tag_; \
};

#define implementUpdateCallback(T) \
UpdateCallback(T)::UpdateCallback(T)( \
    T* obj, ActionMemberFunction(T) func, FrescoObject_ptr attach_to \
) : ActionCallback(T)(obj, func) { \
    attached_to_ = attach_to; \
    tag_ = attached_to_->attach(this); \
} \
UpdateCallback(T)::~UpdateCallback(T)() { \
    attached_to_->detach(tag_); \
} \
\
void UpdateCallback(T)::update() { execute(); }


// class BoundedField

class BoundedField : public BoundedFloat {
public:
    BoundedField(
	Coord lower, Coord upper, Coord visible_lower, Coord visible_upper
    );
    ~BoundedField();	
    void constrain(Coord& c);
    virtual void span(float);
    virtual void visible_span(float);
protected:
    //
    // This is not returned as cur_length by get_settings, because
    // doing so makes slider (on scrollbar) shoot past right-button's
    // left edge.
    //
    float visible_span_;
};

BoundedField::BoundedField(
    Coord lower, Coord upper, Coord visible_lower, Coord visible_upper
) : BoundedFloat(lower, upper, visible_lower) {
    visible_span_ = visible_upper-visible_lower;
}

BoundedField::~BoundedField() { }
    
void BoundedField::constrain(Coord& c) {
    if (span_ <= visible_span_) {
	c = Coord(0);
    } else {
	BoundedFloat::constrain(c);
    }
}

void BoundedField::span(float s) {
    span_ = s;
    scroll_incr_ = span_ * Coord(0.04);
    page_incr_ = span_ * Coord(0.2);    
    notify_observers();
}

void BoundedField::visible_span(float s) {
    visible_span_ = s;
}

// class Character

class Character : _Ix_public FieldChar, _Ix_public GlyphImpl {
public:
    Character(CharCode, Font_ptr, Color_ptr fg, Color_ptr bg);
    ~Character();

    CharCode code();
    void highlight(Boolean);
    
    /* Glyph */
    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
    void draw(XfGlyphTraversal_ptr t); //+ XfGlyph::draw
    void pick(XfGlyphTraversal_ptr t); //+ XfGlyph::pick
    void extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r); //+ XfGlyph::extension

protected:
    CharCode c_;
    Font_var font_;
    Color_var foreground_;
    Color_var background_;
    Color_ptr color_;
    Boolean highlighted_;
};

Character::Character(CharCode c, Font_ptr f, Color_ptr fg, Color_ptr bg) {
    c_ = c;
    font_ = Font::_duplicate(f);
    foreground_ = Color::_duplicate(fg);
    background_ = Color::_duplicate(bg);
    color_ = fg;
    highlighted_ = false;
}

Character::~Character() { }

CharCode Character::code() { return c_; }

void Character::highlight(Boolean on) {
    if (on != highlighted_) {
	if (on) {
	    color_ = background_;
	} else {
	    color_ = foreground_;
	}
	highlighted_ = on;
	need_redraw();
    }
}

//+ Character(XfGlyph::request)
void Character::request(XfGlyph::Requisition& r) {
    Font::Info info;
    font_->char_info(c_, info);
    Coord h = info.font_ascent + info.font_descent;
    Alignment ay = Alignment(0);
    if (!Math::equal(h, float(0), float(1e-2))) {
	ay = info.font_descent / h;
    }
    GlyphImpl::init_requisition(r);
    GlyphImpl::require(r.x, info.width, Coord(0), Coord(0), Alignment(0));
    GlyphImpl::require(r.y, h, Coord(0), Coord(0), ay);
}

//+ Character(XfGlyph::draw)
void Character::draw(XfGlyphTraversal_ptr t) {
    Painter_var p = t->current_painter();
    Vertex lower, upper, origin;
    t->bounds(lower, upper, origin);
    p->current_color(color_);
    p->current_font(font_);
    Font::Info info;
    font_->char_info(c_, info);
    p->character(c_, info.width, origin.x, origin.y);
}

//+ Character(XfGlyph::pick)
void Character::pick(XfGlyphTraversal_ptr t) {
    Painter_var p = t->current_painter();
    TransformImpl tx;
    tx.copy(Transform_var(p->current_matrix()));
    tx.invert();
    RegionImpl r(Region_var(p->visible()), &tx, nil);
    Region::BoundingSpan s;
    r.span(X_axis, s);
    Coord x = s.begin;
    Region_var a = t->allocation();
    a->span(X_axis, s);
    if (x >= s.begin && x <= s.end) {
	Coord length = s.end - s.begin;
	t->hit_info(x - s.begin > length * 0.5 ); // 0=left, 1=right side
	t->hit();
    }
}

//+ Character(XfGlyph::extension)
void Character::extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r) {
    Vertex v;
    v.x = Coord(0);
    v.y = Coord(0);
    v.z = Coord(0);
    if (is_not_nil(a.allocation)) {
	a.allocation->origin(v);
    }
    Font::Info info;
    font_->char_info(c_, info);
    RegionImpl region;
    region.defined_ = true;
    region.lower_.x = v.x - info.left_bearing;
    region.upper_.x = v.x + info.right_bearing;
    region.lower_.y = v.y - info.font_descent;
    region.upper_.y = v.y + info.font_ascent;
    if (is_not_nil(a.transformation)) {
	region.apply_transform(a.transformation);
    }
    r->merge_union(&region);
}

// class Eoln

class Eoln : public GlyphImpl {
public:
    Eoln(Font_ptr);
    ~Eoln();

    /* Glyph */
    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
    void pick(XfGlyphTraversal_ptr t); //+ XfGlyph::pick
protected:
    CharCode c_;
    Font_var font_;
};

Eoln::Eoln(Font_ptr f) {
    font_ = f;
    CharString_var str = FrescoLib::string_ref(" ");
    CharCode* buf = new CharCode;
    CharString::Data data(1, 1, buf);
    str->get_data(data);  
    c_ = *buf;
}

Eoln::~Eoln() { }

void Eoln::request(Glyph::Requisition& r) {
    Font::Info info;
    font_->char_info(' ', info);
    Coord h = info.font_ascent + info.font_descent;
    Alignment ay = Alignment(0);
    if (!Math::equal(h, float(0), float(1e-2))) {
	ay = info.font_descent / h;
    }
    GlyphImpl::init_requisition(r);
    GlyphImpl::require(r.x, Coord(0), Coord(0), Coord(0), Alignment(0));
    GlyphImpl::require(r.y, h, Coord(0), Coord(0), ay);
}

void Eoln::pick(GlyphTraversal_ptr t) {
    // assumes clip region is point.
    Painter_var p = t->current_painter();
    TransformImpl tx;
    tx.copy(Transform_var(p->current_matrix()));
    tx.invert();
    RegionImpl r(Region_var(p->visible()), &tx, nil);
    Region::BoundingSpan s;
    r.span(X_axis, s);
    Coord x = s.begin;
    Region_var a = t->allocation();
    a->span(X_axis, s);
    if (x >= s.begin) {
	t->hit_info(0); // 0=left
	t->hit();
    }
}

/* Clipper */

class Clipper : public MonoGlyph {
public:
    Clipper(Glyph_ptr g);
    ~Clipper();
    
    void traverse(GlyphTraversal_ptr t);
};

Clipper::Clipper(Glyph_ptr g) {
    body(g);
}

Clipper::~Clipper() { }

void Clipper::traverse(GlyphTraversal_ptr t) {
    switch (t->op()) {
    case GlyphTraversal::draw:
	if (t->allocation_is_visible()) {
	    Painter_var p = t->current_painter();
	    Vertex v0, v1, v;
	    t->bounds(v0, v1, v);
	    p->push_clipping();
	    p->clip_rect(v0.x, v0.y, v1.x, v1.y);
	    MonoGlyph::traverse(t);
	    p->pop_clipping();
	}
	break;
    default:
	MonoGlyph::traverse(t);
	break;
    }
}

declareList(PositionList,Coord)
implementList(PositionList,Coord)

//- CharBox
class CharBox : public PolyGlyph {
    //. Simple box assumes children regard only origin/span during traversal.
public:
    CharBox(Coord width, Coord height, Alignment valign);
    ~CharBox();
    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
    void traverse(XfGlyphTraversal_ptr t); //+ XfGlyph::traverse
    void child_allocate(Long index, Glyph::AllocationInfo& a);
    void modified();

    Coord total_char_width();
    Coord scroll_pos;
protected:
    void update_positions();
    Coord width_, height_;
    Alignment valign_;
    Boolean modified_;
    PositionList positions_;
};

CharBox::CharBox(Coord width, Coord height, Alignment valign) {
    scroll_pos = Coord(0);
    width_ = width;
    height_ = height;
    valign_ = valign;
    modified_ = false;
}
CharBox::~CharBox() { }

Coord CharBox::total_char_width() {
    return positions_.item(children_.count()-1);
}

void CharBox::update_positions() {
    if (modified_) {
	positions_.remove_all();
	positions_.append(Coord(0));
	Long n = children_.count();
	for (Long i = 0; i < n; i++) {
	    Glyph::Requisition r;
	    PolyGlyphOffset* p = children_.item(i);
	    Glyph_var(p->child())->request(r);
	    positions_.append(r.x.natural + positions_.item(i));
	}
	modified_ = false;
    }
}

void CharBox::modified() {
    PolyGlyph::modified();
    modified_ = true;
}

//+ CharBox(XfGlyph::request)
void CharBox::request(XfGlyph::Requisition& r) {
    GlyphImpl::init_requisition(r);
    Coord hfil = Coord(10e6);
    GlyphImpl::require(r.x, width_, hfil, Coord(0), Alignment(0)); 
    GlyphImpl::require(r.y, height_, Coord(0), Coord(0), valign_);
}

//+ CharBox(XfGlyph::traverse)
void CharBox::traverse(XfGlyphTraversal_ptr t) {
    Long i_start, i_stop, i_incr;
    Boolean i_picking;
    Long count = children_.count();
    GlyphTraversal::Operation op = t->op();
    switch (op) {
    case GlyphTraversal::pick_top:
    case GlyphTraversal::pick_any:
	i_picking = true;
	i_start = count - 1;
	i_stop = -1;
	i_incr = -1;
	break;
    default:
	i_picking = false;
	i_start = 0;
	i_stop = count;
	i_incr = +1;
	break;
    }
    update_positions();
    RegionImpl* r = new RegionImpl;
    r->copy(Region_var(t->allocation()));
    r->lower_.x += scroll_pos;
    Coord span = r->upper_.x - r->lower_.x;
    for (Long i = i_start; i != i_stop; i += i_incr) {
	Coord left = -scroll_pos;
	Coord right = left + span;
	Coord char_left = positions_.item(i);
	Coord char_right = (i < count-1)?
	    positions_.item(i+1) : positions_.item(count-1);
	Coord span = char_right - char_left;
	Coord dx;
	if (i == i_start) {
	    r->lower_.x += char_left;
	} else {
	    dx = i_picking? -span : span;
	    r->lower_.x += dx;
	}
	r->upper_.x = r->lower_.x + span;
	if (char_right >= left && char_left <= right) { 
	    t->traverse_child(children_.item(i), r);
	    if (i_picking && is_not_nil(GlyphTraversal_var(t->picked()))) {
		break;
	    }
	} 
    }
    CORBA::release(r);
}

void CharBox::child_allocate(Long index, Glyph::AllocationInfo& a) {
    update_positions();
    Long count = children_.count();
    Coord char_left = positions_.item(index);
    Coord char_right = (index < count-1)?
	positions_.item(index+1) : positions_.item(count-1);
    Coord span = char_right - char_left;
    RegionImpl r;
    r.copy(a.allocation);
    r.lower_.x += (scroll_pos + char_left);
    r.upper_.x = (r.lower_.x + span);
    a.allocation->copy(&r);
}


/* class FieldEditorImpl */

declareActionCallback(FieldEditorImpl)
implementActionCallback(FieldEditorImpl)
declareUpdateCallback(FieldEditorImpl)
implementUpdateCallback(FieldEditorImpl)

FieldEditorImpl::FieldEditorImpl(
    ExclusiveEditMgr_ptr mgr, Tag tag, FieldEditorData_ptr fe_data
) : ViewerImpl() {
    threads_ = ThreadLib::thread_kit_ptr();
    FrescoContext* fc = FrescoLib::current_context();
    layouts_ = LayoutKit::_narrow(
        fc->resolve(*NamingLib::to_name("layouts_impl"))
    );
    WidgetKit_var widgets = WidgetKit::_narrow(
        fc->resolve(*NamingLib::to_name("widgets_impl"))
    );
    menus_ = widgets->menu_kit();
    DrawingKit_var drawings = fc->default_drawing_kit();
    StyleContext_var style = drawings->style();
    foreground_ = drawings->foreground(style);
    background_ = drawings->background(style);
    font_ = drawings->default_font(style);
    max_chars_ = fe_data->max_chars_allowed();
    Long width_in_chars = fe_data->init_width_in_chars();
    char *buf = new char[width_in_chars+1];
    for (Long i=0; i<width_in_chars; i++) {
	buf[i] = 'M';
    }
    buf[width_in_chars] = '\0';
    CharString_var sample = FrescoLib::string_copy(buf);
    delete buf;
    Font::Info info;
    font_->string_info(sample, info);
    lmargin_ = Coord(2); rmargin_ = Coord(4); tmargin_ = Coord(1); bmargin_ = Coord(1);
    visible_span_ = info.width;
    Coord font_height = info.font_descent + info.font_ascent;
    Coord vheight = font_height + bmargin_+tmargin_;
    Alignment valign = Alignment(0);
    if (!Math::equal(vheight, float(0), float(1e-2))) {
	valign = (bmargin_+info.font_descent) / vheight;
    }
    charbox_ = new CharBox(visible_span_, vheight, valign);
    ViewerImpl::body(
	Glyph_var(layouts_->hmargin_lr(
	    new Clipper(charbox_),
	    lmargin_, rmargin_
	))
    );
    charbox_->append(Glyph_var(new Eoln(font_)));
    point_ = charbox_->first_child_offset();
    mark_ = point_;
    left_ = lmargin_;
    right_ = left_ + visible_span_;
    bfield_ = new BoundedField(Coord(0), Coord(0), Coord(0), info.width);    
    ticker_ = nil;
    lock_ = threads_->mutex_ptr();
    tickwait_ = threads_->condition();
    delay_ = 0.04f;
    scroll_sensitivity_ = 10.0f; 
    char_count_ = 0;
    have_focus_ = false;
    have_access_ = false;
    editmgr_watcher_ = nil;
    text_change_ = false;
    ticking_ = false;
    edit_brush_ = drawings->simple_brush(Coord(1));
    // fix this
    busy_color_ = drawings->color_rgb(Color::Intensity(0.44), Color::Intensity(0.44), Color::Intensity(0.44));
    edit_color_ = drawings->color_rgb(Color::Intensity(1), Color::Intensity(1), Color::Intensity(1));
    nofocus_color_ = drawings->color_rgb(Color::Intensity(0.76), Color::Intensity(0.76), Color::Intensity(0.76));
    scroll_callback_ = new UpdateCallback(FieldEditorImpl)(
    	this, &FieldEditorImpl::scroll_update, bfield_
    );
    editmgr_ = ExclusiveEditMgr::_duplicate(mgr);
    fsubj_ = FieldSubject::_narrow(
	FrescoObject_var(editmgr_->subject(tag))
    );
    if (is_nil(fsubj_)) {
	abort();  // throw exception
    }
    ftag_ = tag;
    fsubj_callback_ = new UpdateCallback(FieldEditorImpl)(
	this, &FieldEditorImpl::fsubj_update, fsubj_
    );
    fsubj_update();
}

FieldEditorImpl::~FieldEditorImpl() {
    CORBA::release(bfield_);
    CORBA::release(charbox_);
}

OxExchange* FieldEditorImpl::_exchange() { return ViewerImpl::_exchange(); }

//+ FieldEditorImpl(XfFrescoObject::=ViewerImpl::)
XfTag FieldEditorImpl::attach(XfFrescoObject_ptr observer) {
    return ViewerImpl::attach(observer);
}
void FieldEditorImpl::detach(XfTag attach_tag) {
    ViewerImpl::detach(attach_tag);
}
void FieldEditorImpl::disconnect() {
    ViewerImpl::disconnect();
}
void FieldEditorImpl::notify_observers() {
    ViewerImpl::notify_observers();
}
void FieldEditorImpl::update() {
    ViewerImpl::update();
}
//+

//+ FieldEditorImpl(XfGlyph::clone_glyph)
XfGlyph_ptr FieldEditorImpl::clone_glyph() {
    return ViewerImpl::clone_glyph();
}
//+ FieldEditorImpl(XfGlyph::style?)
XfStyleContext_ptr FieldEditorImpl::style() {
    return ViewerImpl::style();
}
//+ FieldEditorImpl(XfGlyph::style=s)
void FieldEditorImpl::style(XfStyleContext_ptr s) {
    ViewerImpl::style(s);
}
//+ FieldEditorImpl(XfGlyph::transformation)
XfTransform_ptr FieldEditorImpl::transformation() {
    return ViewerImpl::transformation();
}
//+ FieldEditorImpl(XfGlyph::request)
void FieldEditorImpl::request(XfGlyph::Requisition& r) {
    ViewerImpl::request(r);
}
//+ FieldEditorImpl(XfGlyph::extension)
void FieldEditorImpl::extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r) {
    ViewerImpl::extension(a, r);
}
//+ FieldEditorImpl(XfGlyph::shape)
void FieldEditorImpl::shape(XfRegion_ptr r) {
    ViewerImpl::shape(r);
}
//+ FieldEditorImpl(XfGlyph::draw)
void FieldEditorImpl::draw(XfGlyphTraversal_ptr t) {
    ViewerImpl::draw(t);
}
//+ FieldEditorImpl(XfGlyph::pick)
void FieldEditorImpl::pick(XfGlyphTraversal_ptr t) {
    ViewerImpl::pick(t);
}
//+ FieldEditorImpl(XfGlyph::body?)
XfGlyph_ptr FieldEditorImpl::body() {
    return ViewerImpl::body();
}
//+ FieldEditorImpl(XfGlyph::body=g)
void FieldEditorImpl::body(XfGlyph_ptr g) {
    ViewerImpl::body(g);
}
//+ FieldEditorImpl(XfGlyph::append)
void FieldEditorImpl::append(XfGlyph_ptr g) {
    ViewerImpl::append(g);
}
//+ FieldEditorImpl(XfGlyph::prepend)
void FieldEditorImpl::prepend(XfGlyph_ptr g) {
    ViewerImpl::prepend(g);
}
//+ FieldEditorImpl(XfGlyph::add_parent)
XfTag FieldEditorImpl::add_parent(XfGlyphOffset_ptr parent_offset) {
    return ViewerImpl::add_parent(parent_offset);
}
//+ FieldEditorImpl(XfGlyph::remove_parent)
void FieldEditorImpl::remove_parent(XfTag add_tag) {
    ViewerImpl::remove_parent(add_tag);
}
//+ FieldEditorImpl(XfGlyph::first_child_offset)
XfGlyphOffset_ptr FieldEditorImpl::first_child_offset() {
    return ViewerImpl::first_child_offset();
}
//+ FieldEditorImpl(XfGlyph::last_child_offset)
XfGlyphOffset_ptr FieldEditorImpl::last_child_offset() {
    return ViewerImpl::last_child_offset();
}
//+ FieldEditorImpl(XfGlyph::parent_offsets)
void FieldEditorImpl::parent_offsets(XfGlyph::OffsetSeq& parents) {
    ViewerImpl::parent_offsets(parents);
}
//+ FieldEditorImpl(XfGlyph::allocations)
void FieldEditorImpl::allocations(XfGlyph::AllocationInfoSeq& a) {
    ViewerImpl::allocations(a);
}
//+ FieldEditorImpl(XfGlyph::need_redraw)
void FieldEditorImpl::need_redraw() {
    ViewerImpl::need_redraw();
}
//+ FieldEditorImpl(XfGlyph::need_redraw_region)
void FieldEditorImpl::need_redraw_region(XfRegion_ptr r) {
    ViewerImpl::need_redraw_region(r);
}
//+ FieldEditorImpl(XfGlyph::need_resize)
void FieldEditorImpl::need_resize() {
    ViewerImpl::need_resize();
}
//+ FieldEditorImpl(XfGlyph::restore_trail)
Boolean FieldEditorImpl::restore_trail(XfGlyphTraversal_ptr t) {
    return ViewerImpl::restore_trail(t);
}

//+ FieldEditorImpl(XfViewer::parent_viewer)
XfViewer_ptr FieldEditorImpl::parent_viewer() {
    return ViewerImpl::parent_viewer();
}
//+ FieldEditorImpl(XfViewer::next_viewer)
XfViewer_ptr FieldEditorImpl::next_viewer() {
    return ViewerImpl::next_viewer();
}
//+ FieldEditorImpl(XfViewer::prev_viewer)
XfViewer_ptr FieldEditorImpl::prev_viewer() {
    return ViewerImpl::prev_viewer();
}
//+ FieldEditorImpl(XfViewer::first_viewer)
XfViewer_ptr FieldEditorImpl::first_viewer() {
    return ViewerImpl::first_viewer();
}
//+ FieldEditorImpl(XfViewer::last_viewer)
XfViewer_ptr FieldEditorImpl::last_viewer() {
    return ViewerImpl::last_viewer();
}
//+ FieldEditorImpl(XfViewer::append_viewer)
void FieldEditorImpl::append_viewer(XfViewer_ptr v) {
    ViewerImpl::append_viewer(v);
}
//+ FieldEditorImpl(XfViewer::prepend_viewer)
void FieldEditorImpl::prepend_viewer(XfViewer_ptr v) {
    ViewerImpl::prepend_viewer(v);
}
//+ FieldEditorImpl(XfViewer::insert_viewer)
void FieldEditorImpl::insert_viewer(XfViewer_ptr v) {
    ViewerImpl::insert_viewer(v);
}
//+ FieldEditorImpl(XfViewer::replace_viewer)
void FieldEditorImpl::replace_viewer(XfViewer_ptr v) {
    ViewerImpl::replace_viewer(v);
}
//+ FieldEditorImpl(XfViewer::remove_viewer)
void FieldEditorImpl::remove_viewer() {
    ViewerImpl::remove_viewer();
}
//+ FieldEditorImpl(XfViewer::set_viewer_links)
void FieldEditorImpl::set_viewer_links(XfViewer_ptr parent, XfViewer_ptr prev, XfViewer_ptr next) {
    ViewerImpl::set_viewer_links(parent, prev, next);
}
//+ FieldEditorImpl(XfViewer::set_first_viewer)
void FieldEditorImpl::set_first_viewer(XfViewer_ptr v) {
    ViewerImpl::set_first_viewer(v);
}
//+ FieldEditorImpl(XfViewer::set_last_viewer)
void FieldEditorImpl::set_last_viewer(XfViewer_ptr v) {
    ViewerImpl::set_last_viewer(v);
}
//+ FieldEditorImpl(XfViewer::request_focus)
XfFocus_ptr FieldEditorImpl::request_focus(XfViewer_ptr requestor, Boolean temporary) {
    return ViewerImpl::request_focus(requestor, temporary);
}
//+ FieldEditorImpl(XfViewer::first_focus)
Boolean FieldEditorImpl::first_focus() {
    return ViewerImpl::first_focus();
}
//+ FieldEditorImpl(XfViewer::last_focus)
Boolean FieldEditorImpl::last_focus() {
    return ViewerImpl::last_focus();
}
//+ FieldEditorImpl(XfViewer::next_focus)
Boolean FieldEditorImpl::next_focus() {
    return ViewerImpl::next_focus();
}
//+ FieldEditorImpl(XfViewer::prev_focus)
Boolean FieldEditorImpl::prev_focus() {
    return ViewerImpl::prev_focus();
}
//+ FieldEditorImpl(XfViewer::handle)
Boolean FieldEditorImpl::handle(XfGlyphTraversal_ptr t, XfEvent_ptr e) {
    return ViewerImpl::handle(t, e);
}
//+ FieldEditorImpl(XfViewer::close)
void FieldEditorImpl::close() {
    ViewerImpl::close();
}

//+ FieldEditorImpl(XfViewer::receive_focus) !primary
Boolean FieldEditorImpl::receive_focus(XfFocus_ptr f, Boolean) {
    MenuItem_var menu_bar = f->menu_bar();
    if (! is_nil(menu_bar)) {
        MenuItem_var edit = menus_->menu_item_string(
    	new CharStringImpl("Edit"), 0
        );
        edit->append(MenuItem_var(menus_->button_item_string(
            CharString_var(new CharStringImpl("Copy")), 0, nil,
            Action_var(new CopyAction)
        )));
        edit->append(MenuItem_var(menus_->button_item_string(
            CharString_var(new CharStringImpl("Cut")), 0, nil,
            Action_var(new CutAction)
        )));
        edit->append(MenuItem_var(menus_->button_item_string(
            CharString_var(new CharStringImpl("Paste")), 0, nil,
            Action_var(new PasteAction)
        )));
        menu_bar->insert(1, edit);
    }
    have_focus_ = true;
    if (!have_access_) {
	if (editmgr_->begin_edit(ftag_)) {
	    have_access_ = true;
	} else {
	    if (is_nil(editmgr_watcher_)) {
		editmgr_watcher_ = new UpdateCallback(FieldEditorImpl)(
		    this, &FieldEditorImpl::notify_ok_to_edit, editmgr_
	        );
	    }
	}
    }
    need_redraw();

    return true;
}

void FieldEditorImpl::notify_ok_to_edit() {
    Boolean keep_watching = false;
    assert(have_access_ != true);
    if (have_focus_) {
	if (editmgr_->begin_edit(ftag_)) {
	    have_access_ = true;
	    need_redraw();
	} else {
	    keep_watching = true;
	}
    }
    if (!keep_watching) {
	delete editmgr_watcher_;
	editmgr_watcher_ = nil;
    }
}

//+ FieldEditorImpl(XfViewer::lose_focus) !temporary
void FieldEditorImpl::lose_focus(XfFocus_ptr f, Boolean primary, Boolean) {
    have_focus_ = false;
    if (have_access_) {
	editmgr_->end_edit();
	have_access_ = false;
    }
    need_redraw();
}

//+ FieldEditorImpl(FieldEditor::field_edit_manager)
ExclusiveEditMgr_ptr FieldEditorImpl::field_edit_manager() {
    return ExclusiveEditMgr::_duplicate(editmgr_);
}

//+ FieldEditorImpl(FieldEditor::field_subject)
FieldSubject_ptr FieldEditorImpl::field_subject() {
    return FieldSubject::_duplicate(fsubj_);
}

//+ FieldEditorImpl(FieldEditor::scroll_adjustment)
Adjustment_ptr FieldEditorImpl::scroll_adjustment() {
    return Adjustment::_duplicate(bfield_);
}

Boolean FieldEditorImpl::press(GlyphTraversal_ptr t, Event_ptr) {
    if (!have_focus_ && t->allocation_is_visible()) {
        request_focus((Viewer*) this, false);
    }
    Boolean picked = t->forward();
    if (picked) {
	t->backward();
	long count = 0;
	while (t->forward()) count++;  
	point_ = mark_ = current_point(t);
	while (count--) t->backward();
	need_redraw();
    } else {
	point_ = mark_ = nil;
    }
    return true;
}

Boolean FieldEditorImpl::drag(GlyphTraversal_ptr t, Event_ptr e) {
    Vertex orig;
    Vertex p;
    p.x = e->pointer_x();
    p.y = e->pointer_y();
    Region_var(t->allocation())->origin(orig);
    Transform_var tx = t->current_transform();
    tx->inverse_transform_vertex(p);
    Coord x = p.x + orig.x;
    lock_->acquire();
    ticking_ = ( x > (edge_ = right_) || x < (edge_ = left_) );
    if (ticking_) {
	if (is_nil(ticker_)) {  
	    ticker_ = start_ticker();
	}
	if (is_not_nil(ticker_)) {  
	    delta_ = (x-edge_) * scroll_sensitivity_;
	    edge_pt_.x = edge_-orig.x;
	    edge_pt_.y = Coord(0);
	    tx->transform_vertex(edge_pt_);
	    tickwait_->notify();
	} else {
	    Coord d = Coord(1e6);    // just scroll to end if no threads
	    bfield_->scroll_by((edge_ == left_)? -d : d);
	    need_redraw();
	}
    } else {
	if (find_point_during_drag(t)) {
	    need_redraw();
	}
    }
    lock_->release();
    return true;
}

void FieldEditorImpl::tick() {
    for (;;) {
	lock_->acquire();
	while (!ticking_) {
	    tickwait_->wait(lock_);
	}
	Coord d = delta_;
	lock_->release();
	bfield_->scroll_by(d);
	need_redraw();
	FrescoLib::delay(delay_);
    }
}

Boolean FieldEditorImpl::release(GlyphTraversal_ptr, Event_ptr) {
    lock_->acquire();
    ticking_ = false;
    lock_->release();
    return true;
}

//+ FieldEditorImpl(XfGlyph::traverse)
void FieldEditorImpl::traverse(XfGlyphTraversal_ptr t) {
    if (!t->allocation_is_visible()) {
	return;
    }
    Region_var a = t->allocation();
    Region::BoundingSpan s;
    a->span(X_axis, s);
    right_ = s.length - rmargin_;
    visible_span_ = right_ - lmargin_;
    bfield_->visible_span(visible_span_);
    a->span(Y_axis, s);
    bottom_ = s.begin+bmargin_;
    top_ = s.end-tmargin_;    
    
    switch (t->op()) {
    case GlyphTraversal::draw: {	    
	if (ticking_) {
	    t->swap_op(GlyphTraversal::pick_top);
	    Painter_var p = t->current_painter();
	    p->push_clipping();
	    Region_var clip = p->visible();
	    RegionImpl r;
	    r.defined_ = true;
	    r.lower_.x = r.upper_.x = edge_pt_.x;
	    r.lower_.y = r.upper_.y = edge_pt_.y;
	    clip->copy(&r);
	    t->begin_viewer(FieldEditor_ptr(this));
	    t->traverse_child(&offset_, a);
	    t->end_viewer();
	    p->pop_clipping();
	    GlyphTraversal_var picked = t->picked();
	    if (is_nil(picked)) {
		point_ = charbox_->first_child_offset();
		if (is_nil(mark_)) {
		    mark_ = point_;
		}
	    } else {
		point_ = current_point(picked);
		t->clear();
	    }
	    t->swap_op(GlyphTraversal::draw);
	}
	if (is_nil(point_)) {
	    ViewerImpl::traverse(t);
	} else {
	    Painter_var p = t->current_painter();
	    if (is_equal(point_, mark_)) {
		ViewerImpl::traverse(t);
		draw_caret(p);
	    } else {
		if (is_not_nil(mark_)) {
		    draw_region(p);
		}
		ViewerImpl::traverse(t);
	    }
	}
	break;
    }
    case GlyphTraversal::pick_top:
    case GlyphTraversal::pick_any:
    case GlyphTraversal::pick_all: {  
	ViewerImpl::traverse(t);
	break;
    }
    default:
	ViewerImpl::traverse(t);
	break;
    }
}
    
Boolean FieldEditorImpl::find_point_during_drag(GlyphTraversal_ptr t) {
    Painter_var p = t->current_painter();
    p->push_clipping();
    Region_var clip = p->visible();
    RegionImpl r(clip, nil, nil);
    Region::BoundingSpan s;
    Region_var a = t->allocation();
    a->span(Y_axis, s);
    r.lower_.y = r.upper_.y = (s.begin + s.end) * Coord(0.5);
    r.yalign_ = Alignment(0);
    clip->copy(&r);
    t->begin_viewer(FieldEditor_ptr(this));
    t->traverse_child(&offset_, a);
    t->end_viewer();
    p->pop_clipping();
    GlyphTraversal_var picked = t->picked();
    if (is_nil(picked)) {
//	printf("no hit during drag\n");
	return false;
    } else {
	point_ = current_point(picked);
	t->clear();
	return true;
    }
}

void FieldEditorImpl::set_color(Painter_ptr p) {
    if (!have_focus_) {
	p->current_color(nofocus_color_);
    } else {
	if (!have_access_) {
	    p->current_color(busy_color_);
	} else {
	    p->current_color(edit_color_);
	}
    }
}

void FieldEditorImpl::draw_region(Painter_ptr p) {
    Color_var c = p->current_color();
    set_color(p);
    Coord x1 = find_pos(mark_) + lmargin_;
    Coord x2 = ticking_? edge_ : find_pos(point_) + lmargin_;
    p->push_clipping();
    p->clip_rect(left_, bottom_, right_, top_);
    if (x1 < x2) {
	x1 = Math::max(x1, left_);
	x2 = Math::min(x2, right_);
	p->fill_rect(x1, bottom_, x2, top_);
    } else {
	x1 = Math::min(x1, right_);
	x2 = Math::max(x2, left_);
	p->fill_rect(x2, bottom_, x1, top_);
    }
    p->pop_clipping();
    p->current_color(c);
}

void FieldEditorImpl::draw_caret(Painter_ptr p) {
    Color_var c = p->current_color();
    Brush_var b = p->current_brush();
    p->current_brush(edit_brush_);
    set_color(p);
    Coord x = find_pos(point_) + lmargin_;
    p->push_clipping();
    p->clip_rect(left_-lmargin_, bottom_, right_+rmargin_, top_);
    p->line(x, bottom_, x, top_);
    p->pop_clipping();
    p->current_color(c);
    if (is_not_nil(b)) {
	p->current_brush(b);
    }
}

Thread_ptr FieldEditorImpl::start_ticker() {
    Thread_ptr ticker = threads_->thread_ptr(
	new ActionCallback(FieldEditorImpl)(this, &FieldEditorImpl::tick)
    );
    if (is_not_nil(ticker)) {
	ticker->run();
    }
    return ticker;
}

void FieldEditorImpl::scroll_update() {
    if (!text_change_) {
	Adjustment::Settings s;
	bfield_->get_settings(s);
	Alignment a =  s.cur_lower / s.upper;
	scroll_to((a * visible_span_) - (a * s.upper));
	if (!ticking_) {
	    need_redraw();
	}
    }
}

GlyphOffset_ptr FieldEditorImpl::current_point(GlyphTraversal_ptr t) {
    GlyphOffset_var o = t->current_offset();
    return t->hit_info() ? o->next_child() : GlyphOffset::_duplicate(o);
}
    
Coord FieldEditorImpl::find_pos(GlyphOffset_ptr o) {
    Coord left, d;
    local_edges(o, left, d);
//  return left;
    return left - 1;
	/* This -1 makes the cursor position look correct on high-DPI
	 * displays but will not be correct for displays with DPI close
	 * to that of the X font bitmap. (Because the font bitmap is in a
	 * lower DPI, there is a gap to the right of characters on
	 * high-DPI displays.)
	 */
}

void FieldEditorImpl::local_edges(GlyphOffset_ptr o, Coord& lt, Coord &rt) {
    // Local_edges returns coordinates that are relative to right side
    // of left margin.
    RegionImpl* d = new RegionImpl;
    d->defined_ = true;
    Glyph::AllocationInfo a;
    a.allocation = d;
    a.transformation = nil;
    a.damaged = nil;
    o->child_allocate(a);
    RegionImpl r;
    GlyphImpl::default_extension(a, &r);
    lt = r.lower_.x;
    rt = r.upper_.x;
}

static Boolean any_modifier(Event_ptr e) {
    return
        e->modifier_is_down(Event::control) ||
        e->modifier_is_down(Event::shift) ||
        e->modifier_is_down(Event::capslock) ||
        e->modifier_is_down(Event::meta) ||
        e->modifier_is_down(Event::alt);
}

Boolean FieldEditorImpl::key_press(GlyphTraversal_ptr, Event_ptr e) {
    if (!have_access_) {
	beep();
	return false;
    }
    Boolean result = true;
    begin_text_change();
    CharCode code = e->character();
    Event::KeySym keysym = e->key();
    if (keysym == XK_Left || keysym == XK_Prior) {
	code = SEPreviousCharacter;
    } else if (keysym == XK_Right || keysym == XK_Next) {
	code = SENextCharacter;
    }
    if (code == SEDeleteNextCharacter) {
	delete_next_char();
    } else if (
	code == SEDeletePreviousCharacter ||
	code == SEDeletePreviousCharacterAlt
    ) {
	delete_prev_char();
    } else if (code == SEPreviousCharacter) {
	move_left();
    } else if (code == SENextCharacter) {
	move_right();
    } else if (code == SEBeginningOfLine) {
	move_to_beg_line();
    } else if (code == SEEndOfLine) {
	move_to_end_line();
    } else if (code == SESelectAll) {
	select_all();
//    } else if (code == SEESCCode || code == SERETURN) {
    } else if (code == SERETURN) {
	change_fsubj();
//    } else if ( ! iscntrl( code & 0x7f ) ){
    } else if (!any_modifier(e) && isprint( code & 0x7f ) ){
	insert_char(code);
    } else {
        result = false;
    }
    end_text_change();
    return result;
}

void FieldEditorImpl::delete_mark_region() {
    GlyphOffset_var left, right;
    Coord point_pos = find_pos(point_);
    Coord mark_pos = find_pos(mark_);
    if (point_pos < mark_pos) {
	left = point_;
	right = mark_;
    } else {
	left = mark_;
	right = point_;
    }
    while (!is_equal(left, right)) {
	GlyphOffset_var tmp = left->next_child();
	left->remove();
	char_count_--;
	left = tmp;
    }
    point_ = mark_ = left;
    point_pos = find_pos(point_);    
    if (point_pos < 0) {
	scroll_to(scroll_pos() - point_pos);
    } else if (point_pos > visible_span_) {
	scroll_to(visible_span_ - point_pos);
    }
}

void FieldEditorImpl::insert_char(XfCharCode c) {
    if (!is_equal(point_, mark_)) {
	delete_mark_region();
    }
    if (!max_chars_ || (char_count_ < max_chars_)) {
	point_->insert(new Character(c, font_, foreground_, background_));
	char_count_++;
	adjust_for_movement_fwd();
    } else {
	beep();
    }
}

void FieldEditorImpl::delete_prev_char() {
    if (!is_equal(point_, mark_)) {
	delete_mark_region();
    } else {
	GlyphOffset_var tmp = point_->prev_child();
	if (is_not_nil(tmp)) {
	    tmp->remove();
	    char_count_--;
	    adjust_for_movement_bwd();
	}
    }
}

void FieldEditorImpl::delete_next_char() {
    if (!is_equal(point_, mark_)) {
	delete_mark_region();
    } else {
	GlyphOffset_var tmp = point_->next_child();
	if (is_not_nil(tmp)) {
	    point_->remove();
	    point_ = mark_ = tmp;
	    char_count_--;
	    adjust_for_movement_fwd();
	}
    }
}

void FieldEditorImpl::move_left() {
    GlyphOffset_var prev = point_->prev_child();
    if (is_not_nil(prev)) {
	point_ = mark_ = prev;
	adjust_for_movement_bwd();
    }
}

void FieldEditorImpl::move_right() {
    GlyphOffset_var next = point_->next_child();
    if (is_not_nil(next)) {
	point_ = mark_ = next;
	adjust_for_movement_fwd();
    }
}

void FieldEditorImpl::move_to_end_line() {
    point_= mark_ = charbox_->last_child_offset();
    adjust_for_movement_fwd();
}

void FieldEditorImpl::move_to_beg_line() {
    point_ = mark_ = charbox_->first_child_offset();
    adjust_for_movement_bwd();
}

void FieldEditorImpl::select_all() {
    mark_ = charbox_->first_child_offset();
    point_= charbox_->last_child_offset();
}

void FieldEditorImpl::adjust_for_movement_fwd() {    
    Coord max = visible_span_ * Coord(1.00);
    Coord point_pos = find_pos(point_);
    if (point_pos > max) {
	scroll_by(max - point_pos);
    }
}

void FieldEditorImpl::adjust_for_movement_bwd() {
    Coord min = Coord(0);
    Coord new_pos = visible_span_ * Coord(0.25);
    Coord point_pos = find_pos(point_);
    if (point_pos < min) {
	scroll_by(new_pos - point_pos);
    }
    GlyphOffset_var first = charbox_->first_child_offset();
    Coord first_pos = find_pos(first);
    if (first_pos > Coord(0)) {
	scroll_to(Coord(0));
    }    
}

void FieldEditorImpl::begin_text_change() { }

void FieldEditorImpl::end_text_change() {
    text_change_ = true;
    Coord point_pos = find_pos(point_);
    Coord w = charbox_->total_char_width();
    bfield_->span(w);
    bfield_->scroll_to(point_pos - scroll_pos());
    if (w < visible_span_) {
	scroll_to(Coord(0));
    }	
    text_change_ = false;
    need_redraw();
}

void FieldEditorImpl::fsubj_update() {
	begin_text_change();
	select_all();
	delete_mark_region();
	Long count = fsubj_->count();
	char* str = fsubj_->get_all();
	for (long i=0; i<count; i++) {
	    point_->insert(
		new Character(str[i], font_, foreground_,background_)
	    );
	    char_count_++;
	}
	delete str;
	end_text_change();
}

void FieldEditorImpl::change_fsubj() {
    CharCode* buf = new CharCode[char_count_];
    GlyphOffset_var o = charbox_->first_child_offset();
    long cur = 0;
    for (long i=0; i < char_count_; i++) {
	Glyph_var g = o->child();
	FieldChar_var c = FieldChar::_narrow(g);
	if (is_not_nil(c)) {
	    buf[cur++] = c->code();
	}
	o = o->next_child();
    }

    CharString_var charstr = new CharStringImpl;
    CharString::Data data(char_count_, cur, buf);
    charstr->put_data(data);
    Long count = fsubj_->count();
    FieldCommand_var paste_cmd = new PasteCommand(0, count-1, charstr);
    paste_cmd->execute(fsubj_);
    editmgr_->log(paste_cmd);

    have_access_ = false;
    editmgr_->end_edit();

    if (editmgr_->begin_edit(ftag_)) {
	have_access_ = true;
    } else {
	if (is_nil(editmgr_watcher_)) {
	    editmgr_watcher_ = new UpdateCallback(FieldEditorImpl)(
		this, &FieldEditorImpl::notify_ok_to_edit, editmgr_
	    );
	}
    }
}

void FieldEditorImpl::scroll_to(Coord p) { charbox_->scroll_pos = p; }

void FieldEditorImpl::scroll_by(Coord d) { charbox_->scroll_pos += d; }

Coord FieldEditorImpl::scroll_pos() { return charbox_->scroll_pos; }

void FieldEditorImpl::beep() {
    Display_var d = FrescoLib::current_context()->display_ptr();
    d->ring_bell(0.75f);
}

Boolean FieldEditorImpl::is_equal(GlyphOffset_ptr o1, GlyphOffset_ptr o2) {
    // We know CharBox (and offsets) are not remote.
    return o1 == o2;
}

// class FieldSubjectImpl

implementList(CharList, CharCode)

FieldSubjectImpl::FieldSubjectImpl() { }

FieldSubjectImpl::~FieldSubjectImpl() { }

OxExchange* FieldSubjectImpl::_exchange() { return &object_; }

//+ FieldSubjectImpl(XfFrescoObject::=object_.)
XfTag FieldSubjectImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void FieldSubjectImpl::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void FieldSubjectImpl::disconnect() {
    object_.disconnect();
}
void FieldSubjectImpl::notify_observers() {
    object_.notify_observers();
}
void FieldSubjectImpl::update() {
    object_.update();
}
//+

//+ FieldSubjectImpl(FieldSubject::count)
Long FieldSubjectImpl::count() {
    return chars_.count();
}



//+ FieldSubjectImpl(FieldSubject::append_char)
void FieldSubjectImpl::append_char(XfCharCode c) {
    chars_.append(c);
    notify_observers();
}

//+ FieldSubjectImpl(FieldSubject::prepend_char)
void FieldSubjectImpl::prepend_char(XfCharCode c) {
    chars_.prepend(c);
    notify_observers();
}

//+ FieldSubjectImpl(FieldSubject::insert_char)
void FieldSubjectImpl::insert_char(Long index, XfCharCode c) {
    chars_.insert(index, c);
    notify_observers();
}

void FieldSubjectImpl::do_insert_chars(Long index, XfCharString_ptr str) {
    char* s = str->string_ptr();
    char* p = s;
    while (*p) {
	chars_.append(XfCharCode(*p++));
    }
    delete s; 
}

//+ FieldSubjectImpl(FieldSubject::insert_chars)
void FieldSubjectImpl::insert_chars(Long index, XfCharString_ptr str) {
    do_insert_chars(index, str);
    notify_observers();
}

//+ FieldSubjectImpl(FieldSubject::replace_char)
void FieldSubjectImpl::replace_char(Long index, XfCharCode c) {
    chars_.remove(index);
    chars_.insert(index, c);
    notify_observers();
}

//+ FieldSubjectImpl(FieldSubject::replace_chars)
void FieldSubjectImpl::replace_chars(Long first, Long last, XfCharString_ptr str) {
    do_delete_chars(first, last);
    do_insert_chars(first, str);
    notify_observers();
}

//+ FieldSubjectImpl(FieldSubject::replace_all)
void FieldSubjectImpl::replace_all(XfCharString_ptr str) {
    do_delete_chars(0, chars_.count()-1);
    do_insert_chars(0, str);
    notify_observers();
}

void FieldSubjectImpl::do_delete_chars(Long first, Long last) {
    for (Long i = first; i<= last; i++) {
	chars_.remove(first);
    }
}

//+ FieldSubjectImpl(FieldSubject::delete_chars)
void FieldSubjectImpl::delete_chars(Long first, Long last) {
    do_delete_chars(first, last);
    notify_observers();
}

//+ FieldSubjectImpl(FieldSubject::delete_all)
void FieldSubjectImpl::delete_all() {
    do_delete_chars(0, chars_.count()-1);
    notify_observers();
}

//+ FieldSubjectImpl(FieldSubject::get_char)
XfCharCode FieldSubjectImpl::get_char(Long index) {
    return chars_.item(index);
}

//+ FieldSubjectImpl(FieldSubject::get_chars)
char* FieldSubjectImpl::get_chars(Long first, Long last) {
    char* str = new char[(last-first)+1 +1];
    char* p = str;
    for (Long i = first; i<= last; i++) {
	*p++ = chars_.item(i);
    }
    *p = '\0';
    return str;
}

//+ FieldSubjectImpl(FieldSubject::get_all)
char* FieldSubjectImpl::get_all() {
    return get_chars(0, chars_.count()-1);
}

//+ FieldSubjectImpl(FieldSubject::copy)
void FieldSubjectImpl::copy(FieldSubject_ptr fs) {
    do_delete_chars(0, chars_.count()-1);
    Long count = fs->count();
    for (Long i=0; i<count; i++) {
	chars_.append(fs->get_char(i));
    }
    notify_observers();
}

// class FieldEditorDataImpl

FieldEditorDataImpl::FieldEditorDataImpl(Long init_width, Long max_chars) {
    init_width_ = init_width;
    max_chars_ = max_chars;
}
    
FieldEditorDataImpl::~FieldEditorDataImpl() { }

OxExchange* FieldEditorDataImpl::_exchange() { return &object_; }

//+ FieldEditorDataImpl(XfFrescoObject::=object_.)
XfTag FieldEditorDataImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void FieldEditorDataImpl::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void FieldEditorDataImpl::disconnect() {
    object_.disconnect();
}
void FieldEditorDataImpl::notify_observers() {
    object_.notify_observers();
}
void FieldEditorDataImpl::update() {
    object_.update();
}
//+

//+ FieldEditorDataImpl(ViewerData::foo)
void FieldEditorDataImpl::foo() {
   // interface needs an op due to ixx bug
}

//+ FieldEditorDataImpl(FieldEditorData::init_width_in_chars?)
Long FieldEditorDataImpl::init_width_in_chars() {
    return init_width_;
}

//+ FieldEditorDataImpl(FieldEditorData::init_width_in_chars=w)
void FieldEditorDataImpl::init_width_in_chars(Long w) {
    init_width_ = w;
}

//+ FieldEditorDataImpl(FieldEditorData::max_chars_allowed?)
Long FieldEditorDataImpl::max_chars_allowed() {
    return max_chars_;
}

//+ FieldEditorDataImpl(FieldEditorData::max_chars_allowed=m)
void FieldEditorDataImpl::max_chars_allowed(Long m) {
    max_chars_ = m;
}


// class FieldEditorCreator

FieldEditorCreator::FieldEditorCreator() { }
FieldEditorCreator::~FieldEditorCreator() { }

OxExchange* FieldEditorCreator::_exchange() { return &object_; }

//+ FieldEditorCreator(XfFrescoObject::=object_.)
XfTag FieldEditorCreator::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void FieldEditorCreator::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void FieldEditorCreator::disconnect() {
    object_.disconnect();
}
void FieldEditorCreator::notify_observers() {
    object_.notify_observers();
}
void FieldEditorCreator::update() {
    object_.update();
}
//+

//+ FieldEditorCreator(ViewerCreator::create_viewer)
XfViewer_ptr FieldEditorCreator::create_viewer(Inset_ptr ins) {
    FieldEditorData_var fe_data = FieldEditorData::_narrow(
	ViewerData_var(ins->viewer_data())
    );    
    FieldSubject_var fsubj = FieldSubject::_narrow(
	FrescoObject_var(ins->subject())
    );
    ExclusiveEditMgr_var fe_mgr = ExclusiveEditMgr::_narrow(
	EditMgr_var(ins->edit_manager())
    );
    WidgetKit_var widgets = Fresco_resolve_object(WidgetKit, "widgets_impl");
    FieldSubject_var newsubj = widgets->field_subject();
    newsubj->copy(fsubj);
    Tag ftag = fe_mgr->attach_subject(newsubj);
    return Viewer_ptr(new FieldEditorImpl(fe_mgr, ftag, fe_data)); 
}

// class FieldCommandImpl

FieldCommandImpl::FieldCommandImpl() { }

FieldCommandImpl::~FieldCommandImpl() { }

OxExchange* FieldCommandImpl::_exchange() { return &object_; }

//+ FieldCommandImpl(XfFrescoObject::=object_.)
XfTag FieldCommandImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void FieldCommandImpl::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void FieldCommandImpl::disconnect() {
    object_.disconnect();
}
void FieldCommandImpl::notify_observers() {
    object_.notify_observers();
}
void FieldCommandImpl::update() {
    object_.update();
}
//+

//+ FieldCommandImpl(Command::reversible) 
Boolean FieldCommandImpl::reversible() { return false; }

//+ FieldCommandImpl(FieldCommand::execute) !fs
void FieldCommandImpl::execute(FieldSubject_ptr) { }

//+ FieldCommandImpl(FieldCommand::unexecute) !fs
void FieldCommandImpl::unexecute(FieldSubject_ptr) { }

// class PasteCommand

PasteCommand::PasteCommand(Long first, Long last, CharString_ptr str) {
    first_ = first;
    last_ = last;
    str_ = CharString::_duplicate(str);
}

PasteCommand::~PasteCommand() { }

//+ PasteCommand(Command::reversible) 
Boolean PasteCommand::reversible() { return true; }

//+ PasteCommand(FieldCommand::execute)
void PasteCommand::execute(FieldSubject_ptr fs) {
    fs->replace_chars(first_, last_, str_);
}

//+ PasteCommand(FieldCommand::unexecute)
void PasteCommand::unexecute(FieldSubject_ptr fs) {
    abort(); // not implemented.
}


// class CutCommand

CutCommand::CutCommand(Long first, Long last) {
    first_ = first;
    last_ = last;
}

CutCommand::~CutCommand() { }

//+ CutCommand(Command::reversible) 
Boolean CutCommand::reversible() { return true; }

//+ CutCommand(FieldCommand::execute)
void CutCommand::execute(FieldSubject_ptr fs) {
    fs->delete_chars(first_, last_);
}

//+ CutCommand(FieldCommand::unexecute)
void CutCommand::unexecute(FieldSubject_ptr fs) {
    abort(); // not implemented.
}

// class FieldEditMgr

void FieldEditMgr::execute(Command_ptr cmd, FrescoObject_ptr subj) {
    FieldSubject_var fsubj = FieldSubject::_narrow(subj);
    if (is_not_nil(fsubj)) {
	FieldCommand_var fcmd = FieldCommand::_narrow(cmd);
	if (is_not_nil(fcmd)) {
	    fcmd->execute(fsubj);
	}
    }
}
