//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-93 Silicon Graphics, Inc.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Stanford and Silicon Graphics may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Stanford and Silicon Graphics.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD OR SILICON GRAPHICS BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#include <assert.h>
#include <Fresco/defs/display.h>
#include <Fresco/defs/drawing.h>
#include <Fresco/defs/fcontext.h>
#include <Fresco/defs/fresco.h>
#include <Fresco/defs_impl/glyphs.h>
#include <Fresco/defs_impl/names.h>
#include <Fresco/widgets/widgets.h>
#include "bevel.h"
#include "bfloat.h"
#include "buttons.h"
#include "dialog_kit.h"
#include "field.h"
#include "kit.h"
#include "menu_kit.h"
#include "slider.h"
#include "stepper.h"
#include "telltales.h"

Fresco_file_id(Fresco_widgets_impl_kit_cxx)

#define infinite_coord Coord(10e6)

TelltaleImpl::TelltaleImpl(TelltaleImpl::FlagSet f) {
    flags_ = f;
    current_ = nil;
}

TelltaleImpl::~TelltaleImpl() { }

OxExchange* TelltaleImpl::_exchange() { return &object_; }

//+ TelltaleImpl(XfFrescoObject::=object_.)
XfTag TelltaleImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void TelltaleImpl::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void TelltaleImpl::disconnect() {
    object_.disconnect();
}
void TelltaleImpl::notify_observers() {
    object_.notify_observers();
}
void TelltaleImpl::update() {
    object_.update();
}
//+

//+ TelltaleImpl(XfTelltale::set)
void TelltaleImpl::set(XfTelltale::Flag f) { modify(f, true); }

//+ TelltaleImpl(XfTelltale::clear)
void TelltaleImpl::clear(XfTelltale::Flag f) { modify(f, false); }

//+ TelltaleImpl(XfTelltale::test)
Boolean TelltaleImpl::test(XfTelltale::Flag f) {
    return (flags_ & (1 << f)) != 0;
}

//+ TelltaleImpl(XfTelltale::current=g)
void TelltaleImpl::current(XfTelltale_ptr g) {
    CORBA::release(current_);
    current_ = Telltale::_duplicate(g);
}

//+ TelltaleImpl(XfTelltale::current?)
XfTelltale_ptr TelltaleImpl::current() {
    return Telltale::_duplicate(current_);
}

void TelltaleImpl::modify(Telltale::Flag f, Boolean on) {
    TelltaleImpl::FlagSet fs = (1 << f);
    TelltaleImpl::FlagSet newflags = on ? (flags_ | fs) : (flags_ & ~fs);
    if (newflags != flags_) {
	if (on && f == Telltale::chosen && is_not_nil(current_)) {
	    Telltale_var t = current_->current();
	    if (is_not_nil(t)) {
		t->clear(Telltale::chosen);
	    }
	    current_->current(this);
	}
	flags_ = newflags;
	notify_observers();
    }
}

TelltaleImpl::FlagSet TelltaleImpl::flags_of(Telltale_ptr t) {
    return
        (t->test(Telltale::enabled) ? TelltaleImpl::enabled_bit : 0) |
        (t->test(Telltale::visible) ? TelltaleImpl::visible_bit : 0) |
        (t->test(Telltale::active) ? TelltaleImpl::active_bit : 0) |
        (t->test(Telltale::chosen) ? TelltaleImpl::chosen_bit : 0);
}

#define hilite_border Coord(4)
#define button_border Coord(4)
#define arrow_border Coord(6)

static const unsigned int checkmark_width = 32;
static const unsigned int checkmark_height = 12;

static unsigned char checkmark_bits[] = {
    0x00, 0x00, 0x3e, 0x00, 0x00, 0xc0, 0x0f, 0x00,
    0x04, 0xf0, 0x03, 0x00, 0x1e, 0xf8, 0x00, 0x00,
    0x3f, 0x3e, 0x00, 0x00, 0xbf, 0x1f, 0x00, 0x00,
    0xf8, 0x07, 0x00, 0x00, 0xf0, 0x03, 0x00, 0x00,
    0xf0, 0x01, 0x00, 0x00, 0xe0, 0x00, 0x00, 0x00,
    0x60, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

static const unsigned int shadow1_width = 32;
static const unsigned int shadow1_height = 13;

static unsigned char shadow1_bits[] = {
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0xc0, 0x00, 0x00, 0x00, 0xe0, 0x00, 0x00,
    0x07, 0x38, 0x00, 0x00, 0x0f, 0x1c, 0x00, 0x00,
    0x08, 0x06, 0x00, 0x00, 0x10, 0x03, 0x00, 0x00,
    0x90, 0x01, 0x00, 0x00, 0xe0, 0x00, 0x00, 0x00,
    0x60, 0x00, 0x00, 0x00
};

static const unsigned int shadow2_width = 32;
static const unsigned int shadow2_height = 13;

static unsigned char shadow2_bits[] = {
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x30, 0x00,
    0x00, 0x00, 0x3c, 0x00, 0x00, 0x00, 0x0f, 0x00,
    0x00, 0x00, 0x03, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00
};

class WKHiliteFrame : public WKFrame {
public:
    WKHiliteFrame(
	Glyph_ptr, Telltale_ptr, WidgetKitImpl::Info*,
	Coord thickness = hilite_border, float xalign = 0.5, float yalign = 0.5
    );
    ~WKHiliteFrame();

    virtual void draw(GlyphTraversal_ptr t);
};

class WKButtonFrame : public WKFrame {
public:
    WKButtonFrame(
	Glyph_ptr, TelltaleImpl*, WidgetKitImpl::Info*,
	Coord thickness = button_border, float xalign = 0.5, float yalign = 0.5
    );
    ~WKButtonFrame();

    virtual void draw_frame(GlyphTraversal_ptr t);
};

class WKPushButtonFrame : public WKButtonFrame {
public:
    WKPushButtonFrame(
	Glyph_ptr, TelltaleImpl*, WidgetKitImpl::Info*,
	Coord thickness = button_border, float xalign = 0.5, float yalign = 0.5
    );
    ~WKPushButtonFrame();

    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
};

class WKFiller : public WKGlyph {
public:
    WKFiller(WidgetKitImpl::Info*, Color_ptr c);
    ~WKFiller();

    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
    void draw(XfGlyphTraversal_ptr t); //+ XfGlyph::draw
protected:
    Color_ptr color_;
};

class WKHSeparator : public WKGlyph {
public:
    WKHSeparator(WidgetKitImpl::Info*);
    ~WKHSeparator();
 
    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
    void draw(XfGlyphTraversal_ptr t); //+ XfGlyph::draw
};
 
class WKVSeparator : public WKGlyph {
public:
    WKVSeparator(WidgetKitImpl::Info*);
    ~WKVSeparator();
 
    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
    void draw(XfGlyphTraversal_ptr t); //+ XfGlyph::draw
};

class WKCheckmark : public WKGlyph {
public:
    WKCheckmark(WidgetKitImpl::Info*, Telltale_ptr);
    ~WKCheckmark();

    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
    void draw(XfGlyphTraversal_ptr t); //+ XfGlyph::draw
protected:
    Telltale_var state_;
};

class WKIndicator : public WKGlyph {
public:
    WKIndicator(WidgetKitImpl::Info*, TelltaleImpl*);
    ~WKIndicator();

    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
    void draw(XfGlyphTraversal_ptr t); //+ XfGlyph::draw
protected:
    TelltaleImpl* state_;
};

class WKRadioFlag : public WKGlyph {
public:
    WKRadioFlag(WidgetKitImpl::Info*, Telltale_ptr, Font_ptr);
    ~WKRadioFlag();

    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
    void draw(XfGlyphTraversal_ptr t); //+ XfGlyph::draw
protected:
    Telltale_var state_;
    Font_ptr font_;
};

class WKRadioItem : public WKRadioFlag {
public:
    WKRadioItem(WidgetKitImpl::Info*, TelltaleImpl*, Font_ptr);
    ~WKRadioItem();

    void draw(XfGlyphTraversal_ptr t); //+ XfGlyph::draw
};

class WKDefaultArrow : public WKGlyph {
public:
    WKDefaultArrow(WidgetKitImpl::Info*, Font_ptr);
    ~WKDefaultArrow();

    void request(XfGlyph::Requisition& r); //+ XfGlyph::request
    void draw(XfGlyphTraversal_ptr t); //+ XfGlyph::draw
protected:
    Coord width_;
    Coord height_;
    Coord ascent_;
    Alignment align_;
};

class WKThumb : public WKGlyph {
public:
    WKThumb(WidgetKitImpl::Info*, Axis, long ridges, TelltaleImpl*);
    ~WKThumb();

    void draw(XfGlyphTraversal_ptr t); //+ XfGlyph::draw
protected:
    Axis axis_;
    long ridges_;
    TelltaleImpl* state_;
};

struct StyleData {
    const char* path;
    const char* value;
};

static StyleData kit_attributes[] = {
    { "*WidgetKit*background", "#aaaaaa" },
    { "*WidgetKit*checkScale", "0.7" },
    { "*WidgetKit*frameThickness", "2.0" },
    { "*WidgetKit*moverSize", "20.0" },
    { "*WidgetKit*radioScale", "0.9" },
    { "*WidgetKit*sliderSize", "20.0" },
    { "*WidgetKit*FileChooser*filter", "off" },
    { "*WidgetKit*FieldEditor*beveled", "on" },
    { "*WidgetKit*FieldEditor*background", "#b88d8d" },
    { "*WidgetKit*FieldEditor*flat", "#b88d8d" },
    { "*WidgetKit*minButtonWidth", "72.0" },
    { "*WidgetKit*minThumbSize", "28.0" },
    { "*WidgetKit*thumbRidges", "3" },
    { "*Slider*thumbRidges", "1" },
    { nil }
};

dload(WidgetKitImpl)

WidgetKitImpl::WidgetKitImpl() {
    FrescoContext* fc = FrescoLib::current_context();
    layouts_ = LayoutKit::_narrow(
	fc->resolve(*NamingLib::to_name("layouts_impl"))
    );
    assert(is_not_nil(layouts_));
    dialogs_ = nil;
    menus_ = nil;
    drawing_ = fc->default_drawing_kit();
    assert(is_not_nil(drawing_));
    StyleContext_var s = fc->style();
    style_ = s->new_style();
    style_->alias(FrescoLib::tmp_string_ref("WidgetKit"));
    style_->link_parent(s);
    for (StyleData* d = kit_attributes; d->path != nil; d++) {
	StyleValue_var a = s->bind(FrescoLib::tmp_string_ref(d->path));
	if (a->uninitialized() || a->priority() <= -10) {
	    a->write_string(FrescoLib::tmp_string_ref(d->value));
	    a->priority(-10);
	}
    }
    load();
}

WidgetKitImpl::~WidgetKitImpl() { }

OxExchange* WidgetKitImpl::_exchange() { return &object_; }

//+ WidgetKitImpl(XfFrescoObject::=object_.)
XfTag WidgetKitImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void WidgetKitImpl::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void WidgetKitImpl::disconnect() {
    object_.disconnect();
}
void WidgetKitImpl::notify_observers() {
    object_.notify_observers();
}
void WidgetKitImpl::update() {
    object_.update();
}
//+

//+ WidgetKitImpl(WidgetKit::outset_frame)
XfGlyph_ptr WidgetKitImpl::outset_frame(XfGlyph_ptr g) {
    return new WKFrame(
	g,
	new TelltaleImpl(TelltaleImpl::enabled_bit | TelltaleImpl::active_bit),
	&info_, info_.thickness
    );
}

//+ WidgetKitImpl(WidgetKit::inset_frame)
XfGlyph_ptr WidgetKitImpl::inset_frame(XfGlyph_ptr g) {
    return new WKFrame(
	g,
	new TelltaleImpl(TelltaleImpl::enabled_bit | TelltaleImpl::chosen_bit),
	&info_, info_.thickness
    );
}

//+ WidgetKitImpl(WidgetKit::bright_inset_frame)
XfGlyph_ptr WidgetKitImpl::bright_inset_frame(XfGlyph_ptr g) {
    return new WKFrame(
	g,
	new TelltaleImpl(
	    TelltaleImpl::enabled_bit | TelltaleImpl::visible_bit |
	    TelltaleImpl::active_bit
	),
	&info_, Coord(0.5) * info_.thickness, Coord(0), Coord(0), false, false
    );
}

//+ WidgetKitImpl(WidgetKit::hilite_frame)
XfGlyph_ptr WidgetKitImpl::hilite_frame(XfGlyph_ptr g, XfTelltale_ptr t) {
    return new WKHiliteFrame(g, t, &info_);
}

//+ WidgetKitImpl(WidgetKit::filler)
XfGlyph_ptr WidgetKitImpl::filler() {
    return color_filler(info_.color[light_gray]);
}

//+ WidgetKitImpl(WidgetKit::color_filler)
XfGlyph_ptr WidgetKitImpl::color_filler(XfColor_ptr c) {
    return new WKFiller(&info_, c);
}

//+ WidgetKitImpl(WidgetKit::hseparator)
XfGlyph_ptr WidgetKitImpl::hseparator() {
    return new WKHSeparator(&info_);
}
 
//+ WidgetKitImpl(WidgetKit::vseparator)
XfGlyph_ptr WidgetKitImpl::vseparator() {
    return new WKVSeparator(&info_);
}

//+ WidgetKitImpl(WidgetKit::field_editor)
FieldEditor_ptr WidgetKitImpl::field_editor(FieldSubject_ptr subj, Long width, Long maximum) {
    ExclusiveEditMgr_var mgr = field_edit_manager();
    Tag s = mgr->attach_subject(subj);
    FieldEditorData_var fe_data = field_editor_data(width, maximum);
    return managed_field_editor(mgr, s, fe_data);
}

//+ WidgetKitImpl(WidgetKit::managed_field_editor)
FieldEditor_ptr WidgetKitImpl::managed_field_editor(ExclusiveEditMgr_ptr mgr, XfTag subj, FieldEditorData_ptr fe_data) {
    return new FieldEditorImpl(mgr, subj, fe_data);
}

//+ WidgetKitImpl(WidgetKit::field_edit_manager)
ExclusiveEditMgr_ptr WidgetKitImpl::field_edit_manager() {
    return new FieldEditMgr();
}

//+ WidgetKitImpl(WidgetKit::field_editor_data)
FieldEditorData_ptr WidgetKitImpl::field_editor_data(Long width, Long maximum) {
    return new FieldEditorDataImpl(width, maximum);
}

//+ WidgetKitImpl(WidgetKit::field_subject)
FieldSubject_ptr WidgetKitImpl::field_subject() {
    return new FieldSubjectImpl();
}

//+ WidgetKitImpl(WidgetKit::field_inset)
Inset_ptr WidgetKitImpl::field_inset(FieldSubject_ptr subj, FieldEditorData_ptr vdata, ExclusiveEditMgr_ptr mgr) {
    return new InsetImpl(
	subj, vdata, mgr, FrescoLib::string_ref("FieldEditor")
    );
}

//+ WidgetKitImpl(WidgetKit::field_editor_creator)
ViewerCreator_ptr WidgetKitImpl::field_editor_creator() {
    return new FieldEditorCreator();
}

//+ WidgetKitImpl(WidgetKit::_registry_ptr)
Registry_ptr WidgetKitImpl::_registry_ptr() {
    return new RegistryImpl();
}
//+ WidgetKitImpl(WidgetKit::_persistent_naming_context)
Naming::NamingContext_ptr WidgetKitImpl::_persistent_naming_context() {
    return new PersistentNamingContext();
}


//+ WidgetKitImpl(WidgetKit::telltale_group)
XfTelltale_ptr WidgetKitImpl::telltale_group() {
    return new TelltaleImpl;
}

//+ WidgetKitImpl(WidgetKit::push_button)
Button_ptr WidgetKitImpl::push_button(XfGlyph_ptr g, XfAction_ptr a) {
    TelltaleImpl* t = new TelltaleImpl(TelltaleImpl::enabled_bit);
    Button_ptr b = new ButtonImpl(t, a);
    b->body(new WKPushButtonFrame(g, t, &info_));
    return b;
}

//+ WidgetKitImpl(WidgetKit::default_button)
Button_ptr WidgetKitImpl::default_button(XfGlyph_ptr g, XfAction_ptr a) {
    Font_ptr f = info_.font;
    Glyph_var hbox = layouts_->hbox();
    hbox->append(g);
    hbox->append(Glyph_var(layouts_->hspace(Coord(3))));
    hbox->append(Glyph_var(new WKDefaultArrow(&info_, f)));
    return push_button(hbox, a);
}

//+ WidgetKitImpl(WidgetKit::checkmark)
XfGlyph_ptr WidgetKitImpl::checkmark(XfTelltale_ptr t) {
    return new WKCheckmark(&info_, t);
}

//+ WidgetKitImpl(WidgetKit::check_box)
Button_ptr WidgetKitImpl::check_box(XfGlyph_ptr g, XfAction_ptr a) {
    TelltaleImpl* t = new TelltaleImpl(
	TelltaleImpl::enabled_bit | TelltaleImpl::toggle_bit
    );
    Button_ptr b = new ButtonImpl(t, a);
    Glyph_var box = layouts_->hbox_align_elements(Coord(0.5));
    box->append(
	Glyph_var(
	    new WKButtonFrame(
		Glyph_var(new WKCheckmark(&info_, t)), t, &info_
	    )
	)
    );
    box->append(Glyph_var(layouts_->hspace(Coord(6))));
    box->append(g);
    b->body(box);
    return b;
}

//+ WidgetKitImpl(WidgetKit::palette_button)
Button_ptr WidgetKitImpl::palette_button(XfGlyph_ptr g, XfAction_ptr a) {
    TelltaleImpl* t = new TelltaleImpl(
	TelltaleImpl::enabled_bit | TelltaleImpl::toggle_bit
    );
    Button_ptr b = new ButtonImpl(t, a);
    Glyph_var box = layouts_->hbox();
    box->append(Glyph_var(layouts_->hspace(Coord(3))));
    box->append(Glyph_var(new WKIndicator(&info_, t)));
    box->append(Glyph_var(layouts_->hspace(Coord(6))));
    box->append(g);
    b->body(
	Glyph_var(
	    new WKPushButtonFrame(box, t, &info_, button_border, Coord(0), Coord(0.5))
	)
    );
    return b;
}

//+ WidgetKitImpl(WidgetKit::radio_flag)
XfGlyph_ptr WidgetKitImpl::radio_flag(XfTelltale_ptr t) {
    Font_ptr f = info_.font;
    return new WKRadioFlag(&info_, t, f);
}

//+ WidgetKitImpl(WidgetKit::radio_button)
Button_ptr WidgetKitImpl::radio_button(XfGlyph_ptr g, XfAction_ptr a, XfTelltale_ptr group) {
    TelltaleImpl* t = new TelltaleImpl(
	TelltaleImpl::enabled_bit | TelltaleImpl::choosable_bit
    );
    t->current(group);
    Button_ptr b = new ButtonImpl(t, a);
    Glyph_var box = layouts_->hbox_align_elements(Coord(0.5));
    Font_ptr f = info_.font;
    box->append(Glyph_var(new WKRadioFlag(&info_, t, f)));
    box->append(Glyph_var(layouts_->hspace(Coord(6))));
    box->append(g);
    b->body(box);
    return b;
}

//+ WidgetKitImpl(WidgetKit::slider)
XfViewer_ptr WidgetKitImpl::slider(XfAxis a, Adjustment_ptr adj) {
    Slider* slider;
    switch (a) {
    case X_axis:
	slider = new XSlider(this, style_, adj);
	break;
    case Y_axis:
	slider = new YSlider(this, style_, adj);
	break;
    default:
	return nil;
    }
    make_thumb(slider, a);
    TelltaleImpl* t = new TelltaleImpl(TelltaleImpl::enabled_bit);
     Viewer_ptr v = new SliderViewer(
	slider,
	Glyph_var(new WKFrame(slider, t, &info_, info_.thickness))
     );
    return v;
}

//+ WidgetKitImpl(WidgetKit::scroll_bar)
XfViewer_ptr WidgetKitImpl::scroll_bar(XfAxis a, Adjustment_ptr adj) {
    Coord xspan, yspan;
    Glyph_var box, body;
    Viewer_var mover1, mover2;
    Slider* slider;
    Boolean hmargin, vmargin;
    switch (a) {
    case X_axis:
	xspan = info_.mover_size;
	yspan = info_.slider_size;
	box = layouts_->hbox();
	body = layouts_->vfixed(box, yspan);
	mover1 = left_mover(adj);
	slider = new XSlider(this, style_, adj);
	mover2 = right_mover(adj);
	hmargin = false;
	vmargin = true;
	break;
    case Y_axis:
	xspan = info_.slider_size;
	yspan = info_.mover_size;
	box = layouts_->vbox();
	body = layouts_->hfixed(box, xspan);
	mover1 = up_mover(adj);
	slider = new YSlider(this, style_, adj);
	mover2 = down_mover(adj);
	hmargin = true;
	vmargin = false;
	break;
    default:
	return nil;
    }

    make_thumb(slider, a);
    Viewer_ptr slider_viewer = new SliderViewer(
	slider,
	Glyph_var(new WKFrame(
	    slider,
	    new TelltaleImpl(TelltaleImpl::enabled_bit),
	    &info_, info_.thickness, Coord(0), Coord(0), hmargin, vmargin
	))
    );
    box->append(Glyph_var(layouts_->fixed(mover1, xspan, yspan)));
    box->append(slider_viewer);
    box->append(Glyph_var(layouts_->fixed(mover2, xspan, yspan)));
    Viewer_var v = viewer_group(body);
    mover1->append_viewer(v);
    slider_viewer->append_viewer(v);
    mover2->append_viewer(v);
    return v;
}

//+ WidgetKitImpl(WidgetKit::panner)
XfViewer_ptr WidgetKitImpl::panner(Adjustment_ptr x, Adjustment_ptr y) {
    StyleContext_var style = style_->new_style();
    style->alias(FrescoLib::tmp_string_ref("Panner"));
    style->link_parent(style_);
    Slider* s = new XYSlider(this, style, x, y);
    Viewer_ptr v = new SliderViewer(s, s);
    make_thumb(s, Z_axis);
    return v;
}

//+ WidgetKitImpl(WidgetKit::zoomer) !scale, !x, !y, !z
Button_ptr WidgetKitImpl::zoomer(XfCoord, Adjustment_ptr, Adjustment_ptr, Adjustment_ptr) {
    return nil;
}

//+ WidgetKitImpl(WidgetKit::up_mover)
Button_ptr WidgetKitImpl::up_mover(Adjustment_ptr a) {
    TelltaleImpl* t = new TelltaleImpl(TelltaleImpl::enabled_bit);
    Button_ptr b = new ForwardScroller(style_, t, a);
    b->body(
	Glyph_var(new WKButtonFrame(
	    new UpArrow(info_.color[dark_gray]), t, &info_, arrow_border
	))
    );
    return b;
}

//+ WidgetKitImpl(WidgetKit::down_mover)
Button_ptr WidgetKitImpl::down_mover(Adjustment_ptr a) {
    TelltaleImpl* t = new TelltaleImpl(TelltaleImpl::enabled_bit);
    Button_ptr b = new BackwardScroller(style_, t, a);
    b->body(
	Glyph_var(new WKButtonFrame(
	    new DownArrow(info_.color[dark_gray]), t, &info_, arrow_border
	))
    );
    return b;
}

//+ WidgetKitImpl(WidgetKit::left_mover)
Button_ptr WidgetKitImpl::left_mover(Adjustment_ptr a) {
    TelltaleImpl* t = new TelltaleImpl(TelltaleImpl::enabled_bit);
    Button_ptr b = new BackwardScroller(style_, t, a);
    b->body(
	Glyph_var(new WKButtonFrame(
	    new LeftArrow(info_.color[dark_gray]), t, &info_, arrow_border
	))
    );
    return b;

}

//+ WidgetKitImpl(WidgetKit::right_mover)
Button_ptr WidgetKitImpl::right_mover(Adjustment_ptr a) {
    TelltaleImpl* t = new TelltaleImpl(TelltaleImpl::enabled_bit);
    Button_ptr b = new ForwardScroller(style_, t, a);
    b->body(
	Glyph_var(new WKButtonFrame(
	    new RightArrow(info_.color[dark_gray]), t, &info_, arrow_border
	))
    );
    return b;
}

//+ WidgetKitImpl(WidgetKit::viewer_group)
XfViewer_ptr WidgetKitImpl::viewer_group(XfGlyph_ptr g) {
    Viewer_ptr v = new ViewerImpl(true);
    if (is_not_nil(g)) {
	v->body(g);
    }
    return v;
}

//+ WidgetKitImpl(WidgetKit::dialog_kit)
DialogKit_ptr WidgetKitImpl::dialog_kit() {
    if (is_nil(dialogs_)) {
        dialogs_ = new DialogKitImpl(this);
    }
    return dialogs_;
}

//+ WidgetKitImpl(WidgetKit::menu_kit)
MenuKit_ptr WidgetKitImpl::menu_kit() {
    if (is_nil(menus_)) {
        menus_ = new MenuKitImpl(this);
    }
    return menus_;
}

//+ WidgetKitImpl(WidgetKit::bounded_float)
Adjustment_ptr WidgetKitImpl::bounded_float(Float lower, Float upper, Float initial) {
    return new BoundedFloat(lower, upper, initial);
}

Color_ptr WidgetKitImpl::brightness(Color_ptr c, float adjust) {
    Color::Intensity r, g, b;
    c->rgb(r, g, b);
    if (adjust >= 0) {
	r += (1 - r) * adjust;
	g += (1 - g) * adjust;
	b += (1 - b) * adjust;
    } else {
	float f = adjust + 1.0f;
	r *= f;
	g *= f;
	b *= f;
    }
    return drawing()->color_rgb(r, g, b);
}

void WidgetKitImpl::load() {
    WidgetKitImpl::Info* i = &info_;
    load_coord("frameThickness", i->thickness);
    load_float("checkScale", i->check_size);
    load_float("radioScale", i->radio_scale);
    load_coord("moverSize", i->mover_size);
    load_coord("sliderSize", i->slider_size);
    load_coord("minButtonWidth", i->min_button_width);

    DrawingKit_ptr d = drawing();
    i->font = d->default_font(style_);
    i->num_colors = 13;
    Color_var* c = new Color_var[i->num_colors];
    i->color = c;
    c[light_gray] = d->background(style_);
    if (is_nil(c[light_gray])) {
	c[light_gray] = d->color_rgb(Color::Intensity(0.7), Color::Intensity(0.7), Color::Intensity(0.7));
    }
    Color_ptr bg = c[light_gray];
    c[black] = brightness(bg, -0.85f);
    c[very_dark_gray] = brightness(bg, -0.66f);
    c[dark_gray] = brightness(bg, -0.5f);
    c[medium_gray] = brightness(bg, -0.33f);
    c[very_light_gray] = brightness(bg, 0.45f);
    c[white] = brightness(bg, 0.7f);
    c[yellow] = d->color_rgb(Color::Intensity(1), Color::Intensity(1), Color::Intensity(0));
    c[light_yellow] = d->color_rgb(Color::Intensity(1), Color::Intensity(1), Color::Intensity(0.875));
    c[medium_yellow] = brightness(i->color[yellow], -0.3f);
    c[dark_yellow] = brightness(i->color[yellow], -0.5f);

    i->checkmark_color = d->resolve_color(
	style_, FrescoLib::tmp_string_ref("checkmarkColor")
    );
    if (is_nil(i->checkmark_color)) {
	i->checkmark_color = d->color_rgb(Color::Intensity(0.9), Color::Intensity(0), Color::Intensity(0));
    }
    i->checkmark = make_bitmap(
	d, checkmark_bits, checkmark_width, checkmark_height
    );
    i->shadow1 = make_bitmap(
	d, shadow1_bits, shadow1_width, shadow1_height
    );
    i->shadow2 = make_bitmap(
	d, shadow2_bits, shadow2_width, shadow2_height
    );
}

Raster_ptr WidgetKitImpl::make_bitmap(
    DrawingKit_ptr d, unsigned char* data,
    unsigned int width, unsigned int height
) {
    Long n = height * ((width + 7) / 8);
    DrawingKit::Data8 seq(n, n, data);
    return d->bitmap_from_data(seq, height, width, 0, 0);
}

void WidgetKitImpl::load_coord(const char* name, Coord& c) {
    StyleValue_var a = style_->resolve(FrescoLib::tmp_string_ref(name));
    if (is_not_nil(a)) {
	a->read_coord(c);
    }
}

void WidgetKitImpl::load_float(const char* name, float& f) {
    StyleValue_var a = style_->resolve(FrescoLib::tmp_string_ref(name));
    if (is_not_nil(a)) {
	double d;
	if (a->read_real(d)) {
	    f = float(d);
	}
    }
}

void WidgetKitImpl::unload() {
    delete [] info_.color;
}

// class WKFrame

WKFrame::WKFrame(
    Glyph_ptr g, Telltale_ptr t, WidgetKitImpl::Info* info, Coord thickness,
    float xalign, float yalign, Boolean hmargin, Boolean vmargin
) : Beveler(thickness, xalign, yalign, hmargin, vmargin) {
    body(g);
    state_ = Telltale::_duplicate(t);
    info_ = info;
}

WKFrame::~WKFrame() { }

//+ WKFrame(XfGlyph::draw)
void WKFrame::draw(XfGlyphTraversal_ptr t) {
    draw_frame(t);
    if (!state_->test(Telltale::enabled)) {
	Coord th = info_->thickness;
	Painter_var p = t->current_painter();
	Vertex v[3];
	t->bounds(v[0], v[1], v[2]);
	v[0].x += th; v[0].y += th;
	v[1].x -= th; v[1].y -= th;
	// need fill patterns or alpha blending here
	fill_rect(p, v[0], v[1], info_->color[light_gray]);
    }
}

static int inset_colors[] = {
    dark_gray, very_dark_gray, none, very_light_gray, white
};

static int bright_inset_colors[] = {
   dark_gray, very_light_gray, white
};

static int outset_colors[] = {
    black, white, none, medium_gray, black
};

static int trough_colors[] = {
    dark_gray, very_light_gray, light_gray, dark_gray, very_dark_gray
};

static int visible_trough_colors[] = {
    medium_gray, white, very_light_gray, medium_gray, dark_gray
};

static int old_thumb_colors[] = {
    very_dark_gray, dark_gray, medium_gray, light_gray, very_light_gray
};

static int* frame_colors[] = {
    /* 0 */ nil,
    /* is_enabled */ trough_colors,
    /* is_visible */ nil,
    /* is_enabled_visible */ visible_trough_colors,
    /* is_active */ nil,
    /* is_enabled_active */ outset_colors,
    /* is_visible_active */ nil,
    /* is_enabled_visible_active */ bright_inset_colors,
    /* is_chosen */ nil,
    /* is_enabled_chosen */ inset_colors,
    /* is_visible_chosen */ nil,
    /* is_enabled_visible_chosen */ nil,
    /* is_active_chosen */ nil,
    /* is_enabled_active_chosen */ old_thumb_colors,
    /* is_visible_active_chosen */ nil,
    /* is_enabled_visible_active_chosen */ nil
};

void WKFrame::draw_frame(GlyphTraversal_ptr t) {
    Painter_var p = t->current_painter();
    Vertex v[3];
    t->bounds(v[0], v[1], v[2]);
    int* colors = frame_colors[TelltaleImpl::flags_of(state_)];
    if (colors == nil) {
	// fill_rect(p, v[0], v[1], info_->color[light_gray]);
    } else {
	int ncolors;
	Coord thickness = p->to_pixels_coord(info_->thickness, X_axis);
	Coord th[2];
	if (colors == bright_inset_colors) {
	    ncolors = 3;
	    th[0] = thickness;
	} else {
	    ncolors = 5;
	    Coord tt = Coord(0.5) * thickness;
	    th[0] = tt;
	    th[1] = tt;
	}
	shade(p, v[0], v[1], info_, colors, ncolors, th);
    }
}

void WKFrame::fill_rect(
    Painter_ptr p, const Vertex& lower, const Vertex& upper, Color_ptr c
) {
    Beveler::set_color(p, c);
    p->fill_rect(lower.x, lower.y, upper.x, upper.y);
}

void WKFrame::shade(
    Painter_ptr p, const Vertex& lower, const Vertex& upper,
    WidgetKitImpl::Info* info, const int* colors, int ncolors, Coord* th
) {
    Coord x0 = lower.x, y0 = lower.y, x1 = upper.x, y1 = upper.y;
    int nbands = (ncolors - 1) >> 1;
    int b = nbands - 1;
    int n = ncolors - 1;
    for (int i = 0; i < b; i++) {
	Coord t1 = th[i];
	Beveler::rect(
	    p, t1, info->color[colors[i]], nil, info->color[colors[n - i]],
	    x0, y0, x1, y1
	);
	x0 += t1; y0 += t1;
	x1 -= t1; y1 -= t1;
    }
    Beveler::rect(
	p, th[b], info->color[colors[b]], info->color[colors[nbands]],
	info->color[colors[nbands + 1]], x0, y0, x1, y1
    );
}

// class WKHiliteFrame

WKHiliteFrame::WKHiliteFrame(
    Glyph_ptr g, Telltale_ptr t, WidgetKitImpl::Info* info, Coord thickness,
    float xalign, float yalign
) : WKFrame(g, t, info, thickness, xalign, yalign, true, true) { }

WKHiliteFrame::~WKHiliteFrame() { }

static int visible_hilite_colors[] = {
    dark_gray, white, white,
    very_light_gray,
    light_gray, dark_gray, very_dark_gray
};

//+ WKHiliteFrame(XfGlyph::draw)
void WKHiliteFrame::draw(XfGlyphTraversal_ptr t) {
    Painter_var p = t->current_painter();
    Vertex v[3];
    t->bounds(v[0], v[1], v[2]);
    if (state_->test(Telltale::enabled)) {
        if (state_->test(Telltale::visible)) {
            Coord t1 = p->to_pixels_coord(Coord(1), X_axis);
            Coord th[3];
            th[0] = t1; th[1] = t1; th[2] = t1 + t1;
            WKFrame::shade(
	        p, v[0], v[1], info_, visible_hilite_colors, 7, th
            );
        } else {
	    fill_rect(p, v[0], v[1], info_->color[light_gray]);
        }
    } else {
	// need fill patterns or alpha blending here
	fill_rect(p, v[0], v[1], info_->color[medium_gray]);
    }
}

// class WKButtonFrame

WKButtonFrame::WKButtonFrame(
    Glyph_ptr g, TelltaleImpl* t, WidgetKitImpl::Info* info, Coord thickness,
    float xalign, float yalign
) : WKFrame(g, t, info, thickness, xalign, yalign, true, true) { }

WKButtonFrame::~WKButtonFrame() { }

static int disabled_button_colors[] = {
    medium_gray, light_gray, light_gray, light_gray,
    light_gray, light_gray, dark_gray
};

static int enabled_button_colors[] = {
    dark_gray, white, very_light_gray, light_gray,
    medium_gray, dark_gray, very_dark_gray
};

static int visible_button_colors[] = {
    dark_gray, white, white, very_light_gray,
    light_gray, dark_gray, very_dark_gray
};

static int active_button_colors[] = {
    very_dark_gray, dark_gray, very_light_gray, light_gray,
    light_gray, black, white
};

static int visible_active_button_colors[] = {
    very_dark_gray, dark_gray, white, very_light_gray,
    light_gray, black, white
};

static int* button_colors[] = {
    /* 0 */ disabled_button_colors,
    /* is_enabled */ enabled_button_colors,
    /* is_visible */ disabled_button_colors,
    /* is_enabled_visible */ visible_button_colors,
    /* is_active */ disabled_button_colors,
    /* is_enabled_active */ active_button_colors,
    /* is_visible_active */ disabled_button_colors,
    /* is_enabled_visible_active */ visible_active_button_colors,
    /* is_chosen */ disabled_button_colors,
    /* is_enabled_chosen */ enabled_button_colors,
    /* is_visible_chosen */ disabled_button_colors,
    /* is_enabled_visible_chosen */ visible_button_colors,
    /* is_active_chosen */ disabled_button_colors,
    /* is_enabled_active_chosen */ active_button_colors,
    /* is_visible_active_chosen */ disabled_button_colors,
    /* is_enabled_visible_active_chosen */ visible_active_button_colors
};

void WKButtonFrame::draw_frame(GlyphTraversal_ptr t) {
    Painter_var p = t->current_painter();
    Vertex v[3];
    t->bounds(v[0], v[1], v[2]);
    Coord t1 = p->to_pixels_coord(Coord(1), X_axis);
    Coord th[3];
    th[0] = t1; th[1] = t1; th[2] = t1 + t1;
    WKFrame::shade(
	p, v[0], v[1], info_, button_colors[TelltaleImpl::flags_of(state_)], 7, th
    );
}

// class WKPushButtonFrame

WKPushButtonFrame::WKPushButtonFrame(
    Glyph_ptr g, TelltaleImpl* t, WidgetKitImpl::Info* info, Coord thickness,
    float xalign, float yalign
) : WKButtonFrame(g, t, info, thickness, xalign, yalign) { }

WKPushButtonFrame::~WKPushButtonFrame() { }

//+ WKPushButtonFrame(XfGlyph::request)
void WKPushButtonFrame::request(XfGlyph::Requisition& r) {
    WKButtonFrame::request(r);
    if (r.x.natural < info_->min_button_width) {
	r.x.natural = info_->min_button_width;
	if (r.x.maximum < r.x.natural) {
	    r.x.maximum = r.x.natural;
	}
    }
    r.x.natural += button_border;
    r.x.maximum += button_border;
    r.x.minimum += button_border;
    r.y.natural += button_border;
    r.y.maximum += button_border;
    r.y.minimum += button_border;
}

// class WKGlyph

WKGlyph::WKGlyph(WidgetKitImpl::Info* info) {
    info_ = info;
}

WKGlyph::~WKGlyph() { }

// class WKFiller

WKFiller::WKFiller(
    WidgetKitImpl::Info* i, Color_ptr c
) : WKGlyph(i) {
    color_ = Color::_duplicate(c);
}

WKFiller::~WKFiller() {
    CORBA::release(color_);
}

//+ WKFiller(XfGlyph::request)
void WKFiller::request(XfGlyph::Requisition& r) {
    GlyphImpl::require(r.x, Coord(0), infinite_coord, Coord(0), Alignment(0));
    GlyphImpl::require(r.y, Coord(0), infinite_coord, Coord(0), Alignment(0));
}

//+ WKFiller(XfGlyph::draw)
void WKFiller::draw(XfGlyphTraversal_ptr t) {
    Painter_var p = t->current_painter();
    Vertex lower, upper, origin;
    t->bounds(lower, upper, origin);
    p->current_color(color_);
    p->fill_rect(lower.x, lower.y, upper.x, upper.y);
}

// class WKHSeparator
 
WKHSeparator::WKHSeparator(WidgetKitImpl::Info* i) : WKGlyph(i) { }
 
WKHSeparator::~WKHSeparator() { }
 
//+ WKHSeparator(XfGlyph::request)
void WKHSeparator::request(XfGlyph::Requisition& r) {
    r.x.defined = true;
    r.x.natural = r.x.maximum = r.x.minimum = info_->thickness;
    r.x.align = Alignment(0.5);
}

//+ WKHSeparator(XfGlyph::draw)
void WKHSeparator::draw(XfGlyphTraversal_ptr t) {
    Painter_var p = t->current_painter();
    Vertex v[3];
    t->bounds(v[0], v[1], v[2]);
    Coord bs = v[0].y;
    Coord ts = v[1].y;
    Coord ls = p->to_pixels_coord(v[1].x, X_axis);
    Coord d = p->to_pixels_coord((v[1].x - v[0].x) / Coord(2), X_axis);
    Coord m = ls + d;
    Coord rs = m + d;
    p->current_color(info_->color[dark_gray]);
    p->fill_rect(ls, bs, m, ts);
    p->current_color(info_->color[white]);
    p->fill_rect(m, bs, rs, ts);
}

// class WKVSeparator
 
WKVSeparator::WKVSeparator(WidgetKitImpl::Info* i) : WKGlyph(i) { }
 
WKVSeparator::~WKVSeparator() { }
 
//+ WKVSeparator(XfGlyph::request)
void WKVSeparator::request(XfGlyph::Requisition& r) {
    r.y.defined = true;
    r.y.natural = r.y.maximum = r.y.minimum = info_->thickness;
    r.y.align = Alignment(0.5);
}
 
//+ WKVSeparator(XfGlyph::draw)
void WKVSeparator::draw(XfGlyphTraversal_ptr t) {
    Painter_var p = t->current_painter();
    Vertex v[3];
    t->bounds(v[0], v[1], v[2]);
    Coord ls = v[0].x;
    Coord rs = v[1].x;
    Coord bs = p->to_pixels_coord(v[1].y, Y_axis);
    Coord d = p->to_pixels_coord((v[1].y - v[0].y) / Coord(2), Y_axis);
    Coord m = bs + d;
    Coord ts = m + d;
    p->current_color(info_->color[dark_gray]);
    p->fill_rect(ls, ts, rs, m);
    p->current_color(info_->color[white]);
    p->fill_rect(ls, m, rs, bs);
}

// class WKCheckmark

WKCheckmark::WKCheckmark(
    WidgetKitImpl::Info* i, Telltale_ptr t
) : WKGlyph(i) {
    state_ = Telltale::_duplicate(t);
}

WKCheckmark::~WKCheckmark() { }

//+ WKCheckmark(XfGlyph::request)
void WKCheckmark::request(XfGlyph::Requisition& r) {
    WidgetKitImpl::Info* i = info_;
    Font::Info b;
    i->font->font_info(b);
    Coord h = b.font_ascent + b.font_descent;
    Coord size = h * i->check_size;
    Alignment a = (h == 0) ? 0 : b.font_descent / h;
    GlyphImpl::require(r.x, size, Coord(0), Coord(0), Alignment(0));
    GlyphImpl::require(r.y, size, Coord(0), Coord(0), a);
}

//+ WKCheckmark(XfGlyph::draw)
void WKCheckmark::draw(XfGlyphTraversal_ptr t) {
    if (state_->test(Telltale::chosen)) {
	Painter_var p = t->current_painter();
	Region_var a = t->allocation();
	Vertex o;
	a->origin(o);
	WidgetKitImpl::Info* i = info_;
	Beveler::set_color(p, i->color[dark_gray]);
	p->stencil(i->shadow1, o.x, o.y);
	p->stencil(i->shadow2, o.x, o.y);
	Beveler::set_color(p, i->checkmark_color);
	p->stencil(i->checkmark, o.x, o.y);
    }
}

// class WKIndicator

WKIndicator::WKIndicator(
    WidgetKitImpl::Info* i, TelltaleImpl* t
) : WKGlyph(i) {
    state_ = t;
}

WKIndicator::~WKIndicator() {
    CORBA::release(Telltale_ptr(state_));
}

//+ WKIndicator(XfGlyph::request)
void WKIndicator::request(XfGlyph::Requisition& r) {
    GlyphImpl::require(r.x, Coord(6), Coord(0), Coord(0), Alignment(0));
}

static int disabled_indicator_colors[] = {
    dark_gray, medium_gray, white
};

static int disabled_chosen_indicator_colors[] = {
    dark_gray, light_yellow, white
};

static int enabled_indicator_colors[] = {
    black, dark_yellow, white
};

static int active_indicator_colors[] = {
    black, yellow, white
};

static int visible_indicator_colors[] = {
    black, medium_yellow, white
};

static int* indicator_colors[] = {
    /* 0 */ disabled_indicator_colors,
    /* is_enabled */ enabled_indicator_colors,
    /* is_visible */ disabled_indicator_colors,
    /* is_enabled_visible */ visible_indicator_colors,
    /* is_active */ disabled_indicator_colors,
    /* is_enabled_active */ enabled_indicator_colors,
    /* is_visible_active */ disabled_indicator_colors,
    /* is_enabled_visible_active */ visible_indicator_colors,
    /* is_chosen */ disabled_chosen_indicator_colors,
    /* is_enabled_chosen */ active_indicator_colors,
    /* is_visible_chosen */ disabled_chosen_indicator_colors,
    /* is_enabled_visible_chosen */ active_indicator_colors,
    /* is_active_chosen */ disabled_chosen_indicator_colors,
    /* is_enabled_active_chosen */ active_indicator_colors,
    /* is_visible_active_chosen */ disabled_chosen_indicator_colors,
    /* is_enabled_visible_active_chosen */ active_indicator_colors
};

//+ WKIndicator(XfGlyph::draw)
void WKIndicator::draw(XfGlyphTraversal_ptr t) {
    WidgetKitImpl::Info* i = info_;
    Painter_var p = t->current_painter();
    int* c = indicator_colors[state_->flags()];
    Vertex v0, v1, v;
    t->bounds(v0, v1, v);
    Coord h = p->to_pixels_coord(Coord(0.2) * (v1.y - v0.y), X_axis);
    Coord t1 = p->to_pixels_coord(Coord(1), X_axis);
    Coord t2 = t1 + t1;
    if (h < t2) {
	h = t2;
    }
    Beveler::rect(
	p, t1, i->color[c[0]], i->color[c[1]], i->color[c[2]],
	v0.x, v0.y + h, v1.x, v1.y - h
    );
}

// class WKRadioFlag

WKRadioFlag::WKRadioFlag(
    WidgetKitImpl::Info* i, Telltale_ptr t, Font_ptr f
) : WKGlyph(i) {
    state_ = Telltale::_duplicate(t);
    font_ = Font::_duplicate(f);
}

WKRadioFlag::~WKRadioFlag() {
    CORBA::release(font_);
}

//+ WKRadioFlag(XfGlyph::request)
void WKRadioFlag::request(XfGlyph::Requisition& r) {
    Font::Info b;
    font_->char_info('M', b);
    Coord size = b.width * info_->radio_scale;
    GlyphImpl::require(r.x, size, Coord(0), Coord(0), Alignment(0));
    GlyphImpl::require(r.y, size, Coord(0), Coord(0), Alignment(0));
}

static int outset_radio_colors[] = {
    very_light_gray, light_gray, dark_gray
};

static int visible_outset_radio_colors[] = {
    white, very_light_gray, medium_gray
};

static int inset_radio_colors[] = {
    dark_gray, light_gray, very_light_gray
};

static int visible_inset_radio_colors[] = {
    medium_gray, very_light_gray, white
};
    
static int* radio_colors[] = {
    /* 0 */ outset_radio_colors,
    /* is_enabled */ outset_radio_colors,
    /* is_visible */ outset_radio_colors,
    /* is_enabled_visible */ visible_outset_radio_colors,
    /* is_active */ outset_radio_colors,
    /* is_enabled_active */ inset_radio_colors,
    /* is_visible_active */ outset_radio_colors,
    /* is_enabled_visible_active */ visible_inset_radio_colors,
    /* is_chosen */ inset_radio_colors,
    /* is_enabled_chosen */ inset_radio_colors,
    /* is_visible_chosen */ inset_radio_colors,
    /* is_enabled_visible_chosen */ visible_inset_radio_colors,
    /* is_active_chosen */ inset_radio_colors,
    /* is_enabled_active_chosen */ inset_radio_colors,
    /* is_visible_active_chosen */ inset_radio_colors,
    /* is_enabled_visible_active_chosen */ visible_inset_radio_colors
};

//+ WKRadioFlag(XfGlyph::draw)
void WKRadioFlag::draw(XfGlyphTraversal_ptr t) {
    WidgetKitImpl::Info* i = info_;
    Painter_var p = t->current_painter();
    int* colors = radio_colors[TelltaleImpl::flags_of(state_)];
    Vertex v0, v1, v;
    t->bounds(v0, v1, v);
    Beveler::diamond(
	p, i->thickness,
	i->color[colors[0]], i->color[colors[1]], i->color[colors[2]],
	v0.x, v0.y, v1.x, v1.y
    );
}

// class WKRadioItem

WKRadioItem::WKRadioItem(
    WidgetKitImpl::Info* i, TelltaleImpl* t, Font_ptr f
) : WKRadioFlag(i, t, f) { }

WKRadioItem::~WKRadioItem() { }

//+ WKRadioItem(XfGlyph::draw)
void WKRadioItem::draw(XfGlyphTraversal_ptr t) {
    if (state_->test(Telltale::chosen)) {
	WKRadioFlag::draw(t);
    }
}

// class WKDefaultArrow

WKDefaultArrow::WKDefaultArrow(
    WidgetKitImpl::Info* i, Font_ptr f
) : WKGlyph(i) {
    Font::Info b;
    f->char_info('M', b);
    width_ = b.width;
    height_ = b.font_ascent + b.font_descent;
    ascent_ = b.ascent;
    align_ = (height_ == 0) ? 0 : b.font_descent / height_;
}

WKDefaultArrow::~WKDefaultArrow() { }

//+ WKDefaultArrow(XfGlyph::request)
void WKDefaultArrow::request(XfGlyph::Requisition& r) {
    GlyphImpl::require(r.x, Coord(1.25) * width_, Coord(0), Coord(0), Alignment(0));
    GlyphImpl::require(r.y, height_, Coord(0), Coord(0), align_);
}

//+ WKDefaultArrow(XfGlyph::draw)
void WKDefaultArrow::draw(XfGlyphTraversal_ptr t) {
    WidgetKitImpl::Info* i = info_;
    Vertex v0, v1, v;
    t->bounds(v0, v1, v);

    Coord left = v0.x, right = v1.x;
    Coord x0 = left + Coord(0.4) * (right - left);
    Coord bottom = v.y;
    Coord top = bottom + Coord(0.9) * ascent_;
    Coord y0 = Coord(0.5) * (bottom + top);

    Painter_var p = t->current_painter();
    Beveler::set_color(p, i->color[black]);
    p->begin_path();
    p->move_to(left, y0);
    p->line_to(x0, top);
    p->line_to(x0, bottom);
    p->close_path();
    p->fill();

    Coord th = i->thickness;
    y0 -= Coord(0.5) * th;
    Coord y1 = y0 + th;
    p->fill_rect(x0, y0, right, y1);
    p->fill_rect(right - th, y1, right, top);
}

void WidgetKitImpl::make_thumb(Slider* slider, Axis a) {
    long r;
    StyleValue_var v = style_->resolve(
	FrescoLib::tmp_string_ref("thumbRidges")
    );
    if (is_not_nil(v)) {
	Long l;
	if (v->read_integer(l)) {
	    r = l;
	}
    }
    slider->normal_thumb(
	new WKThumb(&info_, a, r, new TelltaleImpl(TelltaleImpl::enabled_bit))
    );
    slider->visible_thumb(
	new WKThumb(&info_, a, r, new TelltaleImpl(TelltaleImpl::visible_bit))
    );
    TelltaleImpl* t = new TelltaleImpl(
	TelltaleImpl::enabled_bit | TelltaleImpl::active_bit |
	TelltaleImpl::chosen_bit
    );
    slider->old_thumb(new WKFrame(nil, t, &info_, info_.thickness));
}

// class WKThumb

WKThumb::WKThumb(
    WidgetKitImpl::Info* info, Axis a, long ridges, TelltaleImpl* t
) : WKGlyph(info) {
    axis_ = a;
    ridges_ = ridges;
    state_ = t;
}

WKThumb::~WKThumb() {
    CORBA::release(Telltale_ptr(state_));
}

static int thumb_colors[] = {
    very_light_gray, very_light_gray, light_gray, medium_gray, dark_gray
};

static int visible_thumb_colors[] = {
    white, white, very_light_gray, light_gray, dark_gray
};

//+ WKThumb(XfGlyph::draw)
void WKThumb::draw(XfGlyphTraversal_ptr t) {
    WidgetKitImpl::Info* i = info_;
    int* colors = (
	state_->test(Telltale::visible) ? visible_thumb_colors : thumb_colors
    );
    Color_ptr c[6];
    for (int c_i = 0; c_i < 5; c_i++) {
	c[c_i] = i->color[colors[c_i]];
    }
    Vertex v0, v1, v;
    t->bounds(v0, v1, v);
    Painter_var p = t->current_painter();
    Coord p1 = p->to_pixels_coord(Coord(1), X_axis);
    Coord p2 = p1 + p1, p3 = p2 + p1, p4 = p3 + p1, p5 = p4 + p1;
    Beveler::set_color(p, i->color[very_dark_gray]);
    switch (axis_) {
    case X_axis:
	p->fill_rect(v0.x, v0.y, v0.x + p1, v1.y);
	p->fill_rect(v1.x - p1, v0.y, v1.x, v1.y);
	v0.x += p1;
	v1.x -= p1;
	break;
    case Y_axis:
	p->fill_rect(v0.x, v1.y - p1, v1.x, v1.y);
	p->fill_rect(v0.x, v0.y, v1.x, v0.y + p1);
	v0.y += p1;
	v1.y -= p1;
	break;
    }
    Beveler::rect(p, p1, c[0], nil, c[4], v0.x, v0.y, v1.x, v1.y);
    Beveler::rect(
	p, p1, c[1], c[2], c[3], v0.x + p1, v0.y + p1, v1.x - p1, v1.y - p1
    );

    Coord mid;
    Coord left[6], bottom[6], right[6], top[6];
    long r;
    switch (axis_) {
    case X_axis:
	mid = p->to_pixels_coord(Coord(0.5) * (v0.x + v1.x), X_axis);
	v0.y += p1; v1.y -= p1;
	c[0] = i->color[very_dark_gray];
	c[1] = i->color[white];
	for (r = 0; r < 6; r++) {
	    bottom[r] = v0.y; top[r] = v1.y;
	}
	left[0] = mid + p4; right[0] = mid + p5;
	left[1] = mid + p3; right[1] = mid + p4;
	left[2] = mid; right[2] = mid + p1;
	left[3] = mid - p1; right[3] = mid;
	left[4] = mid - p4; right[4] = mid - p3;
	left[5] = mid - p5; right[5] = mid - p4;
	break;
    case Y_axis:
	mid = p->to_pixels_coord(Coord(0.5) * (v0.y + v1.y), Y_axis);
	v0.x += p1; v1.x -= p1;
	c[0] = i->color[white];
	c[1] = i->color[very_dark_gray];
	for (r = 0; r < 6; r++) {
	    left[r] = v0.x; right[r] = v1.x;
	}
	bottom[0] = mid + p4; top[0] = mid + p5;
	bottom[1] = mid + p3; top[1] = mid + p4;
	bottom[2] = mid; top[2] = mid + p1;
	bottom[3] = mid - p1; top[3] = mid;
	bottom[4] = mid - p4; top[4] = mid - p3;
	bottom[5] = mid - p5; top[5] = mid - p4;
	break;
    default:
	return;
    }
    c[2] = c[0];
    c[3] = c[1];
    c[4] = c[0];
    c[5] = c[1];
    for (r = 3 - ridges_; r < 3 + ridges_; r++) {
	Beveler::set_color(p, c[r]);
	p->fill_rect(left[r], bottom[r], right[r], top[r]);
    }
}
