/*
 * Copyright (c) 1994-95 Firefly Design
 * Copyright (c) 1994 Silicon Graphics, Inc.
 * Copyright (c) 1994 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THIS SOFTWARE.
 */

#include <ctype.h>
#include <string.h>
#include <X11/keysym.h>
#include <Fresco/defs/drawing.h>
#include <Fresco/defs/fresco.h>
#include <Fresco/defs_impl/charstr.h>
#include <Fresco/defs_impl/glyphs.h>
#include <Fresco/defs_impl/math.h>
#include <Fresco/defs_impl/names.h>
#include <Fresco/figures/figures.h>
#include <Fresco/layouts/layouts.h>
#include <Fresco/widgets/menus.h>
#include <Fresco/widgets/widgets.h>
#include "items.h"
#include "menu_kit.h"

static CharCode char_code_to_lower(CharCode c) {
    return (CharCode) tolower((int) c);
}

// class MKLabel

class MKLabel : public GlyphImpl {
public:
    MKLabel(FigureStyle_ptr, CharString_ptr, Long index = -1);
    ~MKLabel();
 
    void request(Glyph::Requisition& r);
    void draw(GlyphTraversal_ptr t);

protected:
    FigureStyle_ptr style_;
    CharString_var text_;
    Long index_;
    Font::Info info_;
};

MKLabel::MKLabel(FigureStyle_ptr s, CharString_ptr str, Long index) {
    style_ = FigureStyle::_duplicate(s);
    text_ = CharString::_duplicate(str);
    index_ = index;
    Font_var font = style_->font_attr();
    font->string_info(text_, info_);
}
 
MKLabel::~MKLabel() {
    CORBA::release(style_);
}
 
void MKLabel::request(Glyph::Requisition& r) {
    Coord h = info_.font_ascent + info_.font_descent;
    Alignment a;
    if (Math::equal(h, Coord(0), Coord(1e-2))) {
        a = Alignment(0);
    } else {
        a = info_.font_descent / h;
    }
    GlyphImpl::require(r.x, info_.width, Coord(0), Coord(0), Alignment(0));
    GlyphImpl::require(r.y, h, Coord(0), Coord(0), a);
}
 
void MKLabel::draw(GlyphTraversal_ptr t) {
    Painter_var p = t->current_painter();
    Region_var a = t->allocation();

    Color_var foreground = style_->foreground();
    p->current_color(foreground);
//---     Brush_var b = style_->brush_attr();
//---     p->current_brush(b);   $$$ FigStyleImpl is returning nil brush. (SC)
    Font_var f = style_->font_attr();
    p->current_font(f);

    Vertex v;
    a->origin(v);
    Font::Info info;
    CharStringBuffer buf(text_);
    const char* cur = buf.string();
    const char* end = &cur[buf.length()];
    v.y = p->to_pixels_coord(v.y, Y_axis);
    Coord line_x0, line_x1;
    for (Long i = 0; cur < end; cur++, ++i) {
        f->char_info(*cur, info);
        if (index_ == i) {
            line_x0 = v.x;
            line_x1 = v.x + info.width;
        }
        p->character(*cur, info.width, v.x, v.y);
        v.x += info.width;
    }
    if (index_ >= 0) {
        p->current_brush(nil);
        Coord y = p->to_pixels_coord(v.y - Coord(1), Y_axis);
        p->line(line_x0, y, line_x1, y);
    }
}

// class MenuKitImpl

#define arrow_width 16
#define arrow_height 16
static unsigned char arrow_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x0c, 0x00, 0x1c, 0x00, 0x38,
   0x00, 0x70, 0xff, 0xff, 0xff, 0xff, 0x00, 0x70, 0x00, 0x38, 0x00, 0x1c,
   0x00, 0x0c, 0x00, 0x04, 0x00, 0x00, 0x00, 0x00};

struct StyleData {
    const char* path;
    const char* value;
};

static StyleData kit_attributes[] = {
    { "*MenuKit*font", "-adobe-new century schoolbook-medium-r-normal--14-*" },
    { "*MenuKit*menuMargin", "0.0" },
    { "*MenuKit*delay", "0.1" },
//    { "*MenuKit*grab", "true" },
    { "*MenuKit*grab", "false" },
    { nil }
};

MenuKitImpl::MenuKitImpl(WidgetKit* widget) {
    FrescoContext* fc = FrescoLib::current_context();
    info_.drawing = fc->default_drawing_kit();
    info_.layout = LayoutKit::_narrow(
        fc->resolve(*NamingLib::to_name("layouts_impl"))
    );
    info_.figure = FigureKit::_narrow(
        fc->resolve(*NamingLib::to_name("figures_impl"))
    );
    info_.widget = widget;
 
    StyleContext_var s = fc->style();
    style_ = info_.figure->new_style(s);
    style_->alias(FrescoLib::tmp_string_ref("MenuKit"));
    for (StyleData* d = kit_attributes; d->path != nil; d++) {
        StyleValue_var a = s->bind(FrescoLib::tmp_string_ref(d->path));
        if (a->uninitialized() || a->priority() <= -10) {
            a->write_string(FrescoLib::tmp_string_ref(d->value));
            a->priority(-10);
        }
    }
 
    load();
}

MenuKitImpl::~MenuKitImpl() {
    unload();
}

OxExchange* MenuKitImpl::_exchange() { return &object_; }

//+ MenuKitImpl(XfFrescoObject::=object_.)
XfTag MenuKitImpl::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void MenuKitImpl::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void MenuKitImpl::disconnect() {
    object_.disconnect();
}
void MenuKitImpl::notify_observers() {
    object_.notify_observers();
}
void MenuKitImpl::update() {
    object_.update();
}
//+

Glyph_ptr MenuKitImpl::label(
    CharString_ptr name, CharCode mnemonic, Boolean dialog
) {
    // !!! assumes ascii -denis
    Glyph_ptr glyph;
    Long count = name->count();
    CharString::Data data(count + 6, count, new CharCode[count + 6]);
    name->get_data(data);
    if (dialog) {
        data[count] = '.';
        data[count + 1] = '.';
        data[count + 2] = '.';
        data.length(count += 3);
        name->put_data(data);
    }
    if (mnemonic != 0) {
        CharCode m = char_code_to_lower(mnemonic);
        for (Long i = 0; i < count; ++i) {
            if (char_code_to_lower(data[i]) == m) {
                break;
            }
        }
        if (i < count) {
            glyph = new MKLabel(style_, name, i);
        } else {
            data[count] = '(';
            data[count + 1] = mnemonic;
            data[count + 2] = ')';
            data.length(count + 3);
            name->put_data(data);
            glyph = new MKLabel(style_, name);
        }
    } else {
        glyph = new MKLabel(style_, name);
    }
    return glyph;
}

Glyph_ptr MenuKitImpl::shortcut(const Event::KeyChord& accelerator) {
    // !!! hard coded limits -denis
    char buffer[256];
    int length = 0;
    Long count = accelerator.length();
    for (Long i = 0; i < count; ++i) {
        Event::KeySym keysym = accelerator[i];
        const char* string = nil;
        if (i < (count - 1)) {
            switch (keysym) {
                case XK_Control_L: string = "Ctrl+"; break;
                case XK_Shift_L: string = "Shift+"; break;
                case XK_Caps_Lock: string = "Caps+"; break;
                case XK_Meta_L: string = "Meta+"; break;
                case XK_Alt_L: string = "Alt+"; break;
                default: "???+"; break;
            }
        } else {
            string = FrescoLib::keysym_to_string(keysym);
        }
        if (string == nil) {
            // !!! should this produce an error message -denis
            string = "???";
        }
        for (; *string; ++string) {
            buffer[length++] = *string;
        }
    }
    buffer[length] = '\0';
 
    if (length > 0) {
        return info_.figure->label(style_, FrescoLib::tmp_string_copy(buffer));
    }
 
    return nil;
}

Event::KeyChord MenuKitImpl::accelerator(CharString_ptr string) {
    Event::KeyChord keys(6);

    if (is_nil(string)) {
        return keys;
    }

    Long count = string->count();
    if (count == 0) {
        return keys;
    }

    CharStringBuffer b(string);
    char* buffer = b.string();
    
    Long last = 0;
    for (Long i = 0; i < count; ++i) {
        if (buffer[i] == '+') {
            Event::KeySym keysym = 0;
            char* start = buffer + last;
            unsigned int n = i - last;
            if (strncmp(start, "Ctrl", n) == 0) {
                keysym = XK_Control_L;
            } else if (strncmp(start, "Shift", n) == 0) {
                keysym = XK_Shift_L;
            } else if (strncmp(start, "Caps", n) == 0) {
                keysym = XK_Caps_Lock;
            } else if (strncmp(start, "Meta", n) == 0) {
                keysym = XK_Meta_L;
            } else if (strncmp(start, "Alt", n) == 0) {
                keysym = XK_Alt_L;
            }
            keys[keys.length()] = keysym;
            keys.length(keys.length() + 1);
            last = i + 1;
        }
    }

    keys[keys.length()] = FrescoLib::string_to_keysym(buffer + last);
    keys.length(keys.length() + 1);

    return keys;
}

MenuItem_ptr MenuKitImpl::menu_item(
    Glyph_ptr label_glyph, Event::KeySym mnemonic
) {
    return new MKItemMenu(&info_, label_glyph, mnemonic);
}

MenuItem_ptr MenuKitImpl::menu_item_string(
    CharString_ptr label_name, CharCode mnemonic
) {
    return new MKItemMenu(
        &info_, Glyph_var(label(label_name, mnemonic)), Event::KeySym(mnemonic)
    );
}

MenuItem_ptr MenuKitImpl::button_item(
    Glyph_ptr label_glyph, Glyph_ptr shortcut_glyph, Event::KeySym mnemonic,
    const Event::KeyChord& accelerator, Action_ptr action
) {
    return new MKItemButton(
        &info_, nil, label_glyph, shortcut_glyph, mnemonic, accelerator, action
    );
}

MenuItem_ptr MenuKitImpl::button_item_string(
    CharString_ptr label_name, CharCode mnemonic, CharString_ptr accelerator_name,
    Action_ptr action
) {
    Event::KeyChord keys = accelerator(accelerator_name);
    return new MKItemButton(
        &info_, nil, Glyph_var(label(label_name, mnemonic)),
        Glyph_var(shortcut(keys)), Event::KeySym(mnemonic), keys, action
    );
}

MenuItem_ptr MenuKitImpl::dialog_button_item_string(
    CharString_ptr label_name, CharCode mnemonic, CharString_ptr accelerator_name,
    Action_ptr action
) {
    Event::KeyChord keys = accelerator(accelerator_name);
    return new MKItemButton(
        &info_, nil, Glyph_var(label(label_name, mnemonic, true)),
        Glyph_var(shortcut(keys)), Event::KeySym(mnemonic), keys, action
    );
}

MenuItem_ptr MenuKitImpl::mark_item(
    Glyph_ptr mark_glyph, Glyph_ptr label_glyph, Glyph_ptr shortcut_glyph,
    Event::KeySym mnemonic, const Event::KeyChord& accelerator, Action_ptr action,
    Telltale_ptr telltale
) {
    return new MKItemButton(
        &info_, mark_glyph, label_glyph, shortcut_glyph,
        mnemonic, accelerator, action, telltale
    );
}

MenuItem_ptr MenuKitImpl::check_box_item(
    Glyph_ptr label_glyph, Glyph_ptr shortcut_glyph, Event::KeySym mnemonic,
    const Event::KeyChord& accelerator, Action_ptr action
) {
    WidgetKit_ptr widget = info_.widget;
    Telltale_var t = widget->telltale_group();
    Glyph_var checkmark = widget->checkmark(t);
    return new MKItemButton(
        &info_, checkmark, label_glyph, shortcut_glyph, mnemonic,
        accelerator, action, t
    );
}

MenuItem_ptr MenuKitImpl::check_box_item_string(
    CharString_ptr label_name, CharCode mnemonic, CharString_ptr accelerator_name,
    Action_ptr action
) {
    WidgetKit_ptr widget = info_.widget;
    Event::KeyChord keys = accelerator(accelerator_name);
    Telltale_var t = widget->telltale_group();
    Glyph_var checkmark = widget->checkmark(t);
    return new MKItemButton(
        &info_, checkmark, Glyph_var(label(label_name, mnemonic)),
        Glyph_var(shortcut(keys)),
        Event::KeySym(mnemonic), keys, action, t
    );
}

MenuItem_ptr MenuKitImpl::radio_item(
    Glyph_ptr label_glyph, Glyph_ptr shortcut_glyph, Event::KeySym mnemonic,
    const Event::KeyChord& accelerator, Action_ptr action
) {
    WidgetKit_ptr widget = info_.widget;
    Telltale_var t = widget->telltale_group();
    Glyph_var radio_flag = widget->radio_flag(t);
    return new MKItemButton(
        &info_, radio_flag, label_glyph, shortcut_glyph, mnemonic,
        accelerator, action, t
    );
}

MenuItem_ptr MenuKitImpl::radio_item_string(
    CharString_ptr label_name, CharCode mnemonic, CharString_ptr accelerator_name,
    Action_ptr action
) {
    WidgetKit_ptr widget = info_.widget;
    Event::KeyChord keys = accelerator(accelerator_name);
    Telltale_var t = widget->telltale_group();
    Glyph_var radio_flag = widget->radio_flag(t);
    return new MKItemButton(
        &info_, radio_flag, Glyph_var(label(label_name, mnemonic)),
        Glyph_var(shortcut(keys)),
        Event::KeySym(mnemonic), keys, action, t
    );
}

MenuItem_ptr MenuKitImpl::separator_item() {
    return new MKItemSeparator(&info_);
}

void MenuKitImpl::path_append(MenuItem::Path& path, MenuItem_ptr item) {
    if (path.length() >= path.maximum()) {
        unsigned long n = path.maximum() == 0 ? 10 : path.maximum() + path.maximum();
        MenuItem_ptr* buf = new MenuItem_ptr[n];
        for (unsigned long pi = 0; pi < path.length(); pi++) {
            buf[pi] = MenuItem::_duplicate(path[pi]);
        }
        for ( ; pi < n; ++pi) {
            buf[pi] = nil;
        }
        Long len = path.length();
        path.load(n, len, buf, true);
    }
 
    path[path.length()] = MenuItem::_duplicate(item);
    path.length(path.length() + 1);
}
 
void MenuKitImpl::path_reverse(MenuItem::Path& path) {
    ULong length = path.length();
    ULong count = length / 2;
    for (ULong i = 0; i < count; ++i) {
        Long other = length - 1 - i;
        MenuItem_field t = path[i];
        path[i] = path[other];
        path[other] = t;
    }
}
 
MenuItem_ptr MenuKitImpl::first_child(MenuItem_ptr item) {
    if (item->type() == MenuItem::submenu) {
        MenuItem_ptr default_item = item->default_item();
        if (
            is_not_nil(default_item) &&
            ((default_item->type() == MenuItem::submenu) || default_item->enabled())
        ) {
            return default_item;
        }
        Long count = item->count();
        for (Long i = 0; i < count; ++i) {
            MenuItem_ptr mi = item->item(i);
            if ((mi->type() == MenuItem::submenu) || mi->enabled()) {
                return mi;
            }
        }
    }
    return nil;
}

MenuItem_ptr MenuKitImpl::first_item(MenuItem_ptr item) {
    MenuItem_ptr child = first_child(item);
    return is_not_nil(child) ? child : item;
}

static Color_ptr brightness(DrawingKit_ptr drawing, Color_ptr c, float adjust) {
    Color::Intensity r, g, b;
    c->rgb(r, g, b);
    if (adjust >= 0) {
        r += (1 - r) * adjust;
        g += (1 - g) * adjust;
        b += (1 - b) * adjust;
    } else {
        float f = adjust + 1.0f;
        r *= f;
        g *= f;
        b *= f;
    }
    return drawing->color_rgb(r, g, b);
}

void MenuKitImpl::load() {
    load_coord("menuMargin", info_.menu_margin);
    load_float("delay", info_.delay);
    load_boolean("grab", info_.grab);

    LayoutKit_ptr layout = info_.layout;
    Glyph_ptr menu_shortcut;
    load_stencil("arrow", menu_shortcut, arrow_bits, arrow_width, arrow_height);
    info_.menu_shortcut = layout->valign(Glyph_var(menu_shortcut), Alignment(0));
	// !!! Better to use hbox_align_elements and no vcenter here.

    WidgetKit_ptr widget = info_.widget;
    info_.menu_separator = widget->vseparator();

    info_.menu_glue = layout->back(layout->hfil(), widget->filler());
}

void MenuKitImpl::load_coord(const char* name, Coord& c) {
    StyleValue_var a = style_->resolve(FrescoLib::tmp_string_ref(name));
    if (is_not_nil(a)) {
	a->read_coord(c);
    }
}

void MenuKitImpl::load_float(const char* name, float& f) {
    StyleValue_var a = style_->resolve(FrescoLib::tmp_string_ref(name));
    if (is_not_nil(a)) {
	double d;
	if (a->read_real(d)) {
	    f = float(d);
	}
    }
}

void MenuKitImpl::load_boolean(const char* name, Boolean& b) {
    StyleValue_var a = style_->resolve(FrescoLib::tmp_string_ref(name));
    if (is_not_nil(a)) {
	a->read_boolean(b);
    }
}

static Raster_ptr make_bitmap(
    DrawingKit_ptr d,
    unsigned char* data, unsigned int width, unsigned int height
) {
    Long n = height * ((width + 7) / 8);
    DrawingKit::Data8 seq(n, n, data);
    return d->bitmap_from_data(seq, height, width, 0, 0);
}

void MenuKitImpl::load_stencil(
    const char* name, Glyph_ptr& g,
    unsigned char* data, unsigned int width, unsigned int height
) {
    DrawingKit_ptr drawing = info_.drawing;
    FigureKit_ptr figure = info_.figure;
 
    Raster_ptr raster = nil;
    StyleValue_var a = style_->resolve(FrescoLib::tmp_string_ref(name));
    if (is_not_nil(a)) {
        CharString_ptr s;
        if (a->read_string(s)) {
            raster = drawing->raster_open(s);
            CORBA::release(s);
            // !!! should warn user if can't get raster -denis
        }
    }
    if (is_nil(raster)) {
        raster = make_bitmap(drawing, data, width, height);
    }
    g = figure->stencil(style_, Raster_var(raster));
}

void MenuKitImpl::unload() {
}
