/*
 * Copyright (c) 1994-95 Firefly Design
 * Copyright (c) 1994 Silicon Graphics, Inc.
 * Copyright (c) 1994 Fujitsu, Ltd.
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Silicon Graphics and Fujitsu may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Silicon Graphics and Fujitsu.
 *
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 *
 * IN NO EVENT SHALL SILICON GRAPHICS OR FUJITSU BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
 * OF THIS SOFTWARE.
 */

#include "popups.h"
#include <Fresco/defs_impl/math.h>
#include <Fresco/layouts/layouts.h>
#include <Fresco/widgets/menus.h>
#include <Fresco/widgets/widgets.h>

const Coord epsilon = Coord(0.001);

static void flush(Window_ptr window) {
    Screen_var screen = window->screen_ptr();
    Display_var display = screen->display_ptr();
    display->flush();
}

// class MKPopupMenu

MKPopupMenu::MKPopupMenu(MenuKitImpl::Info* info, MenuItem_ptr item) {
    box_ = nil;
    window_ = nil;
    manager_ = nil;
    window_tag_ = 0;

    info_ = info;

    item_ = MenuItem::_duplicate(item);
    item_tag_ = item_->attach(this);
}

MKPopupMenu::~MKPopupMenu() {
    item_->detach(item_tag_);
    CORBA::release(item_);
    CORBA::release(box_);
    CORBA::release(window_);
}

OxExchange* MKPopupMenu::_exchange() { return &object_; }

Tag MKPopupMenu::attach(FrescoObject_ptr observer) {
    return object_.attach(observer);
}
void MKPopupMenu::detach(Tag attach_tag) {
    object_.detach(attach_tag);
}
void MKPopupMenu::disconnect() {
    object_.disconnect();
}
void MKPopupMenu::notify_observers() {
    object_.notify_observers();
}
void MKPopupMenu::update() {
    if (window_tag_ != 0) {
        manager_->close_menus();
    }
    CORBA::release(window_);
    window_ = nil;
    CORBA::release(box_);
    box_ = nil;
}

void MKPopupMenu::menu_manager(MenuManager_ptr m) {
    manager_ = m;
}

void MKPopupMenu::show(Region_ptr a) {
    if (is_nil(window_)) {
        build(manager_->filter_screen());
    }
    Window::Placement p = position(a);
    Window::Placement c;
    window_->get_configuration(true, c);
    if (!(
        Math::equal(c.x, p.x, epsilon) &&
        Math::equal(c.y, p.y, epsilon) &&
        Math::equal(c.width, p.width, epsilon) &&
        Math::equal(c.height, p.height, epsilon) &&
        Math::equal(c.align_x, p.align_x, epsilon) &&
        Math::equal(c.align_y, p.align_y, epsilon)
    )) {
        window_->configure(p);
    }
    if (window_tag_ == 0) {
        window_->map();
        window_tag_ = manager_->add_window(window_);
        flush(window_);
    }
}

void MKPopupMenu::select(const MenuItem::Path& path, Boolean flap) {
    box_->select(path, flap);
}

void MKPopupMenu::selection(MenuItem::Path& path) {
    box_->selection(path);
}

void MKPopupMenu::hide() {
    if (window_tag_ != 0) {
        window_->unmap();
        manager_->remove_window(window_tag_);
        window_tag_ = 0;
        flush(window_);
    }
}

void MKPopupMenu::build(Screen_ptr screen) {
    WidgetKit_ptr widget = info_->widget;
    LayoutKit_ptr layout = info_->layout;

    box_ = new MKVBox(info_, item_);
    box_->menu_manager(manager_);
    ViewerImpl* viewer = new ViewerImpl(true);
    viewer->body(
        Glyph_var(widget->outset_frame(
            Glyph_var(layout->margin(box_, info_->menu_margin))
        ))
    );

    window_ = screen->popup(viewer);
}

Window::Placement MKPopupMenu::position(Region_ptr a) {
    Window::Placement p;
    Vertex lower, upper;
    a->bounds(lower, upper);
    location(lower, upper, p.x, p.y);
    Glyph::Requisition r;
    Viewer_var viewer = window_->main_viewer();
    viewer->request(r);
    p.width = r.x.natural;
    p.height = r.y.natural;
    p.align_x = Alignment(0);
    p.align_y = Alignment(1);
    return p;
}

// class MKPulldownMenu

MKPulldownMenu::MKPulldownMenu(
    MenuKitImpl::Info* info, MenuItem_ptr item
) : MKPopupMenu(info, item) {
}

void MKPulldownMenu::location(
    const Vertex& lower, const Vertex&, Coord& x, Coord& y
) {
    x = lower.x;
    y = lower.y;
}

// class MKPullrightMenu

MKPullrightMenu::MKPullrightMenu(
    MenuKitImpl::Info* info, MenuItem_ptr item
) : MKPopupMenu(info, item) {
}

void MKPullrightMenu::location(
    const Vertex&, const Vertex& upper, Coord& x, Coord& y
) {
    x = upper.x;
    y = upper.y;
}
