//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-93 Silicon Graphics, Inc.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Stanford and Silicon Graphics may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Stanford and Silicon Graphics.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD OR SILICON GRAPHICS BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#include <Fresco/defs/fresco.h>
#include <Fresco/defs_impl/math.h>
#include <Fresco/widgets/widgets.h>
#include "kit.h"
#include "slider.h"
#include "stepper.h"
#include "telltales.h"

#define default_minimum_thumb_size Coord(28)

// class SliderViewer

SliderViewer::SliderViewer(Slider* slider, Glyph_ptr bdy) {
    // this duplicate assumes ref counting
    Glyph::_duplicate(slider);
    body(bdy);
    slider_ = slider;
    dragging_ = false;
    aborted_ = false;
    xoffset_ = Coord(0);
    yoffset_ = Coord(0);
    stepper_ = nil;
}

SliderViewer::~SliderViewer() {
    CORBA::release(slider_);
    CORBA::release(Button_ptr(stepper_));
}

Boolean SliderViewer::move(GlyphTraversal_ptr t, Event_ptr e) {
    slider_->move_thumb(t, e);
    return ActiveViewer::move(t, e);
}

Boolean SliderViewer::press(GlyphTraversal_ptr t, Event_ptr e) {
    const left = 1, middle = 2, right = 3;
    Event::ButtonIndex b = e->pointer_button();
    if (b == right) {
	return false;
    }

    Coord x = e->pointer_x();
    Coord y = e->pointer_y();
    Vertex s0, s1, t0, t1;
    Region_var r = t->allocation();
    r->bounds(s0, s1);
    Region_var thumb = slider_->thumb_allocation(r);
    thumb->bounds(t0, t1);
    long rel = slider_->hit_thumb(t, e);
    if (rel == 0) {
	dragging_ = true;
	slider_->begin_adjustment();
	xoffset_ = s0.x + x - t0.x;
	yoffset_ = s0.y + y - t0.y;
    } else if (b == left) {
	stepper_ = (rel == 1) ?
	    slider_->forward_stepper() : slider_->backward_stepper();
	if (stepper_ != nil) {
	    stepper_->start_stepping();
	}
    } else {
	// middle => drag
	dragging_ = true;
	slider_->begin_adjustment();
	xoffset_ = s0.x + (t1.x - t0.x) * Coord(0.5);
	yoffset_ = s0.y + (t1.y - t0.y) * Coord(0.5);
	slider_->move_to(x - xoffset_, y - yoffset_);
	move(t, e);
    }
    return true;
}

Boolean SliderViewer::drag(GlyphTraversal_ptr, Event_ptr e) {
    if (!aborted_ && dragging_) {
	slider_->show_old_thumb();
	slider_->move_to(e->pointer_x() - xoffset_, e->pointer_y() - yoffset_);
    }
    return true;
}

Boolean SliderViewer::release(GlyphTraversal_ptr t, Event_ptr e) {
    if (dragging_) {
	dragging_ = false;
	slider_->release_thumb();
	if (aborted_) {
	    aborted_ = false;
	} else {
	    slider_->move_to(
		e->pointer_x() - xoffset_, e->pointer_y() - yoffset_
	    );
	    move(t, e);
	    slider_->commit_adjustment();
	}
    } else if (stepper_ != nil) {
	stepper_->stop_stepping();
	CORBA::release(Button_ptr(stepper_));
	stepper_ = nil;
	move(t, e);
    }
    return true;
}

// class Slider

Slider::Slider(
    WidgetKitImpl* wk, StyleContext_ptr style
) : old_thumb_index(0), normal_thumb_index(1), visible_thumb_index(2)
{
    kit_ = wk;
    normal_thumb_ = nil;
    visible_thumb_ = nil;
    old_thumb_ = nil;
    min_thumb_size_ = default_minimum_thumb_size;
    StyleValue_var v = style->resolve(
	FrescoLib::tmp_string_ref("minThumbSize")
    );
    if (is_not_nil(v)) {
	Coord c;
	if (v->read_coord(c)) {
	    min_thumb_size_ = c;
	}
    }
    showing_visible_ = false;
    showing_old_thumb_ = false;
    forward_ = nil;
    backward_ = nil;
}

Slider::~Slider() {
    CORBA::release(normal_thumb_);
    CORBA::release(visible_thumb_);
    CORBA::release(old_thumb_);
    CORBA::release(Button_ptr(forward_));
    CORBA::release(Button_ptr(backward_));
}

void Slider::normal_thumb(Glyph_ptr g) {
    CORBA::release(normal_thumb_);
    normal_thumb_ = new PolyGlyphOffset(this, normal_thumb_index, g);
}

void Slider::visible_thumb(Glyph_ptr g) {
    CORBA::release(visible_thumb_);
    visible_thumb_ = new PolyGlyphOffset(this, visible_thumb_index, g);
}

void Slider::old_thumb(Glyph_ptr g) {
    CORBA::release(old_thumb_);
    old_thumb_ = new PolyGlyphOffset(this, old_thumb_index, g);
}

Stepper* Slider::forward_stepper() {
    // assumes ref counting
    Button::_duplicate(forward_);
    return forward_;
}

Stepper* Slider::backward_stepper() {
    // assumes ref counting
    Button::_duplicate(backward_);
    return backward_;
}

//+ Slider(XfGlyph::request)
void Slider::request(XfGlyph::Requisition& r) {
    r.x.defined = true;
    r.x.natural = Coord(22);
    r.x.maximum = kit_->layouts()->fil();
    r.x.minimum = Coord(0);
    r.x.align = Alignment(0);
    r.y = r.x;
}

//+ Slider(XfGlyph::draw)
void Slider::draw(XfGlyphTraversal_ptr t) {
    Painter_var p = t->current_painter();
    if (t->allocation_is_visible()) {
	Vertex v0, v1, v;
	t->bounds(v0, v1, v);
	p->push_clipping();
	p->clip_rect(v0.x, v0.y, v1.x, v1.y);
	if (showing_old_thumb_) {
	    t->traverse_child(
		old_thumb_, old_thumb_allocation(Region_var(t->allocation()))
	    );
	}
	t->traverse_child(
	    showing_visible_ ? visible_thumb_ : normal_thumb_,
	    thumb_allocation(Region_var(t->allocation()))
	);
	p->pop_clipping();
    }
}

void Slider::child_allocation(Long i, Glyph::AllocationInfo& a) {
    Region_ptr r = a.allocation;
    a.allocation = (
	(i = old_thumb_index) ? old_thumb_allocation(r) : thumb_allocation(r)
    );
}

void Slider::update() {
    if (adjustments_changed()) {
	redraw_thumb();
    }
}

void Slider::move_thumb(GlyphTraversal_ptr t, Event_ptr e) {
    if (is_not_nil(visible_thumb_)) {
	if (hit_thumb(t, e) == 0) {
	    if (!showing_visible_) {
		showing_visible_ = true;
		redraw_thumb();
	    }
	} else {
	    if (showing_visible_) {
		showing_visible_ = false;
		redraw_thumb();
	    }
	}
    }
}

Long Slider::hit_thumb(GlyphTraversal_ptr t, Event_ptr) {
    Painter_var p = t->current_painter();
    Region_var r = p->visible();
    Vertex ev, v;
    r->bounds(ev, v);
    Coord x = ev.x, y = ev.y;
    Region_var thumb = thumb_allocation(t->allocation());
    Vertex t0, t1;
    thumb->bounds(t0, t1);
    if (x >= t0.x && x < t1.x && y >= t0.y && y < t1.y) {
	return 0;
    }
    if (x < t0.x || y < t0.y) {
	return -1;
    }
    return 1;
}

void Slider::show_old_thumb() {
    if (!showing_old_thumb_ && is_not_nil(old_thumb_)) {
	showing_old_thumb_ = true;
    }
}

void Slider::release_thumb() {
    if (showing_old_thumb_) {
	showing_old_thumb_ = false;
	Glyph_var(old_thumb_->child())->need_redraw();
    }
    redraw_thumb();
}

void Slider::redraw_thumb() {
    GlyphOffset_ptr thumb = showing_visible_ ? visible_thumb_ : normal_thumb_;
    Glyph_var(thumb->child())->need_redraw();
}

//
// These should be pure virtual, but that tickles a cfront bug.
//

Region_ptr Slider::thumb_allocation(Region_ptr r) { return r; }
void Slider::move_to(Coord, Coord) { }
void Slider::begin_adjustment() { }
void Slider::commit_adjustment() { }

void Slider::forward_stepper(Stepper* s) {
    // assumes ref counting
    Button::_duplicate(s);
    CORBA::release(Button_ptr(forward_));
    forward_ = s;
}

void Slider::backward_stepper(Stepper* s) {
    // assumes ref counting
    Button::_duplicate(s);
    CORBA::release(Button_ptr(backward_));
    backward_ = s;
}

void Slider::allot_thumb_major_axis(
    Coord slider_size, Coord slider_lower, Adjustment::Settings& s,
    float& scale, Coord& lower, Coord& upper, float& align
) {
    const Coord epsilon = Coord(0.001);
    Coord length = s.length;
    Coord cur_length = s.cur_length;
    Coord thumb_size;
    Coord thumb_start;
    if (
        Math::equal(length, Coord(0), epsilon) ||
        (length < (cur_length + epsilon))
    ) {
	thumb_size = slider_size;
	thumb_start = Coord(0);
	scale = 1.0f;
    } else {
	thumb_size = slider_size * cur_length / length;
	if (thumb_size > slider_size) {
	    thumb_size = slider_size;
	    thumb_start = Coord(0);
	    scale = 1.0f;
	} else {
	    if (thumb_size < min_thumb_size_) {
		thumb_size = min_thumb_size_;
	    }
	    scale = (slider_size - thumb_size) / (length - cur_length);
	    thumb_start = scale * (s.cur_lower - s.lower);
	}
    }
    lower = slider_lower + thumb_start;
    upper = lower + thumb_size;
    align = Alignment(0);
}

void Slider::allot_thumb_minor_axis(
    Coord lower, Coord upper, Coord& new_lower, Coord& new_upper, float& align
) {
    new_lower = lower;
    new_upper = upper;
    align = Alignment(0);
}

Boolean Slider::changed(Adjustment_ptr a, Adjustment::Settings& s) {
    Adjustment::Settings new_s;
    a->get_settings(new_s);
    float tol = float(1e-3);
    Boolean different = (
	!Math::equal(new_s.lower, s.lower, tol) ||
	!Math::equal(new_s.upper, s.upper, tol) ||
	!Math::equal(new_s.length, s.length, tol) ||
	!Math::equal(new_s.cur_lower, s.cur_lower, tol) ||
	!Math::equal(new_s.cur_upper, s.cur_upper, tol) ||
	!Math::equal(new_s.cur_length, s.cur_length, tol)
    );
    if (different) {
	s = new_s;
    }
    return different;
}

//
// These should be pure virtual functions, but that tickles a cfront bug.
//

Region_ptr Slider::old_thumb_allocation(Region_ptr r) { return r; }
Boolean Slider::adjustments_changed() { return false; }

// class XSlider

XSlider::XSlider(
    WidgetKitImpl* wk, StyleContext_ptr s, Adjustment_ptr a
) : Slider(wk, s) {
    adjustment_ = a;
    tag_ = a->attach(this);
    xscale_ = 1.0f;
    a->get_settings(settings_);
    old_settings_ = settings_;

    Telltale_var t = new TelltaleImpl;
    forward_stepper(new ForwardPager(s, t, a));
    backward_stepper(new BackwardPager(s, t, a));
}

XSlider::~XSlider() {
    if (is_not_nil(adjustment_)) {
	adjustment_->detach(tag_);
    }
}

void XSlider::move_to(Coord x, Coord) {
    Adjustment::Settings s;
    adjustment_->get_settings(s);
    adjustment_->scroll_to(s.lower + x / xscale_);
}

Region_ptr XSlider::old_thumb_allocation(Region_ptr a) {
    return allocate_thumb(a, old_settings_);
}

Region_ptr XSlider::thumb_allocation(Region_ptr a) {
    return allocate_thumb(a, settings_);
}

Region_ptr XSlider::allocate_thumb(
    Region_ptr given, Adjustment::Settings& s
) {
    RegionImpl* a = new RegionImpl;
    Vertex v0, v1;
    given->bounds(v0, v1);
    allot_thumb_major_axis(
	v1.x - v0.x, v0.x, s, xscale_, a->lower_.x, a->upper_.x, a->xalign_
    );
    allot_thumb_minor_axis(v0.y, v1.y, a->lower_.y, a->upper_.y, a->yalign_);
    a->defined_ = true;
    return a;
}

void XSlider::disconnect() { adjustment_ = nil; }

void XSlider::begin_adjustment() {
    old_settings_ = settings_;
    adjustment_->begin();
}

void XSlider::commit_adjustment() {
    adjustment_->commit();
}

Boolean XSlider::adjustments_changed() {
    return changed(adjustment_, settings_);
}

// class YSlider

YSlider::YSlider(
    WidgetKitImpl* wk, StyleContext_ptr s, Adjustment_ptr a
) : Slider(wk, s) {
    adjustment_ = a;
    tag_ = a->attach(this);
    yscale_ = 1.0f;
    a->get_settings(settings_);
    old_settings_ = settings_;

    Telltale_var t = new TelltaleImpl;
    forward_stepper(new ForwardPager(s, t, a));
    backward_stepper(new BackwardPager(s, t, a));
}

YSlider::~YSlider() {
    if (is_not_nil(adjustment_)) {
	adjustment_->detach(tag_);
    }
}

void YSlider::move_to(Coord, Coord y) {
    Adjustment::Settings s;
    adjustment_->get_settings(s);
    adjustment_->scroll_to(s.lower + y / yscale_);
}

Region_ptr YSlider::old_thumb_allocation(Region_ptr a) {
    return allocate_thumb(a, old_settings_);
}

Region_ptr YSlider::thumb_allocation(Region_ptr a) {
    return allocate_thumb(a, settings_);
}

Region_ptr YSlider::allocate_thumb(
    Region_ptr given, Adjustment::Settings& s
) {
    RegionImpl* a = new RegionImpl;
    Vertex v0, v1;
    given->bounds(v0, v1);
    allot_thumb_major_axis(
	v1.y - v0.y, v0.y, s, yscale_, a->lower_.y, a->upper_.y, a->yalign_
    );
    allot_thumb_minor_axis(v0.x, v1.x, a->lower_.x, a->upper_.x, a->xalign_);
    a->defined_ = true;
    return a;
}

void YSlider::disconnect() { adjustment_ = nil; }

void YSlider::begin_adjustment() {
    old_settings_ = settings_;
    adjustment_->begin();
}

void YSlider::commit_adjustment() {
    adjustment_->commit();
}

Boolean YSlider::adjustments_changed() {
    return changed(adjustment_, settings_);
}

// class XYSlider

XYSlider::XYSlider(
    WidgetKitImpl* wk, StyleContext_ptr s, Adjustment_ptr x, Adjustment_ptr y
) : Slider(wk, s) {
    x_adjustment_ = x;
    y_adjustment_ = y;
    x_tag_ = x->attach(this);
    y_tag_ = y->attach(this);
    xscale_ = 1.0f;
    yscale_ = 1.0f;
    x->get_settings(x_settings_);
    y->get_settings(y_settings_);
    x_old_settings_ = x_settings_;
    y_old_settings_ = y_settings_;
}

XYSlider::~XYSlider() {
    if (is_not_nil(x_adjustment_)) {
	x_adjustment_->detach(x_tag_);
    }
    if (is_not_nil(y_adjustment_)) {
	y_adjustment_->detach(y_tag_);
    }
}

void XYSlider::move_to(Coord x, Coord y) {
    Adjustment::Settings s;
    x_adjustment_->get_settings(s);
    x_adjustment_->scroll_to(s.lower + x / xscale_);

    y_adjustment_->get_settings(s);
    y_adjustment_->scroll_to(s.lower + y / yscale_);
}

Region_ptr XYSlider::old_thumb_allocation(Region_ptr a) {
    return allocate_thumb(a, x_old_settings_, y_old_settings_);
}

Region_ptr XYSlider::thumb_allocation(Region_ptr a) {
    return allocate_thumb(a, x_settings_, y_settings_);
}

Region_ptr XYSlider::allocate_thumb(
    Region_ptr given, Adjustment::Settings& x_s, Adjustment::Settings& y_s
) {
    RegionImpl* a = new RegionImpl;
    Vertex v0, v1;
    given->bounds(v0, v1);
    allot_thumb_major_axis(
	v1.x - v0.x, v0.x, x_s, xscale_, a->lower_.x, a->upper_.x, a->xalign_
    );
    allot_thumb_major_axis(
	v1.y - v0.y, v0.y, y_s, yscale_, a->lower_.y, a->upper_.y, a->yalign_
    );
    a->defined_ = true;
    return a;
}
void XYSlider::disconnect() {
    x_adjustment_ = nil;
    y_adjustment_ = nil;
}

void XYSlider::begin_adjustment() {
    x_old_settings_ = x_settings_;
    y_old_settings_ = y_settings_;
    x_adjustment_->begin();
    y_adjustment_->begin();
}

void XYSlider::commit_adjustment() {
    x_adjustment_->commit();
    y_adjustment_->commit();
}

Boolean XYSlider::adjustments_changed() {
    Boolean x = changed(x_adjustment_, x_settings_);
    Boolean y = changed(y_adjustment_, y_settings_);
    return x || y;
}
