//
// Copyright (c) 1987-91 Stanford University
// Copyright (c) 1991-93 Silicon Graphics, Inc.
//
// Permission to use, copy, modify, distribute, and sell this software and 
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names of
// Stanford and Silicon Graphics may not be used in any advertising or
// publicity relating to the software without the specific, prior written
// permission of Stanford and Silicon Graphics.
// 
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
//
// IN NO EVENT SHALL STANFORD OR SILICON GRAPHICS BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
// OF THIS SOFTWARE.
//

#include <Fresco/defs/fresco.h>
#include <Fresco/defs_impl/action.h>
#include <Fresco/defs_impl/threads.h>
#include <Fresco/widgets/widgets.h>
#include "bevel.h"
#include "stepper.h"

declareActionCallback(Stepper)
implementActionCallback(Stepper)

Stepper::Stepper(
    StyleContext_ptr s, Telltale_ptr t, Action_ptr a
) : ButtonImpl(t, a) {
    float seconds = 0.25f;
    StyleValue_var v = s->resolve(
	FrescoLib::tmp_string_ref("autorepeatStart")
    );
    if (is_not_nil(v)) {
	double d;
	if (v->read_real(d)) {
	    seconds = float(d);
	}
    }
    start_delay_ = long(seconds * 100);
    seconds = 0.05f;
    v = s->resolve(FrescoLib::tmp_string_ref("autorepeatDelay"));
    if (is_not_nil(v)) {
	double d;
	if (v->read_real(d)) {
	    seconds = float(d);
	}
    }
    next_delay_ = long(seconds * 100);
    lock_ = ThreadLib::thread_kit_ptr()->mutex_ptr();
    ticker_ = nil;
}

Stepper::~Stepper() { }

Boolean Stepper::press(GlyphTraversal_ptr t, Event_ptr e) {
    Boolean b = ButtonImpl::press(t, e);
    start_stepping();
    return b;
}

Boolean Stepper::release(GlyphTraversal_ptr t, Event_ptr e) {
    stop_stepping();
    return ButtonImpl::release(t, e);
}

void Stepper::start_stepping() {
    lock_->acquire();
    state_->set(Telltale::stepping);
    lock_->release();
    adjust();
    if (start_delay_ > 10) {
	ticker_ = ThreadLib::thread_kit_ptr()->thread_ptr(
	    new ActionCallback(Stepper)(this, &Stepper::tick)
	);
	if (is_not_nil(ticker_)) {
	    ticker_->run();
	}
    }
}

void Stepper::stop_stepping() {
    if (is_not_nil(ticker_)) {
	ticker_->terminate();
	ticker_ = nil;
    }
    state_->clear(Telltale::stepping);
}

//
// This should be a pure virtual, but that can tickle a cfront bug.
//
void Stepper::adjust() { }

void Stepper::tick() {
    FrescoLib::delay(float(start_delay_ * 0.01));
    float next_delay = float(next_delay_ * 0.01);
    Boolean stepping = true;
    while (stepping) {
	lock_->acquire();
	stepping = state_->test(Telltale::stepping);
	lock_->release();
	if (stepping) {
	    adjust();
	}
	FrescoLib::delay(next_delay);
    }
}

#define implementAdjustStepper(Name,scroll) \
Name::Name( \
    StyleContext_ptr s, Telltale_ptr t, Adjustment_ptr a \
) : Stepper(s, t) { \
    adjustment_ = a; \
} \
\
Name::~Name() { } \
\
void Name::adjust() { adjustment_->scroll(); }

implementAdjustStepper(ForwardScroller,scroll_forward)
implementAdjustStepper(BackwardScroller,scroll_backward)
implementAdjustStepper(ForwardPager,page_forward)
implementAdjustStepper(BackwardPager,page_backward)

#define implementArrowGlyph(Name) \
Name::Name(Color_ptr c) { \
    color_ = c; \
} \
\
Name::~Name() { } \

implementArrowGlyph(UpArrow)
implementArrowGlyph(DownArrow)
implementArrowGlyph(LeftArrow)
implementArrowGlyph(RightArrow)

//+ UpArrow(XfGlyph::draw)
void UpArrow::draw(XfGlyphTraversal_ptr t) {
    Vertex lower, upper, center;
    t->bounds(lower, upper, center);

    Painter_var p = t->current_painter();
    p->current_color(color_);
    Vertex v[3];
    v[0].x = lower.x; v[0].y = lower.y;
    v[1].x = upper.x; v[1].y = lower.y;
    v[2].x = (lower.x + upper.x) * Coord(0.5); v[2].y = upper.y;
    Beveler::fill_path(p, v, 3);
}

//+ DownArrow(XfGlyph::draw)
void DownArrow::draw(XfGlyphTraversal_ptr t) {
    Vertex lower, upper, center;
    t->bounds(lower, upper, center);

    Painter_var p = t->current_painter();
    p->current_color(color_);
    Vertex v[3];
    v[0].x = lower.x; v[0].y = upper.y;
    v[1].x = upper.x; v[1].y = upper.y;
    v[2].x = (lower.x + upper.x) * Coord(0.5); v[2].y = lower.y;
    Beveler::fill_path(p, v, 3);
}

//+ LeftArrow(XfGlyph::draw)
void LeftArrow::draw(XfGlyphTraversal_ptr t) {
    Vertex lower, upper, center;
    t->bounds(lower, upper, center);

    Painter_var p = t->current_painter();
    p->current_color(color_);
    Vertex v[3];
    v[0].x = upper.x; v[0].y = lower.y;
    v[1].x = upper.x; v[1].y = upper.y;
    v[2].x = lower.x; v[2].y = (lower.y + upper.y) * Coord(0.5);
    Beveler::fill_path(p, v, 3);
}

//+ RightArrow(XfGlyph::draw)
void RightArrow::draw(XfGlyphTraversal_ptr t) {
    Vertex lower, upper, center;
    t->bounds(lower, upper, center);

    Painter_var p = t->current_painter();
    p->current_color(color_);
    Vertex v[3];
    v[0].x = lower.x; v[0].y = lower.y;
    v[1].x = lower.x; v[1].y = upper.y;
    v[2].x = upper.x; v[2].y = (lower.y + upper.y) * Coord(0.5);
    Beveler::fill_path(p, v, 3);
}
