//
// Copyright (c) 1994-95 Firefly Design
// Copyright (c) 1991-94 Silicon Graphics, Inc.
// Copyright (c) 1993-94 Fujitsu, Ltd.
//
// Permission to use, copy, modify, distribute, and sell this software and
// its documentation for any purpose is hereby granted without fee, provided
// that (i) the above copyright notices and this permission notice appear in
// all copies of the software and related documentation, and (ii) the names
// of Stanford, Silicon Graphics, and Fujitsu may not be used in any
// advertising or publicity relating to the software without the specific,
// prior written permission of Stanford, Silicon Graphics, and Fujitsu.
//
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
//
// IN NO EVENT SHALL STANFORD, SILICON GRAPHICS, OR FUJITSU BE LIABLE FOR
// ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
// OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
// WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF
// LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE
// OF THIS SOFTWARE.
//

#include <Fresco/defs_impl/math.h>
#include <Fresco/defs_impl/transform.h>
#include "viewport.h"

#define infinite_coord 10e6
#define epsilon Coord(0.01)

// class WKViewportAdjustment

WKViewportAdjustment::WKViewportAdjustment(WKViewport* viewport, XfAxis axis) {
    viewport_ = viewport;
    axis_ = axis;
    scroll_incr_ = 18.0; // .25 inch
    page_incr_ = 0;
}

WKViewportAdjustment::~WKViewportAdjustment() { }

OxExchange* WKViewportAdjustment::_exchange() { return &object_; }

//+ WKViewportAdjustment(XfFrescoObject::=object_.)
XfTag WKViewportAdjustment::attach(XfFrescoObject_ptr observer) {
    return object_.attach(observer);
}
void WKViewportAdjustment::detach(XfTag attach_tag) {
    object_.detach(attach_tag);
}
void WKViewportAdjustment::disconnect() {
    object_.disconnect();
}
void WKViewportAdjustment::notify_observers() {
    object_.notify_observers();
}
void WKViewportAdjustment::update() {
    object_.update();
}
//+

//+ WKViewportAdjustment(Adjustment::small_scroll=s)
void WKViewportAdjustment::small_scroll(XfCoord s) { scroll_incr_ = s; }

//+ WKViewportAdjustment(Adjustment::small_scroll?)
XfCoord WKViewportAdjustment::small_scroll() { return scroll_incr_; }

//+ WKViewportAdjustment(Adjustment::large_scroll=p)
void WKViewportAdjustment::large_scroll(XfCoord p) { page_incr_ = p; }

//+ WKViewportAdjustment(Adjustment::large_scroll?)
XfCoord WKViewportAdjustment::large_scroll() { return page_incr_; }

//+ WKViewportAdjustment(Adjustment::get_settings)
void WKViewportAdjustment::get_settings(Adjustment::Settings& s) {
    s.lower = viewport_->lower(axis_);
    s.length = viewport_->length(axis_);
    s.upper = s.lower + s.length;
    s.cur_lower = viewport_->cur_lower(axis_);
    s.cur_length = viewport_->cur_length(axis_);
    s.cur_upper = s.cur_lower + s.cur_length;
}

//+ WKViewportAdjustment(Adjustment::begin)
void WKViewportAdjustment::begin() { get_settings(saved_); }

//+ WKViewportAdjustment(Adjustment::commit)
void WKViewportAdjustment::commit() { }

//+ WKViewportAdjustment(Adjustment::cancel)
void WKViewportAdjustment::cancel() {
    viewport_->scroll_to(axis_, saved_.lower);
}

//+ WKViewportAdjustment(Adjustment::scroll_forward)
void WKViewportAdjustment::scroll_forward() {
    scroll_to(viewport_->cur_lower(axis_) + scroll_incr_);
}

//+ WKViewportAdjustment(Adjustment::scroll_backward)
void WKViewportAdjustment::scroll_backward() {
    scroll_to(viewport_->cur_lower(axis_) - scroll_incr_);
}

//+ WKViewportAdjustment(Adjustment::page_forward)
void WKViewportAdjustment::page_forward() {
    scroll_to(viewport_->cur_lower(axis_) + page_amount());
}

//+ WKViewportAdjustment(Adjustment::page_backward)
void WKViewportAdjustment::page_backward() {
    scroll_to(viewport_->cur_lower(axis_) - page_amount());
}

//+ WKViewportAdjustment(Adjustment::scroll_to)
void WKViewportAdjustment::scroll_to(XfCoord lower) {
    constrain(lower);
    if (! Math::equal(lower, viewport_->cur_lower(axis_), epsilon)) {
        viewport_->scroll_to(axis_, lower);
	notify_observers();
    }
}

//+ WKViewportAdjustment(Adjustment::scroll_by)
void WKViewportAdjustment::scroll_by(XfCoord delta) {
    scroll_to(viewport_->cur_lower(axis_) + delta);
}

//+ WKViewportAdjustment(Adjustment::scale_to)
void WKViewportAdjustment::scale_to(XfCoord) { }

//+ WKViewportAdjustment(Adjustment::constrain)
void WKViewportAdjustment::constrain(XfCoord& c) {
    Coord lower = viewport_->lower(axis_);
    Coord upper = lower + Math::max(viewport_->length(axis_) -
        viewport_->cur_length(axis_), Coord(0.0));

    if (c < lower) {
	c = lower;
    } else if (c > upper) {
	c = upper;
    }
}

Coord WKViewportAdjustment::page_amount() {
    return Math::equal(page_incr_, Coord(0), epsilon) ?
        viewport_->cur_length(axis_) : page_incr_;
}

// class WKViewport

WKViewport::WKViewport(
    XfGlyph_ptr g
) {
    x_adjustment_ = new WKViewportAdjustment(this, X_axis);
    y_adjustment_ = new WKViewportAdjustment(this, Y_axis);
    tx_ = new TransformImpl;
    requested_ = false;
    cur_offset_[0] = cur_offset_[1] = 0.0;
    cur_length_[0] = cur_length_[1] = 0.0;
    XfGlyphImpl::default_requisition(requisition_);

    body(g);
}

WKViewport::~WKViewport() {
    CORBA::release(x_adjustment_);
    CORBA::release(y_adjustment_);
    CORBA::release(tx_);
}

OxExchange* WKViewport::_exchange() { return ViewerImpl::_exchange(); }

//+ WKViewport(XfFrescoObject::=ViewerImpl::)
XfTag WKViewport::attach(XfFrescoObject_ptr observer) {
    return ViewerImpl::attach(observer);
}
void WKViewport::detach(XfTag attach_tag) {
    ViewerImpl::detach(attach_tag);
}
void WKViewport::disconnect() {
    ViewerImpl::disconnect();
}
void WKViewport::notify_observers() {
    ViewerImpl::notify_observers();
}
void WKViewport::update() {
    ViewerImpl::update();
}
//+

XfGlyph_ptr WKViewport::clone_glyph() {
    return ViewerImpl::clone_glyph();
}
XfStyleContext_ptr WKViewport::style() {
    return ViewerImpl::style();
}
void WKViewport::style(XfStyleContext_ptr _p) {
    ViewerImpl::style(_p);
}
void WKViewport::extension(const XfGlyph::AllocationInfo& a, XfRegion_ptr r) {
    ViewerImpl::extension(a, r);
}
void WKViewport::shape(XfRegion_ptr r) {
    ViewerImpl::shape(r);
}
void WKViewport::draw(XfGlyphTraversal_ptr t) {
    ViewerImpl::draw(t);
}
void WKViewport::pick(XfGlyphTraversal_ptr t) {
    ViewerImpl::pick(t);
}
XfGlyph_ptr WKViewport::body() {
    return ViewerImpl::body();
}
void WKViewport::body(XfGlyph_ptr _p) {
    ViewerImpl::body(_p);
}
void WKViewport::append(XfGlyph_ptr g) {
    ViewerImpl::append(g);
}
void WKViewport::prepend(XfGlyph_ptr g) {
    ViewerImpl::prepend(g);
}
XfTag WKViewport::add_parent(XfGlyphOffset_ptr parent_offset) {
    return ViewerImpl::add_parent(parent_offset);
}
void WKViewport::remove_parent(XfTag add_tag) {
    ViewerImpl::remove_parent(add_tag);
}
XfGlyphOffset_ptr WKViewport::first_child_offset() {
    return ViewerImpl::first_child_offset();
}
XfGlyphOffset_ptr WKViewport::last_child_offset() {
    return ViewerImpl::last_child_offset();
}
void WKViewport::parent_offsets(XfGlyph::OffsetSeq& parents) {
    ViewerImpl::parent_offsets(parents);
}
void WKViewport::allocations(XfGlyph::AllocationInfoSeq& a) {
    ViewerImpl::allocations(a);
}
void WKViewport::need_redraw() {
    ViewerImpl::need_redraw();
}
void WKViewport::need_redraw_region(XfRegion_ptr r) {
    ViewerImpl::need_redraw_region(r);
}
Boolean WKViewport::restore_trail(XfGlyphTraversal_ptr t) {
    return ViewerImpl::restore_trail(t);
}

//+ WKViewport(XfViewer::=ViewerImpl::)
XfViewer_ptr WKViewport::parent_viewer() {
    return ViewerImpl::parent_viewer();
}
XfViewer_ptr WKViewport::next_viewer() {
    return ViewerImpl::next_viewer();
}
XfViewer_ptr WKViewport::prev_viewer() {
    return ViewerImpl::prev_viewer();
}
XfViewer_ptr WKViewport::first_viewer() {
    return ViewerImpl::first_viewer();
}
XfViewer_ptr WKViewport::last_viewer() {
    return ViewerImpl::last_viewer();
}
void WKViewport::append_viewer(XfViewer_ptr v) {
    ViewerImpl::append_viewer(v);
}
void WKViewport::prepend_viewer(XfViewer_ptr v) {
    ViewerImpl::prepend_viewer(v);
}
void WKViewport::insert_viewer(XfViewer_ptr v) {
    ViewerImpl::insert_viewer(v);
}
void WKViewport::replace_viewer(XfViewer_ptr v) {
    ViewerImpl::replace_viewer(v);
}
void WKViewport::remove_viewer() {
    ViewerImpl::remove_viewer();
}
void WKViewport::set_viewer_links(XfViewer_ptr parent, XfViewer_ptr prev, XfViewer_ptr next) {
    ViewerImpl::set_viewer_links(parent, prev, next);
}
void WKViewport::set_first_viewer(XfViewer_ptr v) {
    ViewerImpl::set_first_viewer(v);
}
void WKViewport::set_last_viewer(XfViewer_ptr v) {
    ViewerImpl::set_last_viewer(v);
}
XfFocus_ptr WKViewport::request_focus(XfViewer_ptr requestor, Boolean temporary) {
    return ViewerImpl::request_focus(requestor, temporary);
}
Boolean WKViewport::receive_focus(XfFocus_ptr f, Boolean primary) {
    return ViewerImpl::receive_focus(f, primary);
}
void WKViewport::lose_focus(Boolean temporary) {
    ViewerImpl::lose_focus(temporary);
}
Boolean WKViewport::first_focus() {
    return ViewerImpl::first_focus();
}
Boolean WKViewport::last_focus() {
    return ViewerImpl::last_focus();
}
Boolean WKViewport::next_focus() {
    return ViewerImpl::next_focus();
}
Boolean WKViewport::prev_focus() {
    return ViewerImpl::prev_focus();
}
Boolean WKViewport::handle(XfGlyphTraversal_ptr t, XfEvent_ptr e) {
    return ViewerImpl::handle(t, e);
}
void WKViewport::close() {
    ViewerImpl::close();
}
//+

//+ WKViewport(XfGlyph::transformation)
XfTransform_ptr WKViewport::transformation() {
    return Transform::_duplicate(tx_);
}

//+ WKViewport(XfGlyph::request)
void WKViewport::request(XfGlyph::Requisition& r) {
    GlyphImpl::require(r.x, 0, infinite_coord, 0, 0);
    GlyphImpl::require(r.y, 0, infinite_coord, 0, 0);
    GlyphImpl::require(r.z, 0, infinite_coord, 0, 0);
}

//+ WKViewport(XfGlyph::traverse)
void WKViewport::traverse(XfGlyphTraversal_ptr t) {
    Painter_var p = t->current_painter();
    Region_var a = t->allocation();

    check_allocation(a);
    Flags flags = push_environment(p, a);

    if (is_not_nil(offset_.child_)) {
        t->traverse_child(&offset_, Region_var(body_allocation(a)));
    }

    pop_environment(p, flags);
}

WKViewport::Flags WKViewport::push_environment(XfPainter_ptr p, XfRegion_ptr a) {
    Flags flags;

    flags.clipping = is_not_nil(a);
    if (flags.clipping) {
        p->push_clipping();
        Vertex lower, upper;
        a->bounds(lower, upper);
        p->clip_rect(lower.x, lower.y, upper.x, upper.y);
    }

    flags.transforming = is_not_nil(p) && !tx_->identity();
    if (flags.transforming) {
        p->push_matrix();
        p->premultiply(tx_);
    }

    return flags;
}

void WKViewport::pop_environment(XfPainter_ptr p, WKViewport::Flags flags) {
    if (flags.transforming) {
        p->pop_matrix();
    }

    if (flags.clipping) {
        p->pop_clipping();
    }
}

//+ WKViewport(XfGlyph::need_resize)
void WKViewport::need_resize() {
    requested_ = false;
    x_adjustment_->notify_observers();
    y_adjustment_->notify_observers();
    ViewerImpl::need_resize();
}

//+ WKViewport(Viewport::scroll_adjustment)
Adjustment_ptr WKViewport::scroll_adjustment(XfAxis a) {
    return Adjustment::_duplicate(a == X_axis ? x_adjustment_ : y_adjustment_);
}

void WKViewport::scroll_to(XfAxis a, Coord lower) {
    if (a == X_axis) {
        cur_offset_[X_axis] = lower - adjustment(a);
    } else {
        cur_offset_[Y_axis] = lower - adjustment(a);
    }
    Vertex o;
    o.x = -cur_offset_[X_axis];
    o.y = -cur_offset_[Y_axis];
    o.z = 0;
    TransformImpl t;
    t.translate(o);
    tx_->copy(&t);
    need_redraw();
}

XfCoord WKViewport::lower(XfAxis a) {
    return 0;
}

XfCoord WKViewport::length(XfAxis a) {
    cache_requisition();
    Requirement& r = (a == X_axis ? requisition_.x : requisition_.y);
    return r.natural;
}

XfCoord WKViewport::cur_lower(XfAxis a) {
    return cur_offset_[a] + adjustment(a);
}

XfCoord WKViewport::cur_length(XfAxis a) {
    return cur_length_[a];
}

void WKViewport::cache_requisition() {
    if (! requested_) {
        requested_ = true;
        offset_.child_->request(requisition_);
    }
}

void WKViewport::check_allocation(Region_ptr a) {
    if (is_not_nil(a)) {
        Vertex lower, upper;
        a->bounds(lower, upper);
        Coord w = upper.x - lower.x;
        if (! Math::equal(w, cur_length_[X_axis], epsilon)) {
            cur_length_[X_axis] = w;
// !!! this currently causes a deadlock in the X11 DisplayImpl (damage during
// a draw traversal) -denis
//            x_adjustment_->scroll_to(cur_lower(X_axis));
//            x_adjustment_->notify_observers();
        }
        Coord h = upper.y - lower.y;
        if (! Math::equal(h, cur_length_[Y_axis], epsilon)) {
            cur_length_[Y_axis] = h;
// !!! this currently causes a deadlock in the X11 DisplayImpl (damage during
// a draw traversal) -denis
//            y_adjustment_->scroll_to(cur_lower(Y_axis));
//            y_adjustment_->notify_observers();
        }
    }
}

Region_ptr WKViewport::body_allocation(Region_ptr a) {
    if (is_not_nil(a)) {
        cache_requisition();
        Vertex lower, upper;
        a->bounds(lower, upper);
        RegionImpl* ca = new RegionImpl;
        ca->defined_ = true;
        Glyph::Requirement& rx = requisition_.x;
        Coord ox = lower.x + (upper.x - lower.x) * rx.align;
        ca->lower_.x = ox - rx.natural * rx.align;
        ca->upper_.x = ox + rx.natural * (1.0 - rx.align);
        ca->xalign_ = rx.align;
        Glyph::Requirement& ry = requisition_.y;
        Coord oy = lower.y + (upper.y - lower.y) * ry.align;
        ca->lower_.y = oy - ry.natural * ry.align;
        ca->upper_.y = oy + ry.natural * (1.0 - ry.align);
        ca->yalign_ = ry.align;
        return ca;
    }
    return nil;
}

Coord WKViewport::adjustment(Axis a) {
    Requirement& r = (a == X_axis ? requisition_.x : requisition_.y);
    return (r.natural - cur_length_[a]) * r.align;
}
